/*
 * bpfx.5.c
 * Copyright (C) 2009-2010, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <bpfx-defs.h>
#include <bpfx-types.h>

#define bpfset(bpfset)			((struct bpfset_type *) (bpfset))

static int line(void *, unsigned, void *, int (*) (void *, void *));
static int mind(void *, unsigned, void *, int (*) (void *, void *));
static int pick(void *, unsigned, unsigned, void *, int (*) (void *, void *));
static int slip(void *, unsigned, unsigned, void *, int (*) (void *, void *));

static int
line(void *node, unsigned size, void *back, int (*call) (void *, void *))
{
    int delete;
    unsigned note;

    note = *(integral_q *) node;
    if (note) {
	do {
	    node = (char *) node + size;
	    note--;
	    delete = call(back, node);
	    if (delete) {
		break;
	    }
	} while (note);
    } else {
	delete = 0;
    }

    return delete;
}


static int
mind(void *node, unsigned size, void *back, int (*call) (void *, void *))
{
    int delete;
    integral_q news;

    news = *(integral_q *) node;
    if (news & 2) {
	node = (char *) node + size;
	delete = call(back, node);
	if (delete) {
	} else {
	    news >>= 2;
	    while (news) {
		node = (char *) node + size;
		delete = call(back, node);
		if (delete) {
		    break;
		} else {
		    node = (char *) node + size;
		    if (news & 2) {
			delete = call(back, node);
			if (delete) {
			    break;
			}
		    }
		}

		news >>= 2;
	    }
	}
    } else {
	node = (char *) node + size;
	news >>= 2;
	do {
	    node = (char *) node + size;
	    delete = call(back, node);
	    if (delete) {
		break;
	    } else {
		node = (char *) node + size;
		if (news & 2) {
		    delete = call(back, node);
		    if (delete) {
			break;
		    }
		}
	    }

	    news >>= 2;
	} while (news);
    }

    return delete;
}


static int
pick(void *node, unsigned rate, unsigned size, void *back,
     int (*call) (void *, void *))
{
    int delete;
    integral_q news;
    void **text;

    news = *(integral_q *) node;

    text = (void **) ((char *) node + size * integral_q_bits);

    rate--;

    if (rate) {
	do {
	    if (news & 2) {
		delete = pick(*text, rate, size, back, call);
		if (delete) {
		    break;
		}
	    }

	    text++;

	    {
		delete = pick(*text, rate, size, back, call);
		if (delete) {
		    break;
		}
	    }

	    text++;

	    news >>= 2;
	} while (news);
    } else {
	do {
	    if (news & 2) {
		delete = mind(*text, size, back, call);
		if (delete) {
		    break;
		}
	    }

	    text++;

	    {
		delete = mind(*text, size, back, call);
		if (delete) {
		    break;
		}
	    }

	    text++;

	    news >>= 2;
	} while (news);
    }

    return delete;
}


static int
slip(void *node, unsigned rate, unsigned size, void *back,
     int (*call) (void *, void *))
{
    int delete;
    unsigned note;
    void **text;

    note = *(integral_q *) node + 1;

    text = (void **) ((char *) node + size * integral_q_bits_and_half);

    rate--;

    if (rate) {
	do {
	    delete = pick(*text, rate, size, back, call);
	    if (delete) {
		break;
	    }

	    note--;

	    text++;
	} while (note);
    } else {
	do {
	    delete = mind(*text, size, back, call);
	    if (delete) {
		break;
	    }

	    note--;

	    text++;
	} while (note);
    }

    return delete;
}


int
x1f4_lime_bpfset(void *bpfset, void *back, int (*call) (void *, void *))
{
    int delete;
    unsigned rate;

    rate = bpfset(bpfset)->link_a.fpnews.link_v.rate;
    if (rate) {
	delete = slip
	    (bpfset(bpfset)->link_a.fpnews.node, rate,
	     bpfset(bpfset)->link_a.fpnews.link_v.size, back, call);
    } else {
	delete = line
	    (bpfset(bpfset)->link_a.fpnews.node,
	     bpfset(bpfset)->link_a.fpnews.link_v.size, back, call);
    }

    return delete;
}
