/*
 * bpfx.f.c
 * Copyright (C) 2008-2010, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <bpfx-defs.h>
#include <bpfx-names.h>
#include <bpfx-types.h>
#include <mask.0.h>
#include <trans.h>

#define bpfset(bpfset)			((struct bpfset_type *) (bpfset))

int
x1f4_fast_bpfset(void *bpfset, unsigned flags,
		 struct screen_type *screen_data)
{
    int (*link) (void *, void **, unsigned), status;
    void *text;

    if (flags & TRANS_MASK) {
	struct trans_type *trans_data;

	trans_data = screen_data->trans_data;

	link = trans_data->link;
	text = trans_data->data;
    } else {
	link = _libx1f4l2_link_data;
	text = (void *) 0;
    }

    status = link
	(text, &text, integral_q_bits_and_half * SIZEOF_integral_q << 1);
    if (status) {
	status = LINK_ERROR;
    } else {
	bpfset(bpfset)->link_a.fpnews.node = text;

	bpfset(bpfset)->link_a.fpnews.link_v.rate = 0;

	bpfset(bpfset)->link_a.fpnews.link_v.size = SIZEOF_integral_q;

	*((integral_q *) text) = 0;

	if (flags & TRANS_MASK) {
	    struct trans_type *trans_data;

	    trans_data = screen_data->trans_data;

	    bpfset(bpfset)->link_m.data = trans_data->data;
	    bpfset(bpfset)->link_m.free = trans_data->free;
	    bpfset(bpfset)->link_m.link = trans_data->link;
	    bpfset(bpfset)->link_m.mode = trans_data->mode;
	} else {
	    bpfset(bpfset)->link_m.free = _libx1f4l2_free_data;
	    bpfset(bpfset)->link_m.link = _libx1f4l2_link_data;
	    bpfset(bpfset)->link_m.mode = _libx1f4l2_mode_data;
	}
    }

    return status;
}
