/*
 * bppx.j.c
 * Copyright (C) 2009-2010, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <string.h>

#include <bpfx-defs.h>
#include <bpfx-lines.h>
#include <bpfx-names.h>
#include <bpfx-types.h>
#include <bppx-types.h>
#include <xfs.h>
#include <xls.h>

#define __SIDE_FILL__			1

#define __HALF_FILL__			1

#define __HEAD_LEFT__			1

#if __SIDE_FILL__
#else
# undef __HALF_FILL__
# define __HALF_FILL__			0
#endif				/* __SIDE_FILL__ */

#undef Q
#define Q(e)				((integral_q) (e))

#define bpfset(bpfset)			((struct bpfset_type *) (bpfset))

static int east(void *, unsigned, unsigned *);
static int edge(void *, unsigned, unsigned *);

static int
east(void *node, unsigned call, unsigned *fare)
{
    int status;
    integral_q news, slip;

    news = *KD(node);

    slip = news & ~Lx55555555;

    if (slip) {
	status = 0;

	if (call ^ (integral_q_bits - 1)) {
	    call++;

	    if (call & 1) {
		*KD(node) = news & ~(Q(1) << call);

		*fare = call - 2;

		KD(node)[call - 1] = KD(node)[call];
	    } else {
		if (news & (Q(1) << (call - 1))) {
		    *KD(node) = news & ~(Q(1) << (call - 1));
		    call--;
		    PD(node)[call] = PD(node)[call - 1];

		    *fare = call;
		} else {
		    if (news & (Q(1) << (call + 1))) {
			*fare = call - 3;

			*KD(node) = news & ~(Q(1) << (call + 1));
			if (call ^ 2) {
			    KD(node)[call - 2] = KD(node)[call];
			    KD(node)[call] = KD(node)[call + 1];
			    PD(node)[call - 1] = PD(node)[call];
			} else {
			    KD(node)[call] = KD(node)[call + 1];
			    PD(node)[call - 1] = PD(node)[call];
			}
		    } else {
			integral_q back;
			unsigned deck, ever;

			back = slip & (~Q(0) << call);
			if (back) {
			    l2_q_xfs(ever, back);
#if __HEAD_LEFT__
			    slip &= ((Q(1) << ever) - (Q(1) << call))
				>> (ever - call);
			    if (slip) {
				l2_q_xls(deck, slip);
				deck ^= integral_q_last;
				if (1) {
				    *KD(node) = news & ~(Q(1) << deck);
				    ever = (call - deck - 3) >> 1;
				    call--;
				    *fare = call;
				    for (; ever; ever--) {
					PD(node)[call] = PD(node)[call - 2];
					call--;
					KD(node)[call] = KD(node)[call - 2];
					call--;
				    }
				    {
					PD(node)[call] = PD(node)[call - 2];
					call--;
					KD(node)[call] = KD(node)[call - 1];
					call--;
					PD(node)[call] = PD(node)[call - 1];
				    }
				}
			    } else {
				if (call ^ 2) {
				    *KD(node) = news & ~(Q(1) << ever);
				    deck = (ever - call) >> 1;
				    call -= 2;
				    *fare = call - 1;
				    for (; deck; deck--) {
					KD(node)[call] = KD(node)[call + 2];
					call++;
					PD(node)[call] = PD(node)[call + 2];
					call++;
				    }
				    {
					KD(node)[call] = KD(node)[call + 2];
					call++;
					PD(node)[call] = PD(node)[call + 1];
					call++;
					KD(node)[call] = KD(node)[call + 1];
				    }
				} else {
				    *KD(node) = news & ~(Q(1) << ever);
				    deck = (ever - call) >> 1;
				    call--;
				    *fare = call - 2;
				    for (; deck; deck--) {
					PD(node)[call] = PD(node)[call + 2];
					call++;
					KD(node)[call] = KD(node)[call + 2];
					call++;
				    }
				    {
					PD(node)[call] = PD(node)[call + 1];
					call++;
					KD(node)[call] = KD(node)[call + 1];
				    }
				}
			    }
#else
			    l2_q_xls(deck, slip & ((Q(1) << call) - 1));
			    if (deck ^ integral_q_bits) {
				deck ^= integral_q_last;
				if (call - deck < ever - call) {
				    *KD(node) = news & ~(Q(1) << deck);
				    ever = (call - deck - 3) >> 1;
				    call--;
				    *fare = call;
				    for (; ever; ever--) {
					PD(node)[call] = PD(node)[call - 2];
					call--;
					KD(node)[call] = KD(node)[call - 2];
					call--;
				    }
				    {
					PD(node)[call] = PD(node)[call - 2];
					call--;
					KD(node)[call] = KD(node)[call - 1];
					call--;
					PD(node)[call] = PD(node)[call - 1];
				    }
				} else {
				    *KD(node) = news & ~(Q(1) << ever);
				    deck = (ever - call) >> 1;
				    call -= 2;
				    *fare = call - 1;
				    for (; deck; deck--) {
					KD(node)[call] = KD(node)[call + 2];
					call++;
					PD(node)[call] = PD(node)[call + 2];
					call++;
				    }
				    {
					KD(node)[call] = KD(node)[call + 2];
					call++;
					PD(node)[call] = PD(node)[call + 1];
					call++;
					KD(node)[call] = KD(node)[call + 1];
				    }
				}
			    } else {
				if (call ^ 2) {
				    *KD(node) = news & ~(Q(1) << ever);
				    deck = (ever - call) >> 1;
				    call -= 2;
				    *fare = call - 1;
				    for (; deck; deck--) {
					KD(node)[call] = KD(node)[call + 2];
					call++;
					PD(node)[call] = PD(node)[call + 2];
					call++;
				    }
				    {
					KD(node)[call] = KD(node)[call + 2];
					call++;
					PD(node)[call] = PD(node)[call + 1];
					call++;
					KD(node)[call] = KD(node)[call + 1];
				    }
				} else {
				    *KD(node) = news & ~(Q(1) << ever);
				    deck = (ever - call) >> 1;
				    call--;
				    *fare = call - 2;
				    for (; deck; deck--) {
					PD(node)[call] = PD(node)[call + 2];
					call++;
					KD(node)[call] = KD(node)[call + 2];
					call++;
				    }
				    {
					PD(node)[call] = PD(node)[call + 1];
					call++;
					KD(node)[call] = KD(node)[call + 1];
				    }
				}
			    }
#endif				/* __HEAD_LEFT__ */
			} else {
			    l2_q_xls(deck, slip & ((Q(1) << call) - 1));
			    if (1) {
				deck ^= integral_q_last;
				if (1) {
				    *KD(node) = news & ~(Q(1) << deck);
				    ever = (call - deck - 3) >> 1;
				    call--;
				    *fare = call;
				    for (; ever; ever--) {
					PD(node)[call] = PD(node)[call - 2];
					call--;
					KD(node)[call] = KD(node)[call - 2];
					call--;
				    }
				    {
					PD(node)[call] = PD(node)[call - 2];
					call--;
					KD(node)[call] = KD(node)[call - 1];
					call--;
					PD(node)[call] = PD(node)[call - 1];
				    }
				}
			    }
			}
		    }
		}
	    }
	} else {
	    unsigned deck, ever;

	    l2_q_xls(deck, slip);
	    if (1) {
		deck ^= integral_q_last;
		if (deck ^ (integral_q_bits - 1)) {
		    *KD(node) = news & ~(Q(1) << deck);

		    PD(node)[integral_q_bits - 1] =
			PD(node)[integral_q_bits - 3];

		    *fare = call;

		    call--;

		    ever = (integral_q_bits - 3 - deck) >> 1;
		    for (; ever; ever--) {
			KD(node)[call] = KD(node)[call - 2];
			call--;
			PD(node)[call] = PD(node)[call - 2];
			call--;
		    }
		    {
			KD(node)[call] = KD(node)[call - 1];
			call--;
			PD(node)[call] = PD(node)[call - 1];
		    }
		} else {
		    *KD(node) = news & ~(Q(1) << deck);

		    *fare = call;

		    PD(node)[integral_q_bits - 1] =
			PD(node)[integral_q_bits - 2];
		}
	    }
	}
    } else {
	*fare = call - 2;

	if (call ^ (integral_q_bits - 1)) {
	    *KD(node) = news & ~(Q(1) << (call - 1));
	} else {
	    *KD(node) = news & ~(Q(1) << (integral_q_bits - 2));

	    PD(node)[integral_q_bits - 1] = PD(node)[integral_q_bits - 3];
	}

	if (1) {
	    status = LAST_ERROR + call;
	}
    }

    return status;
}


static int
edge(void *node, unsigned call, unsigned *fare)
{
    int status;
    integral_q news, slip;

    news = *KD(node);

    slip = news & ~Lx55555555;

    if (slip) {
	status = 0;

	if (1) {
	    call++;

	    if (call & 1) {
		*KD(node) = news & ~(Q(1) << call);

		*fare = call;
	    } else {
		if (news & (Q(1) << (call + 1))) {
		    *fare = call - 1;

		    *KD(node) = news & ~(Q(1) << (call + 1));
		    if (2) {
			KD(node)[call] = KD(node)[call + 1];
			PD(node)[call - 1] = PD(node)[call];
		    }
		} else {
		    if (news & (Q(1) << (call - 1))) {
			*KD(node) = news & ~(Q(1) << (call - 1));
			KD(node)[call] = KD(node)[call - 1];
			call--;
			PD(node)[call] = PD(node)[call - 1];

			*fare = call + 2;
		    } else {
			integral_q back;
			unsigned deck, ever;

			back = slip & (~Q(0) << call);
			if (back) {
			    l2_q_xfs(ever, back);
#if __HEAD_LEFT__
			    slip &= ((Q(1) << ever) - (Q(1) << call))
				>> (ever - call);
			    if (slip) {
				l2_q_xls(deck, slip);
				deck ^= integral_q_last;
				if (1) {
				    *KD(node) = news & ~(Q(1) << deck);
				    ever = (call - deck) >> 1;
				    *fare = call + 1;
				    for (; ever; ever--) {
					KD(node)[call] = KD(node)[call - 2];
					call--;
					PD(node)[call] = PD(node)[call - 2];
					call--;
				    }
				    {
					KD(node)[call] = KD(node)[call - 1];
					call--;
					PD(node)[call] = PD(node)[call - 1];
				    }
				}
			    } else {
				if (2) {
				    *KD(node) = news & ~(Q(1) << ever);
				    deck = (ever - call) >> 1;
				    call--;
				    *fare = call;
				    for (; deck; deck--) {
					PD(node)[call] = PD(node)[call + 2];
					call++;
					KD(node)[call] = KD(node)[call + 2];
					call++;
				    }
				    {
					PD(node)[call] = PD(node)[call + 1];
					call++;
					KD(node)[call] = KD(node)[call + 1];
				    }
				}
			    }
#else
			    l2_q_xls(deck, slip & ((Q(1) << call) - 1));
			    if (deck ^ integral_q_bits) {
				deck ^= integral_q_last;
				if (call - deck < ever - call) {
				    *KD(node) = news & ~(Q(1) << deck);
				    ever = (call - deck) >> 1;
				    *fare = call + 1;
				    for (; ever; ever--) {
					KD(node)[call] = KD(node)[call - 2];
					call--;
					PD(node)[call] = PD(node)[call - 2];
					call--;
				    }
				    {
					KD(node)[call] = KD(node)[call - 1];
					call--;
					PD(node)[call] = PD(node)[call - 1];
				    }
				} else {
				    *KD(node) = news & ~(Q(1) << ever);
				    deck = (ever - call) >> 1;
				    call--;
				    *fare = call;
				    for (; deck; deck--) {
					PD(node)[call] = PD(node)[call + 2];
					call++;
					KD(node)[call] = KD(node)[call + 2];
					call++;
				    }
				    {
					PD(node)[call] = PD(node)[call + 1];
					call++;
					KD(node)[call] = KD(node)[call + 1];
				    }
				}
			    } else {
				if (2) {
				    *KD(node) = news & ~(Q(1) << ever);
				    deck = (ever - call) >> 1;
				    call--;
				    *fare = call;
				    for (; deck; deck--) {
					PD(node)[call] = PD(node)[call + 2];
					call++;
					KD(node)[call] = KD(node)[call + 2];
					call++;
				    }
				    {
					PD(node)[call] = PD(node)[call + 1];
					call++;
					KD(node)[call] = KD(node)[call + 1];
				    }
				}
			    }
#endif				/* __HEAD_LEFT__ */
			} else {
			    l2_q_xls(deck, slip & ((Q(1) << call) - 1));
			    if (1) {
				deck ^= integral_q_last;
				if (1) {
				    *KD(node) = news & ~(Q(1) << deck);
				    ever = (call - deck) >> 1;
				    *fare = call + 1;
				    for (; ever; ever--) {
					KD(node)[call] = KD(node)[call - 2];
					call--;
					PD(node)[call] = PD(node)[call - 2];
					call--;
				    }
				    {
					KD(node)[call] = KD(node)[call - 1];
					call--;
					PD(node)[call] = PD(node)[call - 1];
				    }
				}
			    }
			}
		    }
		}
	    }
	}
    } else {
	*fare = call + 2;

	if (1) {
	    if (call ^ 1) {
		*KD(node) = news & ~(Q(1) << (call - 1));

		KD(node)[call + 1] = KD(node)[call - 1];
	    } else {
		*KD(node) = news & ~(Q(1) << (call + 1));
	    }
	}

	if (1) {
	    status = LAST_ERROR + call;
	}
    }

    return status;
}


int
_libx1f4l2_tale_bppset(void *bpfset, void *date, integral_q *data,
		       integral_q **path, integral_q **lead, integral_q pipe,
		       void **lift, unsigned *mile, integral_q **lock,
		       integral_q **lane, integral_q **turn, integral_q *text,
		       void *near, integral_q mode)
{
    int excess, status;
    integral_q *club, *file, *tier;
    unsigned call, deck, ever;
#if __HALF_FILL__
    unsigned fill = 0;
#endif				/* __HALF_FILL__ */
    void *node;

    node = *lift;

    call = *mile;

    club = *lead;

    file = *lock;

    tier = *lane;

    if (tier) {
	if (call ^ integral_q_last) {
	    ever = 2;

	    if (pipe & ((Q(1) << call) - 1) << 1) {
		deck = 2;
	    } else {
		if (club) {
#if __HALF_FILL__
		    deck = 2;
#else
		    deck = 3;
#endif				/* __HALF_FILL__ */
#if __HALF_FILL__
		    if (mode < file[half_integral_q_bits]) {
			fill = 1;
		    } else {
		    }
#endif				/* __HALF_FILL__ */
		} else {
		    deck = 4;
		}
	    }
	} else {
	    deck = 2;
#if __HALF_FILL__
	    ever = 2;
#else
	    ever = 3;
#endif				/* __HALF_FILL__ */
#if __HALF_FILL__
	    if (mode < file[half_integral_q_bits]) {
	    } else {
		fill = 2;
	    }
#endif				/* __HALF_FILL__ */
	}
    } else {
	deck = 2;
	ever = 4;
    }

    if (deck < ever) {
	integral_q *fast, *side;

	fast = club;

	side = fast;

	if (1) {
	    *side = ~Q(0);

	    side++;

	    fast += 2;

	    deck = half_integral_q_bits - 1;
	    for (; deck; deck--) {
		*side = *fast;
		fast += 2;
		side++;
	    }

	    if (0) {
	    } else {
		*side = **path;
	    }

	    side++;

	    fast = file;

	    fast += 2;

	    deck = half_integral_q_bits - 1;
	    for (; deck; deck--) {
		*side = *fast;
		fast += 2;
		side++;
	    }

	    fast = side + 1;

	    deck = half_integral_q_bits;
	    for (; deck; deck--) {
		*side = *fast;
		fast += 2;
		side++;
	    }

	    fast = file + integral_q_bits + 1;

	    deck = half_integral_q_bits;
	    for (; deck; deck--) {
		*side = *fast;
		fast += 2;
		side++;
	    }
	}

	if (1) {
	    status = east(node, call, mile);

	    excess = bpfset(bpfset)->link_m.free
		(bpfset(bpfset)->link_m.data, file);
	    if (excess) {
		status = FREE_ERROR;

		/*
		 * TODO
		 *
		 * restore state
		 */
	    } else {
		call = *mile;

		if (call & 1) {
		    if (*KD(node) & Q(1) << call) {
			*lead = PD(node)[call - 1];
			*path = KD(node) + call;
		    } else {
			if (call ^ 1) {
			    *lead = PD(node)[call - 2];
			    *path = KD(node) + call - 1;
			} else {
			    *path = data;
			    if (date) {
				*lead = PD(date)[integral_q_last];
			    } else {
				*lead = NULL;
			    }
			}
		    }
		} else {
		    if (call) {
			*lead = PD(node)[call - 1];
			*path = KD(node) + call;
		    } else {
			*path = data;
			if (date) {
			    *lead = PD(date)[integral_q_last];
			} else {
			    *lead = NULL;
			}
		    }
		}

		*lock = club;

		if (call ^ integral_q_last) {
		    if (status ^ (LAST_ERROR + call + 2)) {
			*turn = KD(node) + call + 1;
		    } else {
			if (call ^ (integral_q_bits - 3)) {
			    *turn = KD(node) + call + 3;
			} else {
			    *turn = text;
			}
		    }
		} else {
		    *turn = text;
		}
	    }
	}
    } else {
#if __SIDE_FILL__
	int half;
#endif				/* __SIDE_FILL__ */
	integral_q *fast, *side;

#if __SIDE_FILL__
	if (deck ^ 2) {
#elif 0
	}
#else
	{
#endif				/* __SIDE_FILL__ */
	    fast = tier;

	    side = fast;

	    *side = ~Q(0);

	    fast += integral_q_bits - 2;

	    side = fast + 1;

	    deck = half_integral_q_bits - 1;
	    for (; deck; deck--) {
		*side = *fast;
		fast -= 2;
		side--;
	    }

	    if (0) {
	    } else {
		*side = **turn;
	    }

	    side--;

	    fast = file;

	    fast += integral_q_bits - 2;

	    deck = half_integral_q_bits - 1;
	    for (; deck; deck--) {
		*side = *fast;
		fast -= 2;
		side--;
	    }

	    side += (integral_q_bits << 1) - 2;

	    fast = side - 1;

	    deck = half_integral_q_bits - 1;
	    for (; deck; deck--) {
		*side = *fast;
		fast -= 2;
		side--;
	    }

	    fast = file + (integral_q_bits << 1) - 1;

	    deck = half_integral_q_bits;
	    for (; deck; deck--) {
		*side = *fast;
		fast -= 2;
		side--;
	    }

#if __SIDE_FILL__
	    half = 0;
	} else {
	    integral_q away;

	    fast = club;

	    *fast = Lxdddddddd;

	    fast += 4;

	    side = fast - 1;

	    {
		*side = *fast;
		fast += 2;
		side++;
	    }

	    deck = integral_q_bits / 6 - 1;
	    for (; deck; deck--) {
		*side = *fast;
		fast += 2;
		side += 2;
		*side = *fast;
		fast += 2;
		side++;
		*side = *fast;
		fast += 2;
		side++;
	    }

	    if (half_integral_q_bits % 3 ^ 2) {
		*side = *fast;
		side += 2;
		*side = **path;

		fast = file;

		fast += 2;
		side++;
		*side = *fast;
		side++;
	    } else {
		*side = *fast;
		fast += 2;
		side += 2;
		*side = *fast;
		side++;
		*side = **path;

		fast = file;

		fast += 2;
		side++;
	    }

	    deck = integral_q_bits / 4 - integral_q_bits / 6 - 1;
	    for (; deck; deck--) {
		fast += 2;
		*side = *fast;
		side += 2;
		fast += 2;
		*side = *fast;
		side++;
		fast += 2;
		*side = *fast;
		side++;
	    }

	    fast += 2;

	    away = *fast;

	    if (fill) {
		half = fill - 1;
	    } else {
		half = mode < away;
	    }

	    fast = club + integral_q_bits;

	    side = fast;

	    fast--;

	    deck = integral_q_bits / 6;
	    for (; deck; deck--) {
		side++;
		fast += 2;
		*side = *fast;
		side++;
		fast += 2;
		*side = *fast;
		side++;
		fast += 2;
		*side = *fast;
		side++;
	    }

	    if (half_integral_q_bits % 3 ^ 2) {
		side++;
		fast += 2;
		*side = *fast;

		fast = file + integral_q_bits + 1;

		side++;
		*side = *fast;
		side++;
		fast += 2;
		*side = *fast;
		side++;
	    } else {
		side++;
		fast += 2;
		*side = *fast;
		side++;
		fast += 2;
		*side = *fast;

		fast = file + integral_q_bits + 1;

		side++;
		*side = *fast;
		side++;
	    }

	    deck = integral_q_bits / 4 - integral_q_bits / 6 - 1;
	    for (; deck; deck--) {
		side++;
		fast += 2;
		*side = *fast;
		side++;
		fast += 2;
		*side = *fast;
		side++;
		fast += 2;
		*side = *fast;
		side++;
	    }

	    fast = tier + (integral_q_bits << 1);

	    side = fast;

	    fast++;

	    deck = integral_q_bits / 6;
	    for (; deck; deck--) {
		side--;
		fast -= 2;
		*side = *fast;
		side--;
		fast -= 2;
		*side = *fast;
		side--;
		fast -= 2;
		*side = *fast;
		side--;
	    }

	    if (half_integral_q_bits % 3 ^ 2) {
		side--;
		fast -= 2;
		*side = *fast;

		fast = file + (integral_q_bits << 1) - 1;

		side--;
		*side = *fast;
		side--;
		fast -= 2;
		*side = *fast;
		side--;
	    } else {
		side--;
		fast -= 2;
		*side = *fast;
		side--;
		fast -= 2;
		*side = *fast;

		fast = file + (integral_q_bits << 1) - 1;

		side--;
		*side = *fast;
		side--;
	    }

	    deck = integral_q_bits / 4 - integral_q_bits / 6 - 1;
	    for (; deck; deck--) {
		side--;
		fast -= 2;
		*side = *fast;
		side--;
		fast -= 2;
		*side = *fast;
		side--;
		fast -= 2;
		*side = *fast;
		side--;
	    }

	    fast = tier + integral_q_bits;

	    side = fast;

	    deck = integral_q_bits / 6;
	    for (; deck; deck--) {
		side--;
		fast -= 2;
		*side = *fast;
		side--;
		fast -= 2;
		*side = *fast;
		side -= 2;
		fast -= 2;
		*side = *fast;
	    }

	    if (half_integral_q_bits % 3 ^ 2) {
		fast = file + integral_q_bits - 2;

		side--;
		*side = **turn;
		side--;
		*side = *fast;
		side -= 2;
		fast -= 2;
		*side = *fast;
	    } else {
		side--;
		fast -= 2;
		*side = *fast;
		side--;
		*side = **turn;

		fast = file + integral_q_bits - 2;

		side -= 2;
		*side = *fast;
	    }

	    if (half) {
		**turn = away;
	    } else {
		**path = away;
	    }

	    deck = integral_q_bits / 4 - integral_q_bits / 6 - 2;
	    for (; deck; deck--) {
		side--;
		fast -= 2;
		*side = *fast;
		side--;
		fast -= 2;
		*side = *fast;
		side -= 2;
		fast -= 2;
		*side = *fast;
	    }

	    {
		side--;
		fast -= 2;
		*side = *fast;
		side--;
		fast -= 2;
		*side = *fast;
		side -= 2;
	    }

	    *side = Lxdddddddd;
#endif				/* __SIDE_FILL__ */
	}

#if __SIDE_FILL__
	if (half) {
#elif 0
	}
#else
	{
#endif				/* __SIDE_FILL__ */
#if __SIDE_FILL__
	    status = east(node, call, mile);

	    excess = bpfset(bpfset)->link_m.free
		(bpfset(bpfset)->link_m.data, file);
	    if (excess) {
		status = FREE_ERROR;

		/*
		 * TODO
		 *
		 * restore state
		 */
	    } else {
#endif				/* __SIDE_FILL__ */
#if __HALF_FILL__
		if (fill) {
		    *lift = near;

		    *lock = tier;

		    *path = *turn;

		    if (*KD(near) & Q(2)) {
			*mile = 0;

			*lane = PD(near)[1];

			*turn = KD(near) + 1;
		    } else {
			*mile = 1;

			*lane = PD(near)[*KD(near) & Q(8) ? 2 : 3];

			*turn = KD(near) + 2;
		    }
		} else {
#endif				/* __HALF_FILL__ */
#if __SIDE_FILL__
		    call = *mile;

		    if (call & 1) {
			if (*KD(node) & Q(1) << call) {
			    *lead = PD(node)[call - 1];
			    *path = KD(node) + call;
			} else {
			    if (call ^ 1) {
				*lead = PD(node)[call - 2];
				*path = KD(node) + call - 1;
			    } else {
				*path = data;
				if (date) {
				    *lead = PD(date)[integral_q_last];
				} else {
				    *lead = NULL;
				}
			    }
			}
		    } else {
			if (call) {
			    *lead = PD(node)[call - 1];
			    *path = KD(node) + call;
			} else {
			    *path = data;
			    if (date) {
				*lead = PD(date)[integral_q_last];
			    } else {
				*lead = NULL;
			    }
			}
		    }

		    *lock = club;

		    if (call ^ integral_q_last) {
			if (status ^ (LAST_ERROR + call + 2)) {
			    *turn = KD(node) + call + 1;
			} else {
			    if (call ^ (integral_q_bits - 3)) {
				*turn = KD(node) + call + 3;
			    } else {
				*turn = text;
			    }
			}
		    } else {
			*turn = text;
		    }
#endif				/* __SIDE_FILL__ */
#if __HALF_FILL__
		}
#endif				/* __HALF_FILL__ */
#if __SIDE_FILL__
	    }
	} else {
#endif				/* __SIDE_FILL__ */
	    status = edge(node, call, mile);

	    excess = bpfset(bpfset)->link_m.free
		(bpfset(bpfset)->link_m.data, file);
	    if (excess) {
		status = FREE_ERROR;

		/*
		 * TODO
		 *
		 * restore state
		 */
	    } else {
#if __HALF_FILL__
		if (fill) {
		    *lift = date;

		    *lock = club;

		    *turn = *path;

		    *mile = integral_q_last;

		    if (*KD(date) & Q(1) << integral_q_last) {
			*lead = PD(date)[integral_q_bits - 2];

			*path = KD(date) + integral_q_last;
		    } else {
			*lead = PD(date)[integral_q_bits - 3];

			*path = KD(date) + integral_q_bits - 2;
		    }
		} else {
#endif				/* __HALF_FILL__ */
		    call = *mile;

		    if (call & 1) {
			if (*KD(node) & Q(1) << call) {
			    *path = KD(node) + call;
			} else {
			    if (call ^ 1) {
				if (status ^ (LAST_ERROR + call - 2)) {
				    *path = KD(node) + call - 1;
				} else {
				    if (call ^ 3) {
					*path = KD(node) + call - 1;
				    } else {
					*path = data;
				    }
				}
			    } else {
				*path = data;
			    }
			}
		    } else {
			if (call) {
			    *path = KD(node) + call;
			} else {
			    *path = data;
			}
		    }

		    *lock = tier;

		    if (call ^ integral_q_last) {
			*turn = KD(node) + call + 1;
			*lane = PD(node)[call + 1];
		    } else {
			*turn = text;
			if (near) {
			    if (*KD(near) & 2) {
				*lane = PD(near)[0];
			    } else {
				*lane = PD(near)[1];
			    }
			} else {
			    *lane = NULL;
			}
		    }
#if __HALF_FILL__
		}
#endif				/* __HALF_FILL__ */
	    }
	}
    }

    return status;
}
