/*
 * dflist.h
 * Copyright (C) 2008-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __LIBx1f4_DFLIST_H__
#define __LIBx1f4_DFLIST_H__

#include <dfxx-names.h>
#include <errors.h>

/*
 * attribute bit definitions
 */
#define X1f4_DFLIST_TRANS_MASK		(1u << 0x00)

/*
 * error definitions
 */
#define X1f4_DFLIST_CALL_ERROR		LIBx1f4L2_CALL_ERROR
#define X1f4_DFLIST_DECK_ERROR		LIBx1f4L2_DECK_ERROR
#define X1f4_DFLIST_FREE_ERROR		LIBx1f4L2_FREE_ERROR
#define X1f4_DFLIST_LINK_ERROR		LIBx1f4L2_LINK_ERROR
#define X1f4_DFLIST_MODE_ERROR		LIBx1f4L2_MODE_ERROR
#define X1f4_DFLIST_SIZE_ERROR		LIBx1f4L2_SIZE_ERROR

typedef struct x1f4_dflist_type {
    void *trans;
} x1f4_dflist_type;

#if !defined __x1f4_trans_type_defined__
typedef struct x1f4_trans_type {
    int (*free) (void *, void *), (*link) (void *, void **, unsigned),
	(*mode) (void *, void **, unsigned);
    void *context;
} x1f4_trans_type;
#endif				/* !__x1f4_trans_type_defined__ */

#if !defined __x1f4_trans_type_defined__
# define __x1f4_trans_type_defined__
#endif				/* !__x1f4_trans_type_defined__ */

extern int x1f4_ever_dflist(void *, void **);
extern int x1f4_fast_dflist(void *, unsigned, struct x1f4_dflist_type *);
extern int x1f4_fini_dflist(void **);
extern int x1f4_flat_dflist(void *);
extern int x1f4_flip_dflist(void *);
extern int x1f4_head_dflist(void *, unsigned, void **);
extern int x1f4_high_dflist(void *);
extern int x1f4_init_dflist(void **, unsigned, struct x1f4_dflist_type *);
extern int x1f4_last_dflist(void **, void *);
extern int x1f4_lead_dflist(void *, void **);
extern int x1f4_lime_dflist(void *, void *, int (*) (void *, void *));
extern int x1f4_mind_dflist(void *);
extern int x1f4_next_dflist(void **, void *);
extern int x1f4_size_dflist(void *, unsigned *);
extern int x1f4_slip_dflist(void *);
extern int x1f4_tail_dflist(void *, unsigned, void **);

#endif				/* __LIBx1f4_DFLIST_H__ */
