/*
 * l2list.h
 * Copyright (C) 2008-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __LIBx1f4_L2LIST_H__
#define __LIBx1f4_L2LIST_H__

#include <errors.h>
#include <l2xx-names.h>

/*
 * attribute bit definitions
 */
#define X1f4_L2LIST_TRANS_MASK		(1u << 0x00)

/*
 * error definitions
 */
#define X1f4_L2LIST_CALL_ERROR		LIBx1f4L2_CALL_ERROR
#define X1f4_L2LIST_DECK_ERROR		LIBx1f4L2_DECK_ERROR
#define X1f4_L2LIST_FREE_ERROR		LIBx1f4L2_FREE_ERROR
#define X1f4_L2LIST_LINK_ERROR		LIBx1f4L2_LINK_ERROR
#define X1f4_L2LIST_MODE_ERROR		LIBx1f4L2_MODE_ERROR
#define X1f4_L2LIST_SIZE_ERROR		LIBx1f4L2_SIZE_ERROR

typedef struct x1f4_l2list_type {
    void *trans;
} x1f4_l2list_type;

#if !defined __x1f4_trans_type_defined__
typedef struct x1f4_trans_type {
    int (*free) (void *, void *), (*link) (void *, void **, unsigned),
	(*mode) (void *, void **, unsigned);
    void *context;
} x1f4_trans_type;
#endif				/* !__x1f4_trans_type_defined__ */

#if !defined __x1f4_trans_type_defined__
# define __x1f4_trans_type_defined__
#endif				/* !__x1f4_trans_type_defined__ */

extern int x1f4_ever_l2list(void *, void **);
extern int x1f4_fast_l2list(void *, unsigned, struct x1f4_l2list_type *);
extern int x1f4_fini_l2list(void **);
extern int x1f4_flat_l2list(void *);
extern int x1f4_flip_l2list(void *);
extern int x1f4_head_l2list(void *, unsigned, void **);
extern int x1f4_high_l2list(void *);
extern int x1f4_init_l2list(void **, unsigned, struct x1f4_l2list_type *);
extern int x1f4_last_l2list(void **, void *);
extern int x1f4_lead_l2list(void *, void **);
extern int x1f4_lime_l2list(void *, void *, int (*) (void *, void *));
extern int x1f4_mind_l2list(void *);
extern int x1f4_next_l2list(void **, void *);
extern int x1f4_reel_l2list(void *);
extern int x1f4_roll_l2list(void *);
extern int x1f4_size_l2list(void *, unsigned *);
extern int x1f4_slip_l2list(void *);
extern int x1f4_tail_l2list(void *, unsigned, void **);

#endif				/* __LIBx1f4_L2LIST_H__ */
