/*
 * l4list.h
 * Copyright (C) 2008-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __LIBx1f4_L4LIST_H__
#define __LIBx1f4_L4LIST_H__

#include <errors.h>

/*
 * attribute bit definitions
 */
#define X1f4_L4LIST_TRANS_MASK		(1u << 0x00)

/*
 * error definitions
 */
#define X1f4_L4LIST_DECK_ERROR		LIBx1f4L2_DECK_ERROR
#define X1f4_L4LIST_FREE_ERROR		LIBx1f4L2_FREE_ERROR
#define X1f4_L4LIST_LINK_ERROR		LIBx1f4L2_LINK_ERROR
#define X1f4_L4LIST_MODE_ERROR		LIBx1f4L2_MODE_ERROR
#define X1f4_L4LIST_SIZE_ERROR		LIBx1f4L2_SIZE_ERROR

typedef struct x1f4_l4list_type {
    void *trans;
} x1f4_l4list_type;

#if !defined __x1f4_trans_type_defined__
typedef struct x1f4_trans_type {
    int (*free) (void *, void *), (*link) (void *, void **, unsigned),
	(*mode) (void *, void **, unsigned);
    void *context;
} x1f4_trans_type;
#endif				/* !__x1f4_trans_type_defined__ */

#if !defined __x1f4_trans_type_defined__
# define __x1f4_trans_type_defined__
#endif				/* !__x1f4_trans_type_defined__ */

extern int x1f4_copy_l4list
    (void **, void *, int (*) (unsigned *, const void *));
extern int x1f4_deck_l4list(void *, int);
extern int x1f4_deux_l4list(void *, unsigned, void *, unsigned);
extern int x1f4_fini_l4list(void **);
extern int x1f4_head_l4list(void *, unsigned, void **);
extern int x1f4_init_l4list(void **, unsigned, struct x1f4_l4list_type *);
extern int x1f4_lime_l4list(void *, void *, int (*) (void *, void *));
extern int x1f4_line_l4list
    (void *, void *, int (*) (void *, unsigned *, const void *),
     int (*) (void *, void *, const void *), int (*) (void *, void *), void *);
extern int x1f4_miss_l4list(void *, unsigned);
extern int x1f4_mode_l4list(void *, unsigned, unsigned, void **);
extern int x1f4_move_l4list(void *, unsigned, unsigned);
extern int x1f4_peek_l4list(void *, unsigned, void **);
extern int x1f4_push_l4list(void *, unsigned, unsigned, void **);
extern int x1f4_spin_l4list(void *, unsigned, unsigned);
extern int x1f4_tail_l4list(void *, unsigned, void **);
extern int x1f4_text_l4list
    (void **, void *, int (*) (void *, unsigned *, const void *),
     int (*) (void *, void *, const void *), int (*) (void *, void *), void *);
extern int x1f4_wind_l4list(void *, unsigned, void *, unsigned);

#endif				/* __LIBx1f4_L4LIST_H__ */
