/*
 * l4xx.e.c
 * Copyright (C) 2008-2009, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <l4xx-defs.h>
#include <l4xx-inter.h>
#include <l4xx-types.h>

#define l4list(l4list)			((struct l4list_type *) (l4list))

static int fast_list(struct l4list_type *, void **, unsigned);
static int link_list(struct l4list_type *, struct list64_type *);
static int miss_list(struct l4list_type *, void **);
static int push_list(struct l4list_type *, struct list64_type *, unsigned);
static int wind_list(struct l4list_type *, unsigned, struct l4list_type *,
		     unsigned);

static int
fast_list(struct l4list_type *l4list_data, void **list64,
	  unsigned index)
{
    _libx1f4l2_seek_l4list(l4list_data, index, list64);

    _libx1f4l2_fast_l4list(l4list_data, index, *list64);

    l4list_data->link_a.list88.count--;

    _libx1f4l2_tack_l4list(l4list_data);

    return 0;
}


static int
link_list(struct l4list_type *l4list_data, struct list64_type *list64_data)
{
    l4list_data->link_a.list88.count = 1;
    l4list_data->link_a.list88.head64_data = list64_data;
    l4list_data->link_a.list88.stride = 1;
    l4list_data->link_a.list88.tail64_data = list64_data;

    l4list_data->link_c.class = 0;
    l4list_data->link_c.list64_data = list64_data;

    list64_data->last_0 = NULL;
    list64_data->last_1 = NULL;
    list64_data->next_1 = NULL;
    list64_data->next_7 = NULL;

    return 0;
}


static int
miss_list(struct l4list_type *l4list_data, void **list64)
{
    *list64 = l4list_data->link_c.list64_data;

    l4list_data->link_a.list88.count = 0;

    return 0;
}


static int
push_list(struct l4list_type *l4list_data, struct list64_type *list64_data,
	  unsigned index)
{
    unsigned Count, count, stride;

    count = l4list_data->link_a.list88.count;

    Count = count + 1;

    l4list_data->link_a.list88.count = Count;

    stride = l4list_data->link_a.list88.stride;

    if (!index) {
	_libx1f4l2_hfix_l4list(l4list_data, stride == count, list64_data);
    } else {
	if (index == count) {
	    _libx1f4l2_tfix_l4list(l4list_data, stride == count, list64_data);
	} else {
	    _libx1f4l2_mfix_l4list
		(l4list_data, count, stride, index, list64_data);
	}
    }

    l4list_data->link_c.class = index;
    l4list_data->link_c.list64_data = list64_data;

    _libx1f4l2_tack_l4list(l4list_data);

    return 0;
}


static int
wind_list(struct l4list_type *l4list_data, unsigned index,
	  struct l4list_type *l4list_text, unsigned class)
{
    void *list64;

    if (l4list_data->link_a.list88.count == 1) {
	miss_list(l4list_data, &list64);
    } else {
	fast_list(l4list_data, &list64, index);
    }

    if (!l4list_text->link_a.list88.count) {
	link_list(l4list_text, list64);
    } else {
	push_list(l4list_text, list64, class);
    }

    return 0;
}


int
x1f4_wind_l4list(void *l4data, unsigned index, void *l4text, unsigned class)
{
    int status;

    if (l4data == l4text) {
	status = _libx1f4l2_move_l4list(l4data, index, class);
    } else {
	if (!(index < l4list(l4data)->link_a.list88.count)) {
	    status = DECK_ERROR;
	} else {
	    if (l4list(l4text)->link_a.list88.count < class) {
		status = DECK_ERROR;
	    } else {
		status = wind_list(l4data, index, l4text, class);
	    }
	}
    }

    return status;
}
