/*
 * l4xx.s.c
 * Copyright (C) 2008-2009, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <math.h>
#include <stddef.h>

#include <l4xx-defs.h>
#include <l4xx-inter.h>
#include <l4xx-types.h>

#define l4list(list) \
    ((struct l4list_type *) (list))

int
_libx1f4l2_rail_l4list(void *l4data, void *l4text, unsigned count,
		       int (*size) (void *, unsigned *, const void *),
		       int (*copy) (void *, void *, const void *),
		       int (*free) (void *, void *), void *context)
{
    int (*link) (void *, void **, unsigned), status;
    struct list64_type *list64_data, *list64_miss = NULL;
    unsigned class;
    void *data;

    list64_data = l4list(l4text)->link_a.list88.head64_data;

    data = l4list(l4data)->link_m.data;
    link = l4list(l4data)->link_m.link;

    class = count;

    while (1) {
	unsigned down;
	void *list64;

	size(context, &down, list64_data + 1);

	status = link(data, &list64, sizeof(struct list64_type) + down);
	if (status) {
	    status = LINK_ERROR;

	    _libx1f4l2_land_l4list(l4data, list64_miss, free, context);

	    break;
	} else {
	    struct list64_type *list64_copy;

	    list64_copy = list64;

	    if (list64_miss) {
		list64_copy->last_1 = list64_miss;
		list64_miss->next_1 = list64_copy;
	    } else {
		list64_copy->last_1 = NULL;

		l4list(l4data)->link_a.list88.head64_data = list64_copy;
	    }

	    list64_miss = list64_copy;

	    status = copy(context, list64_copy + 1, list64_data + 1);
	    if (status) {
		status = LINK_ERROR;

		l4list(l4data)->link_m.free(data, list64_miss);

		_libx1f4l2_land_l4list
		    (l4data, list64_miss->last_1, free, context);

		break;
	    }
	}

	class--;
	if (!class) {
	    unsigned stride;

	    stride = sqrt(count);
	    if (stride < 1) {
		stride = 1;
	    } else {
		if (stride < count) {
		} else {
		    if (count == 1) {
			stride = 1;
		    } else {
			stride = count - 1;
		    }
		}
	    }

	    l4list(l4data)->link_a.list88.count = count;

	    l4list(l4data)->link_a.list88.stride = stride;

	    l4list(l4data)->link_a.list88.tail64_data = list64;

	    _libx1f4l2_link_l4list
		(l4list(l4data)->link_a.list88.head64_data, count, stride,
		 list64);

	    break;
	} else {
	    list64_data = list64_data->next_1;
	}
    }

    return status;
}
