/*
 * lsaf.i.c
 * Copyright (C) 2011-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <lsas-config.h>

#include <stddef.h>
#include <string.h>

#include <lsas-class.h>
#include <lsas-defs.h>
#include <lsas-inter.h>
#include <lsas-names.h>
#include <lsas-types.h>

#define beta_type			quarter_type

#define lsnote(lsnote)			((struct lsnote_type *) (lsnote))

#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
static int link_list(void *, void *, unsigned, unsigned);
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */

#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
static int
link_list(void *lsnote, void *node, unsigned near, unsigned post)
{
    unsigned *slip;

    slip = node;
    slip[1] = post;

    do {
	slip = (void *) ((char *) node + post);
	post = *slip;
	*slip = near;
	slip = (void *) ((char *) node + near);
	near = *slip;
	*slip = post;
    } while (post);

    return 0;
}
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */


int
_libx1f4l2_beta_lsbill(void *lsnote, const byte *name, unsigned size,
		       unsigned decq, void **post)
{
    int status;
    struct beta_type beta;
    void *node;

    node = lsnote(lsnote)->link_a.fplist.fpnode;

    status = _libx1f4l2_even_lsbill
	(&beta, node, (const byte *) name, size, post);
    if (status) {
    } else {
	unsigned line;
	void *sect;

	line = vert_noir(size);

	status = lsnote(lsnote)->link_m.link
	    (lsnote(lsnote)->link_m.data, &sect, line + decq);
	if (status) {
	    status = LINK_ERROR;
	} else {
	    void *root, *side;

	    status = _libx1f4l2_fall_lsnote(lsnote, &root, &side);
	    if (status) {
		lsnote(lsnote)->link_m.free(lsnote(lsnote)->link_m.data, sect);
	    } else {
		byte *text;
		unsigned *dana, ever, half;

		text = sect;

		memcpy(text, name, size);

		text[size] = 0;

		ever = lsnote(lsnote)->link_a.fplist.dana;

		*post = (char *) sect + line;

		half = ever >> 1;

		beta.half = half + 1;

		beta.deck = 0;

		beta.slip = 0;

		beta.lead = 0;
		beta.shut = 0;

		beta.fast = 0;
#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
		beta.mind = 0;
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */
		beta.near = 0;
#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
		beta.post = 0;
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */

		memcpy(side, node, dana_sizeof * ever);

		_libx1f4l2_moon_lsnote(node, side, *(unsigned *) node, &beta);

		if (beta.shut) {
		    unsigned *near;

		    near = side;
		    *near = beta.shut;
		}
		if (beta.lead) {
		    unsigned *near;

		    near = node;
		    *near = beta.lead;
		}

		if (beta.deck ^ half) {
		    struct fpfast_type *fpfast_data;
		    unsigned club, deck, *fail, rate, call, last = 0, *late,
			*near, pick = 0, slip;

		    if (beta.deck < half) {
			unsigned *sail;

#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
			sail = (void *) ((char *) side + beta.call);
			*sail = beta.mind;
			beta.mind = beta.call;
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */
			sail = (void *) ((char *) side + beta.tear);
			*sail = beta.fast;
			beta.fast = beta.tear;

			dana = (void *) ((char *) node + beta.call);
			fpfast_data = (void *) ((char *) node + beta.tear);

			late = node;
		    } else {
			unsigned *sail;

#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
			sail = (void *) ((char *) node + beta.call);
			*sail = beta.post;
			beta.post = beta.call;
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */
			sail = (void *) ((char *) node + beta.tear);
			*sail = beta.near;
			beta.near = beta.tear;

			dana = (void *) ((char *) side + beta.call);
			fpfast_data = (void *) ((char *) side + beta.tear);

			late = side;
		    }

		    text = fpfast_data->data[0];

		    fpfast_data->data[0] = sect;

		    fail = (void *) fpfast_data;

		    call = size;
		    size = *fail;
		    *fail = call;

		    _libx1f4l2_even_lsbill(&beta, late, name, call, NULL);

		    slip = beta.pick;

		    class_nine(name, call, beta.side, fail);

		    club = point_rate(slip);

		    class_find(name, club, dana);

		    dana[0] = slip;

		    fail = late;
		    deck = *fail;
		    near = (void *) ((char *) late + deck);
		    for (;;) {
			call = *near;
			if (call & POINT_CAST) {
			    rate = point_rate(call);
			    if (club < rate) {
				break;
			    } else {
				if (club == rate) {
				    if (slip & POINT_SEEK) {
					break;
				    }
				}

				fail = near;
				if (call & POINT_SEEK) {
				    pick = 1;
				    deck = F2ND(near);
				    last = rate;
				} else {
				    pick = name[rate >> 3] >> (7 - (rate & 7))
					& 1;
				    deck = near[1] >> (pick << 4) & 65535;
				    last = rate + 1;
				}

				near = (void *) ((char *) late + deck);
			    }
			} else {
			    break;
			}
		    }

		    if (fail == late) {
			fail[0] = beta.call;
		    } else {
			if (pick) {
			    fail[1] &= 65535;
			    fail[1] |= beta.call << 16;
			} else {
			    fail[1] &= ~65535;
			    fail[1] |= beta.call;
			}
		    }

		    if (beta.tier) {
			dana[1] = beta.tear | deck << 16;
		    } else {
			dana[1] = beta.tear << 16 | deck;
		    }

		    class_rail(late, deck, beta.side - last);
		} else {
		}

		lsnote(lsnote)->link_a.fplist.fpnode = root;

		if (1) {
		    struct fpcase_type *fpcase_data;

		    fpcase_data = (void *)
			((char *) root
			 + (lead_switch
			    * lsnote(lsnote)->link_a.fplist.call));

		    fpcase_data->data[1] = node;
		}

		lsnote(lsnote)->link_a.fplist.rate = 1;

		{
		    struct fpcase_type *fpcase_data;
		    unsigned mind, turn;

		    fpcase_data = (void *)
			((char *) root + *(unsigned *) root);

		    dana = (void *) fpcase_data;

		    *dana = size;

		    turn = size;
		    if (3 < turn) {
			turn = 3;
		    }

		    if (turn) {
			mind = text[0] << 020;
			if (1 < turn) {
			    mind |= text[1] << 010;
			    if (2 < turn) {
				mind |= text[2];
			    }
			}
		    } else {
			mind = 0;
		    }

		    fpcase_data->miss[1] = mind | turn << 27;

		    fpcase_data->data[0] = text;

		    fpcase_data->data[1] = side;
		}

#if __POST_SPLIT_CURE__ == __TREE_CURE__
		_libx1f4l2_bend_lsnote(lsnote, node, beta.near, 0);
#else
		link_list(lsnote, node, beta.near, beta.post);
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */

#if __POST_SPLIT_CURE__ == __TREE_CURE__
		_libx1f4l2_bend_lsnote(lsnote, side, beta.fast, 0);
#else
		link_list(lsnote, side, beta.fast, beta.mind);
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */
	    }
	}
    }

    return status;
}
