/*
 * lsaf.j.c
 * Copyright (C) 2011-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <lsas-config.h>

#include <stddef.h>
#include <string.h>

#include <lsas-class.h>
#include <lsas-defs.h>
#include <lsas-inter.h>
#include <lsas-names.h>
#include <lsas-types.h>

#define __ELSE_DANA__			1

#define beta_type			quarter_type

#define lsnote(lsnote)			((struct lsnote_type *) (lsnote))

#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
static int link_list(void *, void *, unsigned, unsigned);
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */

#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
static int
link_list(void *lsnote, void *node, unsigned near, unsigned post)
{
    unsigned *slip;

    slip = node;
    slip[1] = post;

    do {
	slip = (void *) ((char *) node + post);
	post = *slip;
	*slip = near;
	slip = (void *) ((char *) node + near);
	near = *slip;
	*slip = post;
    } while (post);

    return 0;
}
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */


int
_libx1f4l2_dana_lsbill(void *lsnote, void *join, unsigned lock, void *node,
		       const byte *name, unsigned size, unsigned decq,
		       void **post, unsigned take)
{
    int status;
    struct beta_type beta;
#if __ELSE_DANA__
    unsigned scan = 0;
#endif				/* __ELSE_DANA__ */

#if __ELSE_DANA__
    status = _libx1f4l2_else_lsbill
	(&beta, lock, node, (const byte *) name, size, &scan, post);
#else
    status = _libx1f4l2_even_lsbill
	(&beta, node, (const byte *) name, size, post);
#endif				/* __ELSE_DANA__ */
    if (status) {
    } else {
	unsigned line;
	void *sect;

	line = vert_noir(size);

	status = lsnote(lsnote)->link_m.link
	    (lsnote(lsnote)->link_m.data, &sect, line + decq);
	if (status) {
	    status = LINK_ERROR;
	} else {
	    unsigned ever;
	    void *side;

	    ever = lsnote(lsnote)->link_a.fplist.dana;

	    status = lsnote(lsnote)->link_m.link
		(lsnote(lsnote)->link_m.data, &side, dana_sizeof * ever);
	    if (status) {
		lsnote(lsnote)->link_m.free(lsnote(lsnote)->link_m.data, sect);
	    } else {
		byte *text;
		unsigned *dana, half;
#if __ELSE_DANA__
		unsigned mile, *port;
#endif				/* __ELSE_DANA__ */

		text = sect;

		*post = (char *) sect + line;

		memcpy(text, name, size);

		text[size] = 0;

		half = ever >> 1;

		beta.half = half + 1;

		beta.deck = 0;

		beta.slip = 0;

		beta.lead = 0;
		beta.shut = 0;

		beta.fast = 0;
#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
		beta.mind = 0;
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */
		beta.near = 0;
#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
		beta.post = 0;
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */

#if __ELSE_DANA__
		if (scan) {
		    mile = *(unsigned *) ((char *) node + scan);
		} else {
		    mile = 0;
		}
#endif				/* __ELSE_DANA__ */

		memcpy(side, node, dana_sizeof * ever);

		_libx1f4l2_moon_lsnote(node, side, *(unsigned *) node, &beta);

		if (beta.shut) {
		    unsigned *near;

		    near = side;
		    *near = beta.shut;
		}
		if (beta.lead) {
		    unsigned *near;

		    near = node;
		    *near = beta.lead;
		}

		if (beta.deck ^ half) {
		    struct fpfast_type *fpfast_data;
		    unsigned club, deck, *fail, rate, call, last = 0, *late,
			*near, pick = 0, slip;

		    if (beta.deck < half) {
			unsigned *sail;

#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
			sail = (void *) ((char *) side + beta.call);
			*sail = beta.mind;
			beta.mind = beta.call;
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */
			sail = (void *) ((char *) side + beta.tear);
			*sail = beta.fast;
			beta.fast = beta.tear;

			dana = (void *) ((char *) node + beta.call);
			fpfast_data = (void *) ((char *) node + beta.tear);

			late = node;
#if __ELSE_DANA__
			port = side;
#endif				/* __ELSE_DANA__ */
		    } else {
			unsigned *sail;

#if __POST_SPLIT_CURE__ == __TREE_CURE__
#else
			sail = (void *) ((char *) node + beta.call);
			*sail = beta.post;
			beta.post = beta.call;
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */
			sail = (void *) ((char *) node + beta.tear);
			*sail = beta.near;
			beta.near = beta.tear;

			dana = (void *) ((char *) side + beta.call);
			fpfast_data = (void *) ((char *) side + beta.tear);

			late = side;
#if __ELSE_DANA__
			port = node;
#endif				/* __ELSE_DANA__ */
		    }

		    text = fpfast_data->data[0];

		    fpfast_data->data[0] = sect;

		    fail = (void *) fpfast_data;

		    call = size;
		    size = *fail;
		    *fail = call;

#if __ELSE_DANA__
		    do {
			if (scan) {
			    /*
			     * if the cardinal at [late + scan] was overwritten
			     * a new search is needed.
			     * the cardinal might not have been overwritten
			     * yet or it may have overwritten with same value.
			     */

			    /*
			     * check for not yet
			     */
			    if (scan != beta.tear && scan != beta.call) {
				if (*(unsigned *) ((char *) late + scan)
				    == mile) {
				    /*
				     * if the other side has the same value it
				     * may be the cardinal has been overwritten
				     * with same value (the old value would be
				     * a critical bit position, while the new a
				     * byte position indicating the next item
				     * in the linked list of freed slots)
				     */
				    if (*(unsigned *) ((char *) port + scan)
					== mile) {
				    } else {
					break;
				    }
				}
			    }
			}

			_libx1f4l2_else_lsbill
			    (&beta, lock, late, name, call, &scan, NULL);
#else
			_libx1f4l2_even_lsbill(&beta, late, name, call, NULL);
#endif				/* __ELSE_DANA__ */
#if __ELSE_DANA__
		    } while (0);
#endif				/* __ELSE_DANA__ */

		    slip = beta.pick;

		    class_nine(name, call, beta.side, fail);

		    dana[0] = slip;

		    club = point_rate(slip);

		    class_find(name, club, dana);

		    fail = late;
		    deck = *fail;
		    near = (void *) ((char *) late + deck);
		    for (;;) {
			call = *near;
			if (call & POINT_CAST) {
			    rate = point_rate(call);
			    if (club < rate) {
				break;
			    } else {
				if (club == rate) {
				    if (slip & POINT_SEEK) {
					break;
				    }
				}

				fail = near;
				if (call & POINT_SEEK) {
				    pick = 1;
				    deck = F2ND(near);
				    last = rate;
				} else {
				    pick = name[rate >> 3] >> (7 - (rate & 7))
					& 1;
				    deck = near[1] >> (pick << 4) & 65535;
				    last = rate + 1;
				}

				near = (void *) ((char *) late + deck);
			    }
			} else {
			    break;
			}
		    }

		    if (fail == late) {
			fail[0] = beta.call;
		    } else {
			if (pick) {
			    fail[1] &= 65535;
			    fail[1] |= beta.call << 16;
			} else {
			    fail[1] &= ~65535;
			    fail[1] |= beta.call;
			}
		    }

		    if (beta.tier) {
			dana[1] = beta.tear | deck << 16;
		    } else {
			dana[1] = beta.tear << 16 | deck;
		    }

		    class_rail(late, deck, beta.side - last);
		} else {
		}

		_libx1f4l2_join_lsbill
		    (lsnote, join, side, (void *) text, size, take);

#if __POST_SPLIT_CURE__ == __TREE_CURE__
		_libx1f4l2_bend_lsnote(lsnote, node, beta.near, 0);
#else
		link_list(lsnote, node, beta.near, beta.post);
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */

#if __POST_SPLIT_CURE__ == __TREE_CURE__
		_libx1f4l2_bend_lsnote(lsnote, side, beta.fast, 0);
#else
		link_list(lsnote, side, beta.fast, beta.mind);
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */
	    }
	}
    }

    return status;
}
