/*
 * lsas.w.c
 * Copyright (C) 2011-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <lsas-config.h>

#include <string.h>

#include <lsas-defs.h>
#include <lsas-names.h>
#include <lsas-types.h>

#define true(e)				1

#define lsnote(lsnote)			((struct lsnote_type *) (lsnote))

typedef struct long_type {
    unsigned **date, **deck, *fast, *fine, *late, *line, *mind, *pick;
} long_type;

#if __POST_SPLIT_CURE__ == __TREE_CURE__
static void link_list(void *, void *, unsigned, int);
static void shut_node(void *, struct long_type *, unsigned *, unsigned);
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */

#if __POST_SPLIT_CURE__ == __TREE_CURE__
static void
link_list(void *lsnote, void *node, unsigned near, int internal)
{
    unsigned post, *slip;

    if (internal) {
	post = lsnote(lsnote)->link_a.fplist.call * lead_switch >> 1;
    } else {
	post = lsnote(lsnote)->link_a.fplist.dana * lead_switch >> 1;
    }

    slip = node;
    slip++;

    do {
	*slip = post;
	slip = (void *) ((char *) node + post);
	post += lead_switch;
	*slip = near;
	slip = (void *) ((char *) node + near);
	near = *slip;
    } while (near);
}
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */


#if __POST_SPLIT_CURE__ == __TREE_CURE__
static void
shut_node(void *node, struct long_type *long_data, unsigned *slip,
	  unsigned rate)
{
    if (rate == 1) {
	unsigned call, *down, i, l1st, l2nd, lock, *mind, *near;

	mind = long_data->mind;

	i = slip - long_data->fast;
	down = long_data->deck[i];
	if (long_data->late[i]) {
	    *down = M1ST(L1ST(*down))
		| M2ND((char *) mind - (char *) long_data->pick);
	} else {
	    *down = M2ND(L2ND(*down))
		| M1ST((char *) mind - (char *) long_data->pick);
	}

	lock = *slip;
	near = (void *) ((char *) node + lock);

	{
	    *mind++ = *near++;
	}

	call = *near;
	l1st = L1ST(call);
	l2nd = L2ND(call);
	if (*(unsigned *) ((char *) node + l1st) & POINT_CAST) {
	    *long_data->fine++ = l1st;
	    *long_data->date++ = mind;
	    *long_data->line++ = 0;
	}
	if (*(unsigned *) ((char *) node + l2nd) & POINT_CAST) {
	    *long_data->fine++ = l2nd;
	    *long_data->date++ = mind;
	    *long_data->line++ = 1;
	}

	for (i = lead_shrews - 1; i; i--) {
	    *mind++ = *near++;
	}

	long_data->mind = mind;

	*slip = 0;
    } else {
	unsigned *file, i;

	rate >>= 1;
	file = long_data->fine;
	shut_node(node, long_data, slip, rate);
	i = long_data->fine - file;
	for (; i; i--) {
	    if (*file) {
		shut_node(node, long_data, file, rate);
	    }

	    file++;
	}
    }
}
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */


int
_libx1f4l2_bend_lsnote(void *lsnote, void *node, unsigned head, int internal)
{
#if __POST_SPLIT_CURE__ == __TREE_CURE__
    struct long_type bend;
    unsigned *file, *near, note;

    bend.deck = lsnote(lsnote)->link_e.mail;

    note = lsnote(lsnote)->link_e.lane;

    bend.fast = (void *) (bend.deck + note);
    bend.late = bend.fast + note;
    bend.pick = bend.late + note;

    bend.mind = bend.pick + lead_shrews;

    near = node;
    note = *near;
    if (true(*(unsigned *) ((char *) node + note) & POINT_CAST)) {
	unsigned rate = 0, much;

	bend.fast[1] = note;
	bend.deck[1] = &much;
	bend.late[1] = 0;

	file = bend.fast + 1;

	bend.fine = bend.fast + 2;
	bend.date = bend.deck + 2;
	bend.line = bend.late + 2;

	do {
	    unsigned i;

	    i = bend.fine - file;
	    for (; i; i--) {
		if (*file) {
		    shut_node(node, &bend, file, rate ? rate : 1);
		}

		file++;
	    }

	    rate <<= 1;
	    rate = rate ? rate : 1;
	} while (file != bend.fine);

	*near = lead_shrews * sizeof(unsigned);
	memcpy(near + lead_shrews, bend.pick + lead_shrews,
	       (bend.mind - (bend.pick + lead_shrews)) * sizeof(unsigned));
    }

    link_list(lsnote, node, head, internal);
#endif				/* __POST_SPLIT_CURE__ == __TREE_CURE__ */

    return 0;
}
