/*
 * qccc.7.c
 * Copyright (C) 2010-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <qccc-bits.h>
#include <qccc-names.h>
#include <qccc-types.h>

#define M_TEXT(___v, ___t) \
    ((void *) ((char *) (___v) + (___t)))

#define qcrate(qcrate)			((struct qcrate_type *) (qcrate))

static int back_node
    (void *, unsigned, unsigned, void *, int (*) (void *, void *));
static int jack_node
    (void *, unsigned, unsigned, unsigned, void *, int (*) (void *, void *));
static int pick_node
    (void *, unsigned, unsigned, void *, int (*) (void *, void *));

static int
back_node(void *node, unsigned violet, unsigned pick, void *back,
	  int (*call) (void *, void *))
{
    int delete;

    do {
	unsigned slip, text;

	text = Q_NODE(M_TEXT(node, pick));

	slip = L_LAST(text);
	if (slip < violet) {
	    delete = back_node(node, violet, slip, back, call);
	    if (delete) {
		break;
	    }
	} else {
	    delete = call(back, M_TEXT(node, slip));
	    if (delete) {
		break;
	    }
	}

	pick = L_NEXT(text);
    } while (pick < violet);
    if (delete) {
    } else {
	if (1) {
	    delete = call(back, M_TEXT(node, pick));
	}
    }

    return delete;
}


static int
jack_node(void *node, unsigned violet, unsigned pick, unsigned rate,
	  void *back, int (*call) (void *, void *))
{
    int delete;

    do {
	unsigned slip, text;

	text = Q_NODE(M_TEXT(node, pick));

	slip = L_LAST(text);
	if (slip < violet) {
	    delete = jack_node(node, violet, slip, rate, back, call);
	    if (delete) {
		break;
	    }
	} else {
	    delete = pick_node
		(*(void **) M_TEXT(node, slip), violet, rate, back, call);
	    if (delete) {
		break;
	    }
	}

	pick = L_NEXT(text);
    } while (pick < violet);
    if (delete) {
    } else {
	if (1) {
	    delete = pick_node
		(*(void **) M_TEXT(node, pick), violet, rate, back, call);
	}
    }

    return delete;
}


static int
pick_node(void *node, unsigned violet, unsigned rate, void *back,
	  int (*call) (void *, void *))
{
    int delete;

    if (rate) {
	delete = jack_node
	    (node, violet, P_ROOT(Q_NODE(node)), rate - 1, back, call);
    } else {
	delete = back_node(node, violet, P_ROOT(Q_NODE(node)), back, call);
    }

    return delete;
}


int
x1f4_lime_qcrate(void *qcrate, void *back, int (*call) (void *, void *))
{
    int delete;
    void *node;

    node = qcrate(qcrate)->link_a.fplist.fpnode_data;

    if (Q_NODE(M_TEXT(node, sizeof(unsigned))) ^ 1) {
	delete = pick_node
	    (node, sizeof(unsigned) << 1 << qcrate(qcrate)->link_a.fplist.lock,
	     qcrate(qcrate)->link_a.fplist.rate, back, call);
    } else {
	if (Q_NODE(M_TEXT(node, sizeof(unsigned)))) {
	    delete = call(back, M_TEXT(node, P_ROOT(Q_NODE(node))));
	} else {
	    delete = 0;
	}
    }

    return delete;
}
