/*
 * qccc.p.c
 * Copyright (C) 2009-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <qccc-config.h>

#include <string.h>

#include <qccc-bits.h>
#include <qccc-names.h>
#include <qccc-types.h>

#define true(e)				1

#define M_TEXT(___v, ___t) \
    ((void *) ((char *) (___v) + (___t)))

static void copy_data(void *, unsigned, unsigned, void **, void **, unsigned);
static void pick_mini(unsigned *, unsigned);
static void type_list(void *, unsigned);
static void type_mini(void *, unsigned);

static void
copy_data(void *node, unsigned violet, unsigned pick,
	  void **lane, void **back, unsigned size)
{
    do {
	unsigned left, slip, text;

	text = Q_NODE(M_TEXT(node, pick));
	left = Q_NODE(M_TEXT(node, pick + sizeof(unsigned)));

	size -= left;

	slip = L_LAST(text);
	if (slip < violet) {
	    copy_data(node, violet, slip, lane, back, left);
	} else {
	    unsigned *p;
	    void **u, **v;

	    v = M_TEXT(node, slip);
	    u = *back;

	    *u++ = *v;

	    *back = u;

	    p = *lane;

	    *p = left;

	    *lane = p + 2;
	}

	pick = L_NEXT(text);
    } while (pick < violet);
    if (0) {
    } else {
	if (1) {
	    unsigned *p;
	    void **u, **v;

	    v = M_TEXT(node, pick);
	    u = *back;

	    *u++ = *v;

	    *back = u;

	    p = *lane;

	    *p = size;

	    *lane = p + 2;
	}
    }
}


static void
pick_mini(unsigned *mini, unsigned rate)
{
    int pick = 1;
#if ____REFORM_MEMORY_LAYOUT__ == __BREADTH_FIRST_DOUBLE_LAYERED__
    unsigned nine;
#elif __REFORM_MEMORY_LAYOUT__ == __BREADTH_FIRST_TRIPLE_LAYERED__
    unsigned nine;
#endif				/* ____REFORM_MEMORY_LAYOUT__ == __BREADT... */

#if ____REFORM_MEMORY_LAYOUT__ == __BREADTH_FIRST_DOUBLE_LAYERED__
    nine = rate;
#elif __REFORM_MEMORY_LAYOUT__ == __BREADTH_FIRST_TRIPLE_LAYERED__
    nine = rate;
#endif				/* ____REFORM_MEMORY_LAYOUT__ == __BREADT... */

    while (rate) {
	unsigned *call, i, *slip;

	if (pick) {
	    call = mini + (1 << rate) - 2;
	    mini += 2 << rate;
	    slip = mini;
	    i = 1 << rate >> 1;
	    for (; i; i--) {
		*call = mini[-1] + mini[-3];
		slip[-1] = mini[-3];
		call -= 2;
		slip -= 2;
		mini -= 4;
	    }
	} else {
	    call = mini + (1 << rate) - 2;
	    i = 1 << rate;
	    for (; i; i--) {
		mini[1] = *mini;
		mini += 2;
	    }

	    slip = mini;
	    i = 1 << rate >> 1;
	    for (; i; i--) {
		*call = mini[-1] + mini[-3];
		slip[-1] = mini[-3];
		call -= 2;
		slip -= 2;
		mini -= 4;
	    }
	}

	rate--;

	pick = 0;
    }

#if ____REFORM_MEMORY_LAYOUT__ == __BREADTH_FIRST_DOUBLE_LAYERED__
    nine &= ~1;
    while (2 < nine) {
	unsigned *call, i, *slip, *type;

	call = mini + (1 << nine >> 1);
	i = 1 << nine >> 2;
	for (; i; i--) {
	    *call = call[1];
	    call += 2;
	}

	call = mini + (1 << nine >> 1);
	slip = mini + (1 << nine) + 1;
	type = call + 1;
	i = 1 << nine >> 2;
	for (; i; i--) {
	    *type = *call;
	    type += 2;
	    call += 2;
	    *type = *slip;
	    type += 2;
	    slip += 2;
	    *type = *slip;
	    type += 2;
	    slip += 2;
	}

	nine -= 2;
    }
#elif __REFORM_MEMORY_LAYOUT__ == __BREADTH_FIRST_TRIPLE_LAYERED__
    rate = nine % 3;
    if (rate == 2) {
	unsigned *call, i, *slip, *type;

	call = mini + (1 << nine >> 1);
	i = 1 << nine >> 2;
	for (; i; i--) {
	    *call = call[1];
	    call += 2;
	}

	call = mini + (1 << nine >> 1);
	slip = mini + (1 << nine) + 1;
	type = call + 1;
	i = 1 << nine >> 2;
	for (; i; i--) {
	    *type = *call;
	    type += 2;
	    call += 2;
	    *type = *slip;
	    type += 2;
	    slip += 2;
	    *type = *slip;
	    type += 2;
	    slip += 2;
	}

	nine -= 2;
    } else {
	nine -= rate;
    }
    while (3 < nine) {
	unsigned *call, *half, i, *slip, *type;

	call = mini + (1 << nine >> 2);
	i = 1 << nine >> 3;
	for (; i; i--) {
	    *call = call[1];
	    call += 2;
	}
	half = mini + (1 << nine >> 1);
	i = 1 << nine >> 2;
	for (; i; i--) {
	    *half = half[1];
	    half += 2;
	}

	call = mini + (1 << nine >> 2);
	half = mini + (1 << nine >> 1);
	slip = mini + (1 << nine) + 1;
	type = call + 1;
	i = 1 << nine >> 3;
	for (; i; i--) {
	    *type = *call;
	    type += 2;
	    call += 2;
	    *type = *half;
	    type += 2;
	    half += 2;
	    *type = *half;
	    type += 2;
	    half += 2;
	    *type = *slip;
	    type += 2;
	    slip += 2;
	    *type = *slip;
	    type += 2;
	    slip += 2;
	    *type = *slip;
	    type += 2;
	    slip += 2;
	    *type = *slip;
	    type += 2;
	    slip += 2;
	}

	nine -= 3;
    }
#endif				/* ____REFORM_MEMORY_LAYOUT__ == __BREADT... */
}


static void
type_list(void *node, unsigned rate)
{
    unsigned copy, list, text;
    void *side;

    text = (sizeof(unsigned) << 1 << rate) + (sizeof(void *) << rate >> 1);

    side = M_TEXT(node, text);

    list = sizeof(unsigned) << rate;

    S_NODE(node, B_ROOT(sizeof(unsigned) << 1) | B_LIST(list));

    node = M_TEXT(node, list);

    copy = 1 << rate;
    copy >>= 1;
    copy--;
    for (; copy; copy--) {
	S_NODE(node, text);
	text += sizeof(void *);
	list += sizeof(unsigned) << 1;
	S_NODE(side, list);
	side = M_TEXT(side, sizeof(void *));
	node = M_TEXT(node, sizeof(unsigned) << 1);
    }
    {
	S_NODE(node, text);
	S_NODE(side, 0);
    }
}


static void
type_mini(void *node, unsigned rate)
{
#if ____REFORM_MEMORY_LAYOUT__ == __BREADTH_FIRST__
    unsigned *line, pick, type = 0;

    pick = sizeof(unsigned) << 2;

    line = node;

    rate -= 2;

    line += 2;

    for (; type < rate; type++) {
	unsigned i;

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + (sizeof(unsigned) << 1))
		   | B_TILT(0));
	    pick += sizeof(unsigned) << 2;
	    line += 2;
	}
    }

    pick = sizeof(unsigned) << 3 << rate;

    {
	unsigned i;

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(void *)) | B_TILT(0));
	    pick += sizeof(void *) << 1;
	    line += 2;
	}
    }
#elif __REFORM_MEMORY_LAYOUT__ == __BREADTH_FIRST_DOUBLE_LAYERED__
    unsigned beta, *line, pick, slip, type = 0;

    pick = sizeof(unsigned) << 3;
    slip = sizeof(unsigned) << 2;

    line = node;

    line += 2;

    beta = (rate - 3) >> 1;
    for (; beta; beta--) {
	unsigned i;

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 2)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 006;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 6)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 014;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 6)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 014;
	    line += 2;
	}

	type += 2;
    }
    if (rate & 1) {
	unsigned i;

	pick = sizeof(unsigned) << 1 << rate;

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 2)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 006;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(void *)) | B_TILT(0));
	    pick += sizeof(void *) << 1;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(void *)) | B_TILT(0));
	    pick += sizeof(void *) << 1;
	    line += 2;
	}
    } else {
	unsigned i;

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 2)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 006;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 2)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 004;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 2)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 004;
	    line += 2;
	}

	pick = sizeof(unsigned) << 1 << rate;

	i = 4 << type;
	for (; i; i--) {
	    S_NODE(line,
		    B_LAST(pick) | B_NEXT(pick + sizeof(void *)) | B_TILT(0));
	    pick += sizeof(void *) << 1;
	    line += 2;
	}
    }
#elif __REFORM_MEMORY_LAYOUT__ == __BREADTH_FIRST_TRIPLE_LAYERED__
    unsigned beta, *line, mass, pick, slip, type = 0;

    pick = sizeof(unsigned) << 4;
    slip = sizeof(unsigned) << 2;

    line = node;

    line += 2;

    beta = (rate - 4) / 3;

    mass = rate - beta * 3;

    for (; beta; beta--) {
	unsigned i;

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 004;
	    line += 2;
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 004;
	    line += 2;
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 006;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 016)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 034;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 016)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 034;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 016)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 034;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 016)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 034;
	    line += 2;
	}

	type += 3;
    }
    if (mass == 4) {
	unsigned i;

	pick = sizeof(unsigned) << 1 << rate;

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 004;
	    line += 2;
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 004;
	    line += 2;
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 006;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(void *)) | B_TILT(0));
	    pick += sizeof(void *) << 1;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(void *)) | B_TILT(0));
	    pick += sizeof(void *) << 1;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(void *)) | B_TILT(0));
	    pick += sizeof(void *) << 1;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(void *)) | B_TILT(0));
	    pick += sizeof(void *) << 1;
	    line += 2;
	}
    } else {
	if (mass == 5) {
	    unsigned i;

	    i = 1 << type;
	    for (; i; i--) {
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 004;
		line += 2;
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 004;
		line += 2;
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 006;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 002)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 004;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 002)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 004;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 002)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 004;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 002)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 004;
		line += 2;
	    }

	    pick = sizeof(unsigned) << 1 << rate;

	    i = 8 << type;
	    for (; i; i--) {
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(void *))
		       | B_TILT(0));
		pick += sizeof(void *) << 1;
		line += 2;
	    }
	} else {
	    unsigned i;

	    i = 1 << type;
	    for (; i; i--) {
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 004;
		line += 2;
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 004;
		line += 2;
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 006;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 006)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 014;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 006)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 014;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 006)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 014;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 006)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 014;
		line += 2;
	    }

	    pick = sizeof(unsigned) << 1 << rate;

	    i = 8 << type;
	    for (; i; i--) {
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 006;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(void *))
		       | B_TILT(0));
		pick += sizeof(void *) << 1;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(void *))
		       | B_TILT(0));
		pick += sizeof(void *) << 1;
		line += 2;
	    }
	}
    }
#endif				/* ____REFORM_MEMORY_LAYOUT__ == __BREADT... */
}


int
_libx1f4l2_mold_qcrate(void *node, void *side, unsigned rate, unsigned *left)
{
    unsigned violet;
    void *mini, *slip;

    violet = sizeof(unsigned) << 1 << rate;

    mini = M_TEXT(side, sizeof(unsigned));
    slip = M_TEXT(side, violet);

    copy_data(node, violet, P_ROOT(Q_NODE(node)), &mini, &slip,
	      Q_NODE(M_TEXT(node, sizeof(unsigned))));

    memcpy(node, (char *) side + (sizeof(unsigned) << rate),
	   sizeof(unsigned) << rate);

    memcpy((char *) node + violet,
	   (char *) side + violet + (sizeof(void *) << rate >> 1),
	   sizeof(void *) << rate >> 1);

    pick_mini(node, rate - 1);
    S_NODE(M_TEXT(node, sizeof(unsigned)), Q_NODE(node));
    type_mini(node, rate);

    type_list(node, rate);

    pick_mini(side, rate - 1);
    *left = Q_NODE(side);
    S_NODE(M_TEXT(side, sizeof(unsigned)), Q_NODE(side));
    type_mini(side, rate);

    type_list(side, rate);

    return 1;
}
