/*
 * qxcb.y.c
 * Copyright (C) 2009-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <qxcc-config.h>

#include <qxcc-bits.h>
#include <qxcc-names.h>
#include <qxcc-types.h>

int
_libx1f4l2_sail_qxrate(void *node, unsigned rate, unsigned sail)
{
#if ____REFORM_MEMORY_LAYOUT__ == __BREADTH_FIRST__
    unsigned *line, pick, type = 0;

    pick = sizeof(unsigned) << 2;

    line = node;

    rate -= 2;

    line += 2;

    for (; type < rate; type++) {
	unsigned i;

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + (sizeof(unsigned) << 1))
		   | B_TILT(0));
	    pick += sizeof(unsigned) << 2;
	    line += 2;
	}
    }

    pick = sail;

    {
	unsigned i;

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(void *)) | B_TILT(0));
	    pick += sizeof(void *) << 1;
	    line += 2;
	}
    }
#elif __REFORM_MEMORY_LAYOUT__ == __BREADTH_FIRST_DOUBLE_LAYERED__
    unsigned beta, *line, pick, slip, type = 0;

    pick = sizeof(unsigned) << 3;
    slip = sizeof(unsigned) << 2;

    line = node;

    line += 2;

    beta = (rate - 3) >> 1;
    for (; beta; beta--) {
	unsigned i;

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 2)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 006;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 6)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 014;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 6)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 014;
	    line += 2;
	}

	type += 2;
    }
    if (rate & 1) {
	unsigned i;

	pick = sail;

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 2)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 006;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(void *)) | B_TILT(0));
	    pick += sizeof(void *) << 1;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(void *)) | B_TILT(0));
	    pick += sizeof(void *) << 1;
	    line += 2;
	}
    } else {
	unsigned i;

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 2)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 006;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 2)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 004;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 2)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 004;
	    line += 2;
	}

	pick = sail;

	i = 4 << type;
	for (; i; i--) {
	    S_NODE(line,
		    B_LAST(pick) | B_NEXT(pick + sizeof(void *)) | B_TILT(0));
	    pick += sizeof(void *) << 1;
	    line += 2;
	}
    }
#elif __REFORM_MEMORY_LAYOUT__ == __BREADTH_FIRST_TRIPLE_LAYERED__
    unsigned beta, *line, mass, pick, slip, type = 0;

    pick = sizeof(unsigned) << 4;
    slip = sizeof(unsigned) << 2;

    line = node;

    line += 2;

    beta = (rate - 4) / 3;

    mass = rate - beta * 3;

    for (; beta; beta--) {
	unsigned i;

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 004;
	    line += 2;
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 004;
	    line += 2;
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 006;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 016)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 034;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 016)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 034;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 016)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 034;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 016)
		   | B_TILT(0));
	    pick += sizeof(unsigned) * 034;
	    line += 2;
	}

	type += 3;
    }
    if (mass == 4) {
	unsigned i;

	pick = sail;

	i = 1 << type;
	for (; i; i--) {
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 004;
	    line += 2;
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 004;
	    line += 2;
	    S_NODE(line,
		   B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		   | B_TILT(0));
	    slip += sizeof(unsigned) * 006;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(void *)) | B_TILT(0));
	    pick += sizeof(void *) << 1;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(void *)) | B_TILT(0));
	    pick += sizeof(void *) << 1;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(void *)) | B_TILT(0));
	    pick += sizeof(void *) << 1;
	    line += 2;
	    S_NODE(line,
		   B_LAST(pick) | B_NEXT(pick + sizeof(void *)) | B_TILT(0));
	    pick += sizeof(void *) << 1;
	    line += 2;
	}
    } else {
	if (mass == 5) {
	    unsigned i;

	    i = 1 << type;
	    for (; i; i--) {
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 004;
		line += 2;
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 004;
		line += 2;
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 006;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 002)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 004;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 002)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 004;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 002)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 004;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 002)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 004;
		line += 2;
	    }

	    pick = sail;

	    i = 8 << type;
	    for (; i; i--) {
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(void *))
		       | B_TILT(0));
		pick += sizeof(void *) << 1;
		line += 2;
	    }
	} else {
	    unsigned i;

	    i = 1 << type;
	    for (; i; i--) {
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 004;
		line += 2;
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 004;
		line += 2;
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 006;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 006)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 014;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 006)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 014;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 006)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 014;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(unsigned) * 006)
		       | B_TILT(0));
		pick += sizeof(unsigned) * 014;
		line += 2;
	    }

	    pick = sail;

	    i = 8 << type;
	    for (; i; i--) {
		S_NODE(line,
		       B_LAST(slip) | B_NEXT(slip + sizeof(unsigned) * 002)
		       | B_TILT(0));
		slip += sizeof(unsigned) * 006;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(void *))
		       | B_TILT(0));
		pick += sizeof(void *) << 1;
		line += 2;
		S_NODE(line,
		       B_LAST(pick) | B_NEXT(pick + sizeof(void *))
		       | B_TILT(0));
		pick += sizeof(void *) << 1;
		line += 2;
	    }
	}
    }
#endif				/* ____REFORM_MEMORY_LAYOUT__ == __BREADT... */

    return 0;
}
