/*
 * qxcc.2.c
 * Copyright (C) 2010-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <qxcc-config.h>

#include <stddef.h>
#include <string.h>

#include <qxcc-bits.h>
#include <qxcc-defs.h>
#include <qxcc-inter.h>
#include <qxcc-types.h>

#define M_TEXT(___v, ___t) \
    ((void *) ((char *) (___v) + (___t)))

#define qxrate(qxrate)			((struct qxrate_type *) (qxrate))

#if __DELAY_SPLIT__
static int part_late(struct qxrate_type *, unsigned, void *, unsigned,
		     unsigned *, void **);
#endif				/* __DELAY_SPLIT__ */

#if __DELAY_SPLIT__
static int
part_late(struct qxrate_type *qxrate_data, unsigned rate, void *node,
	  unsigned index, unsigned *text, void **push)
{
    int status = 0;

    while (1) {
	int excess;

	excess = _libx1f4l2_gray_qxrate
	    (qxrate_data, rate, node, index, &index, &node);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}

	if (rate) {
	    rate--;
	} else {
	    break;
	}
    }

    *push = node;
    *text = index;

    return status;
}
#endif				/* __DELAY_SPLIT__ */


int
x1f4_push_qxrate(void *qxrate, unsigned index, void **push)
{
    int status;
    unsigned count;

    count = qxrate(qxrate)->link_a.fplist.size;

    if (count < index) {
	status = DECK_ERROR;
    } else {
	unsigned rate;
	void *node;

	node = qxrate(qxrate)->link_a.fplist.fpnode_data;

	qxrate(qxrate)->link_a.fplist.size = count + 1;

	status = 0;

	rate = qxrate(qxrate)->link_a.fplist.rate;

	do {
	    if (rate) {
		if (P_LIST(Q_NODE(node))) {
		} else {
		    status = _libx1f4l2_fall_qxrate(qxrate);
		    if (status) {
			break;
		    } else {
			rate++;
			node = qxrate(qxrate)->link_a.fplist.fpnode_data;
		    }
		}
	    } else {
		if (count ^ 3 * qxrate(qxrate)->link_a.fplist.line >> 1) {
		} else {
		    status = _libx1f4l2_fall_qxrate(qxrate);
		    if (status) {
			break;
		    } else {
			rate++;
			node = qxrate(qxrate)->link_a.fplist.fpnode_data;
		    }
		}
	    }

	    if (rate) {
#if __DELAY_SPLIT__
		unsigned black = 0, late = 0;
#endif				/* __DELAY_SPLIT__ */
		unsigned violet;
#if __DELAY_SPLIT__
		void *hook;
#endif				/* __DELAY_SPLIT__ */

#if __DELAY_SPLIT__
		hook = NULL;
#endif				/* __DELAY_SPLIT__ */

		violet = sizeof(unsigned) << 1
		    << qxrate(qxrate)->link_a.fplist.lock;

		rate--;

		while (rate) {
		    unsigned epnode, *lpnode, trans;
		    void *pick;

		    trans = index;

		    lpnode = node;

		    epnode = P_ROOT(Q_NODE(node));

		    lpnode[1]++;

		    do {
			unsigned class, epslip;

			lpnode = M_TEXT(node, epnode);

			epslip = *lpnode;

			lpnode++;

			class = *lpnode;
			if (trans < class) {
			    epnode = L_LAST(epslip);
			    *lpnode = class + 1;
			} else {
			    epnode = L_NEXT(epslip);
			    trans -= class;
			}
		    } while (epnode < violet);

		    pick = *(void **) M_TEXT(node, epnode);

		    if (P_LIST(Q_NODE(pick))) {
			index = trans;

#if __DELAY_SPLIT__
			hook = NULL;
#endif				/* __DELAY_SPLIT__ */
		    } else {
#if __DELAY_SPLIT__
			if (hook) {
			} else {
			    hook = node;
			    late = rate;
			    if (1) {
				black = index;
			    }
			}

			index = trans;
#else
			int excess;

			excess = _libx1f4l2_gray_qxrate
			    (qxrate, rate, node, index, &index, &pick);
			if (excess) {
			    if (status) {
			    } else {
				status = excess;
			    }
			}
#endif				/* __DELAY_SPLIT__ */
		    }

		    rate--;

		    node = pick;
		}
		{
		    unsigned epnode, *lpnode, trans;
		    void *pick;

		    trans = index;

		    lpnode = node;

		    epnode = P_ROOT(Q_NODE(node));

		    lpnode++;

		    count = *lpnode;

		    *lpnode = count + 1;

		    do {
			unsigned class, epslip;

			lpnode = M_TEXT(node, epnode);

			epslip = *lpnode;

			lpnode++;

			class = *lpnode;
			if (trans < class) {
			    epnode = L_LAST(epslip);
			    *lpnode = class + 1;
			    count = class;
			} else {
			    epnode = L_NEXT(epslip);
			    trans -= class;
			    count -= class;
			}
		    } while (epnode < violet);

		    pick = *(void **) M_TEXT(node, epnode);

		    if (count ^ qxrate(qxrate)->link_a.fplist.line) {
			index = trans;

#if __DELAY_SPLIT__
			hook = NULL;
#endif				/* __DELAY_SPLIT__ */
		    } else {
#if __DELAY_SPLIT__
			if (hook) {
			} else {
			    hook = node;
			    late = 0;
			    if (1) {
				black = index;
			    }
			}

			index = trans;
#else
			int excess;

			excess = _libx1f4l2_gray_qxrate
			    (qxrate, rate, node, index, &index, &pick);
			if (excess) {
			    if (status) {
			    } else {
				status = excess;
			    }
			} else {
			    count = qxrate(qxrate)->link_a.fplist.line >> 1;
			}
#endif				/* __DELAY_SPLIT__ */
		    }

		    node = pick;
		}

#if __DELAY_SPLIT__
		if (hook) {
		    int excess;

		    excess = part_late
			(qxrate, late, hook, black, &index, &node);
		    if (excess) {
			if (status) {
			} else {
			    status = excess;
			}
		    } else {
			count = qxrate(qxrate)->link_a.fplist.line >> 1;
		    }
		}
#endif				/* __DELAY_SPLIT__ */
	    }

	    {
		unsigned mall;

		mall = qxrate(qxrate)->link_a.fplist.mall;

		node = (char *) node + index * mall;

		count -= index;
		if (count) {
		    memmove((char *) node + mall, node, count * mall);
		} else {
		}

		*push = node;
	    }
	} while (0);
    }

    return status;
}
