/*
 * qxcc.7.c
 * Copyright (C) 2010-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <qxcc-bits.h>
#include <qxcc-names.h>
#include <qxcc-types.h>

#define M_TEXT(___v, ___t) \
    ((void *) ((char *) (___v) + (___t)))

#define qxrate(qxrate)			((struct qxrate_type *) (qxrate))

static int back_node
    (void *, unsigned, unsigned, void *, int (*) (void *, void *));
static int fast_node
    (void *, unsigned, unsigned, unsigned, unsigned, void *,
     int (*) (void *, void *));
static int jack_node
    (void *, unsigned, unsigned, unsigned, unsigned, void *,
     int (*) (void *, void *));
static int pick_node
    (void *, unsigned, unsigned, unsigned, void *, int (*) (void *, void *));

static int
back_node(void *node, unsigned pick, unsigned mall, void *back,
	  int (*call) (void *, void *))
{
    int delete;

    do {
	delete = call(back, node);
	if (delete) {
	    break;
	}

	pick--;

	node = (char *) node + mall;
    } while (pick);

    return delete;
}


static int
fast_node(void *node, unsigned count, unsigned violet, unsigned pick,
	  unsigned mall, void *back, int (*call) (void *, void *))
{
    int delete;

    do {
	unsigned *lpnode, slip, text;

	lpnode = M_TEXT(node, pick);

	text = *lpnode;

	lpnode++;

	slip = L_LAST(text);
	if (slip < violet) {
	    delete = fast_node(node, *lpnode, violet, slip, mall, back, call);
	    if (delete) {
		break;
	    }
	} else {
	    delete = back_node
		(*(void **) M_TEXT(node, slip), *lpnode, mall, back, call);
	    if (delete) {
		break;
	    }
	}

	count -= *lpnode;

	pick = L_NEXT(text);
    } while (pick < violet);
    if (delete) {
    } else {
	if (1) {
	    delete = back_node
		(*(void **) M_TEXT(node, pick), count, mall, back, call);
	}
    }

    return delete;
}


static int
jack_node(void *node, unsigned violet, unsigned pick, unsigned rate,
	  unsigned mall, void *back, int (*call) (void *, void *))
{
    int delete;

    do {
	unsigned slip, text;

	text = Q_NODE(M_TEXT(node, pick));

	slip = L_LAST(text);
	if (slip < violet) {
	    delete = jack_node(node, violet, slip, rate, mall, back, call);
	    if (delete) {
		break;
	    }
	} else {
	    delete = pick_node
		(*(void **) M_TEXT(node, slip), violet, rate, mall, back,
		 call);
	    if (delete) {
		break;
	    }
	}

	pick = L_NEXT(text);
    } while (pick < violet);
    if (delete) {
    } else {
	if (1) {
	    delete = pick_node
		(*(void **) M_TEXT(node, pick), violet, rate, mall, back,
		 call);
	}
    }

    return delete;
}


static int
pick_node(void *node, unsigned violet, unsigned rate, unsigned mall,
	  void *back, int (*call) (void *, void *))
{
    int delete;

    if (rate ^ 1) {
	delete = jack_node
	    (node, violet, P_ROOT(Q_NODE(node)), rate - 1, mall, back, call);
    } else {
	delete = fast_node
	    (node, Q_NODE(M_TEXT(node, sizeof(unsigned))), violet,
	     P_ROOT(Q_NODE(node)), mall, back, call);
    }

    return delete;
}


int
x1f4_lime_qxrate(void *qxrate, void *back, int (*call) (void *, void *))
{
    int delete;
    unsigned mall, rate;
    void *node;

    node = qxrate(qxrate)->link_a.fplist.fpnode_data;

    mall = qxrate(qxrate)->link_a.fplist.mall;

    rate = qxrate(qxrate)->link_a.fplist.rate;

    if (rate) {
	unsigned violet;

	violet = sizeof(unsigned) << 1 << qxrate(qxrate)->link_a.fplist.lock;

	if (rate ^ 1) {
	    delete = jack_node
		(node, violet, P_ROOT(Q_NODE(node)), rate - 1, mall, back,
		 call);
	} else {
	    delete = fast_node
		(node, qxrate(qxrate)->link_a.fplist.size, violet,
		 P_ROOT(Q_NODE(node)), mall, back, call);
	}
    } else {
	unsigned size;

	size = qxrate(qxrate)->link_a.fplist.size;
	if (size) {
	    delete = back_node(node, size, mall, back, call);
	} else {
	    delete = 0;
	}
    }

    return delete;
}
