/*
 * qxcc.v.c
 * Copyright (C) 2010-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>

#include <qxcc-bits.h>
#include <qxcc-defs.h>
#include <qxcc-inter.h>
#include <qxcc-types.h>

#define M_TEXT(___v, ___t) \
    ((void *) ((char *) (___v) + (___t)))

#define last_index(index, count) \
    ((index) == (count) - 1 ? 1 : 0)

#define qxrate(qxrate)			((struct qxrate_type *) (qxrate))

#define set_list(list, lock, bare) \
    {									      \
	(list) = P_LIST(Q_NODE((bare))) + sizeof(unsigned);		      \
	if ((list) ^ sizeof(unsigned)) {				      \
	    (list) =							      \
		Q_NODE(M_TEXT((bare), (list))) ^ ((1 << lock) - 3);	      \
	}								      \
    }

int
_libx1f4l2_long_qxrate(void *qxrate, unsigned violet, unsigned trans,
		       unsigned level, unsigned index, unsigned count,
		       unsigned upper, unsigned epslip, unsigned *lpnode,
		       void *node)
{
    int status = 0;
    unsigned line;

    line = qxrate(qxrate)->link_a.fplist.line;

    if (level < count) {
	unsigned esnode;

	esnode = L_NEXT(epslip);
	if (esnode < violet) {
	    if (upper ^ line * 3 >> 1) {
		unsigned ennode, mall, side;
		void *fpnode, *fqnode, *lsnode;

		mall = qxrate(qxrate)->link_a.fplist.mall;

		fpnode = *(void **) M_TEXT(node, L_LAST(epslip));

		lsnode = M_TEXT(node, esnode);

		ennode = Q_NODE(lsnode);

		fqnode = *(void **) M_TEXT(node, L_LAST(ennode));

		fpnode = M_TEXT(fpnode, level * mall);
		level = count - level - 1;
		memmove(fpnode, M_TEXT(fpnode, mall), level * mall);
		fpnode = M_TEXT(fpnode, level * mall);
		lsnode = M_TEXT(lsnode, sizeof(unsigned));
		side = Q_NODE(lsnode);
		if (side << 1 ^ line) {
		    unsigned seek;

		    seek = side - (line >> 1);
		    if (2 < seek) {
			seek++;
			seek >>= 1;
			memcpy(fpnode, fqnode, seek * mall);
			count += seek;
			side -= seek;
			memmove(fqnode, M_TEXT(fqnode, seek * mall),
				side * mall);
			count--;
		    } else {
			memcpy(fpnode, fqnode, mall);
			side--;
			memmove(fqnode, M_TEXT(fqnode, mall), side * mall);
		    }

		    S_NODE(lsnode, side);
		} else {
		    unsigned seek;
		    void *ftnode;

		    ftnode = *(void **) M_TEXT(node, L_NEXT(ennode));

		    seek = upper - (line * 3 >> 1);
		    if (2 < seek) {
			unsigned mail;

			mail = seek / 3;
			seek <<= 1;
			seek /= 3;
			memcpy(fpnode, fqnode, mall * mail);
			count--;
			count += mail;
			line >>= 1;
			memmove(fqnode, M_TEXT(fqnode, mall * mail),
				(line - mail) * mall);
			memcpy(M_TEXT(fqnode, (line - mail) * mall), ftnode,
			       mall * seek);
			memmove(ftnode, M_TEXT(ftnode, mall * seek),
				(upper - (line << 1) - seek) * mall);

			S_NODE(lsnode, line - mail + seek);
		    } else {
			memcpy(fpnode, fqnode, mall);
			line >>= 1;
			memmove(fqnode, M_TEXT(fqnode, mall),
				(line - 1) * mall);
			memcpy(M_TEXT(fqnode, (line - 1) * mall), ftnode,
			       mall);
			memmove(ftnode, M_TEXT(ftnode, mall),
				(upper - (line << 1) - 1) * mall);
		    }
		}

		*lpnode = count;
	    } else {
		status = _libx1f4l2_beta_qxrate
		    (qxrate, 0, node, last_index(index, count), trans, &index,
		     &node);
	    }
	} else {
	    if (upper ^ line) {
		unsigned mall, seek;
		void *fpnode, *fqnode;

		mall = qxrate(qxrate)->link_a.fplist.mall;

		fpnode = *(void **) M_TEXT(node, L_LAST(epslip));
		fqnode = *(void **) M_TEXT(node, L_NEXT(epslip));

		fpnode = M_TEXT(fpnode, level * mall);
		level = count - level - 1;
		memmove(fpnode, M_TEXT(fpnode, mall), level * mall);
		fpnode = M_TEXT(fpnode, level * mall);
		seek = upper - line;
		if (2 < seek) {
		    seek++;
		    seek >>= 1;
		    memcpy(fpnode, fqnode, seek * mall);
		    count += seek;
		    memmove(fqnode, M_TEXT(fqnode, seek * mall),
			    (upper - count) * mall);
		    count--;
		} else {
		    memcpy(fpnode, fqnode, mall);
		    memmove(fqnode, M_TEXT(fqnode, mall),
			    (upper - count - 1) * mall);
		}

		*lpnode = count;
	    } else {
		unsigned list;

		set_list(list, qxrate(qxrate)->link_a.fplist.lock, node);
		if (list) {
		    status = _libx1f4l2_beta_qxrate
			(qxrate, 0, node, last_index(index, count), trans,
			 &index, &node);
		} else {
		    status = _libx1f4l2_trap_qxrate(qxrate, trans);
		}
	    }
	}
    } else {
	unsigned esnode;

	esnode = L_LAST(epslip);
	if (esnode < violet) {
	    if (upper ^ line * 3 >> 1) {
		unsigned ennode, mall, post, side;
		void *fpnode, *fqnode, *lsnode;

		mall = qxrate(qxrate)->link_a.fplist.mall;

		fpnode = *(void **) M_TEXT(node, L_NEXT(epslip));

		lsnode = M_TEXT(node, esnode);

		ennode = Q_NODE(lsnode);

		fqnode = *(void **) M_TEXT(node, L_NEXT(ennode));

		lsnode = M_TEXT(lsnode, sizeof(unsigned));
		side = upper - count - Q_NODE(lsnode);
		if (side << 1 ^ line) {
		    unsigned seek;

		    seek = side - (line >> 1);
		    if (2 < seek) {
			seek++;
			seek >>= 1;
			memmove(M_TEXT(fpnode, (seek + index) * mall),
				M_TEXT(fpnode, (index + 1) * mall),
				(count - index - 1) * mall);
			memmove(M_TEXT(fpnode, seek * mall), fpnode,
				index * mall);
			post = upper - count - seek;
			memcpy(fpnode, M_TEXT(fqnode, (side - seek) * mall),
			       seek * mall);
		    } else {
			memmove(M_TEXT(fpnode, mall), fpnode, index * mall);
			post = upper - count - 1;
			memcpy(fpnode, M_TEXT(fqnode, (side - 1) * mall),
			       mall);
		    }
		} else {
		    unsigned seek;
		    void *ftnode;

		    ftnode = *(void **) M_TEXT(node, L_LAST(ennode));

		    seek = upper - (line * 3 >> 1);
		    if (2 < seek) {
			unsigned mail;

			mail = seek / 3;
			seek <<= 1;
			seek /= 3;
			line >>= 1;
			memmove(M_TEXT(fpnode, (index + mail) * mall),
				M_TEXT(fpnode, (index + 1) * mall),
				(line - index - 1) * mall);
			memmove(M_TEXT(fpnode, mall * mail), fpnode,
				index * mall);
			memcpy(fpnode, M_TEXT(fqnode, (line - mail) * mall),
			       mail * mall);
			memmove(M_TEXT(fqnode, seek * mall), fqnode,
				(line - mail) * mall);
			side = upper - (line << 1) - seek;
			memcpy(fqnode, M_TEXT(ftnode, side * mall),
			       seek * mall);

			post = upper - line - mail;
		    } else {
			memmove(M_TEXT(fpnode, mall), fpnode, index * mall);
			line >>= 1;
			memcpy(fpnode, M_TEXT(fqnode, (line - 1) * mall),
			       mall);
			memmove(M_TEXT(fqnode, mall), fqnode,
				(line - 1) * mall);
			side = upper - (line << 1) - 1;
			memcpy(fqnode, M_TEXT(ftnode, side * mall), mall);

			post = upper - line - 1;
		    }

		    S_NODE(lsnode, side);
		}

		*lpnode = post;
	    } else {
		status = _libx1f4l2_beta_qxrate
		    (qxrate, 0, node, last_index(index, count), trans, &index,
		     &node);
	    }
	} else {
	    if (upper ^ line) {
		unsigned mall, post, seek;
		void *fpnode, *fqnode;

		mall = qxrate(qxrate)->link_a.fplist.mall;

		fpnode = *(void **) M_TEXT(node, L_NEXT(epslip));
		fqnode = *(void **) M_TEXT(node, L_LAST(epslip));

		seek = upper - line;
		if (2 < seek) {
		    seek++;
		    seek >>= 1;
		    memmove(M_TEXT(fpnode, (seek + index) * mall),
			    M_TEXT(fpnode, (index + 1) * mall),
			    (count - index - 1) * mall);
		    memmove(M_TEXT(fpnode, seek * mall), fpnode, index * mall);
		    post = upper - count - seek;
		    memcpy(fpnode, M_TEXT(fqnode, post * mall), seek * mall);
		} else {
		    memmove(M_TEXT(fpnode, mall), fpnode, index * mall);
		    post = upper - count - 1;
		    memcpy(fpnode, M_TEXT(fqnode, post * mall), mall);
		}

		*lpnode = post;
	    } else {
		unsigned list;

		set_list(list, qxrate(qxrate)->link_a.fplist.lock, node);
		if (list) {
		    status = _libx1f4l2_beta_qxrate
			(qxrate, 0, node, last_index(index, count), trans,
			 &index, &node);
		} else {
		    status = _libx1f4l2_trap_qxrate(qxrate, trans);
		}
	    }
	}
    }

    return status;
}
