/*
 * s3xv.4.c
 * Copyright (C) 2009-2010, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <string.h>

#include <s3xx-defs.h>
#include <s3xx-inter.h>
#include <s3xx-names.h>
#include <s3xx-types.h>
#include <trans.h>

#define back(miss)			((const char *) (miss))

#define s3deck(s3deck)			((struct s3deck_type *) (s3deck))

static int back_aime(struct s3deck_type *, struct fpnode_type *, unsigned);
static int back_node(struct s3deck_type *, struct fpnode_type *);
static int copy_aime(struct s3deck_type *, struct fpnode_type *,
		     const struct fpnode_type *);
static int copy_back(struct s3deck_type *, struct fpnode_type *,
		     const struct fpnode_type *);
static int copy_deck(struct s3deck_type *, struct s3deck_type *);
static int fast_deck(struct s3deck_type *, struct s3deck_type *);

static int
back_aime(struct s3deck_type *s3deck_data, struct fpnode_type *fpnode_data,
	  unsigned spread)
{
    while (spread) {
	spread--;
	fpnode_data--;
	back_node(s3deck_data, fpnode_data);
    }

    return -1;
}


static int
back_node(struct s3deck_type *s3deck_data, struct fpnode_type *fpnode_data)
{
    if (1) {
	byte *data;

	data = fpnode_data->base;
	if (data) {
	    s3deck_data->link_m.free(s3deck_data->link_m.data, data);
	}
    }
    if (1) {
	unsigned bits;

	bits = fpnode_data->bits;
	if (Extant(bits)) {
	    unsigned spread;
	    void *data;

	    data = (void *) fpnode_data->data;

	    spread = Spread(bits);

	    back_aime
		(s3deck_data, (struct fpnode_type *) data + spread, spread);

	    s3deck_data->link_m.free(s3deck_data->link_m.data, data);
	}
    }

    return -1;
}


static int
copy_aime(struct s3deck_type *s3deck_data, struct fpnode_type *fpnode_data,
	  const struct fpnode_type *fpnode_fast)
{
    int status;
    unsigned bits;

    bits = fpnode_fast->bits;

    if (Extant(bits)) {
	unsigned spread;
	void *data;

	spread = Spread(bits);

	status = s3deck_data->link_m.link
	    (s3deck_data->link_m.data, &data,
	     spread * sizeof(struct fpnode_type));
	if (status) {
	    status = LINK_ERROR;
	} else {
	    struct fpnode_type *fpnode_slip;
	    const struct fpnode_type *fpnode_text;

	    fpnode_slip = data;
	    fpnode_text = fpnode_fast->data;

	    while (1) {
		status = copy_back(s3deck_data, fpnode_slip, fpnode_text);
		if (status) {
		    break;
		} else {
		    spread--;
		    if (spread) {
			fpnode_slip++;
			fpnode_text++;
		    } else {
			break;
		    }
		}
	    }

	    if (status) {
		back_aime
		    (s3deck_data, fpnode_slip,
		     fpnode_slip - (struct fpnode_type *) data);

		s3deck_data->link_m.free(s3deck_data->link_m.data, data);
	    } else {
		fpnode_data->data = data;
	    }
	}
    } else {
	fpnode_data->data = fpnode_fast->data;
	if (1) {
	    status = 0;
	}
    }

    return status;
}


static int
copy_back(struct s3deck_type *s3deck_data, struct fpnode_type *fpnode_data,
	  const struct fpnode_type *fpnode_fast)
{
    byte *base;
    int status;
    unsigned bits, length;

    bits = fpnode_fast->bits;

    base = fpnode_fast->base;

    length = Length(bits);

    fpnode_data->bits = bits;

    if (length) {
	void *data;

	if (length < Record) {
	} else {
	    length += strlen(back(base) + Record);
	}

	length++;

	status = s3deck_data->link_m.link
	    (s3deck_data->link_m.data, &data, length);
	if (status) {
	    status = LINK_ERROR;
	} else {
	    status = copy_aime(s3deck_data, fpnode_data, fpnode_fast);
	    if (status) {
		s3deck_data->link_m.free(s3deck_data->link_m.data, data);
	    } else {
		fpnode_data->base = data;
		memcpy(data, base, length);
	    }
	}
    } else {
	status = copy_aime(s3deck_data, fpnode_data, fpnode_fast);
	if (status) {
	} else {
	    fpnode_data->base = NULL;
	}
    }

    return status;
}


static int
copy_deck(struct s3deck_type *s3deck_data, struct s3deck_type *s3deck_fast)
{
    int status;
    struct fpnode_type *fpnode_data;

    fpnode_data = s3deck_fast->link_a.fpdeck.fpnode_data;
    if (fpnode_data) {
	void *data;

	status = s3deck_data->link_m.link
	    (s3deck_data->link_m.data, &data, sizeof(struct fpnode_type));
	if (status) {
	    status = LINK_ERROR;
	} else {
	    status = copy_back(s3deck_data, data, fpnode_data);
	    if (status) {
		s3deck_data->link_m.free(s3deck_data->link_m.data, data);
	    } else {
		s3deck_data->link_a.fpdeck.fpnode_data = data;
	    }
	}
    } else {
	status = 0;
    }

    return status;
}


static int
fast_deck(struct s3deck_type *s3deck_data, struct s3deck_type *s3deck_fast)
{
    struct screen_type screen;
    struct trans_type trans;

    trans.data = s3deck_fast->link_m.data;
    trans.free = s3deck_fast->link_m.free;
    trans.link = s3deck_fast->link_m.link;
    trans.mode = s3deck_fast->link_m.mode;

    screen.trans_data = &trans;

    return _libx1f4l2_fast_s3deck(s3deck_data, TRANS_MASK, &screen);
}


int
x1f4_copy_s3deck(void **z3deck, void *s3deck)
{
    int status;
    void *data;

    status = s3deck(s3deck)->link_m.link
	(s3deck(s3deck)->link_m.data, &data, sizeof(struct s3deck_type));
    if (status) {
	status = LINK_ERROR;
    } else {
	fast_deck(data, s3deck);

	status = copy_deck(data, s3deck);
	if (status) {
	    s3deck(s3deck)->link_m.free(s3deck(s3deck)->link_m.data, data);
	} else {
	    *z3deck = data;
	}
    }

    return status;
}
