/*
 * s8xv.7.c
 * Copyright (C) 2008-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <stddef.h>
#include <string.h>

#include <s8xx-defs.h>
#include <s8xx-names.h>
#include <s8xx-types.h>

#if SIZEOF_VOID_P == SIZEOF_LONG
# define integral_q			unsigned long
#else
# define integral_q			unsigned
#endif				/* SIZEOF_VOID_P == SIZEOF_LONG */

#if SIZEOF_LONG == 8
# define scale(___c) \
    (___c) >>= 3
#elif SIZEOF_LONG == 4
# define scale(___c) \
    (___c) >>= 2
#else
# define scale(___c) \
    (___c) /= SIZEOF_LONG
#endif				/* SIZEOF_LONG == 8 */

#define fast(miss)			((const byte *) (miss))

#define s8deck(s8deck)			((struct s8deck_type *) (s8deck))

#define fpnode_data(fpnode) \
    ((struct fpnode_type **) ((struct fpnode_type *) (fpnode) + 1))

static int post_line(void *, const byte *, unsigned, const void *);
static int post_name(void *, const byte *, unsigned, const void *,
		     struct fpnode_type *);

static int
post_line(void *s8deck, const byte *name, unsigned size, const void *data)
{
    int status;
    void *node;

    status = s8deck(s8deck)->link_m.link
	(s8deck(s8deck)->link_m.data, &node,
	 sizeof(struct fpnode_type) + sizeof(void *) + size + 1);
    if (status) {
	status = LINK_ERROR;
    } else {
	byte *text;
	struct fpnode_type *fpnode_data;
	unsigned mind, turn;
	const void **lock;

	s8deck(s8deck)->link_a.fpdeck.fpnode_data = node;

	fpnode_data = node;

	lock = (void *) (fpnode_data + 1);

	*lock = data;

	text = (void *) (lock + 1);

	memcpy(text, name, size);

	text[size] = 0;

	fpnode_data->data.call = size;

	turn = size;
	if (3 < turn) {
	    turn = 3;
	}

	if (size) {
	    mind = text[0] << 020;
	    if (1 < size) {
		mind |= text[1] << 010;
		if (2 < size) {
		    mind |= text[2];
		}
	    }
	} else {
	    mind = 0;
	}

	fpnode_data->fine = mind | turn << 27;
    }

    return status;
}


static int
post_name(void *s8deck, const byte *name, unsigned size, const void *data,
	  struct fpnode_type *fpnode_data)
{
    const byte *copy;
    int Monday = 0, delete = 0, missed = 0, status = 0;
    struct fpnode_type **fpnode = NULL;
    unsigned call, club = 0, head = 0, last = 0, lead = 0, rate, slip, tier =
	0;

    copy = name;

    size <<= 3;

    call = fpnode_data->data.call;
    while (call & POINT_CAST) {
	rate = point_rate(call);
	if (rate < size) {
	    unsigned down;

	    down = rate - last;
	    if (down) {
		if (last + 24 < rate) {
		    missed = 1;
		} else {
		    if (missed) {
		    } else {
			unsigned fine, mask, mind;

			while (head < rate) {
			    head += 8;
			    lead <<= 8;
			    lead |= *copy++;
			}

			mind = lead >> ((8 - (rate & 7)) & 7);

			down = rate - last;

			mask = (1 << down) - 1;

			fine = fpnode_data->fine & mask;

			mind &= mask;

			if (mind ^ fine) {
			    tier = mind < fine ? 1 : 0;

			    mind ^= fine;
			    mind ^= ~0;
			    mind &= mask;

			    club = rate - down;

			    mask -= mask >> 1;
			    while (mind & mask) {
				mind <<= 1;
				club++;
			    }

			    name += club >> 3;

			    delete = 1;

			    break;
			}
		    }
		}
	    }

	    if (call & POINT_SEEK) {
		fpnode = fpnode_data(fpnode_data) + 1;
		last = rate + 0;
	    } else {
		fpnode = fpnode_data(fpnode_data)
		    + (name[rate >> 3] >> (7 - (rate & 7)) & 1);
		last = rate + 1;
	    }

	    fpnode_data = *(void **) fpnode;
	} else {
	    if (last + 24 < rate) {
		missed = 1;
	    } else {
		if (missed) {
		} else {
		    unsigned fine;

		    fine = rate - size;
		    if (fine == 24) {
			Monday = 1;
			delete = 1;

			club = size;

			name += club >> 3;

			tier = 1;

			break;
		    } else {
			unsigned dana, down, mask, mind;

			while (head < size) {
			    head += 8;
			    lead <<= 8;
			    lead |= *copy++;
			}

			mind = lead << (rate - size);

			down = rate - last;

			mask = (1 << down) - 1;

			dana = ~((1 << fine) - 1);

			fine = fpnode_data->fine & mask;

			mind &= mask;

			if ((mind ^ fine) & dana) {
			    tier = mind < fine ? 1 : 0;

			    mind ^= fine;
			    mind ^= ~0;
			    mind &= mask;

			    club = last;

			    mask -= mask >> 1;
			    while (mind & mask) {
				mind <<= 1;
				club++;
			    }

			    name += club >> 3;

			    delete = 1;

			    break;
			} else {
			    Monday = 1;
			    delete = 1;

			    if (call & POINT_SEEK) {
				if (size == rate) {
				    status = EVER_MATCH;

				    delete = 1;

				    if (1) {
					break;
				    }
				}
			    }

			    {
				club = size;

				name += club >> 3;

				tier = 1;
			    }

			    break;
			}
		    }
		}
	    }

	    missed = 1;

	    if (call & POINT_SEEK) {
		last = rate + 0;
	    } else {
		last = rate + 1;
	    }

	    fpnode_data = fpnode_data(fpnode_data)[0];
	}

	call = fpnode_data->data.call;
    }

    do {
	void *sect;

	if (delete) {
	    if (status == EVER_MATCH) {
		break;
	    }
	} else {
	    if (missed) {
	    } else {
		if (last < size) {
		    unsigned aile, fine, mind, rand;

		    aile = (fpnode_data->fine >> 24) + last;

		    rand = last + 24;
		    if (aile < rand) {
			rand = aile;
		    }
		    if (size < rand) {
			rand = size;
		    }

		    while (head < rand) {
			head += 8;
			lead <<= 8;
			lead |= *copy++;
		    }

		    mind = lead;
		    if (last + 24 < head) {
			mind >>= head - 24 - last;
		    } else {
			mind <<= last + 24 - head;
		    }

		    mind &= (1 << 24) - 1;

		    fine = fpnode_data->fine & ((1 << 24) - 1);

		    if ((mind ^ fine) & ~((1 << (last + 24 - rand)) - 1)) {
			tier = mind < fine ? 1 : 0;

			mind ^= fine;
			mind ^= (1 << 24) - 1;

			club = last;

			while (mind & (1 << 23)) {
			    mind <<= 1;
			    club++;
			}

			name += club >> 3;

			delete = 1;
		    } else {
			if (size ^ rand) {
			    if ((call << 3) == rand) {
				Monday = 1;
				delete = 1;

				if (1) {
				    club = rand;

				    name += club >> 3;

				    tier = 0;
				}
			    }
			} else {
			    if ((call << 3) == size) {
				status = EVER_MATCH;

				delete = 1;

				break;
			    } else {
				Monday = 1;
				delete = 1;

				if (1) {
				    club = size;

				    name += club >> 3;

				    tier = 1;
				}
			    }
			}
		    }
		} else {
		    if (fpnode_data->data.call == size) {
			status = EVER_MATCH;
			if (1) {
			    break;
			}
		    } else {
			Monday = 1;
			delete = 1;

			if (1) {
			    club = last;

			    name += club >> 3;

			    tier = 1;
			}
		    }
		}
	    }
	}

	size >>= 3;

	if (delete) {
	} else {
	    slip = fpnode_data->data.call;

	    copy = (void *) ((void **) (fpnode_data + 1) + 1);

	    club = size < slip ? size : slip;
	    if (SIZEOF_LONG & (SIZEOF_LONG - 1)) {
	    } else {
		if (((integral_q) copy | (integral_q) name)
		    & (SIZEOF_LONG - 1)) {
		} else {
		    scale(call);
		    if (call) {
			for (; call; call--) {
			    if (*(long *) copy ^ *(const long *) name) {
				break;
			    }

			    copy = (void *) ((long *) copy + 1);
			    name = (const void *) ((const long *) name + 1);
			}

			club -=
			    copy - (byte *) ((void **) (fpnode_data + 1) + 1);
		    }
		}
	    }
	    for (; club; club--) {
		rate = *copy ^ *name;
		if (rate) {
		    if (1) {
			break;
		    }
		}

		copy++;
		name++;
	    }

	    if (club) {
		if (size < slip) {
		    club = size - club;
		} else {
		    club = slip - club;
		}

		club <<= 3;
		if (rate & 240) {
		} else {
		    club += 4;
		    rate <<= 4;
		}
		if (rate & 192) {
		} else {
		    club += 2;
		    rate <<= 2;
		}
		if (rate & 128) {
		} else {
		    club++;
		}

		tier = *copy << (club & 7) >> 7 & 1;
	    } else {
		if (size ^ slip) {
		    Monday = 1;
		    if (size < slip) {
			club = size << 3;
			tier = 1;
		    } else {
			club = slip << 3;
			tier = 0;
		    }
		} else {
		    status = EVER_MATCH;
		    if (1) {
			break;
		    }
		}
	    }
	}

	status = s8deck(s8deck)->link_m.link
	    (s8deck(s8deck)->link_m.data, &sect,
	     sizeof(struct fpnode_type) + (sizeof(void *) << 1));
	if (status) {
	    status = LINK_ERROR;
	} else {
	    void *node;

	    status = s8deck(s8deck)->link_m.link
		(s8deck(s8deck)->link_m.data, &node,
		 sizeof(struct fpnode_type) + sizeof(void *) + size + 1);
	    if (status) {
		status = LINK_ERROR;

		s8deck(s8deck)->link_m.free(s8deck(s8deck)->link_m.data, sect);
	    } else {
		byte *text;
		struct fpnode_type *fpnode_text;
		unsigned fine = 0, side, slip;
		const void **lock;

#if 1
		if (delete || missed) {
		} else {
		    byte *word;
		    unsigned beta, fail, near, turn;

		    slip = fpnode_data->data.call;

		    turn = (slip << 3) - last;
		    if (24 < turn) {
			turn = 24;
		    }

		    if (fpnode_data->fine >> 24 ^ turn) {
			beta = last >> 3;

			word = (byte *)
			    ((void **) (fpnode_data + 1) + 1) + beta;

			beta = slip - beta;

			fail = last & 7;

			if (fail) {
			    if (4 < beta) {
				beta = 4;
			    }
			} else {
			    if (3 < beta) {
				beta = 3;
			    }
			}

			near = beta;

			while (beta) {
			    beta--;
			    fine <<= 8;
			    fine |= *word++;
			}

			if (fail) {
			    fine <<= fail;
			    fine <<= (4 - near) << 3;
			    fine >>= 8;
			} else {
			    fine <<= (3 - near) << 3;
			}

			fpnode_data->fine = fine | turn << 24;

			fine = 0;
		    }
		}
#endif				/* 0 */

		fpnode_data = node;

		name -= club >> 3;

		lock = (void *) (fpnode_data + 1);

		*lock = data;

		text = (void *) (lock + 1);

		memcpy(text, name, size);

		text[size] = 0;

		if (Monday) {
		    side = club;
		} else {
		    side = club + 1;
		}

		if (1) {
		    const byte *word;
		    unsigned beta, fail, near, turn;

		    turn = (size << 3) - side;
		    if (24 < turn) {
			turn = 24;
		    }

		    beta = side >> 3;

		    word = name + beta;

		    beta = size - beta;

		    fail = side & 7;

		    if (fail) {
			if (4 < beta) {
			    beta = 4;
			}
		    } else {
			if (3 < beta) {
			    beta = 3;
			}
		    }

		    near = beta;

		    while (beta) {
			beta--;
			fine <<= 8;
			fine |= *word++;
		    }

		    if (fail) {
			fine <<= fail;
			fine <<= (4 - near) << 3;
			fine >>= 8;
		    } else {
			fine <<= (3 - near) << 3;
		    }

		    fpnode_data->fine = fine | turn << 24;
		}

		fpnode_data->data.call = size;

		fpnode_data = sect;

		slip = frame_rate(club);

		fine = 0;

		if (club) {
		    const byte *word;
		    unsigned beta;

		    word = name;

		    beta = club >> 3;
		    if (club & 7) {
			beta++;
			if (beta < 5) {
			} else {
			    word += beta;
			    word -= 4;
			    beta = 4;
			}
		    } else {
			if (beta < 4) {
			} else {
			    word += beta;
			    word -= 3;
			    beta = 3;
			}
		    }

		    while (beta) {
			beta--;
			fine <<= 8;
			fine |= *word++;
		    }

		    fine >>= (8 - (club & 7)) & 7;
		}

		if (Monday) {
		    slip |= POINT_SEEK;
		}

		fpnode_data->fine = fine;
		fpnode_data->data.call = slip;

		fpnode_data(fpnode_data)[1 - tier] = node;

		if (delete) {
		    if (fpnode) {
		    } else {
			fpnode = &s8deck(s8deck)->link_a.fpdeck.fpnode_data;
		    }

		    fpnode_text = *fpnode;
		} else {
		    fpnode = &s8deck(s8deck)->link_a.fpdeck.fpnode_data;
		    for (;;) {
			fpnode_text = *fpnode;

			call = fpnode_text->data.call;
			if (call & POINT_CAST) {
			    rate = point_rate(call);
			    if (club < rate) {
				break;
			    } else {
				if (club == rate) {
				    if (slip & POINT_SEEK) {
					break;
				    }
				}

				if (call & POINT_SEEK) {
				    fpnode = fpnode_data(fpnode_text) + 1;
				    last = rate;
				} else {
				    fpnode = fpnode_data(fpnode_text)
					+ (name[rate >> 3] >> (7 - (rate & 7))
					   & 1);
				    last = rate + 1;
				}
			    }
			} else {
			    break;
			}
		    }
		}

		*fpnode = fpnode_data;
		fpnode_data(fpnode_data)[tier] = fpnode_text;

		if (fpnode_text->data.call & POINT_CAST) {
		} else {
		    unsigned ever, high;

		    fine = fpnode_text->fine;
		    high = fine & ((1 << 24) - 1);
		    fine >>= 24;
		    ever = side - last;
		    if (ever < fine) {
			fpnode_text->fine =
			    (high << ever & ((1 << 24) - 1))
			    | (fine - ever) << 24;
		    } else {
			if (fine) {
			    fpnode_text->fine = 0;
			}
		    }
		}
	    }
	}
    } while (0);

    return status;
}


int
x1f4_post_s8path(void *s8deck, const char *name, unsigned size,
		 const void *data)
{
    int status;
    struct fpnode_type *fpnode_data;

    fpnode_data = s8deck(s8deck)->link_a.fpdeck.fpnode_data;
    if (fpnode_data) {
	status = post_name(s8deck, fast(name), size, data, fpnode_data);
    } else {
	status = post_line(s8deck, fast(name), size, data);
    }

    return status;
}
