/*
 * scdeck.h
 * Copyright (C) 2008-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __LIBx1f4_SCDECK_H__
#define __LIBx1f4_SCDECK_H__

#include <errors.h>

/*
 * attribute bit definitions
 */
#define X1f4_SCDECK_TRANS_MASK		(1u << 0x00)

/*
 * error definitions
 */
#define X1f4_SCDECK_DECK_ERROR		LIBx1f4L2_DECK_ERROR
#define X1f4_SCDECK_FREE_ERROR		LIBx1f4L2_FREE_ERROR
#define X1f4_SCDECK_LINK_ERROR		LIBx1f4L2_LINK_ERROR
#define X1f4_SCDECK_MODE_ERROR		LIBx1f4L2_MODE_ERROR
#define X1f4_SCDECK_SIZE_ERROR		LIBx1f4L2_SIZE_ERROR

/*
 * error definitions (continued)
 */
#define X1f4_SCDECK_EVER_MATCH		LIBx1f4L2_EVER_MATCH
#define X1f4_SCDECK_MISS_CLASS		LIBx1f4L2_MISS_CLASS

typedef struct x1f4_scdeck_type {
    void *trans;
} x1f4_scdeck_type;

#if !defined __x1f4_trans_type_defined__
typedef struct x1f4_trans_type {
    int (*free) (void *, void *), (*link) (void *, void **, unsigned),
	(*mode) (void *, void **, unsigned);
    void *context;
} x1f4_trans_type;
#endif				/* !__x1f4_trans_type_defined__ */

#if !defined __x1f4_trans_type_defined__
# define __x1f4_trans_type_defined__
#endif				/* !__x1f4_trans_type_defined__ */

extern int x1f4_bind_scdeck(void *, const char *, const void ***);
extern int x1f4_case_scdeck(void *);
extern int x1f4_cast_scdeck(void *, const char *);
extern int x1f4_copy_scdeck(void **, void *);
extern int x1f4_ever_scdeck(void *, const void **);
extern int x1f4_fast_scdeck(void *, unsigned, struct x1f4_scdeck_type *);
extern int x1f4_fgvq_scdeck
    (void *, const char *, const void ***, char **, unsigned *, unsigned *);
extern int x1f4_fgxd_scdeck(void *, const char *, const void ***);
extern int x1f4_find_scdeck(void *, const char *, const void **);
extern int x1f4_find_scpath(void *, const char *, unsigned, const void **);
extern int x1f4_fini_scdeck(void **);
extern int x1f4_flat_scdeck(void *);
extern int x1f4_high_scdeck(void *);
extern int x1f4_hook_scdeck
    (void *, void *, int (*) (void *, const void *, const char *, unsigned),
     char *);
extern int x1f4_init_scdeck(void **, unsigned, struct x1f4_scdeck_type *);
extern int x1f4_jack_scdeck(void *, void *);
extern int x1f4_lead_scdeck(void *, const void **);
extern int x1f4_lift_scdeck
    (void *, const void **, char **, unsigned *, unsigned *);
extern int x1f4_lime_scdeck(void *, void *, int (*) (void *, const void *));
extern int x1f4_line_scdeck
    (void *, void *, int (*) (void *, const void **, const void *),
     int (*) (void *, const void *), void *);
extern int x1f4_list_scdeck
    (void *, void *, int (*) (void *, const char *, unsigned), char *);
extern int x1f4_llvq_scdeck
    (void *, const char *, const void ***, char **, unsigned *, unsigned *);
extern int x1f4_llxd_scdeck(void *, const char *, const void ***);
extern int x1f4_mind_scdeck(void *);
extern int x1f4_pass_scdeck(void *, const char *);
extern int x1f4_pass_scpath(void *, const char *, unsigned);
extern int x1f4_pick_scdeck(void *, const char *, const void ***);
extern int x1f4_pick_scpath(void *, const char *, unsigned, const void ***);
extern int x1f4_plot_scdeck
    (void *, const void **, char **, unsigned *, unsigned *);
extern int x1f4_post_scdeck(void *, const char *, const void *);
extern int x1f4_post_scpath(void *, const char *, unsigned, const void *);
extern int x1f4_skew_scdeck(void *, void *);
extern int x1f4_slip_scdeck(void *);
extern int x1f4_text_scdeck
    (void **, void *, int (*) (void *, const void **, const void *),
     int (*) (void *, const void *), void *);

#endif				/* __LIBx1f4_SCDECK_H__ */
