/*
 * scxi.a.c
 * Copyright (C) 2009-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>

#include <scxx-defs.h>
#include <scxx-inter.h>
#include <scxx-names.h>
#include <scxx-types.h>
#include <trans.h>

#define back(miss)			((const char *) (miss))

#define null				(&scdeck_data->link_i.zero)

#define scdeck(scdeck)			((struct scdeck_type *) (scdeck))

static int back_aime(struct scdeck_type *, struct fpnode_type *, unsigned,
		     int (*) (void *, const void *), void *);
static int back_node(struct scdeck_type *, struct fpnode_type *,
		     int (*) (void *, const void *), void *);
static int copy_aime(struct scdeck_type *, struct fpnode_type *,
		     const struct fpnode_type *,
		     int (*) (void *, const void **, const void *),
		     int (*) (void *, const void *), void *);
static int copy_back(struct scdeck_type *, struct fpnode_type *,
		     const struct fpnode_type *,
		     int (*) (void *, const void **, const void *),
		     int (*) (void *, const void *), void *);

static int
back_aime(struct scdeck_type *scdeck_data, struct fpnode_type *fpnode_data,
	  unsigned spread, int (*free) (void *, const void *), void *context)
{
    while (spread) {
	spread--;
	fpnode_data--;
	back_node(scdeck_data, fpnode_data, free, context);
    }

    return -1;
}


static int
back_node(struct scdeck_type *scdeck_data, struct fpnode_type *fpnode_data,
	  int (*free) (void *, const void *), void *context)
{
    if (1) {
	byte *data;

	data = fpnode_data->base;
	if (data) {
	    scdeck_data->link_m.free(scdeck_data->link_m.data, data);
	}
    }
    if (1) {
	unsigned bits;

	bits = fpnode_data->bits;
	if (Extant(bits)) {
	    unsigned spread;
	    void *data;

	    data = (void *) fpnode_data->data;

	    spread = Spread(bits);

	    back_aime
		(scdeck_data, (struct fpnode_type *) data + spread, spread,
		 free, context);

	    scdeck_data->link_m.free(scdeck_data->link_m.data, data);
	} else {
	    free(context, fpnode_data->data);
	}
    }

    return -1;
}


static int
copy_aime(struct scdeck_type *scdeck_data, struct fpnode_type *fpnode_data,
	  const struct fpnode_type *fpnode_fast,
	  int (*copy) (void *, const void **, const void *),
	  int (*free) (void *, const void *), void *context)
{
    int status;
    unsigned bits;

    bits = fpnode_fast->bits;

    if (Extant(bits)) {
	unsigned spread;
	void *data;

	spread = Spread(bits);

	status = scdeck_data->link_m.link
	    (scdeck_data->link_m.data, &data,
	     ((spread + 3) & ~3) * sizeof(struct fpnode_type));
	if (status) {
	    status = LINK_ERROR;
	} else {
	    struct fpnode_type *fpnode_slip;
	    const struct fpnode_type *fpnode_text;

	    fpnode_slip = data;
	    fpnode_text = fpnode_fast->data;

	    while (1) {
		status = copy_back
		    (scdeck_data, fpnode_slip, fpnode_text, copy, free,
		     context);
		if (status) {
		    break;
		} else {
		    spread--;
		    if (spread) {
			fpnode_slip++;
			fpnode_text++;
		    } else {
			break;
		    }
		}
	    }

	    if (status) {
		back_aime
		    (scdeck_data, fpnode_slip,
		     fpnode_slip - (struct fpnode_type *) data, free, context);

		scdeck_data->link_m.free(scdeck_data->link_m.data, data);
	    } else {
		fpnode_data->data = data;
	    }
	}
    } else {
	status = copy(context, &fpnode_data->data, fpnode_fast->data);
	if (status) {
	    status = CALL_ERROR;
	}
    }

    return status;
}


static int
copy_back(struct scdeck_type *scdeck_data, struct fpnode_type *fpnode_data,
	  const struct fpnode_type *fpnode_fast,
	  int (*copy) (void *, const void **, const void *),
	  int (*free) (void *, const void *), void *context)
{
    byte *base;
    int status;
    unsigned bits, length;

    bits = fpnode_fast->bits;

    base = fpnode_fast->base;

    length = Length(bits);

    fpnode_data->bits = bits;

    if (length) {
	void *data;

	if (length < Record) {
	} else {
	    length += strlen(back(base) + Record);
	}

	length++;

	status = scdeck_data->link_m.link
	    (scdeck_data->link_m.data, &data, length);
	if (status) {
	    status = LINK_ERROR;
	} else {
	    status = copy_aime
		(scdeck_data, fpnode_data, fpnode_fast, copy, free, context);
	    if (status) {
		scdeck_data->link_m.free(scdeck_data->link_m.data, data);
	    } else {
		fpnode_data->base = data;
		memcpy(data, base, length);
	    }
	}
    } else {
	status = copy_aime
	    (scdeck_data, fpnode_data, fpnode_fast, copy, free, context);
	if (status) {
	} else {
	    fpnode_data->base = null;
	}
    }

    return status;
}


int
_libx1f4l2_rail_scdeck(void *scdata, void *sctext,
		       int (*copy) (void *, const void **, const void *),
		       int (*free) (void *, const void *), void *context)
{
    int status;
    struct fpnode_type *fpnode_data;

    fpnode_data = scdeck(sctext)->link_a.fpdeck.fpnode_data;
    if (fpnode_data) {
	void *data;

	status = scdeck(scdata)->link_m.link
	    (scdeck(scdata)->link_m.data, &data, sizeof(struct fpnode_type));
	if (status) {
	    status = LINK_ERROR;
	} else {
	    status = copy_back(scdata, data, fpnode_data, copy, free, context);
	    if (status) {
		scdeck(scdata)->link_m.free(scdeck(scdata)->link_m.data, data);
	    } else {
		scdeck(scdata)->link_a.fpdeck.fpnode_data = data;
	    }
	}
    } else {
	status = 0;
    }

    return status;
}
