/*
 * scxi.w.c
 * Copyright (C) 2009-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>

#include <scxx-defs.h>
#include <scxx-names.h>
#include <scxx-types.h>

#define back(miss)			((const char *) (miss))

#define Null(___n) \
    ((___n) == (&scdeck(scdeck)->link_i.zero))

#define null				(&scdeck(scdeck)->link_i.zero)

#define scdeck(scdeck)			((struct scdeck_type *) (scdeck))

static int pass_link(void *, struct fpnode_type *, struct fpnode_type *);
static int pass_name(void *, struct fpnode_type *, struct fpnode_type *);

static int
pass_link(void *scdeck, struct fpnode_type *fpnode_slip,
	  struct fpnode_type *fpnode_data)
{
    byte *base, *lock;
    int prefix, status;
    unsigned bits, length, miss, offset;

    bits = fpnode_slip->bits;

    base = fpnode_slip->base;

    length = Length(bits);
    if (length < Record) {
    } else {
	length += strlen(back(base) + Record);
    }

    if (fpnode_data == fpnode_slip->data) {
	fpnode_data++;
    } else {
	fpnode_data--;
    }

    prefix = Prefix(bits);

    miss = fpnode_data->bits;

    lock = fpnode_data->base;

    if (prefix) {
	if (Prefix(miss)) {
	    offset = Length(miss);
	    if (offset < Record) {
	    } else {
		offset += strlen(back(lock) + Record);
	    }

	    offset++;
	} else {
	    offset = 0;
	}
    } else {
	offset = 0;
    }

    if (offset) {
	unsigned mirror;
	void *copy;

	mirror = length + offset;

	if (Null(base)) {
	    status = scdeck(scdeck)->link_m.link
		(scdeck(scdeck)->link_m.data, &copy, mirror + 1);
	} else {
	    copy = base;
	    status = scdeck(scdeck)->link_m.mode
		(scdeck(scdeck)->link_m.data, &copy, mirror + 1);
	}
	if (status) {
	    if (Null(base)) {
		status = LINK_ERROR;
	    } else {
		status = MODE_ERROR;
	    }
	} else {
	    byte *deck;
	    void *slip;

	    fpnode_slip->base = copy;

	    deck = copy;
	    deck += length;
	    *deck = Prefix(miss);
	    deck++;
	    offset--;
	    if (offset) {
		memcpy(deck, lock, offset);
		if (1) {
		    deck += offset;
		}
	    }

	    *deck = 0;

	    slip = (void *) fpnode_slip->data;

	    fpnode_slip->bits = pset(prefix, mirror, miss);
	    fpnode_slip->data = fpnode_data->data;

	    status = scdeck(scdeck)->link_m.free
		(scdeck(scdeck)->link_m.data, slip);
	    if (status) {
		status = FREE_ERROR;
	    } else {
		if (offset) {
		    status = scdeck(scdeck)->link_m.free
			(scdeck(scdeck)->link_m.data, lock);
		    if (status) {
			status = FREE_ERROR;
		    }
		}
	    }
	}
    } else {
	const void *link;

	link = fpnode_data->data;

	status = scdeck(scdeck)->link_m.free
	    (scdeck(scdeck)->link_m.data, (void *) fpnode_slip->data);
	if (status) {
	    status = FREE_ERROR;
	} else {
	    if (prefix) {
		fpnode_slip->bits = pset(prefix, length, miss);
	    } else {
		fpnode_slip->bits = miss;
		fpnode_slip->base = lock;
	    }

	    fpnode_slip->data = link;
	}
    }

    return status;
}


static int
pass_name(void *scdeck, struct fpnode_type *fpnode_slip,
	  struct fpnode_type *fpnode_data)
{
    int status;
    unsigned bits, size;
    void *list;

    bits = fpnode_slip->bits;

    size = Effect(bits);

    fpnode_slip->bits = iset(bits, size);

    list = (void *) fpnode_slip->data;

    memmove(fpnode_data, fpnode_data + 1,
	    (size - (fpnode_data - (struct fpnode_type *) list))
	    * sizeof(struct fpnode_type));

    if (size & 3) {
	status = 0;
    } else {
	status = scdeck(scdeck)->link_m.mode
	    (scdeck(scdeck)->link_m.data, &list,
	     sizeof(struct fpnode_type) * size);
	if (status) {
	    status = MODE_ERROR;
	} else {
	    fpnode_slip->data = list;
	}
    }

    return status;
}


int
_libx1f4l2_blue_scdeck(void *scdeck, struct fpnode_type *fpnode_slip,
		       struct fpnode_type *fpnode_data)
{
    int status;

    do {
	byte *base;

	base = fpnode_data->base;
	if (Null(base)) {
	} else {
	    status = scdeck(scdeck)->link_m.free
		(scdeck(scdeck)->link_m.data, base);
	    if (status) {
		status = FREE_ERROR;

		break;
	    } else {
		fpnode_data->base = null;
	    }
	}

	if (effect(fpnode_slip) == 1) {
	    status = pass_link(scdeck, fpnode_slip, fpnode_data);
	} else {
	    status = pass_name(scdeck, fpnode_slip, fpnode_data);
	}
    } while (0);

    return status;
}
