/*
 * scxt.y.c
 * Copyright (C) 2009-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>
#include <string.h>

#include <scxx-defs.h>
#include <scxx-names.h>
#include <scxx-patch.h>
#include <scxx-types.h>

#define back(miss)			((const char *) (miss))
#define fast(miss)			((const byte *) (miss))

#define scdeck(scdeck)			((struct scdeck_type *) (scdeck))

static int mode_sail(struct scdeck_type *, void **, unsigned *, unsigned,
		     unsigned, unsigned *, char **);

static int
mode_sail(struct scdeck_type *scdeck_data, void **sail, unsigned *ling,
	  unsigned line, unsigned copy, unsigned *news, char **path)
{
    int status;
    unsigned lane;

    lane = copy + line + 2;
    status = scdeck_data->link_m.mode(scdeck_data->link_m.data, sail, lane);
    if (status) {
	status = MODE_ERROR;
    } else {
	*news = 1;
	*ling = lane;
	*path = (char *) *sail + line;
    }

    return status;
}


int
x1f4_fgvq_scdeck(void *scdeck, const char *name, const void ***data,
		 void **sail, unsigned *high, unsigned *ling)
{
    int status;
    const struct fpnode_type *fpnode_data;

    fpnode_data = scdeck(scdeck)->link_a.fpdeck.fpnode_data;
    if (fpnode_data) {
	byte *slip;
	char *path;
	int e;
	const struct fpnode_type *fpnode_pick = NULL;
	unsigned line = 0, news, pick_line = 0;

	news = *ling;
	path = *sail;

	status = 0;

	slip = fpnode_data->base;
	e = prefix(fpnode_data);
	if (e) {
	    slip--;
	}

	while (1) {
	    int c;

	    c = *fast(name);
	    if (c) {
		name++;
		if (e) {
		    if (c == e) {
			slip++;
			if (1) {
			    e = *slip;
			}
		    } else {
			if (c < e) {
			    pick_line = line;
			    fpnode_pick = fpnode_data;
			}

			break;
		    }
		} else {
		    unsigned bits;

		    bits = fpnode_data->bits;
		    if (Extant(bits)) {
			const struct fpnode_type *fpnode_slip;
			unsigned effect;

			effect = Effect(bits);

			if (Prefix(bits)) {
			    byte *base;
			    unsigned copy;

			    base = fpnode_data->base;
			    copy = Length(bits);
			    if (copy < Record) {
			    } else {
				copy += strlen(back(base) + Record);
			    }

			    if (news < copy + 2) {
				status = mode_sail
				    (scdeck, sail, ling, line, copy, &news,
				     &path);
				if (status) {
				    break;
				}
			    } else {
				news--;
				news -= copy;
			    }

			    *path++ = Prefix(bits);
			    memcpy(path, base, copy);
			    path += copy;
			    line += copy + 1;
			}

			fpnode_slip = fpnode_data->data;
			fpnode_data = fpnode_slip;

			Progress_node(fpnode_data, effect);

			e = prefix(fpnode_data);
			if (e ^ c) {
			    if (e < c) {
				if ((fpnode_data - fpnode_slip) ^ effect) {
				    pick_line = line;
				    fpnode_pick = fpnode_data + 1;
				}
			    } else {
				pick_line = line;
				fpnode_pick = fpnode_data;
			    }

			    break;
			} else {
			    slip = fpnode_data->base;
			    if (1) {
				e = *slip;
				if ((fpnode_data - fpnode_slip) ^ effect) {
				    pick_line = line;
				    fpnode_pick = fpnode_data + 1;
				}
			    }
			}
		    } else {
			break;
		    }
		}
	    } else {
		if (extant(fpnode_data)) {
		    unsigned bits;

		    bits = fpnode_data->bits;
		    if (1) {
			if (Prefix(bits)) {
			    byte *base;
			    unsigned copy;

			    base = fpnode_data->base;
			    copy = Length(bits);
			    if (copy < Record) {
			    } else {
				copy += strlen(back(base) + Record);
			    }

			    if (news < copy + 2) {
				status = mode_sail
				    (scdeck, sail, ling, line, copy, &news,
				     &path);
				if (status) {
				    break;
				}
			    } else {
				news--;
				news -= copy;
			    }

			    *path++ = Prefix(bits);
			    memcpy(path, base, copy);
			    path += copy;
			    line += copy + 1;
			}
		    }

		    pick_line = line;
		    fpnode_pick = fpnode_data->data;
		    if (e) {
		    } else {
			if (prefix(fpnode_pick)) {
			} else {
			    fpnode_pick++;
			}
		    }
		} else {
		    if (e) {
			pick_line = line;
			fpnode_pick = fpnode_data;
		    } else {
		    }
		}

		break;
	    }
	}

	if (fpnode_pick) {
	    if (pick_line ^ line) {
		line -= pick_line;
		news += line;
		path -= line;
		line = pick_line;
	    }

	    while (1) {
		unsigned bits;

		bits = fpnode_pick->bits;
		if (Prefix(bits)) {
		    byte *base;
		    unsigned copy;

		    base = fpnode_pick->base;
		    copy = Length(bits);
		    if (copy < Record) {
		    } else {
			copy += strlen(back(base) + Record);
		    }

		    if (news < copy + 2) {
			status = mode_sail
			    (scdeck, sail, ling, line, copy, &news, &path);
			if (status) {
			    break;
			}
		    } else {
			news--;
			news -= copy;
		    }

		    *path++ = Prefix(bits);
		    memcpy(path, base, copy);
		    path += copy;
		    line += copy + 1;
		}

		if (Extant(bits)) {
		    fpnode_pick = fpnode_pick->data;
		} else {
		    *path = 0;
		    *high = line;
		    *data = (const void **) &fpnode_pick->data;
		    if (1) {
			break;
		    }
		}
	    }
	} else {
	    status = MISS_CLASS;
	}
    } else {
	status = MISS_CLASS;
    }

    return status;
}
