/*
 * sfxx.7.c
 * Copyright (C) 2009-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>

#include <sfxx-defs.h>
#include <sfxx-names.h>
#include <sfxx-patch.h>
#include <sfxx-types.h>

#define back(miss)			((const char *) (miss))
#define fast(miss)			((const byte *) (miss))

#define sfdeck(sfdeck)			((struct sfdeck_type *) (sfdeck))

static int case_deck(void *, struct fpnode_type *, const byte *, unsigned, int,
		     int, const byte *, const void *);
static int ever_deck(void *, struct fpnode_type *, const byte *, const void *);
static int fast_deck(void *, struct fpnode_type **, unsigned *, const void *);
static int line_deck(void *, struct fpnode_type **, struct fpnode_type *,
		     unsigned, unsigned, unsigned *, const byte *,
		     const void *);
static int mind_deck(void *, struct fpnode_type *, const byte *, const void *);
static int none_deck(void *, struct fpnode_type *, byte *);
static int post_deck(void *, struct fpnode_type **, unsigned *, const byte *,
		     const void *);
static int post_line(void *, const byte *, const void *);
static int post_name(void *, const byte *, const void *);
static int post_pair(void *, const byte *, const void *);
static int post_slip(void *, const byte *, const void *);

static int
case_deck(void *sfdeck, struct fpnode_type *fpnode_slip, const byte *slip,
	  unsigned suffix, int e, int c, const byte *name, const void *data)
{
    int status;
    void *node;

    status = sfdeck(sfdeck)->link_m.link
	(sfdeck(sfdeck)->link_m.data, &node, sizeof(struct fpnode_type) << 2);
    if (status) {
	status = LINK_ERROR;
    } else {
	unsigned bits, prefix, length;

	prefix = slip - fpnode_slip->base;

	bits = fpnode_slip->bits;

	length = strlen(back(name));
	if (length < sizeof(byte *) + 2) {
	    if (suffix < sizeof(byte *) + 2) {
		struct fpnode_type *fpnode_data;

		fpnode_data = node;

		if (c < e + 1) {
		    if (length) {
			length--;
			memcpy(&fpnode_data->base, name + 1, length);
		    }

		    fpnode_data->bits = dset(c, length);
		    fpnode_data->data = data;

		    fpnode_data++;

		    suffix--;

		    fpnode_data->bits = pset(e, suffix, bits);
		    fpnode_data->data = fpnode_slip->data;

		    if (suffix) {
			memcpy(&fpnode_data->base, slip + 1, suffix);
		    }
		} else {
		    suffix--;

		    fpnode_data->bits = pset(e, suffix, bits);
		    fpnode_data->data = fpnode_slip->data;

		    if (suffix) {
			memcpy(&fpnode_data->base, slip + 1, suffix);
		    }

		    fpnode_data++;

		    if (length) {
			length--;
			memcpy(&fpnode_data->base, name + 1, length);
		    }

		    fpnode_data->bits = dset(c, length);
		    fpnode_data->data = data;
		}

		fpnode_slip->data = node;
	    } else {
		void *flat;

		status = sfdeck(sfdeck)->link_m.link
		    (sfdeck(sfdeck)->link_m.data, &flat, suffix);
		if (status) {
		    status = LINK_ERROR;

		    sfdeck(sfdeck)->link_m.free
			(sfdeck(sfdeck)->link_m.data, node);
		} else {
		    struct fpnode_type *fpnode_data;

		    fpnode_data = node;

		    memcpy(flat, slip + 1, suffix);

		    suffix--;

		    if (c < e + 1) {
			if (length) {
			    length--;
			    memcpy(&fpnode_data->base, name + 1, length);
			}

			fpnode_data->bits = dset(c, length);
			fpnode_data->data = data;

			fpnode_data++;

			fpnode_data->base = flat;
			fpnode_data->bits = pset(e, suffix, bits);
			fpnode_data->data = fpnode_slip->data;
		    } else {
			fpnode_data->base = flat;
			fpnode_data->bits = pset(e, suffix, bits);
			fpnode_data->data = fpnode_slip->data;

			fpnode_data++;

			if (length) {
			    length--;
			    memcpy(&fpnode_data->base, name + 1, length);
			}

			fpnode_data->bits = dset(c, length);
			fpnode_data->data = data;
		    }

		    fpnode_slip->data = node;
		}
	    }
	} else {
	    void *sand;

	    status = sfdeck(sfdeck)->link_m.link
		(sfdeck(sfdeck)->link_m.data, &sand, length);
	    if (status) {
		status = LINK_ERROR;

		sfdeck(sfdeck)->link_m.free(sfdeck(sfdeck)->link_m.data, node);
	    } else {
		if (suffix < sizeof(byte *) + 2) {
		    struct fpnode_type *fpnode_data;

		    fpnode_data = node;

		    memcpy(sand, name + 1, length);

		    length--;

		    if (c < e + 1) {
			fpnode_data->base = sand;
			fpnode_data->bits = dset(c, length);
			fpnode_data->data = data;

			fpnode_data++;

			suffix--;

			fpnode_data->bits = pset(e, suffix, bits);
			fpnode_data->data = fpnode_slip->data;

			if (suffix) {
			    memcpy(&fpnode_data->base, slip + 1, suffix);
			}
		    } else {
			suffix--;

			fpnode_data->bits = pset(e, suffix, bits);
			fpnode_data->data = fpnode_slip->data;

			if (suffix) {
			    memcpy(&fpnode_data->base, slip + 1, suffix);
			}

			fpnode_data++;

			fpnode_data->base = sand;
			fpnode_data->bits = dset(c, length);
			fpnode_data->data = data;
		    }

		    fpnode_slip->data = node;
		} else {
		    void *flat;

		    status = sfdeck(sfdeck)->link_m.link
			(sfdeck(sfdeck)->link_m.data, &flat, suffix);
		    if (status) {
			status = LINK_ERROR;

			sfdeck(sfdeck)->link_m.free
			    (sfdeck(sfdeck)->link_m.data, sand);
			sfdeck(sfdeck)->link_m.free
			    (sfdeck(sfdeck)->link_m.data, node);
		    } else {
			struct fpnode_type *fpnode_data;

			fpnode_data = node;

			memcpy(sand, name + 1, length);

			length--;

			memcpy(flat, slip + 1, suffix);

			suffix--;

			if (c < e + 1) {
			    fpnode_data->base = sand;
			    fpnode_data->bits = dset(c, length);
			    fpnode_data->data = data;

			    fpnode_data++;

			    fpnode_data->base = flat;
			    fpnode_data->bits = pset(e, suffix, bits);
			    fpnode_data->data = fpnode_slip->data;
			} else {
			    fpnode_data->base = flat;
			    fpnode_data->bits = pset(e, suffix, bits);
			    fpnode_data->data = fpnode_slip->data;

			    fpnode_data++;

			    fpnode_data->base = sand;
			    fpnode_data->bits = dset(c, length);
			    fpnode_data->data = data;
			}

			fpnode_slip->data = node;
		    }
		}
	    }
	}
    }

    return status;
}


static int
ever_deck(void *sfdeck, struct fpnode_type *fpnode_slip, const byte *name,
	  const void *data)
{
    int status;
    void *node;

    status = sfdeck(sfdeck)->link_m.link
	(sfdeck(sfdeck)->link_m.data, &node, sizeof(struct fpnode_type) << 2);
    if (status) {
	status = LINK_ERROR;
    } else {
	unsigned length;

	length = strlen(back(name));
	if (length < sizeof(byte *) + 2) {
	    struct fpnode_type *fpnode_data;

	    fpnode_data = node;

	    fpnode_data->bits = dset(0, 0);
	    fpnode_data->data = fpnode_slip->data;

	    fpnode_data++;

	    length--;

	    fpnode_data->bits = dset(*name, length);
	    fpnode_data->data = data;

	    if (length) {
		memcpy(&fpnode_data->base, name + 1, length);
	    }

	    fpnode_slip->bits = jset(fpnode_slip->bits, 2);
	    fpnode_slip->data = node;
	} else {
	    void *sand;

	    status = sfdeck(sfdeck)->link_m.link
		(sfdeck(sfdeck)->link_m.data, &sand, length);
	    if (status) {
		status = LINK_ERROR;

		sfdeck(sfdeck)->link_m.free(sfdeck(sfdeck)->link_m.data, node);
	    } else {
		struct fpnode_type *fpnode_data;

		fpnode_data = node;

		memcpy(sand, name + 1, length);

		length--;

		fpnode_data->bits = dset(0, 0);
		fpnode_data->data = fpnode_slip->data;

		fpnode_data++;

		fpnode_data->base = sand;
		fpnode_data->bits = dset(*name, length);
		fpnode_data->data = data;

		fpnode_slip->bits = jset(fpnode_slip->bits, 2);
		fpnode_slip->data = node;
	    }
	}
    }

    return status;
}


static int
fast_deck(void *sfdeck, struct fpnode_type **fpnode, unsigned *size,
	  const void *data)
{
    int status;
    struct fpnode_type *fpnode_slip;

    fpnode_slip = *fpnode;

    if (prefix(fpnode_slip)) {
	do {
	    unsigned fast;

	    fast = *size;

	    if (fast & 3) {
		status = 0;
	    } else {
		void *node;

		node = fpnode_slip;

		status = sfdeck(sfdeck)->link_m.mode
		    (sfdeck(sfdeck)->link_m.data, &node,
		     (fast + 4) * sizeof(struct fpnode_type));
		if (status) {
		    status = MODE_ERROR;
		    if (1) {
			break;
		    }
		} else {
		    *fpnode = node;

		    fpnode_slip = node;
		}
	    }

	    memmove(fpnode_slip + 1, fpnode_slip,
		    fast * sizeof(struct fpnode_type));

	    fpnode_slip->bits = dset(0, 0);
	    fpnode_slip->data = data;

	    *size = fast + 1;
	} while (0);
    } else {
	status = EVER_MATCH;
    }

    return status;
}


static int
line_deck(void *sfdeck, struct fpnode_type **fpnode,
	  struct fpnode_type *fpnode_data, unsigned line, unsigned fast,
	  unsigned *size, const byte *name, const void *data)
{
    int status;

    do {
	unsigned length;

	if (fast & 3) {
	} else {
	    void *node;

	    node = fpnode_data;

	    status = sfdeck(sfdeck)->link_m.mode
		(sfdeck(sfdeck)->link_m.data, &node,
		 (fast + 4) * sizeof(struct fpnode_type));
	    if (status) {
		status = MODE_ERROR;
		if (1) {
		    break;
		}
	    } else {
		*fpnode = node;

		fpnode_data = node;
	    }
	}

	length = strlen(back(name));
	if (length < sizeof(byte *) + 2) {
	    fpnode_data += line;

	    memmove(fpnode_data + 1, fpnode_data,
		    (fast - line) * sizeof(struct fpnode_type));

	    length--;

	    fpnode_data->bits = dset(*name, length);
	    fpnode_data->data = data;

	    if (length) {
		memcpy(&fpnode_data->base, name + 1, length);
	    }

	    *size = fast + 1;

	    status = 0;
	} else {
	    void *sand;

	    status = sfdeck(sfdeck)->link_m.link
		(sfdeck(sfdeck)->link_m.data, &sand, length);
	    if (status) {
		status = LINK_ERROR;
	    } else {
		fpnode_data += line;

		memmove(fpnode_data + 1, fpnode_data,
			(fast - line) * sizeof(struct fpnode_type));

		memcpy(sand, name + 1, length);

		length--;

		fpnode_data->base = sand;
		fpnode_data->bits = dset(*name, length);
		fpnode_data->data = data;

		*size = fast + 1;
	    }
	}
    } while (0);

    return status;
}


static int
mind_deck(void *sfdeck, struct fpnode_type *fpnode_data, const byte *name,
	  const void *data)
{
    int status;
    struct fpnode_type *fpnode_slip;
    unsigned bits, size;

    bits = fpnode_data->bits;
    fpnode_slip = (struct fpnode_type *) fpnode_data->data;

    size = Spread(bits);

    status = post_deck(sfdeck, &fpnode_slip, &size, name, data);
    if (status) {
    } else {
	fpnode_data->bits = iset(bits, size);
	fpnode_data->data = fpnode_slip;
    }

    return status;
}


static int
none_deck(void *sfdeck, struct fpnode_type *fpnode_slip, byte *slip)
{
    int status;
    unsigned bits, length;

    bits = fpnode_slip->bits;

    length = Length(bits);
    if (sizeof(byte *) < length) {
	unsigned prefix;
	void *miss;

	prefix = slip - fpnode_slip->base;

	fpnode_slip->bits = nset(prefix, bits, 2);

	miss = fpnode_slip->base;

	if (sizeof(byte *) < prefix) {
	    *slip = 0;

	    status = sfdeck(sfdeck)->link_m.mode
		(sfdeck(sfdeck)->link_m.data, &miss, prefix + 1);
	    if (status) {
		status = MODE_ERROR;
	    } else {
		byte *text;

		text = miss;

		text[prefix] = 0;

		fpnode_slip->base = miss;
	    }
	} else {
	    memcpy(&fpnode_slip->base, miss, prefix);

	    status = sfdeck(sfdeck)->link_m.free
		(sfdeck(sfdeck)->link_m.data, miss);
	    if (status) {
		status = FREE_ERROR;
		if (1) {
		    fpnode_slip->base = miss;
		}
	    }
	}
    } else {
	unsigned prefix;

	prefix = slip - (byte *) &fpnode_slip->base;

	fpnode_slip->bits = nset(prefix, bits, 2);

	status = 0;
    }

    return status;
}


static int
post_deck(void *sfdeck, struct fpnode_type **fpnode, unsigned *size,
	  const byte *name, const void *data)
{
    int c, status;

    c = *name;
    if (c) {
	struct fpnode_type *fpnode_data, *fpnode_slip;
	unsigned bits, mind;

	mind = *size;

	fpnode_slip = *fpnode;

	fpnode_data = fpnode_slip;

	Progress_node(fpnode_data, mind - 1);

	bits = fpnode_data->bits;

	if (Prefix(bits) == c) {
	    byte *slip;
	    int e;
	    unsigned length;

	    length = Length(bits);

	    name++;

	    if (sizeof(byte *) < length) {
		slip = fpnode_data->base;
		if (length < Record) {
		} else {
		    length += strlen(back(slip) + Record);
		}
	    } else {
		slip = (byte *) &fpnode_data->base;
	    }

	    if (1) {
		while (1) {
		    if (length) {
			c = *name;
			e = *slip;
			if (c == e) {
			    if (c) {
				slip++;
				name++;
				length--;
			    } else {
				break;
			    }
			} else {
			    break;
			}
		    } else {
			c = *name;
			e = 0;
			if (1) {
			    break;
			}
		    }
		}
	    }
	    if (e) {
		status = case_deck
		    (sfdeck, fpnode_data, slip, length, e, c, name, data);
		if (status) {
		} else {
		    status = none_deck(sfdeck, fpnode_data, slip);
		}
	    } else {
		if (Extant(bits)) {
		    status = mind_deck(sfdeck, fpnode_data, name, data);
		} else {
		    if (c) {
			status = ever_deck(sfdeck, fpnode_data, name, data);
		    } else {
			status = EVER_MATCH;
		    }
		}
	    }
	} else {
	    unsigned line;

	    line = fpnode_data - fpnode_slip;

	    if (c < Prefix(bits)) {
	    } else {
		line++;
	    }

	    status = line_deck
		(sfdeck, fpnode, fpnode_slip, line, mind, size, name, data);
	}
    } else {
	status = fast_deck(sfdeck, fpnode, size, data);
    }

    return status;
}


static int
post_line(void *sfdeck, const byte *name, const void *data)
{
    int status;
    void *node;

    status = sfdeck(sfdeck)->link_m.link
	(sfdeck(sfdeck)->link_m.data, &node, sizeof(struct fpnode_type));
    if (status) {
	status = LINK_ERROR;
    } else {
	unsigned length;

	length = strlen(back(name));
	if (length < sizeof(byte *) + 2) {
	    struct fpnode_type *fpnode_data;

	    sfdeck(sfdeck)->link_a.fpdeck.fpnode_data = node;

	    fpnode_data = node;

	    if (length) {
		length--;
	    }

	    fpnode_data->bits = dset(*name, length);
	    fpnode_data->data = data;

	    if (length) {
		memcpy(&fpnode_data->base, name + 1, length);
	    }
	} else {
	    void *sand;

	    status = sfdeck(sfdeck)->link_m.link
		(sfdeck(sfdeck)->link_m.data, &sand, length);
	    if (status) {
		status = LINK_ERROR;

		sfdeck(sfdeck)->link_m.free(sfdeck(sfdeck)->link_m.data, node);
	    } else {
		struct fpnode_type *fpnode_data;

		sfdeck(sfdeck)->link_a.fpdeck.fpnode_data = node;

		fpnode_data = node;

		memcpy(sand, name + 1, length);

		length--;

		fpnode_data->base = sand;
		fpnode_data->bits = dset(*name, length);
		fpnode_data->data = data;
	    }
	}
    }

    return status;
}


static int
post_name(void *sfdeck, const byte *name, const void *data)
{
    int status;
    struct fpnode_type *fpnode_data, *fpnode_slip;
    unsigned size;

    fpnode_data = sfdeck(sfdeck)->link_a.fpdeck.fpnode_data;

    size = spread(fpnode_data);
    fpnode_slip = (struct fpnode_type *) fpnode_data->data;

    status = post_deck(sfdeck, &fpnode_slip, &size, name, data);
    if (status) {
    } else {
	fpnode_data->bits = eset(size);
	fpnode_data->data = fpnode_slip;
    }

    return status;
}


static int
post_pair(void *sfdeck, const byte *name, const void *data)
{
    int status;
    void *node;

    status = sfdeck(sfdeck)->link_m.link
	(sfdeck(sfdeck)->link_m.data, &node, sizeof(struct fpnode_type) << 2);
    if (status) {
	status = LINK_ERROR;
    } else {
	struct fpnode_type *fpnode_data, *fpnode_slip;
	unsigned size = 1;

	fpnode_data = sfdeck(sfdeck)->link_a.fpdeck.fpnode_data;

	fpnode_slip = node;

	*fpnode_slip = *fpnode_data;

	fpnode_data->bits = eset(1);
	fpnode_data->data = fpnode_slip;

	status = post_deck(sfdeck, &fpnode_slip, &size, name, data);
	if (status) {
	} else {
	    fpnode_data->bits = eset(size);
	    fpnode_data->data = fpnode_slip;
	}
    }

    return status;
}


static int
post_slip(void *sfdeck, const byte *name, const void *data)
{
    int status;
    struct fpnode_type *fpnode_data;
    unsigned size = 1;

    fpnode_data = sfdeck(sfdeck)->link_a.fpdeck.fpnode_data;

    status = post_deck(sfdeck, &fpnode_data, &size, name, data);
    if (status) {
    } else {
	if (0) {
	    sfdeck(sfdeck)->link_a.fpdeck.fpnode_data = fpnode_data;
	}
    }

    return status;
}


int
x1f4_post_sfdeck(void *sfdeck, const char *name, const void *data)
{
    int status;
    struct fpnode_type *fpnode_data;

    fpnode_data = sfdeck(sfdeck)->link_a.fpdeck.fpnode_data;
    if (fpnode_data) {
	int prefix;

	prefix = prefix(fpnode_data);
	if (prefix) {
	    if (prefix == *name) {
		status = post_slip(sfdeck, fast(name), data);
	    } else {
		status = post_pair(sfdeck, fast(name), data);
	    }
	} else {
	    if (extant(fpnode_data)) {
		status = post_name(sfdeck, fast(name), data);
	    } else {
		if (*name) {
		    status = post_pair(sfdeck, fast(name), data);
		} else {
		    status = post_slip(sfdeck, fast(name), data);
		}
	    }
	}
    } else {
	status = post_line(sfdeck, fast(name), data);
    }

    return status;
}
