/*
 * sjxx.7.c
 * Copyright (C) 2009-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <string.h>

#include <sjxx-defs.h>
#include <sjxx-names.h>
#include <sjxx-patch.h>
#include <sjxx-types.h>

#define back(miss)			((const char *) (miss))
#define fast(miss)			((const byte *) (miss))

#define null				(&sjdeck(sjdeck)->link_i.zero)

#define fplead(fplead)			((struct fplead_type *) (fplead))

#define sjdeck(sjdeck)			((struct sjdeck_type *) (sjdeck))

#define violet(fpnode) \
    ((struct fpnode_type *) (fplead((fpnode)) + 1))

#define lead_path(fpnode, ___c) \
    fplead((fpnode))->mirror[(___c) >> 5] |= 1 << ((___c) & 31)

static int case_deck(void *, struct fpnode_type *, const byte *, int, int,
		     const byte *, const void *);
static int ever_deck(void *, struct fpnode_type *, const byte *, const void *);
static int fast_deck(void *, struct fpnode_type **, unsigned *, const void *);
static int line_deck(void *, struct fpnode_type **, struct fpnode_type *,
		     unsigned, unsigned, unsigned *, const byte *,
		     const void *);
static int mind_deck(void *, struct fpnode_type *, const byte *, const void *);
static int none_deck(void *, struct fpnode_type *, byte *);
static int post_deck(void *, struct fpnode_type **, unsigned *, const byte *,
		     const void *);
static int post_line(void *, const byte *, const void *);
static int post_name(void *, const byte *, const void *);
static int post_pair(void *, const byte *, const void *);
static int post_slip(void *, const byte *, const void *);
static int rail_deck(struct fpnode_type *);

static int
case_deck(void *sjdeck, struct fpnode_type *fpnode_slip, const byte *slip,
	  int e, int c, const byte *name, const void *data)
{
    int status;
    void *node;

    status = sjdeck(sjdeck)->link_m.link
	(sjdeck(sjdeck)->link_m.data, &node, sizeof(struct fpnode_type) << 2);
    if (status) {
	status = LINK_ERROR;
    } else {
	unsigned bits, prefix, length;

	prefix = slip - fpnode_slip->base;

	bits = fpnode_slip->bits;

	length = strlen(back(name));
	if (length < 2) {
	    unsigned suffix;

	    suffix = strlen(back(slip));
	    if (suffix == 1) {
		struct fpnode_type *fpnode_data;

		fpnode_data = node;

		if (c < e + 1) {
		    fpnode_data->base = null;
		    fpnode_data->bits = dset(c, 0);
		    fpnode_data->data = data;

		    fpnode_data++;

		    fpnode_data->base = null;
		    fpnode_data->bits = pset(e, 0, bits);
		    fpnode_data->data = fpnode_slip->data;
		} else {
		    fpnode_data->base = null;
		    fpnode_data->bits = pset(e, 0, bits);
		    fpnode_data->data = fpnode_slip->data;

		    fpnode_data++;

		    fpnode_data->base = null;
		    fpnode_data->bits = dset(c, 0);
		    fpnode_data->data = data;
		}

		fpnode_slip->data = node;
	    } else {
		void *flat;

		status = sjdeck(sjdeck)->link_m.link
		    (sjdeck(sjdeck)->link_m.data, &flat, suffix);
		if (status) {
		    status = LINK_ERROR;

		    sjdeck(sjdeck)->link_m.free
			(sjdeck(sjdeck)->link_m.data, node);
		} else {
		    struct fpnode_type *fpnode_data;

		    fpnode_data = node;

		    memcpy(flat, slip + 1, suffix);

		    suffix--;

		    if (c < e + 1) {
			fpnode_data->base = null;
			fpnode_data->bits = dset(c, 0);
			fpnode_data->data = data;

			fpnode_data++;

			fpnode_data->base = flat;
			fpnode_data->bits = pset(e, suffix, bits);
			fpnode_data->data = fpnode_slip->data;
		    } else {
			fpnode_data->base = flat;
			fpnode_data->bits = pset(e, suffix, bits);
			fpnode_data->data = fpnode_slip->data;

			fpnode_data++;

			fpnode_data->base = null;
			fpnode_data->bits = dset(c, 0);
			fpnode_data->data = data;
		    }

		    fpnode_slip->data = node;
		}
	    }
	} else {
	    void *sand;

	    status = sjdeck(sjdeck)->link_m.link
		(sjdeck(sjdeck)->link_m.data, &sand, length);
	    if (status) {
		status = LINK_ERROR;

		sjdeck(sjdeck)->link_m.free(sjdeck(sjdeck)->link_m.data, node);
	    } else {
		unsigned suffix;

		suffix = strlen(back(slip));
		if (suffix == 1) {
		    struct fpnode_type *fpnode_data;

		    fpnode_data = node;

		    memcpy(sand, name + 1, length);

		    length--;

		    if (c < e + 1) {
			fpnode_data->base = sand;
			fpnode_data->bits = dset(c, length);
			fpnode_data->data = data;

			fpnode_data++;

			fpnode_data->base = null;
			fpnode_data->bits = pset(e, 0, bits);
			fpnode_data->data = fpnode_slip->data;
		    } else {
			fpnode_data->base = null;
			fpnode_data->bits = pset(e, 0, bits);
			fpnode_data->data = fpnode_slip->data;

			fpnode_data++;

			fpnode_data->base = sand;
			fpnode_data->bits = dset(c, length);
			fpnode_data->data = data;
		    }

		    fpnode_slip->data = node;
		} else {
		    void *flat;

		    status = sjdeck(sjdeck)->link_m.link
			(sjdeck(sjdeck)->link_m.data, &flat, suffix);
		    if (status) {
			status = LINK_ERROR;

			sjdeck(sjdeck)->link_m.free
			    (sjdeck(sjdeck)->link_m.data, sand);
			sjdeck(sjdeck)->link_m.free
			    (sjdeck(sjdeck)->link_m.data, node);
		    } else {
			struct fpnode_type *fpnode_data;

			fpnode_data = node;

			memcpy(sand, name + 1, length);

			length--;

			memcpy(flat, slip + 1, suffix);

			suffix--;

			if (c < e + 1) {
			    fpnode_data->base = sand;
			    fpnode_data->bits = dset(c, length);
			    fpnode_data->data = data;

			    fpnode_data++;

			    fpnode_data->base = flat;
			    fpnode_data->bits = pset(e, suffix, bits);
			    fpnode_data->data = fpnode_slip->data;
			} else {
			    fpnode_data->base = flat;
			    fpnode_data->bits = pset(e, suffix, bits);
			    fpnode_data->data = fpnode_slip->data;

			    fpnode_data++;

			    fpnode_data->base = sand;
			    fpnode_data->bits = dset(c, length);
			    fpnode_data->data = data;
			}

			fpnode_slip->data = node;
		    }
		}
	    }
	}
    }

    return status;
}


static int
ever_deck(void *sjdeck, struct fpnode_type *fpnode_slip, const byte *name,
	  const void *data)
{
    int status;
    void *node;

    status = sjdeck(sjdeck)->link_m.link
	(sjdeck(sjdeck)->link_m.data, &node, sizeof(struct fpnode_type) << 2);
    if (status) {
	status = LINK_ERROR;
    } else {
	unsigned length;

	length = strlen(back(name));
	if (length == 1) {
	    struct fpnode_type *fpnode_data;

	    fpnode_data = node;

	    fpnode_data->base = null;
	    fpnode_data->bits = dset(0, 0);
	    fpnode_data->data = fpnode_slip->data;

	    fpnode_data++;

	    fpnode_data->base = null;
	    fpnode_data->bits = dset(*name, 0);
	    fpnode_data->data = data;

	    fpnode_slip->bits = jset(fpnode_slip->bits, 2);
	    fpnode_slip->data = node;
	} else {
	    void *sand;

	    status = sjdeck(sjdeck)->link_m.link
		(sjdeck(sjdeck)->link_m.data, &sand, length);
	    if (status) {
		status = LINK_ERROR;

		sjdeck(sjdeck)->link_m.free(sjdeck(sjdeck)->link_m.data, node);
	    } else {
		struct fpnode_type *fpnode_data;

		fpnode_data = node;

		memcpy(sand, name + 1, length);

		length--;

		fpnode_data->base = null;
		fpnode_data->bits = dset(0, 0);
		fpnode_data->data = fpnode_slip->data;

		fpnode_data++;

		fpnode_data->base = sand;
		fpnode_data->bits = dset(*name, length);
		fpnode_data->data = data;

		fpnode_slip->bits = jset(fpnode_slip->bits, 2);
		fpnode_slip->data = node;
	    }
	}
    }

    return status;
}


static int
fast_deck(void *sjdeck, struct fpnode_type **fpnode, unsigned *size,
	  const void *data)
{
    int status;
    struct fpnode_type *fpnode_slip;
    unsigned fast;
    void *node;

    fpnode_slip = *fpnode;

    node = fpnode_slip;

    fast = *size;

    if (fast < MIRROR_1st) {
    } else {
	fpnode_slip = (void *) (fplead(fpnode_slip) + 1);
    }

    if (prefix(fpnode_slip)) {
	do {
	    if (fast & 3) {
		status = 0;
	    } else {
		unsigned rate;

		rate = (fast + 4) * sizeof(struct fpnode_type);
		if (MIRROR_set * sizeof(struct fpnode_type) < rate) {
		    rate += sizeof(struct fplead_type);
		}

		status = sjdeck(sjdeck)->link_m.mode
		    (sjdeck(sjdeck)->link_m.data, &node, rate);
		if (status) {
		    status = MODE_ERROR;
		    if (1) {
			break;
		    }
		} else {
		    *fpnode = node;

		    fpnode_slip = node;

		    if (fast ^ MIRROR_set) {
		    } else {
			rail_deck(fpnode_slip);
		    }

		    if (fast < MIRROR_set) {
		    } else {
			fpnode_slip = (void *) (fplead(fpnode_slip) + 1);
		    }
		}
	    }

	    if (fast < MIRROR_set) {
	    } else {
		fplead(node)->mirror[0] |= 1;
	    }

	    memmove(fpnode_slip + 1, fpnode_slip,
		    fast * sizeof(struct fpnode_type));

	    fpnode_slip->base = null;
	    fpnode_slip->bits = dset(0, 0);
	    fpnode_slip->data = data;

	    *size = fast + 1;
	} while (0);
    } else {
	status = EVER_MATCH;
    }

    return status;
}


static int
line_deck(void *sjdeck, struct fpnode_type **fpnode,
	  struct fpnode_type *fpnode_data, unsigned line, unsigned fast,
	  unsigned *size, const byte *name, const void *data)
{
    int status;

    do {
	unsigned length;

	if (fast & 3) {
	} else {
	    unsigned rate;
	    void *node;

	    node = fpnode_data;

	    rate = (fast + 4) * sizeof(struct fpnode_type);
	    if (MIRROR_set * sizeof(struct fpnode_type) < rate) {
		rate += sizeof(struct fplead_type);
	    }

	    status = sjdeck(sjdeck)->link_m.mode
		(sjdeck(sjdeck)->link_m.data, &node, rate);
	    if (status) {
		status = MODE_ERROR;
		if (1) {
		    break;
		}
	    } else {
		*fpnode = node;

		fpnode_data = node;

		if (fast ^ MIRROR_set) {
		} else {
		    rail_deck(fpnode_data);
		}
	    }
	}

	length = strlen(back(name));
	if (length == 1) {
	    unsigned c;

	    c = *name;

	    if (fast < MIRROR_set) {
	    } else {
		lead_path(fpnode_data, c);

		fpnode_data = violet(fpnode_data);
	    }

	    fpnode_data += line;

	    memmove(fpnode_data + 1, fpnode_data,
		    (fast - line) * sizeof(struct fpnode_type));

	    fpnode_data->base = null;
	    fpnode_data->bits = dset(c, 0);
	    fpnode_data->data = data;

	    *size = fast + 1;

	    status = 0;
	} else {
	    void *sand;

	    status = sjdeck(sjdeck)->link_m.link
		(sjdeck(sjdeck)->link_m.data, &sand, length);
	    if (status) {
		status = LINK_ERROR;
	    } else {
		unsigned c;

		c = *name;

		if (fast < MIRROR_set) {
		} else {
		    lead_path(fpnode_data, c);

		    fpnode_data = violet(fpnode_data);
		}

		fpnode_data += line;

		memmove(fpnode_data + 1, fpnode_data,
			(fast - line) * sizeof(struct fpnode_type));

		memcpy(sand, name + 1, length);

		length--;

		fpnode_data->base = sand;
		fpnode_data->bits = dset(c, length);
		fpnode_data->data = data;

		*size = fast + 1;
	    }
	}
    } while (0);

    return status;
}


static int
mind_deck(void *sjdeck, struct fpnode_type *fpnode_data, const byte *name,
	  const void *data)
{
    int status;
    struct fpnode_type *fpnode_slip;
    unsigned bits, size;

    bits = fpnode_data->bits;
    fpnode_slip = (struct fpnode_type *) fpnode_data->data;

    size = Spread(bits);

    status = post_deck(sjdeck, &fpnode_slip, &size, name, data);
    if (status) {
    } else {
	fpnode_data->bits = iset(bits, size);
	fpnode_data->data = fpnode_slip;
    }

    return status;
}


static int
none_deck(void *sjdeck, struct fpnode_type *fpnode_slip, byte *slip)
{
    int status;
    unsigned prefix;
    void *miss;

    prefix = slip - fpnode_slip->base;

    fpnode_slip->bits = nset(prefix, fpnode_slip->bits, 2);

    miss = fpnode_slip->base;

    if (prefix) {
	*slip = 0;

	status = sjdeck(sjdeck)->link_m.mode
	    (sjdeck(sjdeck)->link_m.data, &miss, prefix + 1);
	if (status) {
	    status = MODE_ERROR;
	} else {
	    byte *text;

	    text = miss;

	    text[prefix] = 0;

	    fpnode_slip->base = miss;
	}
    } else {
	status = sjdeck(sjdeck)->link_m.free
	    (sjdeck(sjdeck)->link_m.data, miss);
	if (status) {
	    status = FREE_ERROR;
	} else {
	    fpnode_slip->base = null;
	}
    }

    return status;
}


static int
post_deck(void *sjdeck, struct fpnode_type **fpnode, unsigned *size,
	  const byte *name, const void *data)
{
    int c, status;

    c = *name;
    if (c) {
	struct fpnode_type *fpnode_data, *fpnode_slip;
	unsigned mind;

	mind = *size;

	fpnode_slip = *fpnode;

	do {
	    byte *slip;
	    int e;

	    if (mind < MIRROR_1st) {
		fpnode_data = fpnode_slip;

		backward_node(fpnode_data, mind - 1);

		if (e ^ c) {
		    unsigned line;

		    line = fpnode_data - fpnode_slip;

		    if (c < prefix(fpnode_data)) {
		    } else {
			line++;
		    }

		    status = line_deck
			(sjdeck, fpnode, fpnode_slip, line, mind, size, name,
			 data);
		    if (1) {
			break;
		    }
		}
	    } else {
		unsigned line;

		transfer_node(fpnode_slip, mind, line);

		if (___c_type_X(___fplead(fpnode_slip)->mirror, c)) {
		    fpnode_data = violet(fpnode_slip) + line;
		} else {
		    status = line_deck
			(sjdeck, fpnode, fpnode_slip, line, mind, size, name,
			 data);
		    if (1) {
			break;
		    }
		}
	    }

	    name++;

	    slip = fpnode_data->base;
	    if (1) {
		while (1) {
		    c = *name;
		    e = *slip;
		    if (c == e) {
			if (c) {
			    slip++;
			    name++;
			} else {
			    break;
			}
		    } else {
			break;
		    }
		}
	    }
	    if (e) {
		status = case_deck
		    (sjdeck, fpnode_data, slip, e, c, name, data);
		if (status) {
		} else {
		    status = none_deck(sjdeck, fpnode_data, slip);
		}
	    } else {
		if (extant(fpnode_data)) {
		    status = mind_deck(sjdeck, fpnode_data, name, data);
		} else {
		    if (c) {
			status = ever_deck(sjdeck, fpnode_data, name, data);
		    } else {
			status = EVER_MATCH;
		    }
		}
	    }
	} while (0);
    } else {
	status = fast_deck(sjdeck, fpnode, size, data);
    }

    return status;
}


static int
post_line(void *sjdeck, const byte *name, const void *data)
{
    int status;
    void *node;

    status = sjdeck(sjdeck)->link_m.link
	(sjdeck(sjdeck)->link_m.data, &node, sizeof(struct fpnode_type));
    if (status) {
	status = LINK_ERROR;
    } else {
	unsigned length;

	length = strlen(back(name));
	if (length < 2) {
	    struct fpnode_type *fpnode_data;

	    sjdeck(sjdeck)->link_a.fpdeck.fpnode_data = node;

	    fpnode_data = node;

	    fpnode_data->base = null;
	    fpnode_data->bits = dset(*name, 0);
	    fpnode_data->data = data;
	} else {
	    void *sand;

	    status = sjdeck(sjdeck)->link_m.link
		(sjdeck(sjdeck)->link_m.data, &sand, length);
	    if (status) {
		status = LINK_ERROR;

		sjdeck(sjdeck)->link_m.free(sjdeck(sjdeck)->link_m.data, node);
	    } else {
		struct fpnode_type *fpnode_data;

		sjdeck(sjdeck)->link_a.fpdeck.fpnode_data = node;

		fpnode_data = node;

		memcpy(sand, name + 1, length);

		length--;

		fpnode_data->base = sand;
		fpnode_data->bits = dset(*name, length);
		fpnode_data->data = data;
	    }
	}
    }

    return status;
}


static int
post_name(void *sjdeck, const byte *name, const void *data)
{
    int status;
    struct fpnode_type *fpnode_data, *fpnode_slip;
    unsigned size;

    fpnode_data = sjdeck(sjdeck)->link_a.fpdeck.fpnode_data;

    size = spread(fpnode_data);
    fpnode_slip = (struct fpnode_type *) fpnode_data->data;

    status = post_deck(sjdeck, &fpnode_slip, &size, name, data);
    if (status) {
    } else {
	fpnode_data->bits = eset(size);
	fpnode_data->data = fpnode_slip;
    }

    return status;
}


static int
post_pair(void *sjdeck, const byte *name, const void *data)
{
    int status;
    void *node;

    status = sjdeck(sjdeck)->link_m.link
	(sjdeck(sjdeck)->link_m.data, &node, sizeof(struct fpnode_type) << 2);
    if (status) {
	status = LINK_ERROR;
    } else {
	struct fpnode_type *fpnode_data, *fpnode_slip;
	unsigned size = 1;

	fpnode_data = sjdeck(sjdeck)->link_a.fpdeck.fpnode_data;

	fpnode_slip = node;

	*fpnode_slip = *fpnode_data;

	fpnode_data->base = null;
	fpnode_data->bits = eset(1);
	fpnode_data->data = fpnode_slip;

	status = post_deck(sjdeck, &fpnode_slip, &size, name, data);
	if (status) {
	} else {
	    fpnode_data->bits = eset(size);
	    fpnode_data->data = fpnode_slip;
	}
    }

    return status;
}


static int
post_slip(void *sjdeck, const byte *name, const void *data)
{
    int status;
    struct fpnode_type *fpnode_data;
    unsigned size = 1;

    fpnode_data = sjdeck(sjdeck)->link_a.fpdeck.fpnode_data;

    status = post_deck(sjdeck, &fpnode_data, &size, name, data);
    if (status) {
    } else {
	if (0) {
	    sjdeck(sjdeck)->link_a.fpdeck.fpnode_data = fpnode_data;
	}
    }

    return status;
}


static int
rail_deck(struct fpnode_type *fpnode_data)
{
    unsigned i = MIRROR_set, *mirror;

    memmove(fplead(fpnode_data) + 1, fpnode_data,
	    sizeof(struct fpnode_type) * MIRROR_set);

    mirror = fplead(fpnode_data)->mirror;

    mirror[0] = 0;
    mirror[1] = 0;
    mirror[2] = 0;
    mirror[3] = 0;
    mirror[4] = 0;
    mirror[5] = 0;
    mirror[6] = 0;
    mirror[7] = 0;

    fpnode_data = violet(fpnode_data);

    for (; i; i--) {
	unsigned c;

	c = prefix(fpnode_data);
	mirror[c >> 5] |= 1 << (c & 31);

	fpnode_data++;
    }

    return 0;
}


int
x1f4_post_sjdeck(void *sjdeck, const char *name, const void *data)
{
    int status;
    struct fpnode_type *fpnode_data;

    fpnode_data = sjdeck(sjdeck)->link_a.fpdeck.fpnode_data;
    if (fpnode_data) {
	int prefix;

	prefix = prefix(fpnode_data);
	if (prefix) {
	    if (prefix == *fast(name)) {
		status = post_slip(sjdeck, fast(name), data);
	    } else {
		status = post_pair(sjdeck, fast(name), data);
	    }
	} else {
	    if (extant(fpnode_data)) {
		status = post_name(sjdeck, fast(name), data);
	    } else {
		if (*name) {
		    status = post_pair(sjdeck, fast(name), data);
		} else {
		    status = post_slip(sjdeck, fast(name), data);
		}
	    }
	}
    } else {
	status = post_line(sjdeck, fast(name), data);
    }

    return status;
}
