/*
 * tclist.h
 * Copyright (C) 2008-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __LIBx1f4_TCLIST_H__
#define __LIBx1f4_TCLIST_H__

#include <errors.h>
#include <tcxx-names.h>

/*
 * attribute bit definitions
 */
#define X1f4_TCLIST_TRANS_MASK		(1u << 0x00)

/*
 * error definitions
 */
#define X1f4_TCLIST_CALL_ERROR		LIBx1f4L2_CALL_ERROR
#define X1f4_TCLIST_DECK_ERROR		LIBx1f4L2_DECK_ERROR
#define X1f4_TCLIST_FREE_ERROR		LIBx1f4L2_FREE_ERROR
#define X1f4_TCLIST_LINK_ERROR		LIBx1f4L2_LINK_ERROR
#define X1f4_TCLIST_MODE_ERROR		LIBx1f4L2_MODE_ERROR
#define X1f4_TCLIST_SIZE_ERROR		LIBx1f4L2_SIZE_ERROR

typedef struct x1f4_tclist_type {
    void *trans;
} x1f4_tclist_type;

#if !defined __x1f4_trans_type_defined__
typedef struct x1f4_trans_type {
    int (*free) (void *, void *), (*link) (void *, void **, unsigned),
	(*mode) (void *, void **, unsigned);
    void *context;
} x1f4_trans_type;
#endif				/* !__x1f4_trans_type_defined__ */

#if !defined __x1f4_trans_type_defined__
# define __x1f4_trans_type_defined__
#endif				/* !__x1f4_trans_type_defined__ */

extern int x1f4_call_tclist(unsigned *);
extern int x1f4_clip_tclist
    (void *, unsigned, void *, int (*) (void *, void *));
extern int x1f4_deck_tclist(void *, int);
extern int x1f4_deux_tclist(void *, unsigned, void *, unsigned);
extern int x1f4_ever_tclist(void *, void **);
extern int x1f4_fast_tclist(void *, unsigned, struct x1f4_tclist_type *);
extern int x1f4_fgxp_tclist
    (void *, void *, unsigned, int (*) (void *, void *), unsigned *);
extern int x1f4_find_tclist
    (void *, void *, unsigned, int (*) (void *, void *), void **);
extern int x1f4_fine_tclist(void *, void *, int (*) (void *, void *));
extern int x1f4_fini_tclist(void **);
extern int x1f4_flat_tclist(void *);
extern int x1f4_flip_tclist(void *);
extern int x1f4_flow_tclist(void *);
extern int x1f4_head_tclist(void *, unsigned, void **);
extern int x1f4_high_tclist(void *);
extern int x1f4_init_tclist(void **, unsigned, struct x1f4_tclist_type *);
extern int x1f4_land_tclist
    (void *, void *, int (*) (void *, void *), unsigned, unsigned, void **);
extern int x1f4_lead_tclist(void *, void **);
extern int x1f4_lime_tclist(void *, void *, int (*) (void *, void *));
extern int x1f4_line_tclist
    (void *, void *, int (*) (void *, unsigned *, const void *),
     int (*) (void *, void *, const void *), int (*) (void *, void *), void *);
extern int x1f4_mind_tclist(void *);
extern int x1f4_miss_tclist(void *, unsigned);
extern int x1f4_mode_tclist(void *, unsigned, unsigned, void **);
extern int x1f4_move_tclist(void *, unsigned, unsigned);
extern int x1f4_nice_tclist(void *, void *, int (*) (void *, void *));
extern int x1f4_peek_tclist(void *, unsigned, void **);
extern int x1f4_pull_tclist
    (void *, unsigned, void *, int (*) (void *, void *));
extern int x1f4_push_tclist(void *, unsigned, unsigned, void **);
extern int x1f4_rand_tclist(void *);
extern int x1f4_seek_tclist
    (void *, void *, unsigned, int (*) (void *, void *), unsigned *);
extern int x1f4_size_tclist(void *, unsigned *);
extern int x1f4_skew_tclist(void *, void *);
extern int x1f4_slip_tclist(void *);
extern int x1f4_sort_tclist
    (void *, void *, unsigned, int (*) (void *, void *, void *));
extern int x1f4_spin_tclist(void *, unsigned, unsigned);
extern int x1f4_tail_tclist(void *, unsigned, void **);
extern int x1f4_text_tclist
    (void **, void *, int (*) (void *, unsigned *, const void *),
     int (*) (void *, void *, const void *), int (*) (void *, void *), void *);
extern int x1f4_tick_tclist
    (void *, void *, unsigned, int (*) (void *, void *), unsigned *, void **);
extern int x1f4_turn_tclist(void *);
extern int x1f4_wind_tclist(void *, unsigned, void *, unsigned);

#endif				/* __LIBx1f4_TCLIST_H__ */
