/*
 * tcmh.4.c
 * Copyright (C) 2008-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#include <tcmn-bits.h>
#include <tcmn-names.h>
#include <tcmn-types.h>

#define d_count				(SIZEOF_UNSIGNED_LONG << 3)

#define d_left				(d_count - 1)

#define d_high				(1ul << d_left)

#define M_TEXT(___v, ___t) \
    ((void *) ((char *) (___v) + (___t)))

static int find_d1db(void *, unsigned, unsigned long, unsigned, void **);
static int find_d1dp(void *, unsigned, unsigned long, unsigned, void **);
static int find_d1qb(void *, unsigned, unsigned long, unsigned, void **);
static int find_d1qp(void *, unsigned, unsigned long, unsigned, void **);

static int
find_d1db(void *tcmini, unsigned node, unsigned long text, unsigned offset,
	  void **find)
{
    int delete;

    while (1) {
	unsigned long diff, slip;
	void *extend;

	extend = M_TEXT(tcmini, node);

	slip = *(unsigned long *) ((char *) extend + offset);

	diff = slip - text;
	if (!diff) {
	    delete = 1;

	    *find = (char *) extend + 4;

	    break;
	} else {
	    node = L_NODE(Q_NODE(extend), (long) diff >> d_left);
	    if (node) {
	    } else {
		delete = 0;
		if (1) {
		    break;
		}
	    }
	}
    }

    return delete;
}


static int
find_d1dp(void *tcmini, unsigned node, unsigned long text, unsigned offset,
	  void **find)
{
    int delete;

    while (1) {
	unsigned long slip;
	void *extend;

	extend = M_TEXT(tcmini, node);

	slip = *(unsigned long *) ((char *) extend + offset);

	if (slip & d_high) {
	    unsigned long diff;

	    diff = slip - text;

	    if (diff >> d_left) {
		node = L_NEXT(Q_NODE(extend));
		if (node) {
		    delete = find_d1db(tcmini, node, text, offset, find);
		    if (1) {
			break;
		    }
		} else {
		    delete = 0;
		    if (1) {
			break;
		    }
		}
	    } else {
		if (!diff) {
		    delete = 1;

		    *find = (char *) extend + 4;

		    break;
		} else {
		    node = L_LAST(Q_NODE(extend));
		    if (node) {
		    } else {
			delete = 0;
			if (1) {
			    break;
			}
		    }
		}
	    }
	} else {
	    node = L_NEXT(Q_NODE(extend));
	    if (node) {
	    } else {
		delete = 0;
		if (1) {
		    break;
		}
	    }
	}
    }

    return delete;
}


static int
find_d1qb(void *tcmini, unsigned node, unsigned long text, unsigned offset,
	  void **find)
{
    int delete;

    while (1) {
	unsigned long diff, slip;
	void *extend;

	extend = M_TEXT(tcmini, node);

	slip = *(unsigned long *) ((char *) extend + offset);

	diff = slip - text;
	if (!diff) {
	    delete = 1;

	    *find = (char *) extend + 4;

	    break;
	} else {
	    node = L_NODE(Q_NODE(extend), (long) diff >> d_left);
	    if (node) {
	    } else {
		delete = 0;
		if (1) {
		    break;
		}
	    }
	}
    }

    return delete;
}


static int
find_d1qp(void *tcmini, unsigned node, unsigned long text, unsigned offset,
	  void **find)
{
    int delete;

    while (1) {
	unsigned long slip;
	void *extend;

	extend = M_TEXT(tcmini, node);

	slip = *(unsigned long *) ((char *) extend + offset);

	if (slip & d_high) {
	    node = L_LAST(Q_NODE(extend));
	    if (node) {
	    } else {
		delete = 0;
		if (1) {
		    break;
		}
	    }
	} else {
	    unsigned long diff;

	    diff = slip - text;

	    if (diff >> d_left) {
		node = L_NEXT(Q_NODE(extend));
		if (node) {
		} else {
		    delete = 0;
		    if (1) {
			break;
		    }
		}
	    } else {
		if (!diff) {
		    delete = 1;

		    *find = (char *) extend + 4;

		    break;
		} else {
		    node = L_LAST(Q_NODE(extend));
		    if (node) {
			delete = find_d1qb(tcmini, node, text, offset, find);
			if (1) {
			    break;
			}
		    } else {
			delete = 0;
			if (1) {
			    break;
			}
		    }
		}
	    }
	}
    }

    return delete;
}


int
x1f4_find_tcnear(void *tcmini, unsigned long text, unsigned slip, void **find)
{
    int delete;
    unsigned root;

    root = P_ROOT(Q_NODE(tcmini));
    if (root) {
	if (text & d_high) {
	    delete = find_d1dp(tcmini, root, text, slip + 4, find);
	} else {
	    delete = find_d1qp(tcmini, root, text, slip + 4, find);
	}
    } else {
	delete = 0;
    }

    return delete;
}
