/*
 * tsqp.v.c
 * Copyright (C) 2008-2010, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <tsqp-defs.h>
#include <tsqp-types.h>

#define SET_U(fpnode_data) \
    (fpnode_data)->fqnode.fbnode.bits &= ~1
#define SET_V(fpnode_data) \
    (fpnode_data)->fqnode.fbnode.bits |= 1

#define GET_T(fpnode_data) \
    ((fpnode_data)->fqnode.fbnode.bits & 1)

#define GET_L(fpnode_data) \
    ((fpnode_data)->fqnode.fbnode.bits & LAST_T)
#define GET_N(fpnode_data) \
    ((fpnode_data)->fqnode.fbnode.bits & NEXT_T)

#define SET_L(fpnode_data) \
    (fpnode_data)->fqnode.fbnode.bits |= LAST_T
#define SET_N(fpnode_data) \
    (fpnode_data)->fqnode.fbnode.bits |= NEXT_T

#define FIX_L(fpnode_data) \
    (fpnode_data)->fqnode.fbnode.bits &= ~LAST_T
#define FIX_N(fpnode_data) \
    (fpnode_data)->fqnode.fbnode.bits &= ~NEXT_T

#define tsline(tsline)			((struct tsline_type *) (tsline))

int
_libx1f4l2_even_tsline(void *tsline, void *text, int (*land) (void *, void *),
		       unsigned offset, unsigned flat, void **push,
		       int (*line) (void *, unsigned, void **))
{
    int status;
    struct fpnode_type *fxnode_data, *fznode_data;
    void *tsnode;

    fxnode_data = tsline(tsline)->link_a.fpline.fplink.fpnode_data[0];
    fznode_data = &tsline(tsline)->link_a.fpline.fplink;

    if (!fxnode_data) {
	status = line(tsline, flat, &tsnode);
	if (status) {
	} else {
	    struct fpnode_type *fnnode_data;

	    fnnode_data = tsnode;

	    *push = fnnode_data + 1;

	    fnnode_data->fpnode_data[0] = NULL;
	    fnnode_data->fpnode_data[1] = NULL;

	    fnnode_data->fqnode.tilt = 0;

	    fnnode_data->fqnode.fbnode.bits = FLAT_T;

	    fznode_data->fpnode_data[0] = fnnode_data;
	}
    } else {
	struct fpnode_type *fpnode_data, *fqnode_data;
	unsigned fast, tilt;
	void *data;

	data = (char *) tsnode + offset;

	fpnode_data = fxnode_data;
	fqnode_data = fznode_data;

	while (1) {
	    int slip;

	    slip = land(text, (char *) fpnode_data + offset);

	    if (1) {
		if (fpnode_data->fqnode.tilt) {
		    fxnode_data = fpnode_data;
		    fznode_data = fqnode_data;
		}

		if (slip < 0) {
		    SET_U(fpnode_data);

		    if (fpnode_data->fqnode.fbnode.bits & LAST_T) {
			fast = LAST_T;
			tilt = 0;
			status = line(tsline, flat, &tsnode);
			if (1) {
			    break;
			}
		    } else {
			fqnode_data = fpnode_data;
			fpnode_data = fpnode_data->fpnode_data[0];
		    }
		} else {
		    SET_V(fpnode_data);

		    if (fpnode_data->fqnode.fbnode.bits & NEXT_T) {
			fast = NEXT_T;
			tilt = 1;
			status = line(tsline, flat, &tsnode);
			if (1) {
			    break;
			}
		    } else {
			fqnode_data = fpnode_data;
			fpnode_data = fpnode_data->fpnode_data[1];
		    }
		}
	    }
	}

	if (status) {
	} else {
	    fqnode_data = tsnode;

	    *push = fqnode_data + 1;

	    fqnode_data->fpnode_data[tilt + 0] =
		fpnode_data->fpnode_data[tilt];
	    fqnode_data->fpnode_data[1 - tilt] = fpnode_data;

	    fqnode_data->fqnode.tilt = 0;

	    fqnode_data->fqnode.fbnode.bits = FLAT_T;

	    fpnode_data->fqnode.fbnode.bits &= ~fast;

	    fpnode_data->fpnode_data[tilt] = fqnode_data;

	    fpnode_data = fxnode_data;

	    do {
		unsigned post;

		post = GET_T(fpnode_data);

		if (post) {
		    fpnode_data->fqnode.tilt++;
		} else {
		    fpnode_data->fqnode.tilt--;
		}

		fpnode_data = fpnode_data->fpnode_data[post];
	    } while (fpnode_data != fqnode_data);

	    if (fxnode_data->fqnode.tilt == -2) {
		if (1) {
		    struct fpnode_type *fvnode_data, *fwnode_data;

		    fvnode_data = fxnode_data->fpnode_data[0];
		    if (fvnode_data->fqnode.tilt == -1) {
			fwnode_data = fvnode_data;
			if (GET_N(fvnode_data)) {
			    FIX_N(fvnode_data);
			    SET_L(fxnode_data);
			    fxnode_data->fpnode_data[0] = fvnode_data;
			} else {
			    struct fpnode_type *fmnode_data;

			    fmnode_data = fvnode_data->fpnode_data[1];
			    fxnode_data->fpnode_data[0] = fmnode_data;
			}
			fvnode_data->fpnode_data[1] = fxnode_data;
			fvnode_data->fqnode.tilt = 0;
			fxnode_data->fqnode.tilt = 0;
		    } else {
			int slip;

			fwnode_data = fvnode_data->fpnode_data[1];
			if (GET_L(fwnode_data)) {
			    SET_N(fvnode_data);
			    fvnode_data->fpnode_data[1] = fwnode_data;
			    FIX_L(fwnode_data);
			} else {
			    struct fpnode_type *fbnode_data;

			    fbnode_data = fwnode_data->fpnode_data[0];
			    fvnode_data->fpnode_data[1] = fbnode_data;
			}
			fwnode_data->fpnode_data[0] = fvnode_data;
			if (GET_N(fwnode_data)) {
			    SET_L(fxnode_data);
			    fxnode_data->fpnode_data[0] = fwnode_data;
			    FIX_N(fwnode_data);
			} else {
			    struct fpnode_type *fdnode_data;

			    fdnode_data = fwnode_data->fpnode_data[1];
			    fxnode_data->fpnode_data[0] = fdnode_data;
			}
			fwnode_data->fpnode_data[1] = fxnode_data;
			slip = fwnode_data->fqnode.tilt;
			fwnode_data->fqnode.tilt = 0;
			if (slip == -1) {
			    fvnode_data->fqnode.tilt = 0;
			    fxnode_data->fqnode.tilt = 1;
			} else {
			    if (!slip) {
				fvnode_data->fqnode.tilt = 0;
				fxnode_data->fqnode.tilt = 0;
			    } else {
				fvnode_data->fqnode.tilt = -1;
				fxnode_data->fqnode.tilt = 0;
			    }
			}
		    }
		    if (fxnode_data == fznode_data->fpnode_data[0]) {
			fznode_data->fpnode_data[0] = fwnode_data;
		    } else {
			fznode_data->fpnode_data[1] = fwnode_data;
		    }
		}
	    } else {
		if (fxnode_data->fqnode.tilt == 2) {
		    struct fpnode_type *fvnode_data, *fwnode_data;

		    fvnode_data = fxnode_data->fpnode_data[1];
		    if (fvnode_data->fqnode.tilt == 1) {
			fwnode_data = fvnode_data;
			if (GET_L(fvnode_data)) {
			    FIX_L(fvnode_data);
			    SET_N(fxnode_data);
			    fxnode_data->fpnode_data[1] = fvnode_data;
			} else {
			    struct fpnode_type *fmnode_data;

			    fmnode_data = fvnode_data->fpnode_data[0];
			    fxnode_data->fpnode_data[1] = fmnode_data;
			}
			fvnode_data->fpnode_data[0] = fxnode_data;
			fvnode_data->fqnode.tilt = 0;
			fxnode_data->fqnode.tilt = 0;
		    } else {
			int slip;

			fwnode_data = fvnode_data->fpnode_data[0];
			if (GET_N(fwnode_data)) {
			    SET_L(fvnode_data);
			    fvnode_data->fpnode_data[0] = fwnode_data;
			    FIX_N(fwnode_data);
			} else {
			    struct fpnode_type *fbnode_data;

			    fbnode_data = fwnode_data->fpnode_data[1];
			    fvnode_data->fpnode_data[0] = fbnode_data;
			}
			fwnode_data->fpnode_data[1] = fvnode_data;
			if (GET_L(fwnode_data)) {
			    SET_N(fxnode_data);
			    fxnode_data->fpnode_data[1] = fwnode_data;
			    FIX_L(fwnode_data);
			} else {
			    struct fpnode_type *fdnode_data;

			    fdnode_data = fwnode_data->fpnode_data[0];
			    fxnode_data->fpnode_data[1] = fdnode_data;
			}
			fwnode_data->fpnode_data[0] = fxnode_data;
			slip = fwnode_data->fqnode.tilt;
			fwnode_data->fqnode.tilt = 0;
			if (slip == 1) {
			    fvnode_data->fqnode.tilt = 0;
			    fxnode_data->fqnode.tilt = -1;
			} else {
			    if (!slip) {
				fvnode_data->fqnode.tilt = 0;
				fxnode_data->fqnode.tilt = 0;
			    } else {
				fvnode_data->fqnode.tilt = 1;
				fxnode_data->fqnode.tilt = 0;
			    }
			}
		    }
		    if (fxnode_data == fznode_data->fpnode_data[0]) {
			fznode_data->fpnode_data[0] = fwnode_data;
		    } else {
			fznode_data->fpnode_data[1] = fwnode_data;
		    }
		} else {
		}
	    }
	}
    }

    return status;
}
