/*
 * tsxv.i.c
 * Copyright (C) 2008-2010, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <tsxx-names.h>
#include <tsxx-types.h>

#define __BREAK_DELETE__		2
#define __TRACE_DELETE__		3

#define __CLASS_DELETE__		__TRACE_DELETE__

#define ____CLASS_DELETE__		__CLASS_DELETE__

#define GET_L(fpnode_data) \
    ((fpnode_data)->fqnode.fbnode.bits & LAST_T)
#define GET_N(fpnode_data) \
    ((fpnode_data)->fqnode.fbnode.bits & NEXT_T)

#define SET_L(fpnode_data) \
    (fpnode_data)->fqnode.fbnode.bits |= LAST_T
#define SET_N(fpnode_data) \
    (fpnode_data)->fqnode.fbnode.bits |= NEXT_T

#define FIX_L(fpnode_data) \
    (fpnode_data)->fqnode.fbnode.bits &= ~LAST_T
#define FIX_N(fpnode_data) \
    (fpnode_data)->fqnode.fbnode.bits &= ~NEXT_T

#if ____CLASS_DELETE__ == __BREAK_DELETE__
# define class_break(fbnode_data, fsnode_data, tslist) \
    {									      \
	if ((fsnode_data)						      \
	    == tslist((tslist))->link_a.fplist.fplink.fpnode_data[0]) {	      \
	    (fbnode_data) = &tslist((tslist))->link_a.fplist.fplink;	      \
	} else {							      \
	    struct fpnode_type *f1node_data, *f2node_data;		      \
									      \
	    f1node_data = (fsnode_data);				      \
	    f2node_data = (fsnode_data);				      \
									      \
	    while (1) {							      \
		if (GET_N(f2node_data)) {				      \
		    if (1) {						      \
			struct fpnode_type *fdnode_data;		      \
									      \
			fdnode_data = f2node_data->fpnode_data[1];	      \
			do {						      \
			    if (fdnode_data) {				      \
				if (fdnode_data->fpnode_data[0]		      \
				    == (fsnode_data)) {			      \
				    (fbnode_data) = fdnode_data;	      \
									      \
				    break;				      \
				}					      \
			    }						      \
									      \
			    while (!GET_L(f1node_data)) {		      \
				f1node_data = f1node_data->fpnode_data[0];    \
			    }						      \
									      \
			    (fbnode_data) = f1node_data->fpnode_data[0];      \
			} while (0);					      \
									      \
			break;						      \
		    }							      \
		} else {						      \
		    if (GET_L(f1node_data)) {				      \
			struct fpnode_type *fdnode_data;		      \
									      \
			fdnode_data = f1node_data->fpnode_data[0];	      \
			do {						      \
			    if (fdnode_data) {				      \
				if (fdnode_data->fpnode_data[1]		      \
				    == (fsnode_data)) {			      \
				    (fbnode_data) = fdnode_data;	      \
									      \
				    break;				      \
				}					      \
			    }						      \
									      \
			    while (!GET_N(f2node_data)) {		      \
				f2node_data = f2node_data->fpnode_data[1];    \
			    }						      \
									      \
			    (fbnode_data) = f2node_data->fpnode_data[1];      \
			} while (0);					      \
									      \
			break;						      \
		    } else {						      \
			f1node_data = f1node_data->fpnode_data[0];	      \
			f2node_data = f2node_data->fpnode_data[1];	      \
		    }							      \
		}							      \
	    }								      \
	}								      \
    }
#endif				/* ____CLASS_DELETE__ == __BREAK_DELETE__ */

#if ____CLASS_DELETE__ == __TRACE_DELETE__
# define shift_break(fsnode_data, fxnode_data, fznode_data) \
    {									      \
	(fznode_data) = (fxnode_data)->fpnode_data[1];			      \
	(fxnode_data)->fpnode_data[1] = (fsnode_data);			      \
    }
#endif				/* ____CLASS_DELETE__ == __TRACE_DELETE__ */

#define tslist(tslist)			((struct tslist_type *) (tslist))

int
_libx1f4l2_away_tslist(void *tslist, void *back, int (*call) (void *, void *))
{
    int delete;
#if ____CLASS_DELETE__ == __TRACE_DELETE__
    struct fpnode_type *ffnode_data;
#endif				/* ____CLASS_DELETE__ == __TRACE_DELETE__ */
    struct fpnode_type *fhnode_data, fpnode, *fpnode_data;

    fhnode_data = &tslist(tslist)->link_a.fplist.fplink;

#if ____CLASS_DELETE__ == __TRACE_DELETE__
    ffnode_data = fhnode_data;
#endif				/* ____CLASS_DELETE__ == __TRACE_DELETE__ */

    fpnode_data = tslist(tslist)->link_a.fplist.fplink.fpnode_data[0];

    while (1) {
	if (fpnode_data->fqnode.fbnode.bits & NEXT_T) {
	    break;
	}

#if ____CLASS_DELETE__ == __TRACE_DELETE__
	if (fpnode_data->fqnode.tilt) {
	} else {
	    ffnode_data = fhnode_data;
	}
#endif				/* ____CLASS_DELETE__ == __TRACE_DELETE__ */

	fpnode_data->fqnode.size--;

	fhnode_data = fpnode_data;
	fpnode_data = fpnode_data->fpnode_data[1];
    }

    fpnode = *fpnode_data;

    delete = call(back, fpnode_data);
    if (delete) {
	struct fpnode_type *fvnode_data;

	fvnode_data = tslist(tslist)->link_a.fplist.fplink.fpnode_data[0];

	while (fvnode_data != fpnode_data) {
	    fvnode_data->fqnode.size++;
	    fvnode_data = fvnode_data->fpnode_data[1];
	}
    } else {
#if ____CLASS_DELETE__ == __TRACE_DELETE__
	struct fpnode_type *fvnode_data;
#endif				/* ____CLASS_DELETE__ == __TRACE_DELETE__ */

	tslist(tslist)->link_a.fplist.fplink.fpnode_data[1] =
	    tslist(tslist)->link_a.fplist.fplink.fpnode_data[0];

#if ____CLASS_DELETE__ == __TRACE_DELETE__
	if (ffnode_data == &tslist(tslist)->link_a.fplist.fplink) {
	    fvnode_data = tslist(tslist)->link_a.fplist.fplink.fpnode_data[1];
	} else {
	    fvnode_data = ffnode_data->fpnode_data[1];
	}
	fhnode_data = ffnode_data;
	if (fvnode_data == fpnode_data) {
	} else {
	    while (1) {
		struct fpnode_type *fqnode_data;

		shift_break(fhnode_data, fvnode_data, fqnode_data);
		if (fqnode_data == fpnode_data) {
		    if (1) {
			fhnode_data = fvnode_data;
		    }

		    break;
		} else {
		    shift_break(fvnode_data, fqnode_data, fhnode_data);
		    if (fhnode_data == fpnode_data) {
			if (1) {
			    fhnode_data = fqnode_data;
			}

			break;
		    } else {
			shift_break(fqnode_data, fhnode_data, fvnode_data);
			if (fvnode_data == fpnode_data) {
			    if (0) {
				fhnode_data = fhnode_data;
			    }

			    break;
			}
		    }
		}
	    }
	}
#endif				/* ____CLASS_DELETE__ == __TRACE_DELETE__ */

	fpnode_data = &fpnode;

	do {
#if ____CLASS_DELETE__ == __TRACE_DELETE__
	    unsigned slip = 0;
#endif				/* ____CLASS_DELETE__ == __TRACE_DELETE__ */

	    if (1) {
#if ____CLASS_DELETE__ == __TRACE_DELETE__
		fvnode_data = fpnode_data->fpnode_data[0];
#endif				/* ____CLASS_DELETE__ == __TRACE_DELETE__ */

		if (GET_L(fpnode_data)) {
#if ____CLASS_DELETE__ == __TRACE_DELETE__
		    fvnode_data = fpnode_data->fpnode_data[1];
#elif __CLASS_DELETE__ == __BREAK_DELETE__
		    fhnode_data->fpnode_data[1] = fpnode_data->fpnode_data[1];
#endif				/* ____CLASS_DELETE__ == __TRACE_DELETE__ */

		    if (fhnode_data != &tslist(tslist)->link_a.fplist.fplink) {
			if (0) {
			} else {
			    SET_N(fhnode_data);
			}
		    }
		} else {
		    struct fpnode_type *ftnode_data;

		    ftnode_data = fpnode_data->fpnode_data[0];

#if ____CLASS_DELETE__ == __TRACE_DELETE__
		    fvnode_data = ftnode_data;
#elif __CLASS_DELETE__ == __BREAK_DELETE__
		    fhnode_data->fpnode_data[1] = ftnode_data;
#endif				/* ____CLASS_DELETE__ == __TRACE_DELETE__ */

		    while (!GET_N(ftnode_data)) {
			ftnode_data = ftnode_data->fpnode_data[1];
		    }

		    ftnode_data->fpnode_data[1] = fpnode_data->fpnode_data[1];
		}
	    }

	    while (1) {
		struct fpnode_type *fqnode_data;

#if ____CLASS_DELETE__ == __BREAK_DELETE__
		fqnode_data = fhnode_data;
#endif				/* ____CLASS_DELETE__ == __BREAK_DELETE__ */

#if ____CLASS_DELETE__ == __BREAK_DELETE__
		if (fqnode_data == &tslist(tslist)->link_a.fplist.fplink) {
		    break;
		}
#endif				/* ____CLASS_DELETE__ == __BREAK_DELETE__ */

#if ____CLASS_DELETE__ == __BREAK_DELETE__
		class_break(fhnode_data, fqnode_data, tslist);
#endif				/* ____CLASS_DELETE__ == __BREAK_DELETE__ */

#if ____CLASS_DELETE__ == __TRACE_DELETE__
		if (fhnode_data == ffnode_data) {
		    fhnode_data->fpnode_data[1] = fvnode_data;

		    break;
		} else {
		    fqnode_data = fhnode_data;
		    fhnode_data = fqnode_data->fpnode_data[1];
		    fqnode_data->fpnode_data[1] = fvnode_data;
		    fvnode_data = fqnode_data;

		    if (slip) {
			continue;
		    }
		}
#endif				/* ____CLASS_DELETE__ == __TRACE_DELETE__ */

#if ____CLASS_DELETE__ == __TRACE_DELETE__
		if (1) {
#elif 0
		}
#elif __CLASS_DELETE__ == __BREAK_DELETE__
		if (1) {
#endif				/* ____CLASS_DELETE__ == __TRACE_DELETE__ */
		    int tilt;

		    tilt = fqnode_data->fqnode.tilt;

		    tilt--;

		    fqnode_data->fqnode.tilt = tilt;

		    if (tilt == -1) {
#if ____CLASS_DELETE__ == __TRACE_DELETE__
			slip |= 4096;
#elif __CLASS_DELETE__ == __BREAK_DELETE__
			break;
#endif				/* ____CLASS_DELETE__ == __TRACE_DELETE__ */
		    } else {
			if (tilt == -2) {
			    struct fpnode_type *fxnode_data;

			    fxnode_data = fqnode_data->fpnode_data[0];
			    if (fxnode_data->fqnode.tilt == 1) {
				struct fpnode_type *fwnode_data;
				unsigned sx_b, sx_d;

				fwnode_data = fxnode_data->fpnode_data[1];
				if (GET_L(fwnode_data)) {
				    FIX_L(fwnode_data);
				    SET_N(fxnode_data);
				    fxnode_data->fpnode_data[1] = fwnode_data;
				    sx_b = 0;
				} else {
				    struct fpnode_type *fbnode_data;

				    fbnode_data = fwnode_data->fpnode_data[0];
				    fxnode_data->fpnode_data[1] = fbnode_data;
				    sx_b = fbnode_data->fqnode.size;
				}
				fwnode_data->fpnode_data[0] = fxnode_data;
				if (GET_N(fwnode_data)) {
				    FIX_N(fwnode_data);
				    SET_L(fqnode_data);
				    fqnode_data->fpnode_data[0] = fwnode_data;
				    sx_d = 0;
				} else {
				    struct fpnode_type *fdnode_data;

				    fdnode_data = fwnode_data->fpnode_data[1];
				    fqnode_data->fpnode_data[0] = fdnode_data;
				    sx_d = fdnode_data->fqnode.size;
				}
				fwnode_data->fpnode_data[1] = fqnode_data;
				tilt = fwnode_data->fqnode.tilt;
				fwnode_data->fqnode.tilt = 0;
				if (tilt == -1) {
				    fxnode_data->fqnode.tilt = 0;
				    fqnode_data->fqnode.tilt = 1;
				} else {
				    if (!tilt) {
					fxnode_data->fqnode.tilt = 0;
					fqnode_data->fqnode.tilt = 0;
				    } else {
					fxnode_data->fqnode.tilt = -1;
					fqnode_data->fqnode.tilt = 0;
				    }
				}
#if ____CLASS_DELETE__ == __TRACE_DELETE__
				fvnode_data = fwnode_data;
#elif __CLASS_DELETE__ == __BREAK_DELETE__
				fhnode_data->fpnode_data[1] = fwnode_data;
#endif				/* ____CLASS_DELETE__ == __TRACE_DELETE__ */
				if (1) {
				    unsigned sx_w, sx_v, sx_x;

				    sx_v = fxnode_data->fqnode.size;
				    sx_w = fwnode_data->fqnode.size;
				    sx_x = fqnode_data->fqnode.size;
				    fxnode_data->fqnode.size =
					sx_v + sx_b - sx_w;
				    fwnode_data->fqnode.size = sx_x;
				    fqnode_data->fqnode.size =
					sx_x + sx_d - sx_v;
				}
			    } else {
				unsigned sx_m;

				if (GET_N(fxnode_data)) {
				    sx_m = 0;
				    if (fxnode_data->fqnode.tilt) {
					FIX_N(fxnode_data);
					SET_L(fqnode_data);
				    } else {
					fqnode_data->fpnode_data[0] =
					    fxnode_data->fpnode_data[1];
				    }
				} else {
				    struct fpnode_type *fmnode_data;

				    fmnode_data = fxnode_data->fpnode_data[1];
				    fqnode_data->fpnode_data[0] = fmnode_data;
				    sx_m = fmnode_data->fqnode.size;
				}
				fxnode_data->fpnode_data[1] = fqnode_data;
#if ____CLASS_DELETE__ == __TRACE_DELETE__
				fvnode_data = fxnode_data;
#elif __CLASS_DELETE__ == __BREAK_DELETE__
				fhnode_data->fpnode_data[1] = fxnode_data;
#endif				/* ____CLASS_DELETE__ == __TRACE_DELETE__ */
				if (1) {
				    unsigned sx_q, sx_x;

				    sx_x = fxnode_data->fqnode.size;
				    sx_q = fqnode_data->fqnode.size;
				    fxnode_data->fqnode.size = sx_q;
				    fqnode_data->fqnode.size =
					sx_q + sx_m - sx_x;
				}
				if (fxnode_data->fqnode.tilt) {
				    fqnode_data->fqnode.tilt = 0;
				    fxnode_data->fqnode.tilt = 0;
				} else {
				    fqnode_data->fqnode.tilt = -1;
				    fxnode_data->fqnode.tilt = 1;
#if ____CLASS_DELETE__ == __TRACE_DELETE__
				    slip |= 4096;
#elif __CLASS_DELETE__ == __BREAK_DELETE__
				    break;
#endif				/* ____CLASS_DELETE__ == __TRACE_DELETE__ */
				}
			    }
			}
		    }
		}
	    }
	} while (0);

	tslist(tslist)->link_a.fplist.fplink.fpnode_data[0] =
	    tslist(tslist)->link_a.fplist.fplink.fpnode_data[1];
    }

    return delete;
}
