/*
 * tvlist.h
 * Copyright (C) 2008-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef __LIBx1f4_TVLIST_H__
#define __LIBx1f4_TVLIST_H__

#include <errors.h>

/*
 * attribute bit definitions
 */
#define X1f4_TVLIST_TRANS_MASK		(1u << 0x00)

/*
 * error definitions
 */
#define X1f4_TVLIST_CALL_ERROR		LIBx1f4L2_CALL_ERROR
#define X1f4_TVLIST_DECK_ERROR		LIBx1f4L2_DECK_ERROR
#define X1f4_TVLIST_FREE_ERROR		LIBx1f4L2_FREE_ERROR
#define X1f4_TVLIST_LINK_ERROR		LIBx1f4L2_LINK_ERROR
#define X1f4_TVLIST_MODE_ERROR		LIBx1f4L2_MODE_ERROR
#define X1f4_TVLIST_SIZE_ERROR		LIBx1f4L2_SIZE_ERROR

/*
 * error definitions (continued)
 */
#define X1f4_TVLIST_EVER_MATCH		LIBx1f4L2_EVER_MATCH

typedef struct x1f4_tvlist_type {
    void *trans;
} x1f4_tvlist_type;

#if !defined __x1f4_trans_type_defined__
typedef struct x1f4_trans_type {
    int (*free) (void *, void *), (*link) (void *, void **, unsigned),
	(*mode) (void *, void **, unsigned);
    void *context;
} x1f4_trans_type;
#endif				/* !__x1f4_trans_type_defined__ */

#if !defined __x1f4_trans_type_defined__
# define __x1f4_trans_type_defined__
#endif				/* !__x1f4_trans_type_defined__ */

extern int x1f4_call_tvlist(unsigned *);
extern int x1f4_clip_tvlist
    (void *, unsigned, void *, int (*) (void *, void *));
extern int x1f4_deck_tvlist(void *, int);
extern int x1f4_deux_tvlist(void *, unsigned, void *, unsigned);
extern int x1f4_ever_tvlist(void *, void **);
extern int x1f4_fast_tvlist(void *, unsigned, struct x1f4_tvlist_type *);
extern int x1f4_fgxp_tvlist
    (void *, void *, unsigned, int (*) (void *, void *), unsigned *);
extern int x1f4_find_tvlist
    (void *, void *, unsigned, int (*) (void *, void *), void **);
extern int x1f4_fine_tvlist(void *, void *, int (*) (void *, void *));
extern int x1f4_fini_tvlist(void **);
extern int x1f4_flat_tvlist(void *);
extern int x1f4_flip_tvlist(void *);
extern int x1f4_flow_tvlist(void *);
extern int x1f4_head_tvlist(void *, unsigned, void **);
extern int x1f4_high_tvlist(void *);
extern int x1f4_init_tvlist(void **, unsigned, struct x1f4_tvlist_type *);
extern int x1f4_land_tvlist
    (void *, void *, int (*) (void *, void *), unsigned, unsigned, void **);
extern int x1f4_last_tvlist(void **, void *);
extern int x1f4_lead_tvlist(void *, void **);
extern int x1f4_lime_tvlist(void *, void *, int (*) (void *, void *));
extern int x1f4_line_tvlist
    (void *, void *, int (*) (void *, unsigned *, const void *),
     int (*) (void *, void *, const void *), int (*) (void *, void *), void *);
extern int x1f4_mind_tvlist(void *);
extern int x1f4_miss_tvlist(void *, unsigned);
extern int x1f4_mode_tvlist(void *, unsigned, unsigned, void **);
extern int x1f4_move_tvlist(void *, unsigned, unsigned);
extern int x1f4_next_tvlist(void **, void *);
extern int x1f4_nice_tvlist(void *, void *, int (*) (void *, void *));
extern int x1f4_peek_tvlist(void *, unsigned, void **);
extern int x1f4_post_tvlist
    (void *, void *, int (*) (void *, void *), unsigned, unsigned, void **);
extern int x1f4_pull_tvlist
    (void *, unsigned, void *, int (*) (void *, void *));
extern int x1f4_push_tvlist(void *, unsigned, unsigned, void **);
extern int x1f4_rand_tvlist(void *);
extern int x1f4_seek_tvlist
    (void *, void *, unsigned, int (*) (void *, void *), unsigned *);
extern int x1f4_side_tvlist(void *, void *);
extern int x1f4_size_tvlist(void *, unsigned *);
extern int x1f4_skew_tvlist(void *, void *);
extern int x1f4_slip_tvlist(void *);
extern int x1f4_sort_tvlist
    (void *, void *, unsigned, int (*) (void *, void *, void *));
extern int x1f4_spin_tvlist(void *, unsigned, unsigned);
extern int x1f4_tail_tvlist(void *, unsigned, void **);
extern int x1f4_text_tvlist
    (void **, void *, int (*) (void *, unsigned *, const void *),
     int (*) (void *, void *, const void *), int (*) (void *, void *), void *);
extern int x1f4_tick_tvlist
    (void *, void *, unsigned, int (*) (void *, void *), unsigned *, void **);
extern int x1f4_tile_tvlist(unsigned *, void *);
extern int x1f4_turn_tvlist(void *);
extern int x1f4_wind_tvlist(void *, unsigned, void *, unsigned);

#endif				/* __LIBx1f4_TVLIST_H__ */
