/*
 * tvqp.i.c
 * Copyright (C) 2008-2010, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <sign.x.h>
#include <tvqp-defs.h>
#include <tvqp-names.h>
#include <tvqp-types.h>

#define __BREAK_RECORD__		1
#define __SLIDE_RECORD__		2

#define __TRACK_METHOD__		__BREAK_RECORD__

#define ____TRACK_METHOD__		__TRACK_METHOD__

#define tvline(tvline)			((struct tvline_type *) (tvline))

int
_libx1f4l2_elif_tvline(void *tvline, void *text, int (*land) (void *, void *),
		       unsigned offset, unsigned flat, void **push,
		       int (*line) (void *, unsigned, void **))
{
    int status = 0;
    struct fpnode_type *fpnode_data, *fqnode_data;
#if ____TRACK_METHOD__ == __BREAK_RECORD__
    struct fpnode_type *fxnode_data;
#elif __TRACK_METHOD__ == __SLIDE_RECORD__
    struct fpnode_type *sxnode[2];
#endif				/* ____TRACK_METHOD__ == __BREAK_RECORD__ */
    unsigned tilt = 0;
    void *tvnode;

#if ____TRACK_METHOD__ == __BREAK_RECORD__
    fxnode_data = tvline(tvline)->link_a.fpline.fplink.fpnode_data[0];
#elif __TRACK_METHOD__ == __SLIDE_RECORD__
    fpnode_data = tvline(tvline)->link_a.fpline.fplink.fpnode_data[0];
#endif				/* ____TRACK_METHOD__ == __BREAK_RECORD__ */

#if ____TRACK_METHOD__ == __SLIDE_RECORD__
    sxnode[1] = fpnode_data;
#endif				/* ____TRACK_METHOD__ == __SLIDE_RECORD__ */

#if ____TRACK_METHOD__ == __BREAK_RECORD__
    fpnode_data = fxnode_data;
#endif				/* ____TRACK_METHOD__ == __BREAK_RECORD__ */
    fqnode_data = &tvline(tvline)->link_a.fpline.fplink;

    if (fpnode_data) {
	while (1) {
	    int slip;

	    slip = land(text, (char *) fpnode_data + offset);

	    if (slip) {
#if ____TRACK_METHOD__ == __BREAK_RECORD__
		if (fpnode_data->fqnode.tilt) {
		    fxnode_data = fpnode_data;
		}
#elif __TRACK_METHOD__ == __SLIDE_RECORD__
		if (1) {
		    sxnode[fpnode_data->fqnode.tilt & 1] = fpnode_data;
		}
#endif				/* ____TRACK_METHOD__ == __BREAK_RECORD__ */

#if 1
		tilt = class_integer(slip);
#else
		if (slip < 0) {
		    tilt = 0;
		} else {
		    tilt = 1;
		}
#endif				/* 1 */

		fqnode_data = fpnode_data;
		fpnode_data = fpnode_data->fpnode_data[tilt];
		if (fpnode_data) {
		} else {
		    if (1) {
			break;
		    }
		}
	    } else {
		status = EVER_MATCH;

		*push = fpnode_data + 1;

		if (1) {
		    break;
		}
	    }
	}
    }

    if (status) {
    } else {
	status = line(tvline, flat, &tvnode);
    }

    if (status) {
    } else {
	struct fpnode_type *fnnode_data;
#if ____TRACK_METHOD__ == __SLIDE_RECORD__
	struct fpnode_type *fxnode_data;
#endif				/* ____TRACK_METHOD__ == __SLIDE_RECORD__ */

	fnnode_data = tvnode;

	*push = fnnode_data + 1;

	fnnode_data->fpnode_data[0] = NULL;
	fnnode_data->fpnode_data[1] = NULL;

	fnnode_data->fqnode.tilt = 0;

	fqnode_data->fpnode_data[tilt] = fnnode_data;

#if ____TRACK_METHOD__ == __SLIDE_RECORD__
	fxnode_data = sxnode[1];
#endif				/* ____TRACK_METHOD__ == __SLIDE_RECORD__ */

	if (!fxnode_data) {
	    fnnode_data->uplink_data = NULL;
	} else {
	    fpnode_data = fnnode_data;

	    fnnode_data->uplink_data = fqnode_data;

	    do {
		fqnode_data = fpnode_data->uplink_data;
		if (fpnode_data == fqnode_data->fpnode_data[0]) {
		    fqnode_data->fqnode.tilt--;
		} else {
		    fqnode_data->fqnode.tilt++;
		}

		fpnode_data = fqnode_data;
	    } while (fpnode_data != fxnode_data);

	    if (fxnode_data->fqnode.tilt == -2) {
		if (1) {
		    struct fpnode_type *fvnode_data, *fwnode_data;

		    fvnode_data = fxnode_data->fpnode_data[0];
		    if (fvnode_data->fqnode.tilt == -1) {
			struct fpnode_type *fmnode_data;

			fwnode_data = fvnode_data;
			fmnode_data = fvnode_data->fpnode_data[1];
			fxnode_data->fpnode_data[0] = fmnode_data;
			fvnode_data->fpnode_data[1] = fxnode_data;
			fvnode_data->fqnode.tilt = 0;
			fxnode_data->fqnode.tilt = 0;
			fvnode_data->uplink_data = fxnode_data->uplink_data;
			fxnode_data->uplink_data = fvnode_data;
			if (fmnode_data) {
			    fmnode_data->uplink_data = fxnode_data;
			}
		    } else {
			int slip;
			struct fpnode_type *fbnode_data, *fdnode_data;

			fwnode_data = fvnode_data->fpnode_data[1];
			fbnode_data = fwnode_data->fpnode_data[0];
			fvnode_data->fpnode_data[1] = fbnode_data;
			fwnode_data->fpnode_data[0] = fvnode_data;
			fdnode_data = fwnode_data->fpnode_data[1];
			fxnode_data->fpnode_data[0] = fdnode_data;
			fwnode_data->fpnode_data[1] = fxnode_data;
			slip = fwnode_data->fqnode.tilt;
			fwnode_data->fqnode.tilt = 0;
			if (slip == -1) {
			    fvnode_data->fqnode.tilt = 0;
			    fxnode_data->fqnode.tilt = 1;
			} else {
			    if (!slip) {
				fvnode_data->fqnode.tilt = 0;
				fxnode_data->fqnode.tilt = 0;
			    } else {
				fvnode_data->fqnode.tilt = -1;
				fxnode_data->fqnode.tilt = 0;
			    }
			}
			fwnode_data->uplink_data = fxnode_data->uplink_data;
			fvnode_data->uplink_data = fwnode_data;
			fxnode_data->uplink_data = fwnode_data;
			if (fbnode_data) {
			    fbnode_data->uplink_data = fvnode_data;
			}
			if (fdnode_data) {
			    fdnode_data->uplink_data = fxnode_data;
			}
		    }
		    fvnode_data = fwnode_data->uplink_data;
		    if (fvnode_data) {
			if (fxnode_data == fvnode_data->fpnode_data[0]) {
			    fvnode_data->fpnode_data[0] = fwnode_data;
			} else {
			    fvnode_data->fpnode_data[1] = fwnode_data;
			}
		    } else {
			tvline(tvline)->link_a.fpline.fplink.fpnode_data[0] =
			    fwnode_data;
		    }
		}
	    } else {
		if (fxnode_data->fqnode.tilt == 2) {
		    struct fpnode_type *fvnode_data, *fwnode_data;

		    fvnode_data = fxnode_data->fpnode_data[1];
		    if (fvnode_data->fqnode.tilt == 1) {
			struct fpnode_type *fmnode_data;

			fwnode_data = fvnode_data;
			fmnode_data = fvnode_data->fpnode_data[0];
			fxnode_data->fpnode_data[1] = fmnode_data;
			fvnode_data->fpnode_data[0] = fxnode_data;
			fvnode_data->fqnode.tilt = 0;
			fxnode_data->fqnode.tilt = 0;
			fvnode_data->uplink_data = fxnode_data->uplink_data;
			fxnode_data->uplink_data = fvnode_data;
			if (fmnode_data) {
			    fmnode_data->uplink_data = fxnode_data;
			}
		    } else {
			int slip;
			struct fpnode_type *fbnode_data, *fdnode_data;

			fwnode_data = fvnode_data->fpnode_data[0];
			fbnode_data = fwnode_data->fpnode_data[1];
			fvnode_data->fpnode_data[0] = fbnode_data;
			fwnode_data->fpnode_data[1] = fvnode_data;
			fdnode_data = fwnode_data->fpnode_data[0];
			fxnode_data->fpnode_data[1] = fdnode_data;
			fwnode_data->fpnode_data[0] = fxnode_data;
			slip = fwnode_data->fqnode.tilt;
			fwnode_data->fqnode.tilt = 0;
			if (slip == 1) {
			    fvnode_data->fqnode.tilt = 0;
			    fxnode_data->fqnode.tilt = -1;
			} else {
			    if (!slip) {
				fvnode_data->fqnode.tilt = 0;
				fxnode_data->fqnode.tilt = 0;
			    } else {
				fvnode_data->fqnode.tilt = 1;
				fxnode_data->fqnode.tilt = 0;
			    }
			}
			fwnode_data->uplink_data = fxnode_data->uplink_data;
			fvnode_data->uplink_data = fwnode_data;
			fxnode_data->uplink_data = fwnode_data;
			if (fbnode_data) {
			    fbnode_data->uplink_data = fvnode_data;
			}
			if (fdnode_data) {
			    fdnode_data->uplink_data = fxnode_data;
			}
		    }
		    fvnode_data = fwnode_data->uplink_data;
		    if (fvnode_data) {
			if (fxnode_data == fvnode_data->fpnode_data[0]) {
			    fvnode_data->fpnode_data[0] = fwnode_data;
			} else {
			    fvnode_data->fpnode_data[1] = fwnode_data;
			}
		    } else {
			tvline(tvline)->link_a.fpline.fplink.fpnode_data[0] =
			    fwnode_data;
		    }
		} else {
		}
	    }
	}
    }

    return status;
}
