/*
 * tvxx.3.c
 * Copyright (C) 2008-2010, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <tvxx-names.h>
#include <tvxx-types.h>

#define tvlist(tvlist)			((struct tvlist_type *) (tvlist))

static int flow_list(struct fpnode_type *);
static int lock_list(struct fpnode_type *, unsigned *);
static int miss_list(struct fpnode_type *, struct fpnode_type *, unsigned *);

static int
flow_list(struct fpnode_type *fpnode_data)
{
    unsigned flow;

    return miss_list(fpnode_data, NULL, &flow);
}


static int
lock_list(struct fpnode_type *fpnode_data, unsigned *fast)
{
    int status;

    do {
	struct fpnode_type *fqnode_data;
	unsigned size = 1;

	fqnode_data = fpnode_data->fpnode_data[0];
	if (fqnode_data) {
	    unsigned fast;

	    status = miss_list(fqnode_data, fpnode_data, &fast);
	    if (status) {
		break;
	    } else {
		size += fast;
	    }
	}

	fqnode_data = fpnode_data->fpnode_data[1];
	if (fqnode_data) {
	    unsigned fast;

	    status = miss_list(fqnode_data, fpnode_data, &fast);
	    if (status) {
		break;
	    } else {
		size += fast;
	    }
	}

	if (size == fpnode_data->fqnode.size) {
	    status = 0;

	    *fast = size;
	} else {
	    status = 2;
	}
    } while (0);

    return status;
}


static int
miss_list(struct fpnode_type *fpnode_data, struct fpnode_type *uplink_data,
	  unsigned *fast)
{
    int status;

    if (fpnode_data->uplink_data != uplink_data) {
	status = 3;
    } else {
	status = lock_list(fpnode_data, fast);
    }

    return status;
}


int
x1f4_flow_tvlist(void *tvlist)
{
    int status;
    struct fpnode_type *fpnode_data;

    fpnode_data = tvlist(tvlist)->link_a.fplist.fplink.fpnode_data[0];
    if (!fpnode_data) {
	if (tvlist(tvlist)->link_a.fplist.fplink.fqnode.size) {
	    status = 1;
	} else {
	    status = 0;
	}
    } else {
	status = flow_list(fpnode_data);
    }

    return status;
}
