/*
 * tvxy.f.c
 * Copyright (C) 2008-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <stddef.h>

#include <tvxx-inter.h>
#include <tvxx-names.h>
#include <tvxx-types.h>

#define __QUICK_SORT__			1
#define __TRANS_SORT__			2

#define __CLASS_SORT__			__TRANS_SORT__

#define ____CLASS_SORT__		__CLASS_SORT__

#define tvlist(tvlist)			((struct tvlist_type *) (tvlist))

#if ____CLASS_SORT__ == __QUICK_SORT__
static int last_node(void **, void *);
static int next_node(void **, void *);
static int sort_list(struct tvlist_type *, void *, unsigned, void *, unsigned,
		     void *, unsigned, int (*) (void *, void *, void *));
#elif __CLASS_SORT__ == __TRANS_SORT__
static int sort_list(struct tvlist_type *, struct fpnode_type *, void *,
		     unsigned, int (*) (void *, void *, void *));
#endif				/* ____CLASS_SORT__ == __QUICK_SORT__ */
#if ____CLASS_SORT__ == __QUICK_SORT__
static int spin_list(struct tvlist_type *, void *, void *);
#endif				/* ____CLASS_SORT__ == __QUICK_SORT__ */

#if ____CLASS_SORT__ == __QUICK_SORT__
static int
last_node(void **last, void *data)
{
    struct fpnode_type *fpnode_data, *fqnode_data;

    fpnode_data = (struct fpnode_type *) data;

    fqnode_data = fpnode_data->fpnode_data[0];
    if (fqnode_data) {
	while (1) {
	    fpnode_data = fqnode_data->fpnode_data[1];
	    if (!fpnode_data) {
		*last = fqnode_data;

		break;
	    } else {
		fqnode_data = fpnode_data->fpnode_data[1];
		if (!fqnode_data) {
		    *last = fpnode_data;

		    break;
		}
	    }
	}
    } else {
	while (1) {
	    fqnode_data = fpnode_data->uplink_data;
	    if (0) {
	    } else {
		if (fpnode_data == fqnode_data->fpnode_data[1]) {
		    *last = fqnode_data;

		    break;
		} else {
		    fpnode_data = fqnode_data->uplink_data;
		    if (0) {
			break;
		    } else {
			if (fqnode_data == fpnode_data->fpnode_data[1]) {
			    *last = fpnode_data;

			    break;
			} else {
			}
		    }
		}
	    }
	}
    }

    return 0;
}
#endif				/* ____CLASS_SORT__ == __QUICK_SORT__ */


#if ____CLASS_SORT__ == __QUICK_SORT__
static int
next_node(void **next, void *data)
{
    struct fpnode_type *fpnode_data, *fqnode_data;

    fpnode_data = (struct fpnode_type *) data;

    fqnode_data = fpnode_data->fpnode_data[1];
    if (fqnode_data) {
	while (1) {
	    fpnode_data = fqnode_data->fpnode_data[0];
	    if (!fpnode_data) {
		*next = fqnode_data;

		break;
	    } else {
		fqnode_data = fpnode_data->fpnode_data[0];
		if (!fqnode_data) {
		    *next = fpnode_data;

		    break;
		}
	    }
	}
    } else {
	while (1) {
	    fqnode_data = fpnode_data->uplink_data;
	    if (0) {
	    } else {
		if (fpnode_data == fqnode_data->fpnode_data[0]) {
		    *next = fqnode_data;

		    break;
		} else {
		    fpnode_data = fqnode_data->uplink_data;
		    if (0) {
		    } else {
			if (fqnode_data == fpnode_data->fpnode_data[0]) {
			    *next = fpnode_data;

			    break;
			} else {
			}
		    }
		}
	    }
	}
    }

    return 0;
}
#endif				/* ____CLASS_SORT__ == __QUICK_SORT__ */


#if ____CLASS_SORT__ == __QUICK_SORT__
static int
sort_list(struct tvlist_type *tvlist_data, void *fphead, unsigned head,
	  void *fptail, unsigned tail, void *text, unsigned slip,
	  int (*land) (void *, void *, void *))
{
    void *fpnode, *fxnode;

    if (fphead) {
	fpnode = fphead;
    } else {
	_libx1f4l2_else_tvlist(tvlist_data, head, &fpnode);
    }

    while ((int) head + 2 < (int) tail) {
	unsigned ever, call, lead;
	void *pick;

	ever = tail;
	lead = head;

	if (fptail) {
	    fxnode = fptail;
	} else {
	    _libx1f4l2_else_tvlist(tvlist_data, tail, &fxnode);
	}

	if (3 < tail - head) {
	    void *qxx1, *qxx2, *qxx3;

	    qxx1 = fpnode;
	    qxx1 = (char *) qxx1 + slip;
	    if (head & 1) {
		last_node(&qxx2, fxnode);
	    } else {
		next_node(&qxx2, fpnode);
	    }
	    qxx2 = (char *) qxx2 + slip;
	    qxx3 = fxnode;
	    qxx3 = (char *) qxx3 + slip;
	    if (land(text, qxx1, qxx2) < 0) {
		if (land(text, qxx2, qxx3) < 0) {
		    fphead = fpnode;
		    fptail = fxnode;
		    pick = qxx2;
		} else {
		    if (land(text, qxx1, qxx3) < 0) {
			fphead = fpnode;
			fptail = NULL;
			pick = qxx3;
		    } else {
			fphead = fxnode;
			fptail = fpnode;
			pick = qxx1;
		    }
		}
	    } else {
		if (land(text, qxx1, qxx3) < 0) {
		    fphead = NULL;
		    fptail = fxnode;
		    pick = qxx1;
		} else {
		    if (land(text, qxx2, qxx3) < 0) {
			fphead = fxnode;
			fptail = fpnode;
			pick = qxx3;
		    } else {
			fphead = fxnode;
			fptail = fpnode;
			pick = qxx2;
		    }
		}
	    }
	} else {
	    fphead = NULL;
	    fptail = NULL;

	    pick = (char *) fpnode + slip;
	}

	while (1) {
	    while (land(text, (char *) fpnode + slip, pick) < 0) {
		head++;
		next_node(&fpnode, fpnode);
	    }
	    while (land(text, pick, (char *) fxnode + slip) < 0) {
		tail--;
		last_node(&fxnode, fxnode);
	    }
	    if (head < tail) {
		void *finode;

		spin_list(tvlist_data, fpnode, fxnode);

		head++;
		last_node(&finode, fpnode);
		tail--;
		next_node(&fpnode, fxnode);

		fxnode = finode;
	    } else {
		call = tail;
		break;
	    }
	}

	if (head ^ (call + 1)) {
	    head = call + 1;
	    next_node(&fpnode, fxnode);
	} else {
	}

	tail = ever;

	sort_list(tvlist_data, fphead, lead, fxnode, call, text, slip, land);
    }

    if ((int) head < (int) tail) {
	unsigned size;

	size = tail - head;
	if (1 < size) {
	    void *fnnode;

	    fnnode = fpnode;

	    next_node(&fpnode, fnnode);

	    next_node(&fxnode, fpnode);

	    if (land(text, (char *) fnnode + slip, (char *) fpnode + slip)
		< 0) {
		if (land(text, (char *) fpnode + slip, (char *) fxnode + slip)
		    < 0) {
		} else {
		    if (land
			(text, (char *) fnnode + slip, (char *) fxnode + slip)
			< 0) {
			spin_list(tvlist_data, fpnode, fxnode);
		    } else {
			spin_list(tvlist_data, fnnode, fxnode);
			spin_list(tvlist_data, fpnode, fnnode);
		    }
		}
	    } else {
		if (land(text, (char *) fpnode + slip, (char *) fxnode + slip)
		    < 0) {
		    if (land
			(text, (char *) fnnode + slip, (char *) fxnode + slip)
			< 0) {
			spin_list(tvlist_data, fnnode, fpnode);
		    } else {
			spin_list(tvlist_data, fnnode, fxnode);
			spin_list(tvlist_data, fxnode, fpnode);
		    }
		} else {
		    spin_list(tvlist_data, fnnode, fxnode);
		}
	    }
	} else {
	    if (size) {
		next_node(&fxnode, fpnode);

		if (land(text, (char *) fpnode + slip, (char *) fxnode + slip)
		    < 0) {
		} else {
		    spin_list(tvlist_data, fpnode, fxnode);
		}
	    }
	}
    }

    return 0;
}
#endif				/* ____CLASS_SORT__ == __QUICK_SORT__ */


#if ____CLASS_SORT__ == __QUICK_SORT__
#elif __CLASS_SORT__ == __TRANS_SORT__
static int
sort_list(struct tvlist_type *tvlist_data, struct fpnode_type *fpnode_data,
	  void *text, unsigned slip, int (*land) (void *, void *, void *))
{
    do {
	struct fpnode_type *fqnode_data;

	fqnode_data = fpnode_data->fpnode_data[0];

	if (fqnode_data) {
	    sort_list(tvlist_data, fqnode_data, text, slip, land);
	}

	fqnode_data = fpnode_data->fpnode_data[1];

	_libx1f4l2_dawn_tvlist(tvlist_data, fpnode_data, text, land, slip);

	tvlist_data->link_a.fplist.fplink.fqnode.size++;

	fpnode_data = fqnode_data;
    } while (fpnode_data);

    return 0;
}
#endif				/* ____CLASS_SORT__ == __QUICK_SORT__ */


#if ____CLASS_SORT__ == __QUICK_SORT__
static int
spin_list(struct tvlist_type *tvlist_data, void *fpnode, void *fxnode)
{
    struct fpnode_type *fpnode_data, *fqnode_data, *fxnode_data, *fznode_data;
    unsigned fast, tilt;

    fpnode_data = fpnode;
    fqnode_data = fpnode_data->uplink_data;
    if (fqnode_data) {
    } else {
	fqnode_data = &tvlist_data->link_a.fplist.fplink;
    }
    if (fpnode_data == fqnode_data->fpnode_data[0]) {
	tilt = 0;
    } else {
	tilt = 1;
    }

    fxnode_data = fxnode;
    fznode_data = fxnode_data->uplink_data;
    if (fznode_data) {
    } else {
	fznode_data = &tvlist_data->link_a.fplist.fplink;
    }
    if (fxnode_data == fznode_data->fpnode_data[0]) {
	fast = 0;
    } else {
	fast = 1;
    }

    fqnode_data->fpnode_data[tilt] = fxnode_data;

    fznode_data->fpnode_data[fast] = fpnode_data;

    if (1) {
	struct fpnode_type fpnode_class;

	fpnode_class = *fpnode_data;
	*fpnode_data = *fxnode_data;
	*fxnode_data = fpnode_class;
    }

    fqnode_data = fpnode_data->fpnode_data[0];
    if (fqnode_data) {
	fqnode_data->uplink_data = fpnode_data;
    }

    fqnode_data = fpnode_data->fpnode_data[1];
    if (fqnode_data) {
	fqnode_data->uplink_data = fpnode_data;
    }

    fznode_data = fxnode_data->fpnode_data[0];
    if (fznode_data) {
	fznode_data->uplink_data = fxnode_data;
    }

    fznode_data = fxnode_data->fpnode_data[1];
    if (fznode_data) {
	fznode_data->uplink_data = fxnode_data;
    }

    return 0;
}
#endif				/* ____CLASS_SORT__ == __QUICK_SORT__ */


int
x1f4_sort_tvlist(void *tvlist, void *text, unsigned slip,
		 int (*land) (void *, void *, void *))
{
    unsigned size;

    size = tvlist(tvlist)->link_a.fplist.fplink.fqnode.size;
    if (1 < size) {
#if ____CLASS_SORT__ == __QUICK_SORT__
	void *fphead, *fptail;

	_libx1f4l2_else_tvlist(tvlist, 0, &fphead);

	_libx1f4l2_else_tvlist(tvlist, size - 1, &fptail);

	slip += sizeof(struct fpnode_type);

	sort_list(tvlist, fphead, 0, fptail, size - 1, text, slip, land);
#elif __CLASS_SORT__ == __TRANS_SORT__
	void *tvnode;

	tvnode = tvlist(tvlist)->link_a.fplist.fplink.fpnode_data[0];

	tvlist(tvlist)->link_a.fplist.fplink.fpnode_data[0] = NULL;
	tvlist(tvlist)->link_a.fplist.fplink.fqnode.size = 0;

	slip += sizeof(struct fpnode_type);

	sort_list(tvlist, tvnode, text, slip, land);
#endif				/* ____CLASS_SORT__ == __QUICK_SORT__ */
    }

    return 0;
}
