/*
 * bqfz.c
 * Copyright (C) 2002-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#if !defined HAVE_LIBLEWD
# include <cardinal-wx.h>
#endif				/* !HAVE_LIBLEWD */
#include <count2.h>
#include <libx1f4l2.h>

#define __UNROLL_PICK__			0

#if SIZEOF_VOID_P == SIZEOF_LONG
# define integral_q			unsigned long
#else
# define integral_q			unsigned
#endif				/* SIZEOF_VOID_P == SIZEOF_LONG */

#if SIZEOF_VOID_P == 8
# define HALF				040
#else
# define HALF				020
#endif				/* SIZEOF_VOID_P == 8 */

#define LX				10

#define VX				1048576

typedef struct context_type {
    const char *self;
    int do_profile;
    unsigned count, size;
} context_type;

static int case_list(struct context_type *, void *, void *);
static int fine_look(void *, void *, unsigned *);
static int fine_fare(void *, void *);
static int fine_move(void *, void *, unsigned);
static int fine_pick(void *, void *, unsigned *);
static int free_list(void *);
static int lain_look(void *, void *, unsigned *);
static int lain_fare(void *, void *);
static int lain_move(void *, void *, unsigned);
static int lain_pick(void *, void *, unsigned *);
static int line_list(struct context_type *, void **);
static int link_list(struct context_type *, void **);

static void usage(void);

static const char digits[] = "0123456789";

static int
case_list(struct context_type *context_data, void *bqfx, void *list)
{
    count2_decls;
    int shift, status;
    unsigned i;

    shift = context_data->size;

    if (context_data->do_profile) {
	start_count2();
    }

    i = context_data->count;
    while (i) {
	void *data;

	i--;
	status = x1f4_find_bqfset(bqfx, list, &data);
	if (status) {
	    status = x1f4_cast_bqfset(bqfx, list);
	    if (status) {
		fprintf(stderr, "%s: cannot delete data\n",
			context_data->self);
		if (1) {
		    break;
		}
	    }
	} else {
	    status = x1f4_post_bqfset(bqfx, list, &data);
	    if (status) {
		fprintf(stderr, "%s: cannot insert data\n",
			context_data->self);
		if (1) {
		    break;
		}
	    } else {
		memcpy(data, list, shift);
	    }
	}

	list = (char *) list + shift;
    }

    if (i) {
	status = 1;
    } else {
	status = 0;

	if (context_data->do_profile) {
	    finish_count2_l("", 55);
	}
    }

    return status;
}


static int
fine_fare(void *text, void *node)
{
    return *(unsigned long *) text < *(unsigned long *) node
	? -1 : *(unsigned long *) text ^ *(unsigned long *) node ? 1 : 0;
}


static int
fine_look(void *text, void *node, unsigned *pick)
{
    int status;
    unsigned size;

    size = *(integral_q *) node;
    if (size) {
	unsigned long *call, fine, *high;

	fine = *(unsigned long *) text;

	high = node;

	call = high + 1;

	size--;
	while (size) {
	    unsigned half;
	    unsigned long *slip;

	    half = (size + 1) >> 1;
	    slip = call + half;
	    if (fine < *slip) {
		size = half - 1;
	    } else {
		call = slip;
		size -= half;
	    }
	}

	if (*call ^ fine) {
	    status = 0;

	    if (fine < *call) {
		call--;
	    } else {
	    }

	    *pick = call - high;
	} else {
	    status = 1;

	    *pick = call - high;
	}
    } else {
	status = 0;

	*pick = 0;
    }

    return status;
}


static int
fine_move(void *target, void *source, unsigned count)
{
#if 1
    unsigned long *s, *t;

    s = source;
    t = target;
    if (count ^ 1) {
	if (t < s) {
	    for (; count; count--) {
		*t++ = *s++;
	    }
	} else {
	    s += count;
	    if (t < s) {
		t += count;
		for (; count; count--) {
		    *--t = *--s;
		}
	    } else {
		s -= count;
		for (; count; count--) {
		    *t++ = *s++;
		}
	    }
	}
    } else {
	*t++ = *s++;
    }
#else
    memmove(target, source, count * sizeof(unsigned long));
#endif				/* 0 */

    return 0;
}


static int
fine_pick(void *text, void *node, unsigned *pick)
{
    int status;
#if __UNROLL_PICK__
#else
    unsigned half = HALF;
#endif				/* __UNROLL_PICK__ */
    unsigned slip = 0;
    unsigned long fine, *high;

    fine = *(unsigned long *) text;

    high = node;

#if __UNROLL_PICK__
# if SIZEOF_VOID_P == 8
    if (fine < high[slip + 0040]) {
    } else {
	slip += 0040;
    }
# endif				/* SIZEOF_VOID_P == 8 */
    if (fine < high[slip + 0020]) {
    } else {
	slip += 0020;
    }
    if (fine < high[slip + 0010]) {
    } else {
	slip += 0010;
    }
    if (fine < high[slip + 0004]) {
    } else {
	slip += 0004;
    }
    if (fine < high[slip + 0002]) {
    } else {
	slip += 0002;
    }
#else
    do {
	if (fine < high[slip + half]) {
	} else {
	    slip += half;
	}

	half >>= 1;
    } while (half ^ 1);
#endif				/* __UNROLL_PICK__ */

    if (*(integral_q *) node & (integral_q) 2 << slip) {
	if (fine < high[slip + 1]) {
	    *pick = slip + 0;
	    if (slip) {
		if (fine ^ high[slip + 0]) {
		    status = 0;
		} else {
		    status = 1;
		}
	    } else {
		status = 0;
	    }
	} else {
	    *pick = slip + 1;
	    if (fine ^ high[slip + 1]) {
		status = 0;
	    } else {
		status = 1;
	    }
	}
    } else {
	*pick = slip + 1;
	if (slip) {
	    if (fine ^ high[slip + 0]) {
		status = 0;
	    } else {
		status = 1;
	    }
	} else {
	    if (1) {
		status = 0;
	    }
	}
    }

    return status;
}


static int
free_list(void *list)
{
    free(list);

    return 0;
}


static int
lain_fare(void *text, void *node)
{
    return memcmp(text, node, LX);
}


static int
lain_look(void *text, void *node, unsigned *pick)
{
    int status;
    unsigned size;

    size = *(integral_q *) node;
    if (size) {
	unsigned look = 1;
	void *call;

	call = (char *) node + LX;

	size--;
	while (size) {
	    unsigned half;
	    void *slip;

	    half = (size + 1) >> 1;
	    slip = (char *) call + LX * half;
	    if (memcmp(text, slip, LX) < 0) {
		size = half - 1;
	    } else {
		look += half;
		call = slip;
		size -= half;
	    }
	}

	if (memcmp(call, text, LX)) {
	    status = 0;

	    if (memcmp(text, call, LX) < 0) {
		look--;
	    } else {
	    }

	    *pick = look;
	} else {
	    status = 1;

	    *pick = look;
	}
    } else {
	status = 0;

	*pick = 0;
    }

    return status;
}


static int
lain_move(void *target, void *source, unsigned count)
{
    memmove(target, source, count * LX);

    return 0;
}


static int
lain_pick(void *text, void *node, unsigned *pick)
{
    int status;
    unsigned half = HALF, slip = 0;

    do {
	if (memcmp(text, (char *) node + LX * (slip + half), LX) < 0) {
	} else {
	    slip += half;
	}

	half >>= 1;
    } while (half ^ 1);

    if (*(integral_q *) node & (integral_q) 2 << slip) {
	if (memcmp(text, (char *) node + LX * (slip + 1), LX) < 0) {
	    *pick = slip + 0;
	    if (slip) {
		if (memcmp(text, (char *) node + LX * (slip + 0), LX)) {
		    status = 0;
		} else {
		    status = 1;
		}
	    } else {
		status = 0;
	    }
	} else {
	    *pick = slip + 1;
	    if (memcmp(text, (char *) node + LX * (slip + 1), LX)) {
		status = 0;
	    } else {
		status = 1;
	    }
	}
    } else {
	*pick = slip + 1;
	if (slip) {
	    if (memcmp(text, (char *) node + LX * (slip + 0), LX)) {
		status = 0;
	    } else {
		status = 1;
	    }
	} else {
	    if (1) {
		status = 0;
	    }
	}
    }

    return status;
}


static int
line_list(struct context_type *context_data, void **link)
{
    int status;
    unsigned size;
    void *list;

    size = context_data->count;

    list = malloc(sizeof(unsigned long) * size);
    if (list) {
	unsigned i;
	unsigned long *line;

	srand48(11);

	status = 0;

	*link = list;

	line = list;

	i = size;
	while (i) {
	    --i;
	    *line++ = (unsigned) (size * drand48() / 4) * 271828183u;
	}
    } else {
	status = -1;
    }

    return status;
}


static int
link_list(struct context_type *context_data, void **link)
{
    int status;
    unsigned size;
    void *list;

    size = context_data->count;

    list = malloc(LX * size);
    if (list) {
	unsigned i;

	srand48(11);

	status = 0;

	*link = list;

	i = size;
	while (i) {
	    char *o;
	    unsigned e;

	    i--;

	    e = (unsigned) (size * drand48() / 4) * 271828183u;

	    o = (char *) list + LX;

	    do {
		--o;
		*o = digits[e % 10];
		e /= 10;
	    } while (e);
	    while (o != list) {
		--o;
		*o = 32;
	    }

	    list = (char *) list + LX;
	}
    } else {
	status = -1;
    }

    return status;
}


static void
usage(void)
{
    puts("Usage: bqfz [OPTIONS]\n\
Test the BQ generic set.\n\
\n\
Options:\n\
  -N, --size SIZE		test SIZE wide sets\n\
  -p, --profile			time operations\n\
      --integer			run positive integer data tests\n\
      --string			run string data tests (default)\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status, string_test = 1;
    struct context_type context;

    context.do_profile = 0;

    context.count = VX;

    unsetenv("POSIXLY_CORRECT");

    {
	const char *size_specs = NULL;
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "integer",	    0x00,   NULL,   0x69    },
		{   "profile",	    0x00,   NULL,   0x70    },
		{   "size",	    0x01,   NULL,   0x4e    },
		{   "string",	    0x00,   NULL,   0x73    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "N:p", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'N':
		size_specs = optarg;
		break;
	    case 'i':
		string_test = 0;
		break;
	    case 'p':
		context.do_profile = 1;
		break;
	    case 's':
		string_test = 1;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}

	if (size_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.count, size_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse test set width specificati"
			"on: `%s'\n", argv[0], size_specs);

		return 1;
	    }
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	struct x1f4_bqfset_type bqfset;
	void *bqfx;

	if (string_test) {
	    context.size = LX;
	} else {
	    context.size = sizeof(unsigned long);
	}

	if (string_test) {
	    bqfset.size = LX;
	} else {
	    bqfset.size = sizeof(unsigned long);
	}

	if (string_test) {
	    bqfset.fare = lain_fare;
	    bqfset.look = lain_look;
	    bqfset.move = lain_move;
	    bqfset.pick = lain_pick;
	} else {
	    bqfset.fare = fine_fare;
	    bqfset.look = fine_look;
	    bqfset.move = fine_move;
	    bqfset.pick = fine_pick;
	}

	status = x1f4_init_bqfset(&bqfx, X1f4_BQFSET_SIZE_FRAME, &bqfset);
	if (status) {
	    perror(argv[0]);
	} else {
	    int excess;
	    void *list;

	    context.self = argv[0];

	    if (string_test) {
		status = link_list(&context, &list);
	    } else {
		status = line_list(&context, &list);
	    }
	    if (status) {
		perror(argv[0]);
	    } else {
		status = case_list(&context, bqfx, list);

		free_list(list);
	    }

	    excess = x1f4_fini_bqfset(&bqfx);
	    if (excess) {
		perror(argv[0]);

		if (status) {
		} else {
		    status = excess;
		}
	    }
	}
    }

    return status;
}
