/*
 * l219.c
 * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, Ciprian Niculescu
 * Copyright (C) 2009, 2010, 2011, 2012, 2013, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBx1f4i0
# include <libx1f4i0.h>
#endif				/* HAVE_LIBx1f4i0 */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

#include <count2.h>
#include <libx1f4l2.h>
#if !defined HAVE_LIBx1f4i0
# include <vkeyhook.h>
# include <vkeytree.h>
#endif				/* !HAVE_LIBx1f4i0 */

#include <l2xx.h>

#if 1
# define __LINK_SLIP_TEST__		0
#endif				/* 1 */

#if __HEAD_NEXT_LIST__
# define Sans_logique_090909_6_		sans_logique_090909_6_
#else
# undef __LINK_SLIP_TEST__
# define __LINK_SLIP_TEST__		1
# define Sans_logique_090909_6_		sans_logique________0_
#endif				/* __HEAD_NEXT_LIST__ */

#if __HEAD_NEXT_LIST__
# define Sans_logique_090909_7_		sans_logique_090909_7_
#else
# undef __LINK_SLIP_TEST__
# define __LINK_SLIP_TEST__		1
# define Sans_logique_090909_7_		sans_logique________0_
#endif				/* __HEAD_NEXT_LIST__ */

#if __HEAD_REEL_LIST__
# define Sans_logique_100926_0_		sans_logique_100926_0_
#else
# undef __LINK_SLIP_TEST__
# define __LINK_SLIP_TEST__		1
# define Sans_logique_100926_0_		sans_logique________0_
#endif				/* __HEAD_REEL_LIST__ */

#if __HEAD_REEL_LIST__
# define Sans_logique_100926_1_		sans_logique_100926_1_
#else
# undef __LINK_SLIP_TEST__
# define __LINK_SLIP_TEST__		1
# define Sans_logique_100926_1_		sans_logique________0_
#endif				/* __HEAD_REEL_LIST__ */

#define SANS_LOGIQUE_ARGS_0 \
    struct context_type *, const struct test_type *
#define SANS_LOGIQUE_ARGS_1 \
    struct context_type *context_data, const struct test_type *test_data
#define SANS_LOGIQUE_ARGS_2 \
    context_data, test_data

#define APPEND_MODE \
    "last position insert 1,000,000 integers"

#define BEHEAD_MODE \
    "forward remove all from 1,000,000 integers list"

#define BETAIL_MODE \
    "backward remove all from 1,000,000 integers list"

#define DESERT_MODE \
    "first position insert 1,000,000 integers"

#define DIRECT_MODE \
    "forward traverse 1,000,000 integers list"

#define EXCESS_MODE \
    "clear 1,000,000 integers list"

#define EXLINE_MODE \
    "callback traverse 1,000,000 integers list"

#define EXPECT_MODE \
    "backward traverse 1,000,000 integers list"

#define INVERT_MODE \
    "reverse 1,000,000 integers list"

#define ORANGE_MODE \
    "circulate 1,000,000 integers list"

#define VIOLET_MODE \
    "circulate back 1,000,000 integers list"

#define STREAM_TEST \
    "all the tests"

typedef struct context_type {
    const char *self;
} context_type;

typedef struct test_type {
    const char *description;
    int (*test) (SANS_LOGIQUE_ARGS_0);
} test_type;

static int fini_list(struct context_type *, void **);
static int init_list(struct context_type *, void **);
static int list_link(void *, const void *, const char *, unsigned);
static int miss_node(void *, void *);
static int sans_logique_090909_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090909_1a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090909_2_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090909_2a(SANS_LOGIQUE_ARGS_0, void *);
#if __HEAD_NEXT_LIST__
static int sans_logique_090909_6_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090909_6a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090909_7_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090909_7a(SANS_LOGIQUE_ARGS_0, void *);
#endif				/* __HEAD_NEXT_LIST__ */
static int sans_logique_090910_3_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090910_3a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_091004_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_091004_0a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_100731_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100731_0a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_100731_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100731_1a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_100813_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100813_0a(SANS_LOGIQUE_ARGS_0, void *);
#if __HEAD_REEL_LIST__
static int sans_logique_100926_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100926_0a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_100926_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100926_1a(SANS_LOGIQUE_ARGS_0, void *);
#endif				/* __HEAD_REEL_LIST__ */
#if __LINK_SLIP_TEST__
static int sans_logique________0_(SANS_LOGIQUE_ARGS_0);
#endif				/* __LINK_SLIP_TEST__ */
static int sans_logique__________(SANS_LOGIQUE_ARGS_0);

static void class(void);
static void usage(void);

static const struct test_type test_1_9[] = {
/* *INDENT-OFF* */
    {	STREAM_TEST,		sans_logique__________			},
    {	APPEND_MODE,		sans_logique_090909_1_			},
    {	DESERT_MODE,		sans_logique_090909_2_			},
    {	DIRECT_MODE,		Sans_logique_090909_6_			},
    {	EXPECT_MODE,		Sans_logique_090909_7_			},
    {	EXCESS_MODE,		sans_logique_090910_3_			},
    {	EXLINE_MODE,		sans_logique_091004_0_			},
    {	BEHEAD_MODE,		sans_logique_100731_0_			},
    {	BETAIL_MODE,		sans_logique_100731_1_			},
    {	INVERT_MODE,		sans_logique_100813_0_			},
    {	ORANGE_MODE,		Sans_logique_100926_0_			},
    {	VIOLET_MODE,		Sans_logique_100926_1_			},
    {   NULL,			NULL					}
/* *INDENT-ON* */
};

static const struct x1f4_vkeytree_type link___c[] = {
/* *INDENT-OFF* */
    {	"all",		{	test_1_9 + 6	},	NULL		},
    {	"lear",		{	test_1_9 + 5	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link___r_e[] = {
/* *INDENT-OFF* */
    {	"el",		{	test_1_9 + 11	},	NULL		},
    {	"verse",	{	test_1_9 + 9	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link___r[] = {
/* *INDENT-OFF* */
    {	"e",		{	NULL		},	link___r_e	},
    {	"oll",		{	test_1_9 + 10	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link___s[] = {
/* *INDENT-OFF* */
    {	"lip",		{	test_1_9 + 8	},	NULL		},
    {	"uffix",	{	test_1_9 + 1	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link_1_9[] = {
/* *INDENT-OFF* */
    {	"all",		{	test_1_9 + 0	},	NULL		},
    {	"c",		{	NULL		},	link___c	},
    {	"backward",	{	test_1_9 + 4	},	NULL		},
    {	"forward",	{	test_1_9 + 3	},	NULL		},
    {	"miss",		{	test_1_9 + 7	},	NULL		},
    {	"prefix",	{	test_1_9 + 2	},	NULL		},
    {	"r",		{	NULL		},	link___r	},
    {	"s",		{	NULL		},	link___s	},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link____[] = {
/* *INDENT-OFF* */
    {	"",		{	NULL		},	link_1_9	},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
};

static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_l2list(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_init_l2list(list, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    } else {
    }

    return status;
}


static int
list_link(void *file, const void *node, const char *data, unsigned size)
{
    fwrite("  ", 2, 1, file);
    fwrite(data, size, 1, file);
    fwrite("                ", 16 - size, 1, file);
    fwrite(" - ", 3, 1, file);
    fputs(((const struct test_type *) node)->description, file);
    fputc('\n', file);

    return 0;
}


static int
miss_node(void *text, void *node)
{
    return 0;
}


static int
sans_logique_090909_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	count2_decls;

	start_count2();

	status = sans_logique_090909_1a(SANS_LOGIQUE_ARGS_2, list);

	if (status) {
	} else {
	    finish_count2_l(test_data->description, 55);
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090909_1a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	void *data;

	i--;

	status = x1f4_tail_l2list(list, sizeof(unsigned), &data);
	if (status) {
	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = i;
	}
    }

    return status;
}


static int
sans_logique_090909_2_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	count2_decls;

	start_count2();

	status = sans_logique_090909_2a(SANS_LOGIQUE_ARGS_2, list);

	if (status) {
	} else {
	    finish_count2_l(test_data->description, 55);
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090909_2a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	void *data;

	i--;

	status = x1f4_head_l2list(list, sizeof(unsigned), &data);
	if (status) {
	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = i;
	}
    }

    return status;
}


#if __HEAD_NEXT_LIST__
static int
sans_logique_090909_6_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_1a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090909_6a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_NEXT_LIST__ */


#if __HEAD_NEXT_LIST__
static int
sans_logique_090909_6a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned e;
    void *data;

    x1f4_size_l2list(list, &e);
    if (e) {
	status = x1f4_lead_l2list(list, &data);
	if (status) {
	} else {
	    do {
	    } while (x1f4_next_l2list(&data, data));
	}
    } else {
	status = 0;
    }

    return status;
}
#endif				/* __HEAD_NEXT_LIST__ */


#if __HEAD_NEXT_LIST__
static int
sans_logique_090909_7_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_1a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090909_7a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_NEXT_LIST__ */


#if __HEAD_NEXT_LIST__
static int
sans_logique_090909_7a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned e;
    void *data;

    x1f4_size_l2list(list, &e);
    if (e) {
	status = x1f4_ever_l2list(list, &data);
	if (status) {
	} else {
	    do {
	    } while (x1f4_last_l2list(&data, data));
	}
    } else {
	status = 0;
    }

    return status;
}
#endif				/* __HEAD_NEXT_LIST__ */


static int
sans_logique_090910_3_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_1a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090910_3a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090910_3a(SANS_LOGIQUE_ARGS_1, void *list)
{
    return x1f4_high_l2list(list);
}


static int
sans_logique_091004_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_1a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_091004_0a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_091004_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    return x1f4_lime_l2list(list, NULL, miss_node);
}


static int
sans_logique_100731_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_1a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100731_0a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_100731_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	i--;

	status = x1f4_mind_l2list(list);
	if (status) {
	    break;
	}
    }

    return status;
}


static int
sans_logique_100731_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_1a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100731_1a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_100731_1a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	i--;

	status = x1f4_slip_l2list(list);
	if (status) {
	    break;
	}
    }

    return status;
}


static int
sans_logique_100813_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_1a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100813_0a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_100813_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    return x1f4_flip_l2list(list);
}


#if __HEAD_REEL_LIST__
static int
sans_logique_100926_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_1a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100926_0a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_REEL_LIST__ */


#if __HEAD_REEL_LIST__
static int
sans_logique_100926_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    unsigned i;

    i = 1000000;
    while (i) {
	i--;

	x1f4_roll_l2list(list);
    }

    return 0;
}
#endif				/* __HEAD_REEL_LIST__ */


#if __HEAD_REEL_LIST__
static int
sans_logique_100926_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_1a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100926_1a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_REEL_LIST__ */


#if __HEAD_REEL_LIST__
static int
sans_logique_100926_1a(SANS_LOGIQUE_ARGS_1, void *list)
{
    unsigned i;

    i = 1000000;
    while (i) {
	i--;

	x1f4_reel_l2list(list);
    }

    return 0;
}
#endif				/* __HEAD_REEL_LIST__ */


#if __LINK_SLIP_TEST__
static int
sans_logique________0_(SANS_LOGIQUE_ARGS_1)
{
    count2_print("%*s: not available\n", 55, test_data->description);

    return 0;
}
#endif				/* __LINK_SLIP_TEST__ */


static int
sans_logique__________(SANS_LOGIQUE_ARGS_1)
{
    int status;

    test_data = test_1_9 + 1;

    do {
	status = test_data->test(SANS_LOGIQUE_ARGS_2);
	if (status) {
	    break;
	} else {
	    test_data++;
	}
    } while (test_data->test);

    return status;
}


static void
class(void)
{
    x1f4_hook_vtkey(stdout, list_link, link____);
}


static void
usage(void)
{
    puts("Usage: l219 [OPTIONS] TEST\n\
Profile the " l2name "list sequence.\n\
\n\
Options:\n\
  --list			lists available tests and exit\n\
  --help			display this help and exit\n\
  --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;

    unsetenv("POSIXLY_CORRECT");

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "list",	    0x00,   NULL,   0x6c    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'h':
	    case 'l':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'l':
		class();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (argc - optind != 1) {
	fprintf(stderr, "%s: wrong number of arguments\nTry `%s --help' for "
		"more information.\n", argv[0], argv[0]);

	return 1;
    } else {
	const void *test;

	if (x1f4_find_vtkey(argv[optind], link____, &test)) {
	    status = 1;

	    fprintf(stderr, "%s: cannot parse test specification: `%s'\n",
		    argv[0], argv[optind]);
	} else {
	    struct context_type context;

	    context.self = argv[0];

	    status = ((struct test_type *) test)->test(&context, test);
	}
    }

    return status;
}
