/*
 * qs11.c
 * Copyright (C) 2002-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#if !defined HAVE_LIBLEWD
# include <cardinal-wx.h>
#endif				/* !HAVE_LIBLEWD */
#include <inter.h>
#include <libx1f4l2.h>
#include <qxcc.h>

#define __TEXT_TEST__			1

#define __PUSH_TEST__			1

#define __HEAD_TEST__			2
#define __TAIL_TEST__			3

#define __QS11_TEST__			__PUSH_TEST__

#define ____QS11_TEST__			__QS11_TEST__

#define QS11				1000000

#if __HAVE_TEXT_RATE__
#else
# undef __TEXT_TEST__
# define __TEXT_TEST__			0
#endif				/* __HAVE_TEXT_RATE__ */

typedef struct context_type {
    const char *self;
    unsigned class, count, frame, width;
} context_type;

static int case_away(struct context_type *);
#if __TEXT_TEST__
static int copy_slip(void *, void *, const void *);
#endif				/* __TEXT_TEST__ */
static int fini_list(struct context_type *, void **);
static int fini_rate(struct context_type *, void **);
#if __TEXT_TEST__
static int free_slip(void *, void *);
#endif				/* __TEXT_TEST__ */
static int init_list(struct context_type *, void **);
static int init_rate(struct context_type *, void **);
static int long_rate(struct context_type *, void *, void *);
static int push_rate(struct context_type *, void *, void *);
#if __TEXT_TEST__
static int text_rate(struct context_type *, void **);
#endif				/* __TEXT_TEST__ */

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	void *rate;

	status = init_rate(context_data, &rate);
	if (status) {
	} else {
	    status = push_rate(context_data, list, rate);
	    if (status) {
	    } else {
#if __TEXT_TEST__
		status = text_rate(context_data, &rate);
		if (status) {
		} else {
#endif				/* __TEXT_TEST__ */
		    status = long_rate(context_data, list, rate);
#if __TEXT_TEST__
		}
#endif				/* __TEXT_TEST__ */
	    }

	    excess = fini_rate(context_data, &rate);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


#if __TEXT_TEST__
static int
copy_slip(void *null, void *t, const void *s)
{
    *(unsigned *) t = *(unsigned *) s;

    return 0;
}
#endif				/* __TEXT_TEST__ */


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_tclist(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
fini_rate(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_qxrate(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


#if __TEXT_TEST__
static int
free_slip(void *null, void *x)
{
    return 0;
}
#endif				/* __TEXT_TEST__ */


static int
init_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_init_tclist(list, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    }

    return status;
}


static int
init_rate(struct context_type *context_data, void **list)
{
    int status;
    struct x1f4_qxrate_type qxrate;

#if __MISS_FOUR_RATE__
#else
    qxrate.quarter = context_data->frame;
#endif				/* __MISS_FOUR_RATE__ */
    qxrate.rate = 7;
    qxrate.size = context_data->width;

    status = x1f4_init_qxrate
	(list, X1f4_QXRATE_FOUR_FRAME | X1f4_QXRATE_RATE_FRAME, &qxrate);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    }

    return status;
}


static int
long_rate(struct context_type *context_data, void *list, void *rate)
{
    int status = 0;
    unsigned i;

    i = context_data->count;
    while (i) {
	void *plus;

	i--;

	status = x1f4_peek_tclist(list, i, &plus);
	if (status) {
	    fprintf(stderr, "%s: cannot retrieve item\n", context_data->self);

	    break;
	} else {
	    void *data;

	    status = x1f4_peek_qxrate(rate, i, &data);
	    if (status) {
		fprintf(stderr, "%s: cannot retrieve item\n",
			context_data->self);

		break;
	    } else {
		if (*(unsigned *) data ^ *(unsigned *) plus) {
		    unsigned size;

		    x1f4_size_tclist(list, &size);

		    fprintf(stderr, "%s: cannot match lists, one must be"
			    " buggy\n", context_data->self);
		    fprintf(stderr, "%s: index was %u, tclist was %u, "
			    qxname "rate was %u\n", context_data->self, i,
			    *(unsigned *) plus, *(unsigned *) data);
		    fprintf(stderr, "%s: size of lists is %u\n",
			    context_data->self, size);

		    status = 1;

		    break;
		}
	    }
	}
    }

    return status;
}


static int
push_rate(struct context_type *context_data, void *list, void *rate)
{
    int status = 0;
    unsigned i;

    srand(context_data->class);

    i = 0;
    while (i < context_data->count) {
#if ____QS11_TEST__ == __PUSH_TEST__
	unsigned e;
#endif				/* ____QS11_TEST__ == __PUSH_TEST__ */
	void *data;

	i++;

#if ____QS11_TEST__ == __PUSH_TEST__
	e = rand() % i;
#endif				/* ____QS11_TEST__ == __PUSH_TEST__ */

#if ____QS11_TEST__ == __PUSH_TEST__
	status = x1f4_push_tclist(list, e, sizeof(unsigned), &data);
#elif __QS11_TEST__ == __HEAD_TEST__
	status = x1f4_head_tclist(list, sizeof(unsigned), &data);
#elif __QS11_TEST__ == __TAIL_TEST__
	status = x1f4_tail_tclist(list, sizeof(unsigned), &data);
#endif				/* ____QS11_TEST__ == __PUSH_TEST__ */
	if (status) {
	    fprintf(stderr, "%s: cannot insert item\n", context_data->self);

	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = i;

#if ____QS11_TEST__ == __PUSH_TEST__
	    status = x1f4_push_qxrate(rate, e, &data);
#elif __QS11_TEST__ == __HEAD_TEST__
	    status = x1f4_head_qxrate(rate, &data);
#elif __QS11_TEST__ == __TAIL_TEST__
	    status = x1f4_tail_qxrate(rate, &data);
#endif				/* ____QS11_TEST__ == __PUSH_TEST__ */
	    if (status) {
		fprintf
		    (stderr, "%s: cannot insert item\n", context_data->self);

		break;
	    } else {
		slip = data;
		*slip = i;
	    }
	}
    }

    return status;
}


#if __TEXT_TEST__
static int
text_rate(struct context_type *context_data, void **rate)
{
    int status;
    void *list;

    status = x1f4_text_qxrate(&list, *rate, copy_slip, free_slip, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot copy list\n", context_data->self);
    } else {
	status = fini_rate(context_data, rate);
	if (1) {
	    *rate = list;
	}
    }

    return status;
}
#endif				/* __TEXT_TEST__ */


static void
usage(void)
{
    puts("Usage: qs11 [OPTIONS]\n\
Test " qxname "_insert() against tc_insert().\n\
\n\
Options:\n\
  -N, --size SIZE		test SIZE wide sets\n\
  -S, --seed SEED		use SEED for random sequence initialization\n\
  -e, --external L		set external nodes width as L * 4\n\
  -w, --width WIDTH		use WIDTH bytes items\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    unsetenv("POSIXLY_CORRECT");

    context.class = 0;
    context.count = QS11;
    context.frame = 1 << 4;
    context.width = sizeof(unsigned);

    {
	const char *frame_specs = NULL, *seed_specs = NULL, *size_specs = NULL,
	    *width_specs = NULL;
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "external",	    0x01,   NULL,   0x65    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   "seed",	    0x01,   NULL,   0x53    },
		{   "size",	    0x01,   NULL,   0x4e    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "width",	    0x01,   NULL,   0x77    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "N:S:e:w:", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'N':
		size_specs = optarg;
		break;
	    case 'S':
		seed_specs = optarg;
		break;
	    case 'e':
		frame_specs = optarg;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case 'w':
		width_specs = optarg;
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}

	if (frame_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.frame, frame_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse external node width specif"
			"ication: `%s'\n", argv[0], frame_specs);

		return 1;
	    } else {
		if (context.frame < 1) {
		    fprintf(stderr, "%s: will not use external node widths l"
			    "ess than 4 (like the indicated %u << 2)\n",
			    argv[0], context.frame);

		    return 1;
		}
	    }
	}

	if (seed_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.class, seed_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse list seed specification: `"
			"%s'\n", argv[0], seed_specs);

		return 1;
	    }
	}

	if (size_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.count, size_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse list size specification: `"
			"%s'\n", argv[0], size_specs);

		return 1;
	    }
	}

	if (width_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.width, width_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse item width specification: "
			"`%s'\n", argv[0], width_specs);

		return 1;
	    } else {
		if (context.width < sizeof(unsigned)) {
		    fprintf(stderr, "%s: will not use item widths less than "
			    "unsigned width (like the indicated %u)\n",
			    argv[0], context.width);

		    return 1;
		}
	    }
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	context.self = argv[0];

	status = case_away(&context);
    }

    return status;
}
