/*
 * qs12.c
 * Copyright (C) 2002-2011, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#if !defined HAVE_LIBLEWD
# include <cardinal-wx.h>
#endif				/* !HAVE_LIBLEWD */
#include <inter.h>
#include <libx1f4l2.h>
#include <qxcc.h>

#define QS11				1000000
#define QS12				10000

typedef struct context_type {
    const char *self;
    unsigned class, count, width;
} context_type;

static int case_away(struct context_type *);
static int fini_list(struct context_type *, void **);
static int fini_rate(struct context_type *, void **);
static int init_list(struct context_type *, void **);
static int init_rate(struct context_type *, void **);
static int miss_rate(struct context_type *, void *, void *);
static int push_rate(struct context_type *, void *, void *);
static int slip_rate(struct context_type *, void *, void *);

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	void *rate;

	status = init_rate(context_data, &rate);
	if (status) {
	} else {
	    status = push_rate(context_data, list, rate);
	    if (status) {
	    } else {
		status = miss_rate(context_data, list, rate);
	    }

	    excess = fini_rate(context_data, &rate);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_tclist(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
fini_rate(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_qxrate(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_init_tclist(list, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    }

    return status;
}


static int
init_rate(struct context_type *context_data, void **list)
{
    int status;
    struct x1f4_qxrate_type qxrate;

#if __MISS_FOUR_RATE__
#else
    qxrate.quarter = 1 << 4;
#endif				/* __MISS_FOUR_RATE__ */
    qxrate.rate = 4;
    qxrate.size = context_data->width;

    status = x1f4_init_qxrate
	(list, X1f4_QXRATE_FOUR_FRAME | X1f4_QXRATE_RATE_FRAME, &qxrate);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    }

    return status;
}


static int
miss_rate(struct context_type *context_data, void *list, void *rate)
{
    int status = 0;
    unsigned class, count;

    count = context_data->count;
    class = context_data->class;

    while (count) {
	status = slip_rate(context_data, list, rate);
	if (status) {
	    break;
	} else {
	    if (class < count) {
		count -= class;
	    } else {
		count = 0;
	    }
	}
    }

    return status;
}


static int
push_rate(struct context_type *context_data, void *list, void *rate)
{
    int status = 0;
    unsigned i;

    srand(0);

    i = 0;
    while (i < context_data->count) {
	unsigned e;
	void *data;

	i++;

	e = rand() % i;

	status = x1f4_push_tclist(list, e, sizeof(unsigned), &data);
	if (status) {
	    fprintf(stderr, "%s: cannot insert item\n", context_data->self);

	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = i;

	    status = x1f4_push_qxrate(rate, e, &data);
	    if (status) {
		fprintf
		    (stderr, "%s: cannot insert item\n", context_data->self);

		break;
	    } else {
		slip = data;
		*slip = i;
	    }
	}
    }

    return status;
}


static int
slip_rate(struct context_type *context_data, void *list, void *rate)
{
    int status = 0;
    unsigned i, size;

    x1f4_size_qxrate(rate, &size);

    i = context_data->class;
    if (i < size) {
    } else {
	i = size;
    }

    while (i) {
	unsigned e;

	i--;

	e = rand() % size;

	size--;

	status = x1f4_miss_tclist(list, e);
	if (status) {
	    fprintf(stderr, "%s: cannot delete item\n", context_data->self);

	    break;
	} else {
	    status = x1f4_miss_qxrate(rate, e);
	    if (status) {
		fprintf
		    (stderr, "%s: cannot delete item\n", context_data->self);

		break;
	    }
	}
    }

    if (status) {
    } else {
	x1f4_size_qxrate(rate, &size);

#if 1
	i = size;
	while (i) {
	    void *plus;

	    i--;

	    status = x1f4_peek_tclist(list, i, &plus);
	    if (status) {
		fprintf
		    (stderr, "%s: cannot retrieve item\n", context_data->self);

		break;
	    } else {
		void *data;

		status = x1f4_peek_qxrate(rate, i, &data);
		if (status) {
		    fprintf(stderr, "%s: cannot retrieve item\n",
			    context_data->self);

		    break;
		} else {
		    if (*(unsigned *) data ^ *(unsigned *) plus) {
			fprintf(stderr, "%s: cannot match lists, one must be"
				" buggy\n", context_data->self);
			fprintf(stderr, "%s: index was %u, tclist was %u, "
				qxname "rate was %u\n", context_data->self, i,
				*(unsigned *) plus, *(unsigned *) data);
			fprintf(stderr, "%s: size of lists is %u\n",
				context_data->self, size);

			status = 1;

			break;
		    }
		}
	    }
	}
#else
	i = 0;
	while (i < size) {
	    void *plus;

	    status = x1f4_peek_tclist(list, i, &plus);
	    if (status) {
		fprintf
		    (stderr, "%s: cannot retrieve item\n", context_data->self);

		break;
	    } else {
		void *data;

		status = x1f4_peek_qxrate(rate, i, &data);
		if (status) {
		    fprintf(stderr, "%s: cannot retrieve item\n",
			    context_data->self);

		    break;
		} else {
		    if (*(unsigned *) data ^ *(unsigned *) plus) {
			fprintf(stderr, "%s: cannot match lists, one must be"
				" buggy\n", context_data->self);
			fprintf(stderr, "%s: index was %u, tclist was %u, "
				qxname "rate was %u\n", context_data->self, i,
				*(unsigned *) plus, *(unsigned *) data);
			fprintf(stderr, "%s: size of lists is %u\n",
				context_data->self, size);

			status = 1;

			break;
		    }
		}
	    }

	    i++;
	}
#endif				/* 1 */
    }

    return status;
}


static void
usage(void)
{
    puts("Usage: qs12 [OPTIONS]\n\
Test " qxname "_delete() against tc_delete().\n\
\n\
Options:\n\
  -N, --size SIZE		test SIZE wide sets\n\
  -R, --run RUN			compare the two sequences every RUN deletes\n\
  -w, --width WIDTH		use WIDTH bytes items\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    unsetenv("POSIXLY_CORRECT");

    context.class = QS12;
    context.count = QS11;
    context.width = sizeof(unsigned);

    {
	const char *mind_specs = NULL, *size_specs = NULL, *width_specs = NULL;
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "run",	    0x01,   NULL,   0x52    },
		{   "size",	    0x01,   NULL,   0x4e    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "width",	    0x01,   NULL,   0x77    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "N:R:w:", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'N':
		size_specs = optarg;
		break;
	    case 'R':
		mind_specs = optarg;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case 'w':
		width_specs = optarg;
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}

	if (size_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.count, size_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse list size specification: `"
			"%s'\n", argv[0], size_specs);

		return 1;
	    } else {
		context.class = context.count >> 2;
		if (QS12 < context.class) {
		    context.class = QS12;
		}
	    }
	}

	if (mind_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.class, mind_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse test run specification: `%"
			"s'\n", argv[0], mind_specs);

		return 1;
	    }
	}

	if (width_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.width, width_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse item width specification: "
			"`%s'\n", argv[0], width_specs);

		return 1;
	    } else {
		if (context.width < sizeof(unsigned)) {
		    fprintf(stderr, "%s: will not use item widths less than "
			    "unsigned width (like the indicated %u)\n",
			    argv[0], context.width);

		    return 1;
		}
	    }
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	context.self = argv[0];

	status = case_away(&context);
    }

    return status;
}
