/*
 * qs14.c
 * Copyright (C) 2002-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#if !defined HAVE_LIBLEWD
# include <cardinal-wx.h>
#endif				/* !HAVE_LIBLEWD */
#include <inter.h>
#include <libx1f4l2.h>
#include <qxcc.h>

#define __BACKWARD_RATE__		1

#define __TAIL_RATE__			0

#define __TEXT_TEST__			1

#define __PUSH_TEST__			1

#define __HEAD_TEST__			2
#define __TAIL_TEST__			3

#define __QS11_TEST__			__PUSH_TEST__

#define ____QS11_TEST__			__QS11_TEST__

#define QS11				1000000

#if __HAVE_TEXT_RATE__
#else
# undef __TEXT_TEST__
# define __TEXT_TEST__			0
#endif				/* __HAVE_TEXT_RATE__ */

typedef struct context_type {
    const char *self;
    unsigned count, width;
} context_type;

static int case_away(struct context_type *);
#if __TEXT_TEST__
static int copy_slip(void *, void *, const void *);
#endif				/* __TEXT_TEST__ */
static int fast_rate(struct context_type *, void *);
static int fini_rate(struct context_type *, void **);
#if __TEXT_TEST__
static int free_slip(void *, void *);
#endif				/* __TEXT_TEST__ */
static int init_rate(struct context_type *, void **);
static int long_rate(struct context_type *, void *);
static int pick_rate(struct context_type *, void *);
static int push_rate(struct context_type *, void *);
#if __TAIL_RATE__
static int tail_rate(struct context_type *, void *);
#endif				/* __TAIL_RATE__ */
#if __TEXT_TEST__
static int text_rate(struct context_type *, void **);
#endif				/* __TEXT_TEST__ */

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *rate;

    status = init_rate(context_data, &rate);
    if (status) {
    } else {
	int excess;

	status = push_rate(context_data, rate);
	if (status) {
	} else {
#if __TEXT_TEST__
	    status = text_rate(context_data, &rate);
	    if (status) {
	    } else {
#endif				/* __TEXT_TEST__ */
		status = long_rate(context_data, rate);
		if (status) {
		} else {
#if __TAIL_RATE__
		    status = tail_rate(context_data, rate);
		    if (status) {
		    } else {
#endif				/* __TAIL_RATE__ */
			status = pick_rate(context_data, rate);
			if (status) {
			} else {
			    status = fast_rate(context_data, rate);
			}
#if __TAIL_RATE__
		    }
#endif				/* __TAIL_RATE__ */
		}
#if __TEXT_TEST__
	    }
#endif				/* __TEXT_TEST__ */
	}

	excess = fini_rate(context_data, &rate);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }


    return status;
}


#if __TEXT_TEST__
static int
copy_slip(void *null, void *t, const void *s)
{
    *(unsigned *) t = *(unsigned *) s;

    return 0;
}
#endif				/* __TEXT_TEST__ */


static int
fast_rate(struct context_type *context_data, void *rate)
{
    int status = 0;
    unsigned i;

#if __BACKWARD_RATE__
    i = context_data->count;
    while (i) {
#elif 0
    }
#else
    i = 0;
    while (i < context_data->count) {
#endif				/* __BACKWARD_RATE__ */
	void *data;

#if __BACKWARD_RATE__
	--i;
#endif				/* __BACKWARD_RATE__ */

#if __BACKWARD_RATE__
	status = x1f4_ever_qxrate(rate, &data);
#else
# if __HEAD_LEAD_RATE__
	status = x1f4_lead_qxrate(rate, &data);
# else
	status = x1f4_peek_qxrate(rate, 0, &data);
# endif				/* __HEAD_LEAD_RATE__ */
#endif				/* __BACKWARD_RATE__ */
	if (status) {
	    fprintf(stderr, "%s: cannot retrieve item\n", context_data->self);

	    break;
	} else {
	    if (*(unsigned *) data ^ i) {
		fprintf(stderr, "%s: surprising value (%u) found (expected %"
			"u)\n", context_data->self, *(unsigned *) data, i);

		status = -1;

		break;
	    } else {
#if __BACKWARD_RATE__
# if __HEAD_SLIP_RATE__
		status = x1f4_slip_qxrate(rate);
# else
		status = x1f4_miss_qxrate(rate, i);
# endif				/* __HEAD_SLIP_RATE__ */
#else
# if __HEAD_MIND_RATE__
		status = x1f4_mind_qxrate(rate);
# else
		status = x1f4_miss_qxrate(rate, 0);
# endif				/* __HEAD_MIND_RATE__ */
#endif				/* __BACKWARD_RATE__ */
		if (status) {
		    fprintf(stderr, "%s: cannot delete item\n",
			    context_data->self);

		    break;
		}
	    }
	}

#if __BACKWARD_RATE__
#else
	i++;
#endif				/* __BACKWARD_RATE__ */
    }

    if (status) {
    } else {
	if (x1f4_case_qxrate(rate)) {
	    fprintf(stderr, "%s: sequence not void\n", context_data->self);
	    status = 1;
	}
    }

    return status;
}


static int
fini_rate(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_qxrate(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


#if __TEXT_TEST__
static int
free_slip(void *null, void *x)
{
    return 0;
}
#endif				/* __TEXT_TEST__ */


static int
init_rate(struct context_type *context_data, void **list)
{
    int status;
    struct x1f4_qxrate_type qxrate;

#if __MISS_FOUR_RATE__
#else
    qxrate.quarter = 1 << 4;
#endif				/* __MISS_FOUR_RATE__ */
    qxrate.rate = 6;
    qxrate.size = context_data->width;

    status = x1f4_init_qxrate
	(list, X1f4_QXRATE_FOUR_FRAME | X1f4_QXRATE_RATE_FRAME, &qxrate);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    }

    return status;
}


static int
long_rate(struct context_type *context_data, void *rate)
{
    int status = 0;
    unsigned count, i;

    count = context_data->count;

    i = count;
    while (i) {
	void *data;

	i--;

#if __BACKWARD_RATE__
	status = x1f4_ever_qxrate(rate, &data);
#else
# if __HEAD_LEAD_RATE__
	status = x1f4_lead_qxrate(rate, &data);
# else
	status = x1f4_peek_qxrate(rate, 0, &data);
# endif				/* __HEAD_LEAD_RATE__ */
#endif				/* __BACKWARD_RATE__ */
	if (status) {
	    fprintf(stderr, "%s: cannot retrieve item\n", context_data->self);

	    break;
	} else {
	    unsigned e;

	    e = *(unsigned *) data;

#if __BACKWARD_RATE__
# if __HAVE_HEAD_RATE__
	    status = x1f4_head_qxrate(rate, &data);
# else
	    status = x1f4_push_qxrate(rate, 0, &data);
# endif				/* __HAVE_HEAD_RATE__ */
#else
# if __HAVE_TAIL_RATE__
	    status = x1f4_tail_qxrate(rate, &data);
# else
	    status = x1f4_push_qxrate(rate, count, &data);
# endif				/* __HAVE_TAIL_RATE__ */
#endif				/* __BACKWARD_RATE__ */
	    if (status) {
		fprintf(stderr, "%s: cannot insert item\n",
			context_data->self);

		break;
	    } else {
		unsigned *slip;

		slip = data;
		*slip = e;

#if __BACKWARD_RATE__
		status = x1f4_miss_qxrate(rate, count);
#else
		status = x1f4_miss_qxrate(rate, 0);
#endif				/* __BACKWARD_RATE__ */
		if (status) {
		    fprintf(stderr, "%s: cannot delete item\n",
			    context_data->self);

		    break;
		}
	    }
	}
    }

    if (status) {
    } else {
	x1f4_size_qxrate(rate, &i);
	if (i ^ count) {
	    fprintf(stderr, "%s: surprising sequence size (%u, expected %u)\n",
		    context_data->self, i, count);
	    status = 1;
	}
    }

    return status;
}


static int
pick_rate(struct context_type *context_data, void *rate)
{
    int status = 0;
    unsigned i;

    i = context_data->count;
    while (i) {
	void *data;

	--i;

	status = x1f4_peek_qxrate(rate, i, &data);
	if (status) {
	    fprintf(stderr, "%s: cannot retrieve item\n", context_data->self);

	    break;
	} else {
	    if (*(unsigned *) data ^ i) {
		fprintf(stderr, "%s: surprising value (%u) found (expected %"
			"u)\n", context_data->self, *(unsigned *) data, i);

		status = -1;

		break;
	    }
	}
    }

    return status;
}


static int
push_rate(struct context_type *context_data, void *rate)
{
    int status = 0;
    unsigned i, count;

    count = context_data->count;

    i = 0;
    while (i < count) {
	unsigned *slip;
	void *data;

#if __HAVE_TAIL_RATE__
	status = x1f4_tail_qxrate(rate, &data);
#else
	status = x1f4_push_qxrate(rate, i, &data);
#endif				/* __HAVE_TAIL_RATE__ */
	if (status) {
	    fprintf(stderr, "%s: cannot insert item\n", context_data->self);

	    break;
	} else {
	    slip = data;
	    *slip = i;
	}

	i++;
    }

    return status;
}


#if __TAIL_RATE__
static int
tail_rate(struct context_type *context_data, void *rate)
{
    int status = 0;
    unsigned i, count;

    count = context_data->count;

    i = count;

    count <<= 1;

    context_data->count = count;

    while (i < count) {
	unsigned *slip;
	void *data;

# if __HAVE_TAIL_RATE__
	status = x1f4_tail_qxrate(rate, &data);
# else
	status = x1f4_push_qxrate(rate, i, &data);
# endif				/* __HAVE_TAIL_RATE__ */
	if (status) {
	    fprintf(stderr, "%s: cannot insert item\n", context_data->self);

	    break;
	} else {
	    slip = data;
	    *slip = i;
	}

	i++;
    }

    return status;
}
#endif				/* __TAIL_RATE__ */


#if __TEXT_TEST__
static int
text_rate(struct context_type *context_data, void **rate)
{
    int status;
    void *list;

    status = x1f4_text_qxrate(&list, *rate, copy_slip, free_slip, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot copy list\n", context_data->self);
    } else {
	status = fini_rate(context_data, rate);
	if (1) {
	    *rate = list;
	}
    }

    return status;
}
#endif				/* __TEXT_TEST__ */


static void
usage(void)
{
    puts("Usage: qs14 [OPTIONS]\n\
Test " qxname "rate.\n\
\n\
Options:\n\
  -N, --size SIZE		test SIZE wide sets\n\
  -w, --width WIDTH		use WIDTH bytes items\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    unsetenv("POSIXLY_CORRECT");

    context.count = QS11;
    context.width = sizeof(unsigned);

    {
	const char *size_specs = NULL, *width_specs = NULL;
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "help",	    0x00,   NULL,   0x68    },
		{   "size",	    0x01,   NULL,   0x4e    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "width",	    0x01,   NULL,   0x77    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "N:w:", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'N':
		size_specs = optarg;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case 'w':
		width_specs = optarg;
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}

	if (size_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.count, size_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse list size specification: `"
			"%s'\n", argv[0], size_specs);

		return 1;
	    }
	}

	if (width_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.width, width_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse item width specification: "
			"`%s'\n", argv[0], width_specs);

		return 1;
	    } else {
		if (context.width < sizeof(unsigned)) {
		    fprintf(stderr, "%s: will not use item widths less than "
			    "unsigned width (like the indicated %u)\n",
			    argv[0], context.width);

		    return 1;
		}
	    }
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	context.self = argv[0];

	status = case_away(&context);
    }

    return status;
}
