/*
 * qs19.c
 * Copyright (C) 2002-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

#if !defined HAVE_LIBLEWD
# include <cardinal-wx.h>
#endif				/* !HAVE_LIBLEWD */
#include <count2.h>
#include <libx1f4l2.h>
#include <qxcc.h>
#if !defined HAVE_LIBLEWD
# include <vkeyhook.h>
# include <vkeytree.h>
#endif				/* !HAVE_LIBLEWD */

#if 1
# define __LINK_SLIP_TEST__		0
#endif				/* 1 */

#if __HEAD_FLIP_RATE__
# define Sans_logique_100813_0_		sans_logique_100813_0_
#else
# undef __LINK_SLIP_TEST__
# define __LINK_SLIP_TEST__		1
# define Sans_logique_100813_0_		sans_logique________0_
#endif				/* __HEAD_FLIP_RATE__ */

#if __HEAD_MISS_RATE__
# define Sans_logique_090909_5_		sans_logique_090909_5_
#else
# undef __LINK_SLIP_TEST__
# define __LINK_SLIP_TEST__		1
# define Sans_logique_090909_5_		sans_logique________0_
#endif				/* __HEAD_MISS_RATE__ */

#if __HEAD_MISS_RATE__
# define Sans_logique_090910_2_		sans_logique_090910_2_
#else
# undef __LINK_SLIP_TEST__
# define __LINK_SLIP_TEST__		1
# define Sans_logique_090910_2_		sans_logique________0_
#endif				/* __HEAD_MISS_RATE__ */

#if __HEAD_MIND_RATE__
# define Sans_logique_100731_0_		sans_logique_100731_0_
#else
# undef __LINK_SLIP_TEST__
# define __LINK_SLIP_TEST__		1
# define Sans_logique_100731_0_		sans_logique________0_
#endif				/* __HEAD_MIND_RATE__ */

#if __HEAD_SLIP_RATE__
# define Sans_logique_100731_1_		sans_logique_100731_1_
#else
# undef __LINK_SLIP_TEST__
# define __LINK_SLIP_TEST__		1
# define Sans_logique_100731_1_		sans_logique________0_
#endif				/* __HEAD_SLIP_RATE__ */

#define SANS_LOGIQUE_ARGS_0 \
    struct context_type *, const struct test_type *
#define SANS_LOGIQUE_ARGS_1 \
    struct context_type *context_data, const struct test_type *test_data
#define SANS_LOGIQUE_ARGS_2 \
    context_data, test_data

#define APPEND_MODE \
    "last position insert 1,000,000 integers"

#define ASPECT_MODE \
    "sequentially fetch all from 1,000,000 integers list"

#define BEHEAD_MODE \
    "forward remove all from 1,000,000 integers list"

#define BETAIL_MODE \
    "backward remove all from 1,000,000 integers list"

#define DELETE_MODE \
    "sequentially remove all from 1,000,000 integers list"

#define DESERT_MODE \
    "first position insert 1,000,000 integers"

#define EXCESS_MODE \
    "clear 1,000,000 integers list"

#define EXLINE_MODE \
    "callback traverse 1,000,000 integers list"

#define INSERT_MODE \
    "random (2nd) insert 1,000,000 integers"

#define INVERT_MODE \
    "reverse 1,000,000 integers list"

#define RANDOM_MODE \
    "fetch 1,000,000 from 1,000,000 integers list"

#define REMOVE_MODE \
    "remove all from 1,000,000 integers list"

#define STREAM_TEST \
    "all the tests"

#define TARGET_MODE \
    "(libc) random insert 1,000,000 integers"

typedef struct context_type {
    const char *self;
    unsigned class, frame, width;
} context_type;

typedef struct test_type {
    const char *description;
    int (*test) (SANS_LOGIQUE_ARGS_0);
} test_type;

static int fini_list(struct context_type *, void **);
static int init_list(struct context_type *, void **);
static int list_link(void *, const void *, const char *, unsigned);
static int miss_node(void *, void *);
static int sans_logique_090909_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090909_0a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090909_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090909_1a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090909_2_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090909_2a(SANS_LOGIQUE_ARGS_0, void *);
#if __HEAD_MISS_RATE__
static int sans_logique_090909_5_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090909_5a(SANS_LOGIQUE_ARGS_0, void *);
#endif				/* __HEAD_MISS_RATE__ */
static int sans_logique_090910_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090910_0a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090910_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090910_1a(SANS_LOGIQUE_ARGS_0, void *);
#if __HEAD_MISS_RATE__
static int sans_logique_090910_2_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090910_2a(SANS_LOGIQUE_ARGS_0, void *);
#endif				/* __HEAD_MISS_RATE__ */
static int sans_logique_090910_3_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090910_3a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_091004_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_091004_0a(SANS_LOGIQUE_ARGS_0, void *);
#if __HEAD_MIND_RATE__
static int sans_logique_100731_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100731_0a(SANS_LOGIQUE_ARGS_0, void *);
#endif				/* __HEAD_MIND_RATE__ */
#if __HEAD_SLIP_RATE__
static int sans_logique_100731_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100731_1a(SANS_LOGIQUE_ARGS_0, void *);
#endif				/* __HEAD_SLIP_RATE__ */
static int sans_logique_100806_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100806_0a(SANS_LOGIQUE_ARGS_0, void *);
#if __HEAD_FLIP_RATE__
static int sans_logique_100813_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100813_0a(SANS_LOGIQUE_ARGS_0, void *);
#endif				/* __HEAD_FLIP_RATE__ */
#if __LINK_SLIP_TEST__
static int sans_logique________0_(SANS_LOGIQUE_ARGS_0);
#endif				/* __LINK_SLIP_TEST__ */
static int sans_logique__________(SANS_LOGIQUE_ARGS_0);

static void class(void);
static void usage(void);

static const struct test_type test_1_9[] = {
/* *INDENT-OFF* */
    {	TARGET_MODE,		sans_logique_100806_0_			},
    {	INSERT_MODE,		sans_logique_090909_0_			},
    {	APPEND_MODE,		sans_logique_090909_1_			},
    {	DESERT_MODE,		sans_logique_090909_2_			},
    {	0,			NULL					},
    {	0,			NULL					},
    {	REMOVE_MODE,		Sans_logique_090909_5_			},
    {	0,			NULL					},
    {	0,			NULL					},
    {	RANDOM_MODE,		sans_logique_090910_0_			},
    {	ASPECT_MODE,		sans_logique_090910_1_			},
    {	DELETE_MODE,		Sans_logique_090910_2_			},
    {	EXCESS_MODE,		sans_logique_090910_3_			},
    {	EXLINE_MODE,		sans_logique_091004_0_			},
    {	0,			NULL					},
    {	0,			NULL					},
    {	0,			NULL					},
    {	BEHEAD_MODE,		Sans_logique_100731_0_			},
    {	BETAIL_MODE,		Sans_logique_100731_1_			},
    {	INVERT_MODE,		Sans_logique_100813_0_			},
    {	0,			NULL					},
    {	STREAM_TEST,		sans_logique__________			}
/* *INDENT-ON* */
};

static const struct x1f4_vkeytree_type link_1_g[] = {
/* *INDENT-OFF* */
    {	"",		{	test_1_9 + 1	},	NULL		},
    {	"2",		{	test_1_9 + 0	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link_1_e[] = {
/* *INDENT-OFF* */
    {	"all",		{	test_1_9 + 13	},	NULL		},
    {	"lear",		{	test_1_9 + 12	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link___r_e[] = {
/* *INDENT-OFF* */
    {	"move",		{	test_1_9 + 6	},	NULL		},
    {	"verse",	{	test_1_9 + 19	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link___r[] = {
/* *INDENT-OFF* */
    {	"andom",	{	test_1_9 + 9	},	NULL		},
    {	"e",		{	NULL		},	link___r_e	},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link_1_a[] = {
/* *INDENT-OFF* */
    {	"lip",		{	test_1_9 + 18	},	NULL		},
    {	"uffix",	{	test_1_9 + 2	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link_1_9[] = {
/* *INDENT-OFF* */
    {	"all",		{	test_1_9 + 21	},	NULL		},
    {	"c",		{	NULL		},	link_1_e	},
    {	"delete",	{	test_1_9 + 11	},	NULL		},
    {	"fetch",	{	test_1_9 + 10	},	NULL		},
    {	"insert",	{	NULL		},	link_1_g	},
    {	"mind",		{	test_1_9 + 17	},	NULL		},
    {	"prefix",	{	test_1_9 + 3	},	NULL		},
    {	"r",		{	NULL		},	link___r	},
    {	"s",		{	NULL		},	link_1_a	},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link____[] = {
/* *INDENT-OFF* */
    {	"",		{	NULL		},	link_1_9	},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
};

static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_qxrate(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;
    struct x1f4_qxrate_type qxrate;

#if __MISS_FOUR_RATE__
#else
    qxrate.quarter = context_data->frame;
#endif				/* __MISS_FOUR_RATE__ */
    qxrate.rate = context_data->class;
    qxrate.size = context_data->width;

    status = x1f4_init_qxrate
	(list, X1f4_QXRATE_FOUR_FRAME | X1f4_QXRATE_RATE_FRAME, &qxrate);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    } else {
    }

    return status;
}


static int
list_link(void *file, const void *node, const char *data, unsigned size)
{
    fwrite("  ", 2, 1, file);
    fwrite(data, size, 1, file);
    fwrite("                ", 16 - size, 1, file);
    fwrite(" - ", 3, 1, file);
    fputs(((const struct test_type *) node)->description, file);
    fputc('\n', file);

    return 0;
}


static int
miss_node(void *text, void *node)
{
    unsigned *sail;

    sail = text;
    *sail += *(unsigned *) node;

    return 0;
}


static int
sans_logique_090909_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	count2_decls;

	start_count2();

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);

	if (status) {
	} else {
	    finish_count2_l(test_data->description, 55);
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090909_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 0;
    while (i < 1000000) {
	unsigned e;
	void *data;

	e = i * 523 + 342;

	i++;

	e ^= e << 003;
	e += e >> 005;
	e ^= e << 002;
	e += e >> 017;
	e ^= e << 012;

	if (e < i) {
	} else {
	    unsigned a;

	    a = i >> 1;

	    a |= a >> 1;
	    a |= a >> 2;

	    e = e & a;
	}

	if (i & 1) {
	    e = i - e - 1;
	}

	status = x1f4_push_qxrate(list, e, &data);
	if (status) {
	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = i;
	}
    }

    return status;
}


static int
sans_logique_090909_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	count2_decls;

	start_count2();

	status = sans_logique_090909_1a(SANS_LOGIQUE_ARGS_2, list);

	if (status) {
	} else {
	    finish_count2_l(test_data->description, 55);
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090909_1a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 0;
    while (i < 1000000) {
	void *data;

#if __HAVE_TAIL_RATE__
	status = x1f4_tail_qxrate(list, &data);
#else
	status = x1f4_push_qxrate(list, i, &data);
#endif				/* __HAVE_TAIL_RATE__ */
	if (status) {
	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = i;
	}

	i++;
    }

    return status;
}


static int
sans_logique_090909_2_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	count2_decls;

	start_count2();

	status = sans_logique_090909_2a(SANS_LOGIQUE_ARGS_2, list);

	if (status) {
	} else {
	    finish_count2_l(test_data->description, 55);
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090909_2a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	void *data;

	i--;

#if __HAVE_HEAD_RATE__
	status = x1f4_head_qxrate(list, &data);
#else
	status = x1f4_push_qxrate(list, 0, &data);
#endif				/* __HAVE_HEAD_RATE__ */
	if (status) {
	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = i;
	}
    }

    return status;
}


#if __HEAD_MISS_RATE__
static int
sans_logique_090909_5_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090909_5a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_MISS_RATE__ */


#if __HEAD_MISS_RATE__
static int
sans_logique_090909_5a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	unsigned e;

	e = i * 523 + 342;

	i--;

	e ^= e << 003;
	e += e >> 005;
	e ^= e << 002;
	e += e >> 017;
	e ^= e << 012;

	if (e < i) {
	} else {
	    unsigned a;

	    a = i >> 1;

	    a |= a >> 1;
	    a |= a >> 2;

	    e = e & a;
	}

	if (i & 1) {
	    e = i - e - 1;
	}

	status = x1f4_miss_qxrate(list, e);
	if (status) {
	    break;
	}
    }

    return status;
}
#endif				/* __HEAD_MISS_RATE__ */


static int
sans_logique_090910_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090910_0a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090910_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
#define E				1000000

    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	unsigned e;
	void *data;

	e = i * 523 + 342;

	i--;

	e ^= e << 003;
	e += e >> 005;
	e ^= e << 002;
	e += e >> 017;
	e ^= e << 012;

	if (e < E) {
	} else {
	    unsigned a;

	    a = E >> 1;

	    a |= a >> 1;
	    a |= a >> 2;

	    e = e & a;
	}

	if (i & 1) {
	    e = E - e - 1;
	}

	status = x1f4_peek_qxrate(list, e, &data);
	if (status) {
	    break;
	}
    }

    return status;

#undef E
}


static int
sans_logique_090910_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090910_1a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090910_1a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	void *data;

	i--;

	status = x1f4_peek_qxrate(list, i, &data);
	if (status) {
	    break;
	}
    }

    return status;
}


#if __HEAD_MISS_RATE__
static int
sans_logique_090910_2_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090910_2a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_MISS_RATE__ */


#if __HEAD_MISS_RATE__
static int
sans_logique_090910_2a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	i--;

	status = x1f4_miss_qxrate(list, i);
	if (status) {
	    break;
	}
    }

    return status;
}
#endif				/* __HEAD_MISS_RATE__ */


static int
sans_logique_090910_3_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090910_3a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090910_3a(SANS_LOGIQUE_ARGS_1, void *list)
{
    return x1f4_high_qxrate(list);
}


static int
sans_logique_091004_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_091004_0a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_091004_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    unsigned text = 0;

    return x1f4_lime_qxrate(list, &text, miss_node);
}


#if __HEAD_MIND_RATE__
static int
sans_logique_100731_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100731_0a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_MIND_RATE__ */


#if __HEAD_MIND_RATE__
static int
sans_logique_100731_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	i--;

	status = x1f4_mind_qxrate(list);
	if (status) {
	    break;
	}
    }

    if (status) {
    } else {
	status = x1f4_case_qxrate(list);
    }

    return status;
}
#endif				/* __HEAD_MIND_RATE__ */


#if __HEAD_SLIP_RATE__
static int
sans_logique_100731_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100731_1a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_SLIP_RATE__ */


#if __HEAD_SLIP_RATE__
static int
sans_logique_100731_1a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	i--;

	status = x1f4_slip_qxrate(list);
	if (status) {
	    break;
	}
    }

    if (status) {
    } else {
	status = x1f4_case_qxrate(list);
    }

    return status;
}
#endif				/* __HEAD_SLIP_RATE__ */


static int
sans_logique_100806_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	count2_decls;

	srand(0);

	start_count2();

	status = sans_logique_100806_0a(SANS_LOGIQUE_ARGS_2, list);

	if (status) {
	} else {
	    finish_count2_l(test_data->description, 55);
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_100806_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 0;
    while (i < 1000000) {
	void *data;

	i++;

	status = x1f4_push_qxrate(list, rand() % i, &data);
	if (status) {
	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = i;
	}
    }

    return status;
}


#if __HEAD_FLIP_RATE__
static int
sans_logique_100813_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100813_0a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_FLIP_RATE__ */


#if __HEAD_FLIP_RATE__
static int
sans_logique_100813_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    return x1f4_flip_qxrate(list);
}
#endif				/* __HEAD_FLIP_RATE__ */


#if __LINK_SLIP_TEST__
static int
sans_logique________0_(SANS_LOGIQUE_ARGS_1)
{
    count2_print("%*s: not available\n", 55, test_data->description);

    return 0;
}
#endif				/* __LINK_SLIP_TEST__ */


static int
sans_logique__________(SANS_LOGIQUE_ARGS_1)
{
    int status = 0;

    test_data = test_1_9;

    do {
	if (test_data->test) {
	    status = test_data->test(SANS_LOGIQUE_ARGS_2);
	    if (status) {
		break;
	    }
	}

	test_data++;
    } while (test_data->test != sans_logique__________);

    return status;
}


static void
class(void)
{
    x1f4_hook_vtkey(stdout, list_link, link____);
}


static void
usage(void)
{
    puts("Usage: qs19 [OPTIONS] TEST\n\
Profile the " qxname "rate sequence.\n\
\n\
Options:\n\
  -e, --external L		set external nodes width as L * 4\n\
  -i, --internal L		set internal nodes width as 2 ^ L\n\
  -w, --width WIDTH		use WIDTH bytes items\n\
      --list			lists available tests and exit\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    context.class = 8;
    context.frame = 1 << 4;
    context.width = sizeof(unsigned);

    unsetenv("POSIXLY_CORRECT");

    {
	const char *class_specs = NULL, *frame_specs = NULL, *width_specs =
	    NULL;
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "external",	    0x01,   NULL,   0x65    },
		{   "internal",	    0x01,   NULL,   0x69    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   "list",	    0x00,   NULL,   0x6c    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "width",	    0x01,   NULL,   0x77    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "e:i:w:", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'e':
		frame_specs = optarg;
		break;
	    case 'h':
	    case 'l':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case 'i':
		class_specs = optarg;
		break;
	    case 'w':
		width_specs = optarg;
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'l':
		class();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}

	if (class_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.class, class_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse internal node width specif"
			"ication: `%s'\n", argv[0], class_specs);

		return 1;
	    } else {
		if (context.class < 4) {
		    fprintf(stderr, "%s: will not use internal node widths l"
			    "ess than 1 << 4 (like the indicated 1 << %u)\n",
			    argv[0], context.class);

		    return 1;
		}
	    }
	}

	if (frame_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.frame, frame_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse external node width specif"
			"ication: `%s'\n", argv[0], frame_specs);

		return 1;
	    } else {
		if (context.frame < 1) {
		    fprintf(stderr, "%s: will not use external node widths l"
			    "ess than 4 (like the indicated %u << 2)\n",
			    argv[0], context.frame);

		    return 1;
		}
	    }
	}

	if (width_specs) {
	    status = x1f4_parse_wxcardinal
		(&context.width, width_specs, NULL, 0);
	    if (status) {
		fprintf(stderr, "%s: cannot parse item width specification: "
			"`%s'\n", argv[0], width_specs);

		return 1;
	    } else {
		if (context.width < sizeof(unsigned)) {
		    fprintf(stderr, "%s: will not use item widths less than "
			    "unsigned width (like the indicated %u)\n",
			    argv[0], context.width);

		    return 1;
		}
	    }
	}
    }

    if (argc - optind != 1) {
	fprintf(stderr, "%s: wrong number of arguments\nTry `%s --help' for "
		"more information.\n", argv[0], argv[0]);

	return 1;
    } else {
	const void *test;

	if (x1f4_find_vtkey(argv[optind], link____, &test)) {
	    status = 1;

	    fprintf(stderr, "%s: cannot parse test specification: `%s'\n",
		    argv[0], argv[optind]);
	} else {
	    context.self = argv[0];

	    status = ((struct test_type *) test)->test(&context, test);
	}
    }

    return status;
}
