/*
 * tc13.c
 * Copyright (C) 2002-2009, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <inter.h>
#include <libx1f4l2.h>
#include <tcxx.h>

typedef struct context_type {
    const char *self;
} context_type;

static int case_away(struct context_type *);
static int dump_list(struct context_type *, void *);
static int fast_line(struct context_type *, void *, char *, unsigned,
		     unsigned);
static int fast_list(struct context_type *, void *, void *, unsigned);
static int fill_spin(struct context_type *, void *, unsigned);
static int fini_list(struct context_type *, void **);
static int head_line(struct context_type *, void *, char *, unsigned);
static int init_list(struct context_type *, void **);
static int lock_line(struct context_type *, void *, char *, unsigned,
		     unsigned *);
static int post_line(struct context_type *, void *, char *, unsigned);
static int read_list(struct context_type *, void *);
static int spin_list(struct context_type *, void *);

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = read_list(context_data, list);
	if (status) {
	} else {
	    status = spin_list(context_data, list);
	    if (status) {
	    } else {
		status = dump_list(context_data, list);
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
dump_list(struct context_type *context_data, void *list)
{
    int status = 0;
    unsigned i = 0, size;

    x1f4_size_tclist(list, &size);
    while (i < size) {
	void *peek;

	status = x1f4_peek_tclist(list, i, &peek);
	if (status) {
	    break;
	} else {
	    puts(peek);
	    i++;
	}
    }

    return status;
}


static int
fast_line(struct context_type *context_data, void *list, char *data,
	  unsigned size, unsigned post)
{
    int status;
    void *slip;

    status = x1f4_push_tclist(list, post, size + 1, &slip);
    if (status) {
	fprintf(stderr, "%s: cannot insert item\n", context_data->self);
    } else {
	if (0) {
	    status = x1f4_flow_tclist(list);
	}
	if (status) {
	    fprintf(stderr, "%s: corrupt list\n", context_data->self);
	} else {
	    memcpy(slip, data, size + 1);
	}
    }

    return status;
}


static int
fast_list(struct context_type *context_data, void *list, void *spin,
	  unsigned size)
{
    int status;
    unsigned fast;

    fast = size;

    size >>= 1;
    if (!size) {
	status = 0;
    } else {
	srand(time(NULL));

	do {
	    unsigned peek;
	    void *data;

	    peek = rand() % size;

	    size--;

	    status = x1f4_peek_tclist(spin, peek, &data);
	    if (status) {
		break;
	    } else {
		unsigned slip;
		unsigned char *miss;

		miss = data;

		slip = miss[0] << 030 | miss[1] << 020 | miss[2] << 010
		    | miss[3];

		status = x1f4_miss_tclist(spin, peek);
		if (status) {
		    break;
		} else {
		    status = x1f4_peek_tclist(list, slip, &data);
		    if (status) {
			break;
		    } else {
			void *spin;

			peek = strlen(data) + 1;

			status = x1f4_push_tclist
			    (list, fast - slip, peek, &spin);
			if (status) {
			    break;
			} else {
			    if (0) {
				status = x1f4_flow_tclist(list);
				if (status) {
				    break;
				}
			    }

			    memcpy(spin, data, peek);

			    status = x1f4_miss_tclist(list, slip);
			    if (status) {
				break;
			    } else {
				if (0) {
				    status = x1f4_flow_tclist(list);
				     if (status) {
					break;
				    }
				}

				status = x1f4_peek_tclist
				    (list, fast - slip - 2, &data);
				if (status) {
				    break;
				} else {
				    peek = strlen(data) + 1;

				    status = x1f4_push_tclist
					(list, slip, peek, &spin);
				    if (status) {
					break;
				    } else {
					if (0) {
					    status = x1f4_flow_tclist(list);
					     if (status) {
						break;
					    }
					}

					memcpy(spin, data, peek);

					status = x1f4_miss_tclist
					    (list, fast - slip - 1);
					if (status) {
					    break;
					} else {
					    if (0) {
						status =
						    x1f4_flow_tclist(list);
						if (status) {
						    break;
						}
					    }
					}
				    }
				}
			    }
			}
		    }
		}
	    }
	} while (size);
    }

    return status;
}


static int
fill_spin(struct context_type *context_data, void *spin, unsigned size)
{
    int status;

    size >>= 1;
    if (!size) {
	status = 0;
    } else {
	do {
	    unsigned char *miss;
	    void *data;

	    size--;

	    status = x1f4_push_tclist(spin, 0, 4, &data);
	    if (status) {
		break;
	    } else {
		miss = data;

		miss[0] = size >> 030;
		miss[1] = size >> 020;
		miss[2] = size >> 010;
		miss[3] = size >> 000;
	    }
	} while (size);
    }

    return status;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_tclist(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
head_line(struct context_type *context_data, void *list, char *data,
	  unsigned size)
{
    int status;
    void *slip;

    status = x1f4_push_tclist(list, 0, size + 1, &slip);
    if (status) {
	fprintf(stderr, "%s: cannot insert item\n", context_data->self);
    } else {
	status = x1f4_flow_tclist(list);
	if (status) {
	    fprintf(stderr, "%s: corrupt list\n", context_data->self);
	} else {
	    memcpy(slip, data, size + 1);
	}
    }

    return status;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_init_tclist(list, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    }

    return status;
}


static int
lock_line(struct context_type *context_data, void *list, char *data,
	  unsigned size, unsigned *lock)
{
    int status = 0;
    unsigned head = 0, line;

    x1f4_size_tclist(list, &line);
    while (head < line) {
	unsigned half;
	void *peek;

	half = (head + line) >> 1;
	status = x1f4_peek_tclist(list, half, &peek);
	if (status) {
	    break;
	} else {
	    int slip;

	    slip = strcmp(data, peek);
	    if (slip < 0) {
		line = half;
	    } else {
		head = half + 1;
	    }
	}
    }

    if (status) {
    } else {
	*lock = head;
    }

    return status;
}


static int
post_line(struct context_type *context_data, void *list, char *data,
	  unsigned size)
{
    int status;
    unsigned post;

    status = lock_line(context_data, list, data, size, &post);
    if (status) {
    } else {
	status = fast_line(context_data, list, data, size, post);
    }

    return status;
}


static int
read_list(struct context_type *context_data, void *list)
{
    int status;

    do {
	char *data;
	int done = 0;
	unsigned size = 0;

	status = libx1f4l2_read_line(context_data->self, &done, &data, &size);
	if (status) {
	    break;
	} else {
	    if (!done || size) {
		status = head_line(context_data, list, data, size);
	    }

	    if (1) {
		free(data);
	    }

	    if (status) {
		break;
	    }

	    if (done) {
		break;
	    } else {
		while (1) {
		    status = libx1f4l2_read_line
			(context_data->self, &done, &data, &size);
		    if (status) {
			break;
		    }

		    if (!done || size) {
			status = post_line(context_data, list, data, size);
		    }

		    if (1) {
			free(data);
		    }

		    if (status) {
			break;
		    }

		    if (done) {
			break;
		    }
		}
	    }
	}
    } while (0);

    return status;
}


static int
spin_list(struct context_type *context_data, void *list)
{
    int status;
    void *spin;

    status = x1f4_init_tclist(&spin, 0, NULL);
    if (status) {
    } else {
	int excess;
	unsigned size;

	x1f4_size_tclist(list, &size);

	status = fill_spin(context_data, spin, size);
	if (status) {
	} else {
	    status = fast_list(context_data, list, spin, size);
	}

	excess = x1f4_fini_tclist(&spin);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static void
usage(void)
{
    puts("Usage: tc13 [OPTIONS]\n\
Test " tcname "_delete(), output reverse sorted input.\n\
\n\
Options:\n\
  --help			display this help and exit\n\
  --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;

    unsetenv("POSIXLY_CORRECT");

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	struct context_type context;

	context.self = argv[0];

	status = case_away(&context);
    }

    return status;
}
