/*
 * tc39.c
 * Copyright (C) 2002-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <errno.h>
#include <fcntl.h>
#include <getopt.h>
#if defined HAVE_LIBx1f4i0
# include <libx1f4i0.h>
#endif				/* HAVE_LIBx1f4i0 */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <count2.h>
#include <inter.h>
#include <libx1f4l2.h>
#include <tcqp.h>
#if !defined HAVE_LIBx1f4i0
# include <vkeyhook.h>
# include <vkeytree.h>
#endif				/* !HAVE_LIBx1f4i0 */

#define __STRING_SELECT__		0

#if __HEAD_FGXD_LINE__
# define Sans_logique_100718_3_		sans_logique_100718_3_
#else
# define Sans_logique_100718_3_		sans_logique________0_
#endif				/* __HEAD_FGXD_LINE__ */

#if __HEAD_FIND_FINE__
# define Sans_logique_090928_0_		sans_logique_090928_0_
#else
# define Sans_logique_090928_0_		sans_logique________0_
#endif				/* __HEAD_FIND_FINE__ */

#if __HEAD_FIND_WORD__
# define Sans_logique_100806_3_		sans_logique_100806_3_
#else
# define Sans_logique_100806_3_		sans_logique________0_
#endif				/* __HEAD_FIND_WORD__ */

#if __HEAD_FMXD_LINE__
# define Sans_logique_100719_0_		sans_logique_100719_0_
#else
# define Sans_logique_100719_0_		sans_logique________0_
#endif				/* __HEAD_FMXD_LINE__ */

#if __HEAD_FXXD_LINE__
# define Sans_logique_100718_2_		sans_logique_100718_2_
#else
# define Sans_logique_100718_2_		sans_logique________0_
#endif				/* __HEAD_FXXD_LINE__ */

#if __HEAD_LLXD_LINE__
# define Sans_logique_100718_0_		sans_logique_100718_0_
#else
# define Sans_logique_100718_0_		sans_logique________0_
#endif				/* __HEAD_LLXD_LINE__ */

#if __HEAD_LMXD_LINE__
# define Sans_logique_100719_1_		sans_logique_100719_1_
#else
# define Sans_logique_100719_1_		sans_logique________0_
#endif				/* __HEAD_LMXD_LINE__ */

#if __HEAD_LXXD_LINE__
# define Sans_logique_100718_1_		sans_logique_100718_1_
#else
# define Sans_logique_100718_1_		sans_logique________0_
#endif				/* __HEAD_LXXD_LINE__ */

#if __HEAD_NEXT_LINE__
# define Sans_logique_090910_6_		sans_logique_090910_6_
#else
# define Sans_logique_090910_6_		sans_logique________0_
#endif				/* __HEAD_NEXT_LINE__ */

#if __HEAD_NEXT_LINE__
# define Sans_logique_090910_7_		sans_logique_090910_7_
#else
# define Sans_logique_090910_7_		sans_logique________0_
#endif				/* __HEAD_NEXT_LINE__ */

#if __HEAD_FIND_WORD__
# define Sans_logique_100819_1_		sans_logique_100819_1_
#else
# define Sans_logique_100819_1_		sans_logique________0_
#endif				/* __HEAD_FIND_WORD__ */

#define SANS_LOGIQUE_ARGS_0 \
    struct context_type *, const struct test_type *
#define SANS_LOGIQUE_ARGS_1 \
    struct context_type *context_data, const struct test_type *test_data
#define SANS_LOGIQUE_ARGS_2 \
    context_data, test_data

#define CKMISS_MODE \
    "miss 1,000,000 times in 1,000,000 integers list"

#define DDEPTH_MODE \
    "balance after 2 * 10^6 insertions, half deleted"

#define DELETE_MODE \
    "(sort of) random delete all in 1,000,000 integers list"

#define DIRECT_MODE \
    "forward traverse 1,000,000 integers list"

#define EXCESS_MODE \
    "clear 1,000,000 integers list"

#define EXLINE_MODE \
    "callback traverse 1,000,000 integers list"

#define EXPECT_MODE \
    "backward traverse 1,000,000 integers list"

#define FRIDAY_MODE \
    "insert 1,000,000 non unique 30bit 13 powers"

#define HIGHER_MODE \
    "insert ever greater 1,000,000 integers"

#define IDEPTH_MODE \
    "balance after 1,000,000 insertions"

#define INSERT_MODE \
    "insert 1,000,000 integers"

#define LESSER_MODE \
    "insert ever lesser 1,000,000 integers"

#define LOOKUP_FINE \
    "random (special string) retrieve all dictionary strings"

#define LOOKUP_MODE \
    "random retrieve all dictionary strings"

#define MONDAY_LANE \
    "(sort of) random (l) fetch all from 10^6 integers list"

#define MONDAY_MODE \
    "(sort of) random (f) fetch all from 10^6 integers list"

#define RANDOM_FINE \
    "(sort of) random special integer fetch, 10^6 items list"

#define RANDOM_MODE \
    "(sort of) random fetch all from 1,000,000 integers list"

#define REMOVE_MODE \
    "forward delete 1,000,000 integers list"

#define SELECT_FINE \
    "random (special string) retrieve all, 10^6 items list"

#define SELECT_MODE \
    "random retrieve all from 1,000,000 strings list"

#define STREAM_TEST \
    "all the tests"

#define STRING_MODE \
    "insert 1,000,000 strings"

#define STRIPE_MODE \
    "insert dictionary strings"

#define SUNDAY_CASE \
    "fetch 10^6 the first not less, 10^6 integers list"

#define SUNDAY_FINE \
    "fetch 10^6 the last not greater, 10^6 integers list"

#define SUNDAY_LANE \
    "fetch 10^6 the first greater, 10^6 integers list"

#define SUNDAY_MODE \
    "fetch 10^6 the last less, 10^6 integers list"

#define TARGET_MODE \
    "(libc) random insert 1,000,000 integers"

#define UNLINK_MODE \
    "backward delete 1,000,000 integers list"

typedef struct context_type {
    const char *dictionary, *self;
} context_type;

#if __STRING_SELECT__
typedef struct lock_type {
    unsigned byte, call, left, plus;
    void *text;
} lock_type;
#endif				/* __STRING_SELECT__ */

typedef struct test_type {
    const char *description;
    int (*test) (SANS_LOGIQUE_ARGS_0);
} test_type;

#if __STRING_SELECT__
static int deck_call(void *, void *);
#endif				/* __STRING_SELECT__ */
static int deck_mode(void *, void *);
static int deck_miss(void *, void *);
static int fini_list(struct context_type *, void **);
static int fini_path(struct context_type *, void **);
static int fini_rate(struct context_type *, void **);
static int init_list(struct context_type *, void **);
static int init_path(struct context_type *, void **);
static int init_rate(struct context_type *, void **);
static int list_link(void *, const void *, const char *, unsigned);
static int mean_list(void *, double *, unsigned *);
static int miss_node(void *, void *);
static int sans_logique_090910_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090910_0a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090910_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090910_1a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090910_2_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090910_2a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090910_3_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090910_3a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090910_4_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090910_4a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090910_5_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090910_5a(SANS_LOGIQUE_ARGS_0, void *);
#if __HEAD_NEXT_LINE__
static int sans_logique_090910_6_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090910_6a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090910_7_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090910_7a(SANS_LOGIQUE_ARGS_0, void *);
#endif				/* __HEAD_NEXT_LINE__ */
static int sans_logique_090911_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090911_0a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090911_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090911_1a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090911_2_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090911_2a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090911_3_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090911_3a(SANS_LOGIQUE_ARGS_0, void *);
#if __HEAD_FIND_FINE__
static int sans_logique_090928_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090928_0a(SANS_LOGIQUE_ARGS_0, void *);
#endif				/* __HEAD_FIND_FINE__ */
static int sans_logique_090929_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090929_0a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090929_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090929_1a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_100124_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100124_0a(SANS_LOGIQUE_ARGS_0, void *);
#if __HEAD_LLXD_LINE__
static int sans_logique_100718_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100718_0a(SANS_LOGIQUE_ARGS_0, void *);
#endif				/* __HEAD_LLXD_LINE__ */
#if __HEAD_LXXD_LINE__
static int sans_logique_100718_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100718_1a(SANS_LOGIQUE_ARGS_0, void *);
#endif				/* __HEAD_LXXD_LINE__ */
#if __HEAD_FXXD_LINE__
static int sans_logique_100718_2_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100718_2a(SANS_LOGIQUE_ARGS_0, void *);
#endif				/* __HEAD_FXXD_LINE__ */
#if __HEAD_FGXD_LINE__
static int sans_logique_100718_3_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100718_3a(SANS_LOGIQUE_ARGS_0, void *);
#endif				/* __HEAD_FGXD_LINE__ */
#if __HEAD_FMXD_LINE__
static int sans_logique_100719_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100719_0a(SANS_LOGIQUE_ARGS_0, void *);
#endif				/* __HEAD_FMXD_LINE__ */
#if __HEAD_LMXD_LINE__
static int sans_logique_100719_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100719_1a(SANS_LOGIQUE_ARGS_0, void *);
#endif				/* __HEAD_LMXD_LINE__ */
static int sans_logique_100806_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100806_0a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_100806_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100806_1a(SANS_LOGIQUE_ARGS_0, void *, void *);
static int sans_logique_100806_2_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100806_2a(SANS_LOGIQUE_ARGS_0, void *, void *);
#if __HEAD_FIND_WORD__
static int sans_logique_100806_3_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100806_3a(SANS_LOGIQUE_ARGS_0, void *, void *);
#endif				/* __HEAD_FIND_WORD__ */
static int sans_logique_100812_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100812_0a(SANS_LOGIQUE_ARGS_0, void *, void *);
static int sans_logique_100812_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100812_1a(SANS_LOGIQUE_ARGS_0, void *, void *);
#if __HEAD_FIND_WORD__
static int sans_logique_100819_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100819_1a(SANS_LOGIQUE_ARGS_0, void *, void *);
#endif				/* __HEAD_FIND_WORD__ */
static int sans_logique________0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique__________(SANS_LOGIQUE_ARGS_0);
static int scan_node(void *, void *, unsigned);

static void class(void);
static void usage(void);

static const char digits[] = "0123456789abcdef";
static const struct test_type test_1_9[] = {
/* *INDENT-OFF* */
    {	TARGET_MODE,		sans_logique_100806_0_			},
    {	INSERT_MODE,		sans_logique_090910_0_			},
    {	LESSER_MODE,		sans_logique_090910_1_			},
    {	HIGHER_MODE,		sans_logique_090910_2_			},
    {	EXCESS_MODE,		sans_logique_090910_3_			},
    {	RANDOM_MODE,		sans_logique_090910_4_			},
    {	RANDOM_FINE,		Sans_logique_090928_0_			},
    {	EXLINE_MODE,		sans_logique_090910_5_			},
    {	DIRECT_MODE,		Sans_logique_090910_6_			},
    {	EXPECT_MODE,		Sans_logique_090910_7_			},
    {	DELETE_MODE,		sans_logique_090911_0_			},
    {	REMOVE_MODE,		sans_logique_090911_1_			},
    {	UNLINK_MODE,		sans_logique_090911_2_			},
    {	CKMISS_MODE,		sans_logique_090911_3_			},
    {	IDEPTH_MODE,		sans_logique_090929_0_			},
    {	DDEPTH_MODE,		sans_logique_090929_1_			},
    {	FRIDAY_MODE,		sans_logique_100124_0_			},
    {	SUNDAY_MODE,		Sans_logique_100718_0_			},
    {	SUNDAY_FINE,		Sans_logique_100718_1_			},
    {	SUNDAY_CASE,		Sans_logique_100718_2_			},
    {	SUNDAY_LANE,		Sans_logique_100718_3_			},
    {	MONDAY_MODE,		Sans_logique_100719_0_			},
    {	MONDAY_LANE,		Sans_logique_100719_1_			},
    {	STRING_MODE,		sans_logique_100806_1_			},
    {	SELECT_MODE,		sans_logique_100806_2_			},
    {	SELECT_FINE,		Sans_logique_100806_3_			},
    {	STRIPE_MODE,		sans_logique_100812_0_			},
    {	LOOKUP_MODE,		sans_logique_100812_1_			},
    {	LOOKUP_FINE,		Sans_logique_100819_1_			},
    {	STREAM_TEST,		sans_logique__________			}
/* *INDENT-ON* */
};

static const struct x1f4_vkeytree_type link___f_i_r[] = {
/* *INDENT-OFF* */
    {	"greater",	{	test_1_9 + 20	},	NULL		},
    {	"matching",	{	test_1_9 + 21	},	NULL		},
    {	"not-less",	{	test_1_9 + 19	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link___f_i[] = {
/* *INDENT-OFF* */
    {	"nd",		{	test_1_9 + 24	},	NULL		},
    {	"rst-",		{	NULL		},	link___f_i_r	},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link_1_h[] = {
/* *INDENT-OFF* */
    {	"",		{	test_1_9 + 1	},	NULL		},
    {	"2",		{	test_1_9 + 0	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link___f[] = {
/* *INDENT-OFF* */
    {	"i",		{	NULL		},	link___f_i	},
    {	"orward",	{	test_1_9 + 8	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link_1_e[] = {
/* *INDENT-OFF* */
    {	"less",		{	test_1_9 + 17	},	NULL		},
    {	"matching",	{	test_1_9 + 22	},	NULL		},
    {	"not-greater",	{	test_1_9 + 18	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link_1_d[] = {
/* *INDENT-OFF* */
    {	"ckward",	{	test_1_9 + 9	},	NULL		},
    {	"lance",	{	test_1_9 + 14	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link___s_minus[] = {
/* *INDENT-OFF* */
    {	"find",		{	test_1_9 + 25	},	NULL		},
    {	"lookup",	{	test_1_9 + 28	},	NULL		},
    {	"random",	{	test_1_9 + 6	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link___s[] = {
/* *INDENT-OFF* */
    {	"-",		{	NULL		},	link___s_minus	},
    {	"tring",	{	test_1_9 + 23	},	NULL		},
    {	"uffix",	{	test_1_9 + 3	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link___r[] = {
/* *INDENT-OFF* */
    {	"andom",	{	test_1_9 + 5	},	NULL		},
    {	"emove",	{	test_1_9 + 11	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link___l[] = {
/* *INDENT-OFF* */
    {	"ast-",		{	NULL		},	link_1_e	},
    {	"ookup",	{	test_1_9 + 27	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link___c[] = {
/* *INDENT-OFF* */
    {	"all",		{	test_1_9 + 7	},	NULL		},
    {	"lear",		{	test_1_9 + 4	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link_1_9[] = {
/* *INDENT-OFF* */
    {	"all",		{	test_1_9 + 29	},	NULL		},
    {	"ba",		{	NULL		},	link_1_d	},
    {	"c",		{	NULL		},	link___c	},
    {	"delete",	{	test_1_9 + 10	},	NULL		},
    {	"f",		{	NULL		},	link___f	},
    {	"half-balance",	{	test_1_9 + 15	},	NULL		},
    {	"insert",	{	NULL		},	link_1_h	},
    {	"l",		{	NULL		},	link___l	},
    {	"miss",		{	test_1_9 + 13	},	NULL		},
    {	"prefix",	{	test_1_9 + 2	},	NULL		},
    {	"r",		{	NULL		},	link___r	},
    {	"s",		{	NULL		},	link___s	},
    {	"thirteen",	{	test_1_9 + 16	},	NULL		},
    {	"unlink",	{	test_1_9 + 12	},	NULL		},
    {	"words",	{	test_1_9 + 26	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link____[] = {
/* *INDENT-OFF* */
    {	"",		{	NULL		},	link_1_9	},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
};

#if __STRING_SELECT__
static int
deck_call(void *deck, void *miss)
{
    int delete;
    unsigned char c, e, *v;

    v = (unsigned char *) miss + ((struct lock_type *) deck)->call;

    c = ((struct lock_type *) deck)->byte;
    e = *v;

    if (c ^ e) {
	delete = (int) c - (int) e;
    } else {
	if (c) {
	    unsigned char *s;

	    s = ((struct lock_type *) deck)->text;

	    do {
		s++;
		v++;

		c = *s;
		e = *v;
		if (c ^ e) {
		    if ((int) c < (int) e) {
			unsigned left, plus;

			plus = ((struct lock_type *) deck)->plus;
			left = v - (unsigned char *) miss;
			delete = left < plus ? left : plus;
			s -= left - delete;
			((struct lock_type *) deck)->byte = *s;
			((struct lock_type *) deck)->left = left;
			((struct lock_type *) deck)->text = s;
			((struct lock_type *) deck)->call = delete;

			delete = -1;
		    } else {
			unsigned left, plus;

			left = ((struct lock_type *) deck)->left;
			plus = v - (unsigned char *) miss;
			delete = left < plus ? left : plus;
			s -= plus - delete;
			((struct lock_type *) deck)->byte = *s;
			((struct lock_type *) deck)->plus = plus;
			((struct lock_type *) deck)->text = s;
			((struct lock_type *) deck)->call = delete;

			delete = 13;
		    }

		    break;
		} else {
		    if (c) {
		    } else {
			delete = 0;

			break;
		    }
		}
	    } while (1);
	} else {
	    delete = 0;
	}
    }

    return delete;
}
#endif				/* __STRING_SELECT__ */


static int
deck_mode(void *deck, void *miss)
{
    return (int) *(unsigned *) deck - (int) *(unsigned *) miss;
}


static int
deck_miss(void *deck, void *miss)
{
    int delete;
    unsigned char *s, *v;

    s = deck;
    v = miss;

    while (1) {
	unsigned char c;

	c = *s;
	if (c ^ *v) {
	    delete = (int) c - (int) *v;

	    break;
	} else {
	    if (c) {
		s++;
		v++;
	    } else {
		delete = 0;

		break;
	    }
	}
    }

    return delete;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_tcline(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy tree\n", context_data->self);
    }

    return status;
}


static int
fini_path(struct context_type *context_data, void **path)
{
    free(*path);

    return 0;
}


static int
fini_rate(struct context_type *context_data, void **rate)
{
    int status;

    status = x1f4_fini_l2list(rate);
    if (status) {
	fprintf(stderr, "%s: cannot destroy dictionary\n", context_data->self);
    }

    return status;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_init_tcline(list, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create tree\n", context_data->self);
    } else {
    }

    return status;
}


static int
init_path(struct context_type *context_data, void **path)
{
#define __DUMP_PROFILE__		0
#define __LAME_MATCHES__		0

    int status;
    void *line;

    status = x1f4_init_tvlist(&line, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create tree\n", context_data->self);
    } else {
	unsigned copy = 0;
#if __LAME_MATCHES__
	unsigned matches = 0;
#endif				/* __LAME_MATCHES__ */
#if __DUMP_PROFILE__
	unsigned profile[24];
#endif				/* __DUMP_PROFILE__ */
	unsigned size = 0;

	srand(0);

#if __DUMP_PROFILE__
	memset(profile, 0, 24 * sizeof(unsigned));
#endif				/* __DUMP_PROFILE__ */

	while (size < 1000000) {
	    char buffer[24], *string;
	    unsigned extend, length;
	    void *data;

	    length = rand() % (256 - 112) + 112;
	    length = length * length * length;
	    length = (1 << 25) / length;

	    extend = length + 1;

	    string = buffer;
	    for (; length; length--) {
		*string++ = digits[rand() & 15];
	    }

	    *string = 0;

	    status = x1f4_post_tvlist
		(line, buffer, deck_miss, 0, extend, &data);
	    if (status == X1f4_TVLINE_EVER_MATCH) {
#if __LAME_MATCHES__
		matches++;
#endif				/* __LAME_MATCHES__ */
	    } else {
		if (status) {
		    break;
		} else {
#if __DUMP_PROFILE__
		    profile[extend - 1]++;
#endif				/* __DUMP_PROFILE__ */
		    memcpy(data, buffer, extend);
		    if (1) {
			size++;
			if (1) {
			    copy += extend;
			}
		    }
		}
	    }
	}

	if (status) {
	} else {
	    x1f4_rand_tvlist(line);

#if __DUMP_PROFILE__
	    if (1) {
		unsigned i = 3, *p;

		p = profile;

		for (; i; i--) {
		    unsigned j = 8;

		    for (; j; j--) {
			fprintf(stderr, " %7u", *p++);
		    }

		    fputc('\n', stderr);
		}
	    }
#endif				/* __DUMP_PROFILE__ */
	}

	if (status) {
#if __LAME_MATCHES__
	} else {
	    fprintf(stderr, "matched for %u times while building dictionary\n",
		    matches);
#endif				/* __LAME_MATCHES__ */
	}

	if (status) {
	} else {
	    void *text;

	    text = malloc(copy + 1000000);
	    if (!text) {
		status = -1;
	    } else {
		unsigned char *slip;
		void *data;

		*path = text;

		slip = text;

		x1f4_lead_tvlist(line, &data);

		do {
		    unsigned extend;

		    extend = strlen(data) + 1;

		    *slip++ = extend;

		    memcpy(slip, data, extend);

		    slip += extend;
		} while (x1f4_next_tvlist(&data, data));
	    }
	}

	x1f4_fini_tvlist(&line);
    }

    return status;

#undef __LAME_MATCHES__
#undef __DUMP_PROFILE__
}


static int
init_rate(struct context_type *context_data, void **rate)
{
    int status;

    status = x1f4_init_l2list(rate, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create dictionary\n", context_data->self);
    } else {
	if (!context_data->dictionary) {
	    fprintf(stderr, "%s: no dictionary defined\n", context_data->self);

	    status = -1;
	} else {
	    int file;

	    file = open(context_data->dictionary, O_RDONLY);
	    if (file == -1) {
		status = 1;

		fprintf(stderr, "%s: cannot open `%s': %s\n",
			context_data->self, context_data->dictionary,
			strerror(errno));
	    } else {
		int done = 0, excess;
		void *list;

		list = *rate;

		while (1) {
		    char *data;
		    unsigned size;

		    status = libx1f4l2_rate_line
			(context_data->self, context_data->dictionary, file,
			 &done, &data, &size);
		    if (status) {
			break;
		    } else {
			if (size) {
			    void *line;

			    status = x1f4_tail_l2list
				(list, sizeof(unsigned) + size + 1, &line);
			    if (status) {
				break;
			    } else {
				unsigned *copy;

				copy = line;

				*copy = size;

				memcpy(copy + 1, data, size + 1);
			    }
			}

			free(data);

			if (done) {
			    break;
			}
		    }
		}

		excess = close(file);
		if (excess) {
		    if (status) {
		    } else {
			status = excess;
		    }

		    fprintf(stderr, "%s: cannot close `%s': %s\n",
			    context_data->self, context_data->dictionary,
			    strerror(errno));
		}
	    }
	}

	if (status) {
	    x1f4_fini_l2list(rate);
	}
    }

    return status;
}


static int
list_link(void *file, const void *node, const char *data, unsigned size)
{
    fwrite("  ", 2, 1, file);
    fwrite(data, size, 1, file);
    fwrite("                ", 16 - size, 1, file);
    fwrite(" - ", 3, 1, file);
    fputs(((const struct test_type *) node)->description, file);
    fputc('\n', file);

    return 0;
}


static int
mean_list(void *list, double *depth, unsigned *maximum)
{
    unsigned miss[64];

    memset(miss, 0, sizeof(unsigned) << 6);

    x1f4_scan_tcline(list, miss, scan_node);

    if (1) {
	double class = 0;
	unsigned i, size = 0, *slip;

	slip = miss;
	i = 0;
	for (; i < 64; i++) {
	    unsigned fast;

	    fast = *slip;
	    if (fast) {
		class += i * fast;
		size += fast;
		slip++;
	    } else {
		break;
	    }
	}

	*maximum = i;

	*depth = class / size;
    }

    return 0;
}


static int
miss_node(void *text, void *node)
{
    return 0;
}


static int
sans_logique_090910_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	count2_decls;

	start_count2();

	status = sans_logique_090910_0a(SANS_LOGIQUE_ARGS_2, list);

	if (status) {
	} else {
	    finish_count2_l(test_data->description, 55);
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090910_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i, v = 1;

    i = 1000000;
    while (i) {
	unsigned e;
	void *data;

	e = i * 523 + 342;

	v *= 1033;

	i--;

	e ^= v;

	e = (e & 0x3fff8000) >> 15 | (e & 0x00007fff) << 15;
	e = (e & 0x3e0f83e0) >> 5 | (e & 0x01f07c1f) << 5;

	status = x1f4_post_tcline
	    (list, &e, deck_mode, 0, sizeof(unsigned), &data);
	if (status) {
	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = e;
	}
    }

    return status;
}


static int
sans_logique_090910_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	count2_decls;

	start_count2();

	status = sans_logique_090910_1a(SANS_LOGIQUE_ARGS_2, list);

	if (status) {
	} else {
	    finish_count2_l(test_data->description, 55);
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090910_1a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	void *data;

	i--;

	status = x1f4_post_tcline
	    (list, &i, deck_mode, 0, sizeof(unsigned), &data);
	if (status) {
	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = i;
	}
    }

    return status;
}


static int
sans_logique_090910_2_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	count2_decls;

	start_count2();

	status = sans_logique_090910_2a(SANS_LOGIQUE_ARGS_2, list);

	if (status) {
	} else {
	    finish_count2_l(test_data->description, 55);
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090910_2a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 0;
    while (i < 1000000) {
	void *data;

	i++;

	status = x1f4_post_tcline
	    (list, &i, deck_mode, 0, sizeof(unsigned), &data);
	if (status) {
	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = i;
	}
    }

    return status;
}


static int
sans_logique_090910_3_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090910_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090910_3a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090910_3a(SANS_LOGIQUE_ARGS_1, void *list)
{
    return x1f4_high_tcline(list);
}


static int
sans_logique_090910_4_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090910_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090910_4a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090910_4a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i, v = 1;

    i = 1000000;
    while (1) {
	unsigned e;
	void *data;

	e = i * 523 + 342;

	v *= 1033;

	i--;

	e ^= v;

	e = (e & 0x3fff8000) >> 15 | (e & 0x00007fff) << 15;
	e = (e & 0x3e0f83e0) >> 5 | (e & 0x01f07c1f) << 5;

	status = x1f4_find_tcline(list, &e, 0, deck_mode, &data);
	if (status) {
	    if (i) {
	    } else {
		status = 0;
		if (1) {
		    break;
		}
	    }
	} else {
	    status = 1;
	    if (1) {
		break;
	    }
	}
    }

    return status;
}


static int
sans_logique_090910_5_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090910_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090910_5a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090910_5a(SANS_LOGIQUE_ARGS_1, void *list)
{
    return x1f4_lime_tcline(list, NULL, miss_node);
}


#if __HEAD_NEXT_LINE__
static int
sans_logique_090910_6_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090910_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090910_6a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_NEXT_LINE__ */


#if __HEAD_NEXT_LINE__
static int
sans_logique_090910_6a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;

    if (x1f4_case_tcline(list)) {
	void *data;

	status = x1f4_lead_tcline(list, &data);
	if (status) {
	} else {
	    do {
	    } while (x1f4_next_tcline(&data, data));
	}
    } else {
	status = 0;
    }

    return status;
}
#endif				/* __HEAD_NEXT_LINE__ */


#if __HEAD_NEXT_LINE__
static int
sans_logique_090910_7_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090910_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090910_7a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_NEXT_LINE__ */


#if __HEAD_NEXT_LINE__
static int
sans_logique_090910_7a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;

    if (x1f4_case_tcline(list)) {
	void *data;

	status = x1f4_ever_tcline(list, &data);
	if (status) {
	} else {
	    do {
	    } while (x1f4_last_tcline(&data, data));
	}
    } else {
	status = 0;
    }

    return status;
}
#endif				/* __HEAD_NEXT_LINE__ */


static int
sans_logique_090911_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090910_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090911_0a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090911_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i, v = 1;

    i = 1000000;
    while (i) {
	unsigned e;

	e = i * 523 + 342;

	v *= 1033;

	i--;

	e ^= v;

	e = (e & 0x3fff8000) >> 15 | (e & 0x00007fff) << 15;
	e = (e & 0x3e0f83e0) >> 5 | (e & 0x01f07c1f) << 5;

	status = x1f4_pass_tcline(list, &e, deck_mode, 0);
	if (status) {
	    break;
	}
    }

    return status;
}


static int
sans_logique_090911_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090910_1a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090911_1a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090911_1a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status = 0;
    unsigned i;

    i = 1000000;
    while (i) {
	i--;

	status = x1f4_pass_tcline(list, &i, deck_mode, 0);
	if (status) {
	    break;
	}
    }

    return status;
}


static int
sans_logique_090911_2_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090910_1a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090911_2a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090911_2a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status = 0;
    unsigned i;

    i = 0;
    while (i < 1000000) {
	i++;

	status = x1f4_pass_tcline(list, &i, deck_mode, 0);
	if (status) {
	    break;
	}
    }

    return status;
}


static int
sans_logique_090911_3_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090910_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090911_3a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090911_3a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i, v = 1;

    i = 1000000;
    while (i) {
	void *data;

	v *= 1033;

	i--;

	status = x1f4_find_tcline(list, &v, 0, deck_mode, &data);
	if (status) {
	    status = 0;
	} else {
	}
    }

    return status;
}


#if __HEAD_FIND_FINE__
static int
sans_logique_090928_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090910_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090928_0a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_FIND_FINE__ */


#if __HEAD_FIND_FINE__
static int
sans_logique_090928_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i, v = 1;

    i = 1000000;
    while (1) {
	unsigned e;
	void *data;

	e = i * 523 + 342;

	v *= 1033;

	i--;

	e ^= v;

	e = (e & 0x3fff8000) >> 15 | (e & 0x00007fff) << 15;
	e = (e & 0x3e0f83e0) >> 5 | (e & 0x01f07c1f) << 5;

	status = x1f4_find_tcfine(list, e, 0, &data);
	if (status) {
	    if (i) {
	    } else {
		status = 0;
		if (1) {
		    break;
		}
	    }
	} else {
	    status = 1;
	    if (1) {
		break;
	    }
	}
    }

    return status;
}
#endif				/* __HEAD_FIND_FINE__ */


static int
sans_logique_090929_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090910_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    status = 0;
	    if (status) {
	    } else {
		double depth;
		unsigned max;

		mean_list(list, &depth, &max);

		fprintf(stdout, "%47s (depth): %2.2f mean, %2u max\n",
			test_data->description, depth, max);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090929_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i, v = 1;

    i = 2000000;
    while (i) {
	unsigned e;
	void *data;

	e = i * 523 + 342;

	v *= 1033;

	i--;

	e ^= v;

	e = (e & 0x3fff8000) >> 15 | (e & 0x00007fff) << 15;
	e = (e & 0x3e0f83e0) >> 5 | (e & 0x01f07c1f) << 5;

	status = x1f4_post_tcline
	    (list, &e, deck_mode, 0, sizeof(unsigned), &data);
	if (status) {
	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = e;
	}
    }

    return status;
}


static int
sans_logique_090929_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090929_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    status = sans_logique_090929_1a(SANS_LOGIQUE_ARGS_2, list);
	    if (status) {
	    } else {
		double depth;
		unsigned max;

		mean_list(list, &depth, &max);

		fprintf(stdout, "%47s (depth): %2.2f mean, %2u max\n",
			test_data->description, depth, max);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090929_1a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i, v = 1;

    i = 2000000;
    while (i) {
	unsigned e;

	e = i * 523 + 342;

	v *= 1033;

	i--;

	e ^= v;

	e = (e & 0x3fff8000) >> 15 | (e & 0x00007fff) << 15;
	e = (e & 0x3e0f83e0) >> 5 | (e & 0x01f07c1f) << 5;

	v *= 1033;

	i--;

	status = x1f4_pass_tcline(list, &e, deck_mode, 0);
	if (status) {
	    break;
	}
    }

    return status;
}


static int
sans_logique_100124_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	count2_decls;

	start_count2();

	status = sans_logique_100124_0a(SANS_LOGIQUE_ARGS_2, list);

	if (status) {
	} else {
	    finish_count2_l(test_data->description, 55);
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_100124_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned c = 1, i;

    i = 1000000;
    while (i) {
	unsigned e;
	void *data;

	e = i * 523 + 342;

	i--;

	c *= 13;
	c &= (1 << 30) - 1;

	status = x1f4_land_tcline
	    (list, &c, deck_mode, 0, sizeof(unsigned), &data);
	if (status) {
	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = c;
	}
    }

    return status;
}


#if __HEAD_LLXD_LINE__
static int
sans_logique_100718_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090910_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100718_0a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_LLXD_LINE__ */


#if __HEAD_LLXD_LINE__
static int
sans_logique_100718_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned e = 0, i;

    i = 1000000;
    while (i) {
	void *data;

	e += 660764183;
	e &= (1 << 30) - 1;

	i--;

	status = x1f4_llxd_tcline(list, &e, 0, deck_mode, &data);
	if (status) {
	    status = 0;
	} else {
	}
    }

    return status;
}
#endif				/* __HEAD_LLXD_LINE__ */


#if __HEAD_LXXD_LINE__
static int
sans_logique_100718_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090910_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100718_1a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_LXXD_LINE__ */


#if __HEAD_LXXD_LINE__
static int
sans_logique_100718_1a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned e = 0, i;

    i = 1000000;
    while (i) {
	void *data;

	e += 660764183;
	e &= (1 << 30) - 1;

	i--;

	status = x1f4_lxxd_tcline(list, &e, 0, deck_mode, &data);
	if (status) {
	    status = 0;
	} else {
	}
    }

    return status;
}
#endif				/* __HEAD_LXXD_LINE__ */


#if __HEAD_FXXD_LINE__
static int
sans_logique_100718_2_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090910_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100718_2a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_FXXD_LINE__ */


#if __HEAD_FXXD_LINE__
static int
sans_logique_100718_2a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned e = 0, i;

    i = 1000000;
    while (i) {
	void *data;

	e += 660764183;
	e &= (1 << 30) - 1;

	i--;

	status = x1f4_fxxd_tcline(list, &e, 0, deck_mode, &data);
	if (status) {
	    status = 0;
	} else {
	}
    }

    return status;
}
#endif				/* __HEAD_FXXD_LINE__ */


#if __HEAD_FGXD_LINE__
static int
sans_logique_100718_3_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090910_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100718_3a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_FGXD_LINE__ */


#if __HEAD_FGXD_LINE__
static int
sans_logique_100718_3a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned e = 0, i;

    i = 1000000;
    while (i) {
	void *data;

	e += 660764183;
	e &= (1 << 30) - 1;

	i--;

	status = x1f4_fgxd_tcline(list, &e, 0, deck_mode, &data);
	if (status) {
	    status = 0;
	} else {
	}
    }

    return status;
}
#endif				/* __HEAD_FGXD_LINE__ */


#if __HEAD_FMXD_LINE__
static int
sans_logique_100719_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090910_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100719_0a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_FMXD_LINE__ */


#if __HEAD_FMXD_LINE__
static int
sans_logique_100719_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i, v = 1;

    i = 1000000;
    while (1) {
	unsigned e;
	void *data;

	e = i * 523 + 342;

	v *= 1033;

	i--;

	e ^= v;

	e = (e & 0x3fff8000) >> 15 | (e & 0x00007fff) << 15;
	e = (e & 0x3e0f83e0) >> 5 | (e & 0x01f07c1f) << 5;

	status = x1f4_fmxd_tcline(list, &e, 0, deck_mode, &data);
	if (status) {
	    if (i) {
	    } else {
		status = 0;
		if (1) {
		    break;
		}
	    }
	} else {
	    status = 1;
	    if (1) {
		break;
	    }
	}
    }

    return status;
}
#endif				/* __HEAD_FMXD_LINE__ */


#if __HEAD_LMXD_LINE__
static int
sans_logique_100719_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090910_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100719_1a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_LMXD_LINE__ */


#if __HEAD_LMXD_LINE__
static int
sans_logique_100719_1a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i, v = 1;

    i = 1000000;
    while (1) {
	unsigned e;
	void *data;

	e = i * 523 + 342;

	v *= 1033;

	i--;

	e ^= v;

	e = (e & 0x3fff8000) >> 15 | (e & 0x00007fff) << 15;
	e = (e & 0x3e0f83e0) >> 5 | (e & 0x01f07c1f) << 5;

	status = x1f4_lmxd_tcline(list, &e, 0, deck_mode, &data);
	if (status) {
	    if (i) {
	    } else {
		status = 0;
		if (1) {
		    break;
		}
	    }
	} else {
	    status = 1;
	    if (1) {
		break;
	    }
	}
    }

    return status;
}
#endif				/* __HEAD_LMXD_LINE__ */


static int
sans_logique_100806_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	count2_decls;

	srand(0);

	start_count2();

	status = sans_logique_100806_0a(SANS_LOGIQUE_ARGS_2, list);

	if (status) {
	} else {
	    finish_count2_l(test_data->description, 55);
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_100806_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i, v = 1;

    i = 1000000;
    while (i) {
	unsigned e;
	void *data;

	i--;

	v *= 1033;

	e = rand() & 1023;
	e <<= 20;
	e += (i * 523 ^ v) & 1048575;

	status = x1f4_post_tcline
	    (list, &e, deck_mode, 0, sizeof(unsigned), &data);
	if (status) {
	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = e;
	}
    }

    return status;
}


static int
sans_logique_100806_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	void *line = NULL;

	status = init_path(context_data, &line);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100806_1a(SANS_LOGIQUE_ARGS_2, line, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }

	    excess = fini_path(context_data, &line);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
sans_logique_100806_1a(SANS_LOGIQUE_ARGS_1, void *line, void *list)
{
    int status;
    unsigned i = 1000000;
    unsigned char *slip;

    slip = line;

    while (i) {
	unsigned extend;
	void *text;

	i--;

	extend = *slip++;

	status = x1f4_post_tcline(list, slip, deck_miss, 0, extend, &text);
	if (status) {
	    break;
	} else {
	    memcpy(text, slip, extend);
	    if (1) {
		slip += extend;
	    }
	}
    }

    return status;
}


static int
sans_logique_100806_2_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	void *line = NULL;

	status = init_path(context_data, &line);
	if (status) {
	} else {
	    status = sans_logique_100806_1a(SANS_LOGIQUE_ARGS_2, line, list);
	    if (status) {
	    } else {
		count2_decls;

		start_count2();

		status = sans_logique_100806_2a
		    (SANS_LOGIQUE_ARGS_2, line, list);

		if (status) {
		} else {
		    finish_count2_l(test_data->description, 55);
		}
	    }

	    excess = fini_path(context_data, &line);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
sans_logique_100806_2a(SANS_LOGIQUE_ARGS_1, void *line, void *list)
{
    int status;
    unsigned i = 1000000;
    unsigned char *slip;

    slip = line;

    while (1) {
#if __STRING_SELECT__
	struct lock_type lock;
#endif				/* __STRING_SELECT__ */
	void *text;

	i--;

#if __STRING_SELECT__
	lock.byte = slip[1];
	lock.call = 0;
	lock.left = 0;
	lock.plus = 0;
	lock.text = slip + 1;
#endif				/* __STRING_SELECT__ */

#if __STRING_SELECT__
	status = x1f4_find_tcline(list, &lock, 0, deck_call, &text);
#else
	status = x1f4_find_tcline(list, slip + 1, 0, deck_miss, &text);
#endif				/* __STRING_SELECT__ */
	if (status) {
	    if (i) {
		slip += *slip + 1;
	    } else {
		status = 0;
		if (1) {
		    break;
		}
	    }
	} else {
	    status = 1;
	    if (1) {
		break;
	    }
	}
    }

    return status;
}


#if __HEAD_FIND_WORD__
static int
sans_logique_100806_3_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	void *line = NULL;

	status = init_path(context_data, &line);
	if (status) {
	} else {
	    status = sans_logique_100806_1a(SANS_LOGIQUE_ARGS_2, line, list);
	    if (status) {
	    } else {
		count2_decls;

		start_count2();

		status = sans_logique_100806_3a
		    (SANS_LOGIQUE_ARGS_2, line, list);

		if (status) {
		} else {
		    finish_count2_l(test_data->description, 55);
		}
	    }

	    excess = fini_path(context_data, &line);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}
#endif				/* __HEAD_FIND_WORD__ */


#if __HEAD_FIND_WORD__
static int
sans_logique_100806_3a(SANS_LOGIQUE_ARGS_1, void *line, void *list)
{
    int status;
    unsigned i = 1000000;
    unsigned char *slip;

    slip = line;

    while (1) {
	void *text;

	i--;

	status = x1f4_find_tcword(list, slip + 1, 0, &text);
	if (status) {
	    if (i) {
		slip += *slip + 1;
	    } else {
		status = 0;
		if (1) {
		    break;
		}
	    }
	} else {
	    status = 1;
	    if (1) {
		break;
	    }
	}
    }

    return status;
}
#endif				/* __HEAD_FIND_WORD__ */


static int
sans_logique_100812_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	void *line = NULL;

	status = init_rate(context_data, &line);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100812_0a(SANS_LOGIQUE_ARGS_2, line, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }

	    excess = fini_rate(context_data, &line);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
sans_logique_100812_0a(SANS_LOGIQUE_ARGS_1, void *line, void *list)
{
    int status;
    void *data;

    status = x1f4_lead_l2list(line, &data);
    if (status == X1f4_L2LIST_DECK_ERROR) {
	status = 0;
    } else {
	do {
	    unsigned *call, extend;
	    void *text;

	    call = data;

	    extend = *call + 1;

	    status = x1f4_post_tcline
		(list, (void *) (call + 1), deck_miss, 0, extend, &text);
	    if (status) {
		if (status == X1f4_L2CLIP_EVER_MATCH) {
		    status = 0;
		} else {
		    break;
		}
	    } else {
		memcpy(text, call + 1, extend);
	    }
	} while (x1f4_next_l2list(&data, data));
    }

    return status;
}


static int
sans_logique_100812_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	void *line = NULL;

	status = init_rate(context_data, &line);
	if (status) {
	} else {
	    status = sans_logique_100812_0a(SANS_LOGIQUE_ARGS_2, line, list);
	    if (status) {
	    } else {
		count2_decls;

		start_count2();

		status = sans_logique_100812_1a
		    (SANS_LOGIQUE_ARGS_2, line, list);

		if (status) {
		} else {
		    finish_count2_l(test_data->description, 55);
		}
	    }

	    excess = fini_rate(context_data, &line);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
sans_logique_100812_1a(SANS_LOGIQUE_ARGS_1, void *line, void *list)
{
    int status;
    void *data;

    status = x1f4_lead_l2list(line, &data);
    if (status == X1f4_L2LIST_DECK_ERROR) {
	status = 0;
    } else {
	do {
#if __STRING_SELECT__
	    struct lock_type lock;
#endif				/* __STRING_SELECT__ */
	    unsigned *call;
	    void *text;

	    call = data;

#if __STRING_SELECT__
	    lock.byte = *(unsigned char *) (call + 1);
	    lock.call = 0;
	    lock.left = 0;
	    lock.plus = 0;
	    lock.text = call + 1;
#endif				/* __STRING_SELECT__ */

#if __STRING_SELECT__
	    status = x1f4_find_tcline(list, &lock, 0, deck_call, &text);
#else
	    status = x1f4_find_tcline
		(list, (void *) (call + 1), 0, deck_miss, &text);
#endif				/* __STRING_SELECT__ */
	    if (status) {
		status = 0;
	    } else {
		status = 1;
		if (1) {
		    break;
		}
	    }
	} while (x1f4_next_l2list(&data, data));
    }

    return status;
}


#if __HEAD_FIND_WORD__
static int
sans_logique_100819_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	void *line = NULL;

	status = init_rate(context_data, &line);
	if (status) {
	} else {
	    status = sans_logique_100812_0a(SANS_LOGIQUE_ARGS_2, line, list);
	    if (status) {
	    } else {
		count2_decls;

		start_count2();

		status = sans_logique_100819_1a
		    (SANS_LOGIQUE_ARGS_2, line, list);

		if (status) {
		} else {
		    finish_count2_l(test_data->description, 55);
		}
	    }

	    excess = fini_rate(context_data, &line);
	    if (excess) {
		if (status) {
		} else {
		    status = excess;
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}
#endif				/* __HEAD_FIND_WORD__ */


#if __HEAD_FIND_WORD__
static int
sans_logique_100819_1a(SANS_LOGIQUE_ARGS_1, void *line, void *list)
{
    int status;
    void *data;

    status = x1f4_lead_l2list(line, &data);
    if (status == X1f4_L2LIST_DECK_ERROR) {
	status = 0;
    } else {
	do {
	    unsigned *call;
	    void *text;

	    call = data;
	    status = x1f4_find_tcword(list, (void *) (call + 1), 0, &text);
	    if (status) {
		status = 0;
	    } else {
		status = 1;
		if (1) {
		    break;
		}
	    }
	} while (x1f4_next_l2list(&data, data));
    }

    return status;
}
#endif				/* __HEAD_FIND_WORD__ */


static int
sans_logique________0_(SANS_LOGIQUE_ARGS_1)
{
    count2_print("%*s: not available\n", 55, test_data->description);

    return 0;
}


static int
sans_logique__________(SANS_LOGIQUE_ARGS_1)
{
    int status;

    test_data = test_1_9;

    do {
	status = test_data->test(SANS_LOGIQUE_ARGS_2);
	if (status) {
	    break;
	} else {
	    test_data++;
	}
    } while (test_data->test != sans_logique__________);

    return status;
}


static int
scan_node(void *text, void *node, unsigned slip)
{
    if (slip < 64) {
	unsigned *miss;

	miss = text;
	miss[slip]++;
    }

    return 0;
}


static void
class(void)
{
    x1f4_hook_vtkey(stdout, list_link, link____);
}


static void
usage(void)
{
    puts("Usage: tc39 [OPTIONS] TEST\n\
Profile the " tcname "line associative array.\n\
\n\
Options:\n\
  -d, --dictionary PATH		set dictionary as PATH\n\
      --list			lists available tests and exit\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    context.dictionary = NULL;

    unsetenv("POSIXLY_CORRECT");

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "dictionary",   0x01,   NULL,   0x64    },
		{   "list",	    0x00,   NULL,   0x6c    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "d:", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'd':
		context.dictionary = optarg;
		break;
	    case 'h':
	    case 'l':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'l':
		class();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (argc - optind != 1) {
	fprintf(stderr, "%s: wrong number of arguments\nTry `%s --help' for "
		"more information.\n", argv[0], argv[0]);

	return 1;
    } else {
	const void *test;

	if (x1f4_find_vtkey(argv[optind], link____, &test)) {
	    status = 1;

	    fprintf(stderr, "%s: cannot parse test specification: `%s'\n",
		    argv[0], argv[optind]);
	} else {
	    context.self = argv[0];

	    status = ((struct test_type *) test)->test(&context, test);
	}
    }

    return status;
}
