/*
 * tc40.c
 * Copyright (C) 2002-2010, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <libx1f4l2.h>
#include <tcqp.h>

typedef struct context_type {
    char **argv, *seek;
    const char *self;
    int argc, do_logique, do_reverse;
} context_type;

static int case_away(struct context_type *);
static int copy_list(struct context_type *, void *);
static int deck_line(struct context_type *, void *, char *);
static int deck_miss(void *, void *);
static int dump_list(struct context_type *, void *);
#if !__HEAD_NEXT_LINE__
static int dump_node(void *, void *);
static int ever_node(void *, void *);
#endif				/* !__HEAD_NEXT_LINE__ */
static int ever_seek(void *, const char *, const char *, void *);
static int fast_line(struct context_type *, void *, char *);
static int fini_list(struct context_type *, void **);
static int init_list(struct context_type *, void **);
static int peek_list(struct context_type *, void *);
static int read_list(struct context_type *, void *);
static int seek_list(struct context_type *, void *);
static int slip_miss(void *, void *);
static int tick_list(struct context_type *, void *);
static int type_list(struct context_type *, void *);

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = read_list(context_data, list);
	if (status) {
	} else {
	    status = 0;
	    if (status) {
	    } else {
		status = dump_list(context_data, list);
		if (status) {
		} else {
		    status = tick_list(context_data, list);
		    if (status) {
		    } else {
			status = dump_list(context_data, list);
			if (status) {
			} else {
			    status = seek_list(context_data, list);
			    if (status) {
			    } else {
				status = dump_list(context_data, list);
				if (status) {
				} else {
				    status = peek_list(context_data, list);
				    if (status) {
				    } else {
					status = dump_list(context_data, list);
					if (status) {
					} else {
					}
				    }
				}
			    }
			}
		    }
		}
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
copy_list(struct context_type *context_data, void *list)
{
    int status;
    unsigned size;

    size = context_data->argc;
    if (size) {
	char **argv;

	argv = context_data->argv;

	while (1) {
	    status = deck_line(context_data, list, *argv);
	    if (status) {
		break;
	    } else {
		size--;
		if (size) {
		    argv++;
		} else {
		    break;
		}
	    }
	}
    } else {
	status = 0;
    }

    return status;
}


static int
deck_line(struct context_type *context_data, void *list, char *deck)
{
    int status;
    void *slip;

    status = x1f4_post_tcline
	(list, deck, slip_miss, 0, sizeof(char *), &slip);
    if (0) {
    } else {
	if (status) {
	    fprintf(stderr, "%s: cannot insert item\n", context_data->self);
	} else {
	    char **miss;

	    miss = slip;
	    *miss = deck;
	}
    }

    return status;
}


static int
deck_miss(void *deck, void *miss)
{
    int delete;
    unsigned char *s, *v;

    s = (unsigned char *) deck;
    v = (unsigned char *) *(char **) miss;

    while (1) {
	unsigned char c;

	c = *s;
	if (c ^ *v) {
	    delete = (int) c - (int) *v;

	    break;
	} else {
	    if (c) {
		s++;
		v++;
	    } else {
		delete = 0;

		break;
	    }
	}
    }

    return delete;
}


static int
dump_list(struct context_type *context_data, void *list)
{
#if !__HEAD_NEXT_LINE__
    x1f4_lime_tcline(list, NULL, dump_node);
#else
    if (x1f4_case_tcline(list)) {
	void *node;

	if (x1f4_lead_tcline(list, &node)) {
	} else {
	    do {
		fputs("   ", stdout);
		fputs(*(char **) node, stdout);
	    } while (x1f4_next_tcline(&node, node));
	}
    }
#endif				/* !__HEAD_NEXT_LINE__ */

    fputc('\n', stdout);

    return 0;
}


#if !__HEAD_NEXT_LINE__
static int
dump_node(void *text, void *node)
{
    fputs("   ", stdout);
    fputs(*(char **) node, stdout);

    return 0;
}
#endif				/* !__HEAD_NEXT_LINE__ */


#if !__HEAD_NEXT_LINE__
static int
ever_node(void *text, void *node)
{
    if (node == *(void **) text) {
	fputs(" ", stdout);
	fputs(*((void **) text + 1), stdout);
    } else {
	fputs("    ", stdout);
    }

    return 0;
}
#endif				/* !__HEAD_NEXT_LINE__ */


static int
ever_seek(void *list, const char *ever, const char *slip, void *seek)
{
#if !__HEAD_NEXT_LINE__
    const char *text[2];
#endif				/* !__HEAD_NEXT_LINE__ */

    fputs(ever, stdout);

#if !__HEAD_NEXT_LINE__
    text[0] = seek;
    text[1] = slip;
#endif				/* !__HEAD_NEXT_LINE__ */

#if !__HEAD_NEXT_LINE__
    x1f4_lime_tcline(list, text, ever_node);
#else
    if (x1f4_case_tcline(list)) {
	void *node;

	if (x1f4_lead_tcline(list, &node)) {
	} else {
	    do {
		if (node == seek) {
		    fputs(" ", stdout);
		    fputs(slip, stdout);
		    if (1) {
			break;
		    }
		} else {
		    fputs("    ", stdout);
		}
	    } while (x1f4_next_tcline(&node, node));
	}
    }
#endif				/* !__HEAD_NEXT_LINE__ */

    fputc('\n', stdout);

    return 0;
}


static int
fast_line(struct context_type *context_data, void *list, char *deck)
{
    int status;
    void *slip;

    status = x1f4_post_tcline
	(list, deck, deck_miss, 0, sizeof(char *), &slip);
    if (status == X1f4_TCLINE_EVER_MATCH) {
	char **miss;

	miss = slip;
	*miss = deck;

	status = 0;
    } else {
	if (status) {
	    fprintf(stderr, "%s: cannot insert item\n", context_data->self);
	} else {
	    char **miss;

	    miss = slip;
	    *miss = deck;
	}
    }

    return status;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_tcline(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_init_tcline(list, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    } else {
    }

    return status;
}


static int
peek_list(struct context_type *context_data, void *list)
{
    void *node;

    if (x1f4_find_tcline(list, context_data->seek, 0, deck_miss, &node)) {
	ever_seek(list, "*", " ^ ", node);
    } else {
	fputs("*\n", stdout);
    }

    return 0;
}


static int
read_list(struct context_type *context_data, void *list)
{
    int status;

    if (context_data->do_logique) {
	status = copy_list(context_data, list);
    } else {
	status = type_list(context_data, list);
    }

    return status;
}


static int
seek_list(struct context_type *context_data, void *list)
{
    void *node;

    if (x1f4_llxd_tcline(list, context_data->seek, 0, deck_miss, &node)) {
	ever_seek(list, "L", " ^>", node);
    } else {
	fputs("L\n", stdout);
    }
    if (x1f4_lxxd_tcline(list, context_data->seek, 0, deck_miss, &node)) {
	ever_seek(list, "l", " ^>", node);
    } else {
	fputs("l\n", stdout);
    }
    if (x1f4_fxxd_tcline(list, context_data->seek, 0, deck_miss, &node)) {
	ever_seek(list, "g", "<^ ", node);
    } else {
	fputs("g\n", stdout);
    }
    if (x1f4_fgxd_tcline(list, context_data->seek, 0, deck_miss, &node)) {
	ever_seek(list, "G", "<^ ", node);
    } else {
	fputs("G\n", stdout);
    }

    return 0;
}


static int
slip_miss(void *deck, void *miss)
{
    int delete;
    unsigned char *s, *v;

    s = (unsigned char *) deck;
    v = (unsigned char *) *(char **) miss;

    while (1) {
	unsigned char c;

	c = *s;
	if (c ^ *v) {
	    delete = (int) c - (int) *v;

	    break;
	} else {
	    if (c) {
		s++;
		v++;
	    } else {
		delete = 1;

		break;
	    }
	}
    }

    return delete;
}


static int
tick_list(struct context_type *context_data, void *list)
{
    void *node;

    if (x1f4_fmxd_tcline(list, context_data->seek, 0, deck_miss, &node)) {
	ever_seek(list, "F", "<^ ", node);
    } else {
	fputs("F\n", stdout);
    }
    if (x1f4_lmxd_tcline(list, context_data->seek, 0, deck_miss, &node)) {
	ever_seek(list, "L", " ^>", node);
    } else {
	fputs("L\n", stdout);
    }

    return 0;
}


static int
type_list(struct context_type *context_data, void *list)
{
    int status;
    unsigned size;

    size = context_data->argc;
    if (size) {
	char **argv;

	argv = context_data->argv;

	while (1) {
	    status = fast_line(context_data, list, *argv);
	    if (status) {
		break;
	    } else {
		size--;
		if (size) {
		    argv++;
		} else {
		    break;
		}
	    }
	}
    } else {
	status = 0;
    }

    return status;
}


static void
usage(void)
{
    puts("Usage: tc40 [OPTIONS] SEEK [STRING]...\n\
Test " tcname "_find()s, look for SEEK in STRING(s).\n\
\n\
Options:\n\
  -d, --preserve		preserve input duplicates\n\
  -r, --reverse			does nothing\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;
    struct context_type context;

    unsetenv("POSIXLY_CORRECT");

    context.do_logique = 0;
    context.do_reverse = 0;

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "preserve",	    0x00,   NULL,   0x64    },
		{   "reverse",	    0x00,   NULL,   0x72    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "dr", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'd':
		context.do_logique = 1;
		break;
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case 'r':
		context.do_reverse = 1;
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if ((argc - optind) < 1) {
	fprintf(stderr, "%s: too few arguments\nTry `%s --help' for more inf"
		"ormation.\n", argv[0], argv[0]);

	return 1;
    } else {
	context.self = argv[0];

	context.seek = argv[optind];

	optind++;

	context.argc = argc - optind;
	context.argv = argv + optind;

	status = case_away(&context);
    }

    return status;
}
