/*
 * ts17.c
 * Copyright (C) 2002-2010, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <inter.h>
#include <libx1f4l2.h>
#include <tcxx.h>

typedef struct context_type {
    const char *self;
} context_type;

static int case_away(struct context_type *);
static int dump_list(struct context_type *, void *);
static int fast_line(struct context_type *, void *, unsigned, char *,
		     unsigned);
static int fini_list(struct context_type *, void **);
static int flip_list(struct context_type *, void *);
static int init_list(struct context_type *, void **);
static int read_list(struct context_type *, void *);

static void usage(void);

static int
case_away(struct context_type *context_data)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = read_list(context_data, list);
	if (status) {
	} else {
	    status = flip_list(context_data, list);
	    if (status) {
	    } else {
		status = dump_list(context_data, list);
	    }
	}

	excess = fini_list(context_data, &list);
	if (excess) {
	    if (status) {
	    } else {
		status = excess;
	    }
	}
    }

    return status;
}


static int
dump_list(struct context_type *context_data, void *list)
{
    unsigned size;

    x1f4_size_tclist(list, &size);
    if (size) {
	unsigned i = 0;

	do {
	    void *peek;

	    x1f4_peek_tclist(list, i, &peek);
	    puts(peek);
	    i++;
	} while (i < size);
    }

    return 0;
}


static int
fast_line(struct context_type *context_data, void *list, unsigned index,
	  char *data, unsigned size)
{
    int status;
    void *slip;

    status = x1f4_tail_tclist(list, size + 1, &slip);
    if (status) {
	fprintf(stderr, "%s: cannot insert item\n", context_data->self);
    } else {
	memcpy(slip, data, size + 1);
    }

    return status;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_tclist(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
flip_list(struct context_type *context_data, void *list)
{
    int status;

    status = x1f4_flip_tclist(list);
    if (status) {
    } else {
	status = x1f4_flow_tclist(list);
	if (status) {
	    fprintf(stderr, "%s: corrupt list after flipping\n",
		    context_data->self);
	}
    }

    return status;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_init_tclist(list, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    } else {
    }

    return status;
}


static int
read_list(struct context_type *context_data, void *list)
{
    char *data;
    int done = 0, status;
    unsigned i = 0, size = 0;

    while (1) {
	status = libx1f4l2_read_line(context_data->self, &done, &data, &size);
	if (status) {
	    break;
	} else {
	    if (!done || size) {
		status = fast_line(context_data, list, i, data, size);
	    }

	    if (1) {
		free(data);
	    }

	    if (status) {
		break;
	    } else {
		i++;
	    }

	    if (done) {
		break;
	    }
	}
    }

    return status;
}


static void
usage(void)
{
    puts("Usage: ts17 [OPTIONS]\n\
Test " tcname "_flip(), output reversed input.\n\
\n\
Options:\n\
  --help			display this help and exit\n\
  --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;

    unsetenv("POSIXLY_CORRECT");

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'h':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (0 < argc - optind) {
	fprintf(stderr, "%s: too many arguments\nTry `%s --help' for more in"
		"formation.\n", argv[0], argv[0]);

	return 1;
    } else {
	struct context_type context;

	context.self = argv[0];

	status = case_away(&context);
    }

    return status;
}
