/*
 * ts19.c
 * Copyright (C) 2002-2012, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#if defined HAVE_LIBLEWD
# include <lewd.h>
#endif				/* HAVE_LIBLEWD */
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>

#include <count2.h>
#include <libx1f4l2.h>
#include <tcxx.h>
#if !defined HAVE_LIBLEWD
# include <vkeyhook.h>
# include <vkeytree.h>
#endif				/* !HAVE_LIBLEWD */

#if 1
# define __LINK_SLIP_TEST__		0
#endif				/* 1 */

#if __HEAD_NEXT_LIST__
# define Sans_logique_090909_6_		sans_logique_090909_6_
#else
# undef __LINK_SLIP_TEST__
# define __LINK_SLIP_TEST__		1
# define Sans_logique_090909_6_		sans_logique________0_
#endif				/* __HEAD_NEXT_LIST__ */

#if __HEAD_NEXT_LIST__
# define Sans_logique_090909_7_		sans_logique_090909_7_
#else
# undef __LINK_SLIP_TEST__
# define __LINK_SLIP_TEST__		1
# define Sans_logique_090909_7_		sans_logique________0_
#endif				/* __HEAD_NEXT_LIST__ */

#define SANS_LOGIQUE_ARGS_0 \
    struct context_type *, const struct test_type *
#define SANS_LOGIQUE_ARGS_1 \
    struct context_type *context_data, const struct test_type *test_data
#define SANS_LOGIQUE_ARGS_2 \
    context_data, test_data

#define APPEND_MODE \
    "last position insert 1,000,000 integers"

#define ASPECT_MODE \
    "sequentially fetch all from 1,000,000 integers list"

#define BEHEAD_MODE \
    "forward remove all from 1,000,000 integers list"

#define BETAIL_MODE \
    "backward remove all from 1,000,000 integers list"

#define DELETE_MODE \
    "sequentially remove all from 1,000,000 integers list"

#define DESERT_MODE \
    "first position insert 1,000,000 integers"

#define DIRECT_MODE \
    "forward traverse 1,000,000 integers list"

#define EXCESS_MODE \
    "clear 1,000,000 integers list"

#define EXLINE_MODE \
    "callback traverse 1,000,000 integers list"

#define EXPECT_MODE \
    "backward traverse 1,000,000 integers list"

#define FRIDAY_MODE \
    "sort insert 1,000,000 non unique 30bit 13 powers"

#define FXRAND_MODE \
    "shuffle 1,000,000 integers list"

#define FXSORT_MODE \
    "sort 1,000,000 integers list"

#define INSERT_MODE \
    "random (2nd) insert 1,000,000 integers"

#define INVERT_MODE \
    "reverse 1,000,000 integers list"

#define JOLENE_MODE \
    "insert 1,000,000 empty records"

#define RANDOM_MODE \
    "fetch 1,000,000 from 1,000,000 integers list"

#define REMOVE_MODE \
    "remove all from 1,000,000 integers list"

#define REVERT_MODE \
    "sequentially twice move reverse 1,000,000 integers list"

#define STREAM_TEST \
    "all the tests"

#define TARGET_MODE \
    "(libc) random insert 1,000,000 integers"

#define UPSORT_MODE \
    "sort insert 1,000,000 integers"

typedef struct context_type {
    const char *self;
} context_type;

typedef struct test_type {
    const char *description;
    int (*test) (SANS_LOGIQUE_ARGS_0);
} test_type;

static int deck_mode(void *, void *);
static int fini_list(struct context_type *, void **);
static int init_list(struct context_type *, void **);
static int less_mode(void *, void *, void *);
static int list_link(void *, const void *, const char *, unsigned);
static int miss_node(void *, void *);
static int sans_logique_090909_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090909_0a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090909_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090909_1a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090909_2_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090909_2a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090909_3_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090909_3a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090909_4_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090909_4a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090909_5_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090909_5a(SANS_LOGIQUE_ARGS_0, void *);
#if __HEAD_NEXT_LIST__
static int sans_logique_090909_6_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090909_6a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090909_7_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090909_7a(SANS_LOGIQUE_ARGS_0, void *);
#endif				/* __HEAD_NEXT_LIST__ */
static int sans_logique_090910_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090910_0a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090910_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090910_1a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090910_2_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090910_2a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_090910_3_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_090910_3a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_091004_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_091004_0a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_091022_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_091022_0a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_100124_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100124_0a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_100124_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100124_1a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_100731_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100731_0a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_100731_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100731_1a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_100806_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100806_0a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_100813_0_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100813_0a(SANS_LOGIQUE_ARGS_0, void *);
static int sans_logique_100813_1_(SANS_LOGIQUE_ARGS_0);
static int sans_logique_100813_1a(SANS_LOGIQUE_ARGS_0, void *);
#if __LINK_SLIP_TEST__
static int sans_logique________0_(SANS_LOGIQUE_ARGS_0);
#endif				/* __LINK_SLIP_TEST__ */
static int sans_logique__________(SANS_LOGIQUE_ARGS_0);

static void class(void);
static void usage(void);

static const struct test_type test_1_9[] = {
/* *INDENT-OFF* */
    {	TARGET_MODE,		sans_logique_100806_0_			},
    {	INSERT_MODE,		sans_logique_090909_0_			},
    {	APPEND_MODE,		sans_logique_090909_1_			},
    {	DESERT_MODE,		sans_logique_090909_2_			},
    {	FXSORT_MODE,		sans_logique_090909_3_			},
    {	FXRAND_MODE,		sans_logique_090909_4_			},
    {	REMOVE_MODE,		sans_logique_090909_5_			},
    {	DIRECT_MODE,		Sans_logique_090909_6_			},
    {	EXPECT_MODE,		Sans_logique_090909_7_			},
    {	RANDOM_MODE,		sans_logique_090910_0_			},
    {	ASPECT_MODE,		sans_logique_090910_1_			},
    {	DELETE_MODE,		sans_logique_090910_2_			},
    {	EXCESS_MODE,		sans_logique_090910_3_			},
    {	EXLINE_MODE,		sans_logique_091004_0_			},
    {	JOLENE_MODE,		sans_logique_091022_0_			},
    {	UPSORT_MODE,		sans_logique_100124_0_			},
    {	FRIDAY_MODE,		sans_logique_100124_1_			},
    {	BEHEAD_MODE,		sans_logique_100731_0_			},
    {	BETAIL_MODE,		sans_logique_100731_1_			},
    {	INVERT_MODE,		sans_logique_100813_0_			},
    {	REVERT_MODE,		sans_logique_100813_1_			},
    {	STREAM_TEST,		sans_logique__________			}
/* *INDENT-ON* */
};

static const struct x1f4_vkeytree_type link_1_g[] = {
/* *INDENT-OFF* */
    {	"",		{	test_1_9 + 1	},	NULL		},
    {	"2",		{	test_1_9 + 0	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link_1_f[] = {
/* *INDENT-OFF* */
    {	"ll",		{	test_1_9 + 21	},	NULL		},
    {	"tom",		{	test_1_9 + 14	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link_1_e[] = {
/* *INDENT-OFF* */
    {	"all",		{	test_1_9 + 13	},	NULL		},
    {	"lear",		{	test_1_9 + 12	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link_1_d[] = {
/* *INDENT-OFF* */
    {	"etch",		{	test_1_9 + 10	},	NULL		},
    {	"orward",	{	test_1_9 + 7	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link_1_c[] = {
/* *INDENT-OFF* */
    {	"",		{	test_1_9 + 5	},	NULL		},
    {	"om",		{	test_1_9 + 9	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link___r_e[] = {
/* *INDENT-OFF* */
    {	"move",		{	test_1_9 + 6	},	NULL		},
    {	"verse",	{	test_1_9 + 19	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link___r[] = {
/* *INDENT-OFF* */
    {	"and",		{	NULL		},	link_1_c	},
    {	"e",		{	NULL		},	link___r_e	},
    {	"oll",		{	test_1_9 + 20	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link_1_a[] = {
/* *INDENT-OFF* */
    {	"lip",		{	test_1_9 + 18	},	NULL		},
    {	"ort",		{	test_1_9 + 4	},	NULL		},
    {	"uffix",	{	test_1_9 + 2	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link_1_9[] = {
/* *INDENT-OFF* */
    {	"a",		{	NULL		},	link_1_f	},
    {	"c",		{	NULL		},	link_1_e	},
    {	"backward",	{	test_1_9 + 8	},	NULL		},
    {	"delete",	{	test_1_9 + 11	},	NULL		},
    {	"f",		{	NULL		},	link_1_d	},
    {	"insert",	{	NULL		},	link_1_g	},
    {	"lock",		{	test_1_9 + 15	},	NULL		},
    {	"mind",		{	test_1_9 + 17	},	NULL		},
    {	"prefix",	{	test_1_9 + 3	},	NULL		},
    {	"r",		{	NULL		},	link___r	},
    {	"s",		{	NULL		},	link_1_a	},
    {	"thirteen",	{	test_1_9 + 16	},	NULL		},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
}, link____[] = {
/* *INDENT-OFF* */
    {	"",		{	NULL		},	link_1_9	},
    {	NULL,		{	NULL		},	NULL		}
/* *INDENT-ON* */
};

static int
deck_mode(void *deck, void *miss)
{
    return (int) *(unsigned *) deck - (int) *(unsigned *) miss;
}


static int
fini_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_fini_tclist(list);
    if (status) {
	fprintf(stderr, "%s: cannot destroy list\n", context_data->self);
    }

    return status;
}


static int
init_list(struct context_type *context_data, void **list)
{
    int status;

    status = x1f4_init_tclist(list, 0, NULL);
    if (status) {
	fprintf(stderr, "%s: cannot create list\n", context_data->self);
    } else {
    }

    return status;
}


static int
less_mode(void *none, void *deck, void *miss)
{
    return (int) *(unsigned *) deck - (int) *(unsigned *) miss;
}


static int
list_link(void *file, const void *node, const char *data, unsigned size)
{
    fwrite("  ", 2, 1, file);
    fwrite(data, size, 1, file);
    fwrite("                ", 16 - size, 1, file);
    fwrite(" - ", 3, 1, file);
    fputs(((const struct test_type *) node)->description, file);
    fputc('\n', file);

    return 0;
}


static int
miss_node(void *text, void *node)
{
    unsigned *sail;

    sail = text;
    *sail += *(unsigned *) node;

    return 0;
}


static int
sans_logique_090909_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	count2_decls;

	start_count2();

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);

	if (status) {
	} else {
	    finish_count2_l(test_data->description, 55);
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090909_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 0;
    while (i < 1000000) {
	unsigned e;
	void *data;

	e = i * 523 + 342;

	i++;

	e ^= e << 003;
	e += e >> 005;
	e ^= e << 002;
	e += e >> 017;
	e ^= e << 012;

	if (e < i) {
	} else {
	    unsigned a;

	    a = i >> 1;

	    a |= a >> 1;
	    a |= a >> 2;

	    e = e & a;
	}

	if (i & 1) {
	    e = i - e - 1;
	}

	status = x1f4_push_tclist(list, e, sizeof(unsigned), &data);
	if (status) {
	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = i;
	}
    }

    return status;
}


static int
sans_logique_090909_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	count2_decls;

	start_count2();

	status = sans_logique_090909_1a(SANS_LOGIQUE_ARGS_2, list);

	if (status) {
	} else {
	    finish_count2_l(test_data->description, 55);
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090909_1a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	void *data;

	i--;

	status = x1f4_tail_tclist(list, sizeof(unsigned), &data);
	if (status) {
	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = i;
	}
    }

    return status;
}


static int
sans_logique_090909_2_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	count2_decls;

	start_count2();

	status = sans_logique_090909_2a(SANS_LOGIQUE_ARGS_2, list);

	if (status) {
	} else {
	    finish_count2_l(test_data->description, 55);
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090909_2a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	void *data;

	i--;

	status = x1f4_head_tclist(list, sizeof(unsigned), &data);
	if (status) {
	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = i;
	}
    }

    return status;
}


static int
sans_logique_090909_3_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090909_3a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090909_3a(SANS_LOGIQUE_ARGS_1, void *list)
{
    return x1f4_sort_tclist(list, NULL, 0, less_mode);
}


static int
sans_logique_090909_4_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090909_4a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090909_4a(SANS_LOGIQUE_ARGS_1, void *list)
{
    return x1f4_rand_tclist(list);
}


static int
sans_logique_090909_5_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090909_5a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090909_5a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	unsigned e;

	e = i * 523 + 342;

	i--;

	e ^= e << 003;
	e += e >> 005;
	e ^= e << 002;
	e += e >> 017;
	e ^= e << 012;

	if (e < i) {
	} else {
	    unsigned a;

	    a = i >> 1;

	    a |= a >> 1;
	    a |= a >> 2;

	    e = e & a;
	}

	if (i & 1) {
	    e = i - e - 1;
	}

	status = x1f4_miss_tclist(list, e);
	if (status) {
	    break;
	}
    }

    return status;
}


#if __HEAD_NEXT_LIST__
static int
sans_logique_090909_6_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090909_6a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_NEXT_LIST__ */


#if __HEAD_NEXT_LIST__
static int
sans_logique_090909_6a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned e;
    void *data;

    x1f4_size_tclist(list, &e);
    if (e) {
	status = x1f4_lead_tclist(list, &data);
	if (status) {
	} else {
	    do {
	    } while (x1f4_next_tclist(&data, data));
	}
    } else {
	status = 0;
    }

    return status;
}
#endif				/* __HEAD_NEXT_LIST__ */


#if __HEAD_NEXT_LIST__
static int
sans_logique_090909_7_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090909_7a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}
#endif				/* __HEAD_NEXT_LIST__ */


#if __HEAD_NEXT_LIST__
static int
sans_logique_090909_7a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned e;
    void *data;

    x1f4_size_tclist(list, &e);
    if (e) {
	status = x1f4_ever_tclist(list, &data);
	if (status) {
	} else {
	    do {
	    } while (x1f4_last_tclist(&data, data));
	}
    } else {
	status = 0;
    }

    return status;
}
#endif				/* __HEAD_NEXT_LIST__ */


static int
sans_logique_090910_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090910_0a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090910_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
#define E				1000000

    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	unsigned e;
	void *data;

	e = i * 523 + 342;

	i--;

	e ^= e << 003;
	e += e >> 005;
	e ^= e << 002;
	e += e >> 017;
	e ^= e << 012;

	if (e < E) {
	} else {
	    unsigned a;

	    a = E >> 1;

	    a |= a >> 1;
	    a |= a >> 2;

	    e = e & a;
	}

	if (i & 1) {
	    e = E - e - 1;
	}

	status = x1f4_peek_tclist(list, e, &data);
	if (status) {
	    break;
	}
    }

    return status;

#undef E
}


static int
sans_logique_090910_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090910_1a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090910_1a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	void *data;

	i--;

	status = x1f4_peek_tclist(list, i, &data);
	if (status) {
	    break;
	}
    }

    return status;
}


static int
sans_logique_090910_2_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090910_2a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090910_2a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	i--;

	status = x1f4_miss_tclist(list, i);
	if (status) {
	    break;
	}
    }

    return status;
}


static int
sans_logique_090910_3_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_090910_3a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_090910_3a(SANS_LOGIQUE_ARGS_1, void *list)
{
    return x1f4_high_tclist(list);
}


static int
sans_logique_091004_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_091004_0a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_091004_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    unsigned text = 0;

    return x1f4_lime_tclist(list, &text, miss_node);
}


static int
sans_logique_091022_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	count2_decls;

	start_count2();

	status = sans_logique_091022_0a(SANS_LOGIQUE_ARGS_2, list);

	if (status) {
	} else {
	    finish_count2_l(test_data->description, 55);
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_091022_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 0;
    while (i < 1000000) {
	unsigned e;
	void *data;

	e = i * 523 + 342;

	i++;

	e ^= e << 003;
	e += e >> 005;
	e ^= e << 002;
	e += e >> 017;
	e ^= e << 012;

	if (e < i) {
	} else {
	    unsigned a;

	    a = i >> 1;

	    a |= a >> 1;
	    a |= a >> 2;

	    e = e & a;
	}

	if (i & 1) {
	    e = i - e - 1;
	}

	status = x1f4_push_tclist(list, e, 0, &data);
	if (status) {
	    break;
	} else {
	}
    }

    return status;
}


static int
sans_logique_100124_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	count2_decls;

	start_count2();

	status = sans_logique_100124_0a(SANS_LOGIQUE_ARGS_2, list);

	if (status) {
	} else {
	    finish_count2_l(test_data->description, 55);
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_100124_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 0;
    while (i < 1000000) {
	unsigned e;
	void *data;

	e = i * 523 + 342;

	i++;

	e ^= e << 003;
	e += e >> 005;
	e ^= e << 002;
	e += e >> 017;
	e ^= e << 012;

	if (e < i) {
	} else {
	    unsigned a;

	    a = i >> 1;

	    a |= a >> 1;
	    a |= a >> 2;

	    e = e & a;
	}

	if (i & 1) {
	    e = i - e - 1;
	}

	status = x1f4_land_tclist
	    (list, &e, deck_mode, 0, sizeof(unsigned), &data);
	if (status) {
	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = e;
	}
    }

    return status;
}


static int
sans_logique_100124_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	count2_decls;

	start_count2();

	status = sans_logique_100124_1a(SANS_LOGIQUE_ARGS_2, list);

	if (status) {
	} else {
	    finish_count2_l(test_data->description, 55);
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_100124_1a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned c = 1, i;

    i = 1000000;
    while (i) {
	void *data;

	i--;

	c *= 13;
	c &= (1 << 30) - 1;

	status = x1f4_land_tclist
	    (list, &c, deck_mode, 0, sizeof(unsigned), &data);
	if (status) {
	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = c;
	}
    }

    return status;
}


static int
sans_logique_100731_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100731_0a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_100731_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	i--;

	status = x1f4_mind_tclist(list);
	if (status) {
	    break;
	}
    }

    return status;
}


static int
sans_logique_100731_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100731_1a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_100731_1a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	i--;

	status = x1f4_slip_tclist(list);
	if (status) {
	    break;
	}
    }

    return status;
}


static int
sans_logique_100806_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;
	count2_decls;

	srand(0);

	start_count2();

	status = sans_logique_100806_0a(SANS_LOGIQUE_ARGS_2, list);

	if (status) {
	} else {
	    finish_count2_l(test_data->description, 55);
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_100806_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 0;
    while (i < 1000000) {
	void *data;

	i++;

	status = x1f4_push_tclist(list, rand() % i, sizeof(unsigned), &data);
	if (status) {
	    break;
	} else {
	    unsigned *slip;

	    slip = data;
	    *slip = i;
	}
    }

    return status;
}


static int
sans_logique_100813_0_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100813_0a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_100813_0a(SANS_LOGIQUE_ARGS_1, void *list)
{
    return x1f4_flip_tclist(list);
}


static int
sans_logique_100813_1_(SANS_LOGIQUE_ARGS_1)
{
    int status;
    void *list;

    status = init_list(context_data, &list);
    if (status) {
    } else {
	int excess;

	status = sans_logique_090909_0a(SANS_LOGIQUE_ARGS_2, list);
	if (status) {
	} else {
	    count2_decls;

	    start_count2();

	    status = sans_logique_100813_1a(SANS_LOGIQUE_ARGS_2, list);

	    if (status) {
	    } else {
		finish_count2_l(test_data->description, 55);
	    }
	}

	excess = fini_list(context_data, &list);
    }

    return status;
}


static int
sans_logique_100813_1a(SANS_LOGIQUE_ARGS_1, void *list)
{
    int status;
    unsigned i;

    i = 1000000;
    while (i) {
	i--;

	status = x1f4_move_tclist(list, i, 1000000 - i);
	if (status) {
	    break;
	}
    }

    return status;
}


#if __LINK_SLIP_TEST__
static int
sans_logique________0_(SANS_LOGIQUE_ARGS_1)
{
    count2_print("%*s: not available\n", 55, test_data->description);

    return 0;
}
#endif				/* __LINK_SLIP_TEST__ */


static int
sans_logique__________(SANS_LOGIQUE_ARGS_1)
{
    int status;

    test_data = test_1_9;

    do {
	status = test_data->test(SANS_LOGIQUE_ARGS_2);
	if (status) {
	    break;
	} else {
	    test_data++;
	}
    } while (test_data->test != sans_logique__________);

    return status;
}


static void
class(void)
{
    x1f4_hook_vtkey(stdout, list_link, link____);
}


static void
usage(void)
{
    puts("Usage: ts19 [OPTIONS] TEST\n\
Profile the " tcname "list sequence.\n\
\n\
Options:\n\
  --list			lists available tests and exit\n\
  --help			display this help and exit\n\
  --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int status = 0;

    unsetenv("POSIXLY_CORRECT");

    {
	int fast = ~0;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "list",	    0x00,   NULL,   0x6c    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "help",	    0x00,   NULL,   0x68    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 'h':
	    case 'l':
	    case 'v':
		if (!~fast) {
		    fast = c;
		}
		break;
	    case '?':
		return 1;
	    }
	}

	if (~fast) {
	    switch (fast) {
	    case 'h':
		usage();
		break;
	    case 'l':
		class();
		break;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);
	    }

	    return 0;
	}
    }

    if (argc - optind != 1) {
	fprintf(stderr, "%s: wrong number of arguments\nTry `%s --help' for "
		"more information.\n", argv[0], argv[0]);

	return 1;
    } else {
	const void *test;

	if (x1f4_find_vtkey(argv[optind], link____, &test)) {
	    status = 1;

	    fprintf(stderr, "%s: cannot parse test specification: `%s'\n",
		    argv[0], argv[optind]);
	} else {
	    struct context_type context;

	    context.self = argv[0];

	    status = ((struct test_type *) test)->test(&context, test);
	}
    }

    return status;
}
